// msTest.C -*-c++-*-

// $Source: /projects/HMS/nsr/udfAncil/msTest/arcRead.cpp,v $
// $Revision: 1.45 $    $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 96/04/30 23:31:41 $

//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <rw/gslist.h>
#include <rw/cstring.h>
#include <rw/rwfile.h>
#include "nsrLib.h"
#include <devs/nsrRSSdv.h>
#include <devs/nsrRMSdv.h>
#include "arcImEx.h"
#include "cmr.h"
#include "cpio.h"
#ifndef _ARCBROWSE_
#include "arcSynth.h"
#endif

//#ifdef _ARCBROWSE_
#include "bsh.h"
//#endif        // _ARCBROWSE_

//-----------------------------------------------------------------------------
// Some macros
//-----------------------------------------------------------------------------

#ifndef DATA_START_SECTOR
#define DATA_START_SECTOR       300
#endif

const char *defaultDBFile = "/tmp/arcRead.swap";

FILE *myIn;
//-----------------------------------------------------------------------------
// Function prototypes for arcRead/arcBrowse related functions
//-----------------------------------------------------------------------------

static int      printFiles(CHAR_PTR dir,TreeNode& node);
static void     printFile(CHAR_PTR name,Attributes& attr,UINT32 filesize);
static int      restoreFiles(TreeNode& node,hpACDev& acdev,CHAR_PTR directory,
                             SidList& sidList);
#ifndef _ARCBROWSE_
       int      synthCheck (TreeNode& root,hpACDev& acdev,SidList& sidList);
static int      synthSidCheck(TreeNode& node,hpACDev& acdev,Sid_t sid);
static int      synthExtentCheck(char* seed, hpACDev& acdev, Sid_t sid,
                                 Extent& ext, int offset);
#endif  // _ARCBROWSE_
static int      recreateDirTree(TreeNode& node, CHAR_PTR dir);
static int      recreateFile(TreeNode& node,CHAR_PTR filepath);
static int      getSymlinkPath(CHAR_PTR path,TreeNode& node);
static int      restoreAttributes(TreeNode& node,CHAR_PTR dir);
static int      restoreData(TreeNode& node,hpACDev& acdev,
                            CHAR_PTR dir,Sid_t sid);
static int      copyExtent(int fd,hpACDev& acdev,Sid_t sid,int offset,
                           Extent& ext);
static void     freeTree( TreeNode& node );
static void     Usage(void);
       void     setParent(TreeNode& parent,TreeNode& node);

//-----------------------------------------------------------------------------
// Function prototypes for udfval related functions
//-----------------------------------------------------------------------------

static void freeTree( TreeNode& node );
static void Usage(void);
void treeAccess ( char *dirName,
                 RWBinaryTree *dirEntries,
                 void **dirPointer);
void rootAccess (RWCString &rootName, Attributes& attr, void **rootPointer);

void *startDirFunc(VoidPtr parent, char *name, Attributes& attr);
void endDirFunc(void *dir);
void fileFunc(char *name, Attributes& attr, ExtentList& elist);
void symlinkFunc(char *name, Attributes& attr, PathElementList& plist);
void endDirectory(void *dir);
static int              writeCmr ( TreeNode& root, CmrWriter& cmrWriter );

static int writeCpio(TreeNode& root,cpioWriter& writer,hpACDev& acdev,
                     char *parentName);
static void nodeToCpio(TreeNode& node,cpioHeader& hdr,char *pname,char *path);

//-----------------------------------------------------------------------------
// Global variables
//-----------------------------------------------------------------------------

int trace = 0;                  // Should go away, but used for now to
                                // communicate to the jukebox object.

// These global variables are needed for the udfval code.

CmrDiff *cmrDiff;
TreeNode* gRootptr;

//-----------------------------------------------------------------------------
// arcRead.cpp:  This file contains source for a simple test to use the NSR
//               MSLIB Archive object in conjunction with the hpACDev 
//               autochanger device object to read an nsr volume.  By default, 
//               it will simply echo to screen a list of filenames and some 
//               simple information about each file.  If the [ -r directory ] 
//               option is specified, the files will actually be restored to 
//               the given directory.
//
// Usage:
//
// arcRead [-a acrcFile] [-s jbstatFile] [-r directory ] [-f fsname] 
//         [-g volname] [-c cmrFile] [-I] [-o outFile] [-P] [-x] 
//         [-w swapFile] [-W] -v vsname
//
// [-a acrcFile]:    Optional.  Pathname of the acrc configuration file.
// [-s jbStatFile]:  Optional.  Pathname of the jukebox status file.
// [-r directory]:   Optional.  If this option is given, the test will
//                   restore the files to the specified directory.  Note
//                   that you must ensure that the target directory has
//                   sufficient free space.
// [-R]:             Optional.  Specifies that the archive was created with
//                   a synth file.  Compare data to the synth pattern, without
//                   doing a restore.
// [-f fsname]:      Optional.  The file set name to import, quoted if 
//                   necessary.  If not specified, the most recent fileset is
//                   imported from the given volume set.
// [-g volname]:     Optional.  Grows the imported volume set by one volume
//                   with the specified name.
// [-c cmrFile]:     Optional.  If specified, reads the specified CMR file
//                   and compares it with the imported metadata.
// [-l logFile]:     Optional.  Specifies the name of the logfile for the
//                   CMR comparison.
// [-I]:             Optional.  No import; read volume labels only.
// [-o outFile]:     Optional.  Overrides all other options.  Inventories 
//                   the jukebox and generates a status file of the specified
//                   name.
// [-P]:             Optional.  Do not print list of files to stdout.
// [-x]:             Optional.  Output imported files as a cpio archive.
// [-w swapFile]:    Optional.  Specifies the pathname of the swapfile.
//                   This option is only useful if the paging tree option
//                   has been compiled in (ROGUEWAVIFIED).  The swapfile,
//                   and a swapfile named swapfile.md, will be created to
//                   store metadata information.  Default:  /tmp/arcRead.swap
// [-W]:             Optional.  If specified, the swap file and the
//                   metadata file will NOT be removed at the completion
//                   of the program.
// -v vsname:        Mandatory.  The volume set name, quoted if necessary.
//
// Example:  
//
// arcRead -a /users/jonh/.acrc -s /tmp/jbstat.1 -v "Volset 1" -f "Fileset 1"
//
// ENVIRONMENT VARIABLES:
//
// The acrc file and the jukebox status file can both be specified by 
// environment variables:
//
// ACRC_FILE:    Environment variable containing the acrc file pathname.
// JBSTAT_FILE:  Environment variable containing the jukebox status file 
//               pathname.
// ARCTRACE:     Environment variable containing the value of the trace var.
//
// Note that the -a and -s options will override any value in the corresponding
// environment variable, and the -a and -s options are mandatory if the
// environment variables are not set.
//-----------------------------------------------------------------------------

declare(RWGSlist,RWCString)

main (int argc,CHAR_PTR argv[])
{
    int                 c;
    INT32               i;
    int                 rval = 0;
    int                 freeSpaceRval = 0;
    int                 usage = 0;
    int                 noImport = 0;
    int                 noPrint = 0;
    int                 outputCpio = 0;
    int                 rmFile = 1;
    CHAR_PTR            cptr;
    RWCString           acrcFile;
    RWCString           jbStatFile;
    RWCString           vsname;
    RWCString           fsname;
    RWCString           restoreDir;
    RWCString           outFile;
    RWCString           *vnptr;
    RWCString           cmrFile;
    RWCString           genCmrFile;
    RWCString           logFile("arcRead.out");
    RWCString           dbFile(defaultDBFile);
    RWCString           freeFileName;
    RWGSlist(RWCString) growVols;
    RWBoolean           doSynthCheck = FALSE;
    int                 killVS = 0;

    extern CHAR_PTR     optarg;                 // For getopt()
    extern int          optind,opterr;          // For getopt()

    myIn = stdin;

    // Get environment variables, if any.

    cptr = getenv("ACRC_FILE");
    if(cptr != NULL) {
        acrcFile = cptr;
    }
    cptr = getenv("JBSTAT_FILE");
    if(cptr != NULL) {
        jbStatFile = cptr;
    }
    cptr = getenv("ARCTRACE");
    if(cptr != NULL) {
        trace = atoi(cptr);
    }

    // Parse command-line arguments.

#ifndef _ARCBROWSE_
    while((c = getopt(argc,argv,"a:s:r:v:f:o:Pg:e:l:c:C:It:w:WxR")) != EOF) {
#else
    while((c = getopt(argc,argv,"a:s:r:v:f:g:c:t:w:W")) != EOF) {
#endif  // _ARCBROWSE_
        switch(c) {
        case 'a':
            acrcFile = optarg;
            break;
        case 's':
            jbStatFile = optarg;
            break;
        case 'r':
            restoreDir = optarg;
            break;
        case 'R':
            doSynthCheck = TRUE;
            break;
        case 'v':
            vsname = optarg;
            break;
        case 'f':
            fsname = optarg;
            break;
#ifndef _ARCBROWSE_
        case 'o':
            outFile = optarg;
            break;
        case 'P':
            noPrint = 1;
            break;
        case 'I':
            noImport = 1;
            break;
#endif  // _ARCBROWSE_
        case 'g':
            vnptr = new RWCString(optarg);
            if(vnptr == NULL) {
                fprintf(stderr,"No free store, aborting!\n");
                return(2);
            }
            growVols.append(vnptr);
            break;
        case 'c':
            cmrFile = optarg;
            break;
        case 'C':
            genCmrFile = optarg;
            break;
        case 'e':
            freeFileName = optarg;
            break;
        case 'l':
            logFile = optarg;
            break;
        case 't':
            trace = atoi(optarg);
            break;
        case 'x':
            outputCpio = 1;
            noPrint = 1;
            break;
        case 'w':
            dbFile = optarg;
            break;
        case 'W':
            rmFile = 0;
            break;
        default:
            usage = 1;
            break;
        }
    }

    if(acrcFile.isNull() || jbStatFile.isNull() || vsname.isNull()) {
        usage = 1;
    } 

    // Make sure that the argument directory for restore, if specified,
    // exists and can be accessed.

    if(!restoreDir.isNull()) {
        if(access((CHAR_PTR) restoreDir,R_OK | W_OK | X_OK | F_OK)) {
            fprintf(stderr,"Restore directory %s not accessible!\n",
                    (CHAR_PTR) restoreDir);
            usage = 1;
        }
    }

    if(usage && outFile.isNull()) {
        Usage();
        return(TERM_ERROR);
    }


    // If we got here, we're ready to start testing.  If the importArchive
    // function succeeds, the root TreeNode will point to the 
    // root node of the imported directory structure, and the TreeDB
    // contain the tree structure and all file information.

    NSRJukeBox          jb(acrcFile);
    TreeDB              db(dbFile,rmFile);
    TreeNode            root;

    if(jb.error() != NSR::error_None) {
        fprintf(stderr,"Construction of NSRJukeBox failed, error is %s\n",
                (jb.error()).message());
        return(TERM_ERROR);
    }

    // If the -o option was specified, we're just cataloging, so do it now
    // and get out.

#ifndef _ARCBROWSE_
    if(!outFile.isNull()) {
        jb.genStatusFile(outFile);
        if(jb.error() != NSR::error_None) {
            fprintf(stderr,"genStatusFile() failed, error is %s\n",
                    (jb.error()).message());
        }
        else {
            printf("Generated status file \"%s\"\n",(CHAR_PTR) outFile);
        }
        return TERM_NOERROR_TEST_PASSED;
    }
#endif  // _ARCBROWSE_

    // Read the status file.

    jb.readStatusFile(jbStatFile);
    if(jb.error() != NSR::error_None) {
        fprintf(stderr,"Read of status file %s failed, error is %s\n",
                (CHAR_PTR) jbStatFile,(jb.error()).message());
        return(TERM_ERROR);
    }

    // Process quoted arguments
    jb.processIdentifier(vsname,IN);
    jb.processIdentifier(fsname,IN);

    // Set the working volume set.

    jb.setExistingWVS(vsname);
    if(jb.error() != NSR::error_None) {
        fprintf(stderr,"Unable to set WVS to volume set \"%s\", error is %s\n",
                (CHAR_PTR) vsname,(jb.error()).message());
        return(TERM_NOERROR_TEST_FAILED);
    }

    // Validate free space..
    if (freeFileName != NULL) {
        #define MAX_FREE_FILESIZE 10240
        // Get the check file into a buffer.
        FILE* freeFile = fopen ((char*)freeFileName, "r");
        char fileBuf[MAX_FREE_FILESIZE];
        fread ((char*) fileBuf, sizeof(char), MAX_FREE_FILESIZE, freeFile);

        // Now get the current free space.
        String localBuf;
        jb.writeAsText (localBuf);

        if (strcmp(fileBuf, (char*)localBuf) != 0) {
            fprintf(stderr,"Imported fileset did not compare with free space file %s.\n",
                    (char*)freeFileName);
            fprintf(stderr, "Freespace buffers, delimited by square brackets:");
            fprintf(stderr, "export: [%s]", (char*)fileBuf);
            fprintf(stderr, "import: [%s]", (char*)localBuf);
            freeSpaceRval = TERM_NOERROR_TEST_FAILED;
        }
    }

    // If we're growing, do it now.

    if(growVols.entries() > 0) {
        jb.beginGrow();
        if(jb.error() != NSR::error_None) {
            fprintf(stderr,"beginGrow() failed, error is %s\n",
                    (jb.error()).message());
            return TERM_NOERROR_TEST_FAILED;
        }
    }

    if(growVols.entries() > 0) {
        for(i = 0; i < growVols.entries(); i++) {
            jb.growWVS(*(growVols.at(i)),DATA_START_SECTOR);
            if(jb.error() != NSR::error_None) {
                fprintf(stderr,"Unable to grow volume set, error is %s\n",
                        (CHAR_PTR) (jb.error()).message());
                return(TERM_NOERROR_TEST_FAILED);
            }
        }
    }

    if(growVols.entries() > 0) {
        jb.completeGrow();
        if(jb.error() != NSR::error_None) {
            fprintf(stderr,"completeGrow() failed, error is %s\n",
                    (jb.error()).message());
            return TERM_NOERROR_TEST_FAILED;
        }
    }

    // Print out the volume names

    fprintf(stderr,"Reading volume set \"%s\"\n",(CHAR_PTR) vsname);
    VolumeFreeSpaceList vfsl;
    VolumeFreeSpace     *vfsptr;

    jb.wvsFreeSpace(vfsl);              // To get the names
    for(i = 0; i < vfsl.entries(); i++) {
        vfsptr = vfsl.at(i);
        fprintf(stderr,"\tVolume \"%s\"\n",(CHAR_PTR) vfsptr->volname_);
    }

    // If we're not doing the import, just quit now.

    if(noImport && !doSynthCheck) {
        return TERM_NOERROR_TEST_PASSED;
    }

    // Get implementation info

    ImplementationInfo  iinfo;

    jb.getImplementationInfo(fsname,iinfo);
    if(jb.error() != NSR::error_None) {
        fprintf(stderr,"WARNING, getImplementInfo() failed, error is %s\n",
                (jb.error()).message());
    }
    else {
        fprintf(stderr,"\n***** Implementation Info *****\n\n");
        fprintf(stderr,"Id:              %s\n",(CHAR_PTR) iinfo.Id);
        fprintf(stderr,"osClassId:       %u\n",iinfo.osClassId);
        fprintf(stderr,"osId:            %u\n",iinfo.osId);
        fprintf(stderr,"librVersionNo:   %u\n",iinfo.librVersionNo);
        fprintf(stderr,"implRegistrNo:   %u\n",iinfo.implRegistrNo);
        fprintf(stderr,"implVersionNo:   %u\n",iinfo.implVersionNo);
        fprintf(stderr,"implRevisionNo:  %u\n\n",iinfo.implRevisionNo);
    }

    // Do the import!

    VolsetNameMapList   vsnmList;
    VolsetNameMap *     vsnmptr;
    SidList             sidList;

    jb.importArchive(fsname,db,root);
#ifndef _ARCBROWSE_
    if(jb.error() == NSR::error_None) {
#ifdef ROGUEWAVIFIED
        db.setRoot(root);
#endif
        if(!restoreDir.isNull()) {
            vsnmList = jb.volsetNameMapList();
            vsnmptr = vsnmList.search(vsname);
            sidList = vsnmptr->sidList_;

            rval = restoreFiles(root,*(jb.acdev()),(CHAR_PTR) restoreDir,
                                sidList);
        }
        else if (doSynthCheck) {
            vsnmList = jb.volsetNameMapList();
            vsnmptr = vsnmList.search(vsname);
            sidList = vsnmptr->sidList_;
            setParent(root,root);

            rval = synthCheck (
                root,*(jb.acdev()),sidList);
            if (rval) {
                fprintf (stderr, "synth check did not compare\n");
                rval = TERM_NOERROR_TEST_FAILED;
            }            
        }
        else {
            if(!noPrint) {
                rval = printFiles("",root);  // Just print their names
            }
            else {
                if(outputCpio) {
                    cpioWriter  writer;
                    rval = writeCpio(root,writer,*(jb.acdev()),".");
                    writer.closeArchive();
                }
            }
        }
        if (!genCmrFile.isNull()) {
            CmrWriter   cmrWriter((CHAR_PTR) genCmrFile);
            cmrWriter.useShortReservedWords();
            cmrWriter.setTabSize(2);

            printf("Writing CMR file (on import) %s\n",(CHAR_PTR) genCmrFile);
            writeCmr(root,cmrWriter);
        }
        
        if(!cmrFile.isNull() && rval != TERM_NOERROR_TEST_FAILED) {
            printf("\n\n-----Validating CMR-----\n\n");
            gRootptr = &root;

            cmrDiff  = new CmrDiff ((RiderAccessFunc) treeAccess,
                                    (RiderRootAccessFunc) rootAccess,
                                    (RiderEndDirectory) endDirectory,
                                    (CHAR_PTR) cmrFile,
                                    (CHAR_PTR) "tree",
                                    (CHAR_PTR) logFile);
            CmrReader cmrReader (cmrFile,
                                 (StartDirFunc) startDirFunc,
                                 (EndDirFunc) endDirFunc,
                                 (FileFunc) fileFunc,
                                 (SymlinkFunc) symlinkFunc);
            cmrReader.cmrparse();

            if(cmrDiff->areEqual() != NSR::True) {
                fprintf(stderr,"Imported fileset did not compare with cmr!\n");
                rval = TERM_NOERROR_TEST_FAILED;
            }
        }
    }
    else {
        fprintf(stderr,"importArchive failed, error is %s\n",
                (jb.error()).message());
        rval = -1;
    }
#else
#ifdef ROGUEWAVIFIED
    db.setRoot(root);
#endif
    // setParent(root,root);
    browseShell shell(jb.acdev(),&jb,root);
    rval = shell.start();
#endif  // _ARCBROWSE_


    if(rmFile) {
        freeTree(root);
    }

    if(rval || freeSpaceRval) {
        if (rval != 0)
            fprintf (stderr, "(%s, %d) Originating arcRead failure value: %d\n",
                     __FILE__, __LINE__, rval);
        rval = TERM_NOERROR_TEST_FAILED;
    }
    else {
        rval = TERM_NOERROR_TEST_PASSED;
    }
    return rval;
}

//-----------------------------------------------------------------------------
// FUNCTION:  printFiles()
//
// Recursively prints file information to screen.
//-----------------------------------------------------------------------------

static int
printFiles( CHAR_PTR dir,TreeNode& root )
{
    int                 i;
    int                 rval;
    char                path[MAX_NAMELENGTH];
    TreeNode            tnode;
    Error               err;

    // Print out information for the current file.

    if((root.filename_)[0] == 0) {      // Parent entry, skip
        return 0;
    }
    if(root.address_.sector == ROOT_ICB_ADDRESS) {
        sprintf(path,"%s",(CHAR_PTR) root.filename_);
    }
   else {
        if(strcmp(dir,"/") == 0) {
            sprintf(path,"%s%s",dir,(CHAR_PTR) root.filename_);
        }
        else {
            sprintf(path,"%s/%s",dir,(CHAR_PTR) root.filename_);
        }
    }
    UINT32 filesize;

    if(root.attr_.fileType() == NSR::file_SymbolicLink) {
        char    buffer[MAX_NAMELENGTH];

        getSymlinkPath(buffer,root);
        filesize = strlen(buffer);
    }
    else {
        filesize = root.elist_.totalBytes();
    }
    printFile(path,root.attr_,filesize);

    // If we're a directory, recur on our children.  

    if(root.attr_.fileType() == NSR::file_Directory) {
        for(i = 0; i < root.getNumChildren(); i++) {
            err = root.getChild(i,tnode);
            if(err != NSR::error_None) {                // Parent entry
                continue;
            }
            rval = printFiles(path,tnode);
            if(rval) {
                return rval;
            }
        }
    }

    return 0;
}

//-----------------------------------------------------------------------------
// FUNCTION:  printFile()
//
// Prints a file's information to screen.  The format is as follows:
//
// Mode bits    Length  Pathname
//
// -rw-r--r--    53202  /nsr/ms/nsrArc.C
//-----------------------------------------------------------------------------

static void
printFile( CHAR_PTR path,Attributes& attr,UINT32 filesize )
{
    mode_t      mode;

    if(path == NULL) {
        printf("ERROR:  NULL path passed to printFile()\n");
        return;
    }

    // First print the mode bits.  

    mode = attr.unixMode();
    if(attr.fileType() == NSR::file_Directory) {
        printf("d");
    }
    else {
        if(attr.fileType() == NSR::file_SymbolicLink) {
            printf("l");
        }
        else {
            printf("-");
        }
    }
        
    // Print out the rwx bits.  Very kludgy, I'm sure there's a better way.

    if(mode & S_IRUSR) {
        printf("r");
    } 
    else {
        printf("-");
    }
    if(mode & S_IWUSR) {
        printf("w");
    } 
    else {
        printf("-");
    }
    switch((mode & S_IXUSR) | (mode & S_ISUID)) {
    case S_IXUSR | S_ISUID:
        printf("s");
        break;
    case S_IXUSR:
        printf("x");
        break;
    case S_ISUID:
        printf("S");
        break;
    default:
        printf("-");
        break;
    }
    if(mode & S_IRGRP) {
        printf("r");
    } 
    else {
        printf("-");
    }
    if(mode & S_IWGRP) {
        printf("w");
    } 
    else {
        printf("-");
    }
    switch((mode & S_IXGRP) | (mode & S_ISGID)) {
    case S_IXGRP | S_ISGID:
        printf("s");
        break;
    case S_IXGRP:
        printf("x");
        break;
    case S_ISGID:
        printf("S");
        break;
    default:
        printf("-");
        break;
    }
    if(mode & S_IROTH) {
        printf("r");
    } 
    else {
        printf("-");
    }
    if(mode & S_IWOTH) {
        printf("w");
    } 
    else {
        printf("-");
    }
    if(mode & S_IXOTH) {
        printf("x");
    } 
    else {
        printf("-");
    }
    
    // Now print out the length of the file, in bytes

    printf("  %8d  ",filesize);

    // Print out the modify time, in some reasonable format

    time_t mtime = (time_t) attr.modificationTime();

    struct tm *tm = localtime(&mtime);
    char date[64];
    if(tm->tm_year < 95) {                      // Horrible. . .
        strcpy(date,nl_asctime(tm,"%b %d %Y",0));
    }
    else {
        strcpy(date,nl_asctime(tm,"%b %d %H:%M",0));
    }
    // Remove trailing \n
    date[strlen(date) - 1] = 0;
    printf("%s ",date);

    // Print out the pathname.

    printf("%s\n",path);
}

#ifndef _ARCBROWSE_

int
synthCheck(TreeNode& node,hpACDev& acdev,SidList& sidList)
{
    int         i;
    int         rval;

    Sid_t       sid;
    for(i = 0; i < sidList.entries(); i++) {
        sid = *(sidList.at(i));
        rval = synthSidCheck(node,acdev,sid);
        if(rval) {
            return rval;
        }
    }
    return 0;    
}

static int      
synthSidCheck(TreeNode& node,hpACDev& acdev,Sid_t sid)
{
    int                 i;
    int                 offset;
    int                 rval;
    TreeNode            tnode;
    Error               err;
    
    if(node.attr_.fileType() == NSR::file_SymbolicLink) {
        return 0;
    }

    // The algorithm for restoring data is as follows:  Go through the
    // file's extent list.  For each extent that is on the current sid,
    // read the data from optical and write it to the file system.

    if((node.attr_.fileType() != NSR::file_Directory)) {  // Regular file.

        offset = 0;
        for(i = 0; i < node.elist_.getCount(); i++) {
            Extent& ext = (node.elist_)[i];
            if(ext.start.sid != sid) {
                offset += ext.len;
                continue;
            }

            // Have to restore data.
            rval = synthExtentCheck(node.filename_, acdev, sid, ext, offset);
            if(rval) {
                printf("SynthCheck data compare of %s failed.\n", (char*)node.filename_);
                return rval;
            }
            offset += ext.len;
        }
        return 0;
    }

    // If we got here, the current file is a directory.  Recur on its
    // children.
    for(i = 0; i < node.getNumChildren(); i++) {
        err = node.getChild(i,tnode);
        if(err != NSR::error_None) {
            continue;
        }
        rval = synthSidCheck(tnode,acdev,sid);
        if(rval) {
            return rval;
        }
    }
    return 0;
} 

static int
synthExtentCheck(char* seed, hpACDev& acdev, Sid_t sid, Extent& ext, int offset)
{
    int         residue;
    int         len;
    int         rval;
    UINT32      sectorSize;
    UINT32      sector_offset;
    static BYTE buf[MAX_IOSIZE];        // Don't reallocate constantly
    int         testFailed = 0;

    sectorSize = acdev.sectorSize(sid);
    if(sectorSize == 0) {
        printf("copyExtent():  bad sectorSize returned!\n");
        return -1;
    }

    memset(buf, '\0', MAX_IOSIZE);
    residue = ext.len;
    if(residue == 0) {
        printf("copyExtent: hmm, we have an extent of 0 length!\n");
    }
    sector_offset = ext.start.sector;
    while(residue > 0) {
        len = ((residue < MAX_IOSIZE) ? residue : MAX_IOSIZE);
        rval = acdev.read(sid,sector_offset,len,buf);
        if(rval < len) {        // Got a short count
            fprintf(stderr,"restoreData() WARNING:  got short count reading \
file data, %d should be %d bytes\n",rval,len);
            fprintf(stderr,"acdev.error is %s\n",(acdev.error()).message());
            testFailed = 1;
        }
        sector_offset += (len / sectorSize);

        // Now write the data to the file, seeking if necessary.
        SynthData checkData(seed, offset, len);
        if (strcmp((char*)checkData.data(), (char*)buf)) {
            fprintf (stderr, "Node %s data does not compare.\n", seed);
            fprintf (stderr, "extent: sid- %d, sector- %d, len- %d\n",
                     ext.start.sid, ext.start.sector, ext.len);
            fprintf (stderr, "surface:\n%s\n\nPattern:\n%s\n\n",
                     (char*)buf, (char*)checkData.data());
            fprintf (stderr, "size of buf: %d: %d\n", strlen((char*) buf), sizeof(buf));
            testFailed = 1;
        }
        
        residue -= len;
        offset += len;
    }
    return testFailed;    
}

#endif  // _ARCBROWSE_

/*
int
nodeToPath(TreeNode& node, RWCString& path)
*/

//-----------------------------------------------------------------------------
// FUNCTION:  restoreFiles()
//
// Restores files to the specified directory.
//
// ALGORITHM:
//
// The primary concern here is to avoid unnecessary swapping.  We use a 
// multi-pass algorithm, traversing the directory tree several times, in
// order to reduce swapping.  First we traverse the tree and recreate the
// directory structure, creating each file with 0 length.
// Then we make as many passes as there are surfaces.  For each pass, we
// restore the data for each file that is on the current surface.  
// Finally, we make one final post-order traversal of the entire tree to
// restore the attributes to the files.  
//
// ARGUMENTS:
//
// TreeNode        root       IN  Root of the n-ary tree
// hpACDev&        acdev      IN  Autochanger device object
// char           *dir        IN  Directory to restore files into
// SidList&        sidList    IN  List of sid's in volume set
//
// RETURNS:
//
// Returns 0 on success, -1 on failure.
//
// PRECONDITIONS:
//
// --The n-ary tree specified by root has been completely constructed
// --The specified directory exists and is accessible (has appropriate
//   permissions)
//
// POSTCONDITIONS:
//
// --The directory structure on the imported NSR volumes will be restored
//   to the specified directory.  All files, directories, file and directory
//   attributes, and file data will be restored.
//-----------------------------------------------------------------------------

static int
restoreFiles(TreeNode& root,hpACDev& acdev,CHAR_PTR  dir,SidList& sidList)
{
    int         i;
    int         rval;

    // First pass:  Restore the directory structure to the argument directory,
    // creating zero-length files.

    rval = recreateDirTree(root,dir);
    if(rval) {
        return rval;
    }

    // Now make as many passes as there are surfaces, restoring data to 
    // files for each surface.
  
    Sid_t       sid;

    for(i = 0; i < sidList.entries(); i++) {
        sid = *(sidList.at(i));
        printf("Restoring file data from Sid_t %d:",sid);
        rval = restoreData(root,acdev,dir,sid);
        printf("\n");
        if(rval) {
            return rval;
        }
    }

    // Now restore attributes.  This is done last because the chown() system
    // call may render a file or directory henceforth unaccessible.

    printf("Restoring file attributes. . .\n");
    rval = restoreAttributes(root,dir);
    if(rval) {
        return rval;
    }

    // We're done.  Easy enough.

    return 0;
}

//-----------------------------------------------------------------------------
// FUNCTION:  recreateDirTree()
//
// Recreates a directory tree by creating zero-length files and subdirectories
// as needed.  Restores file attributes.
//
// This function is recursive.
//
// ARGUMENTS:
//
// TreeNode&       root       IN  The node of a directory to be restored.
// char           *dir        IN  Directory in which to restore the directory.
//
// PRECONDITIONS:  
//
// --The argument TreeNode corresponds to a directory.
// --Dir corresponds to the directory pathname in which to restore the 
//   directory given by the argument nsrImportTree.
//
// POSTCONDITIONS:
// --All files and subdirectories in the given directory, as represented by
//   the argument nsrImportTree, will be restored.  Files and directories will
//   minimally have user RWX permissions, but attributes will not otherwise
//   be restored.
// --All restored files will have size 0.
// --Symbolic links will be restored.
//-----------------------------------------------------------------------------

static int
recreateDirTree(TreeNode& root,CHAR_PTR dir) 
{
    int                 i;
    int                 rval;
    TreeNode            node;
    Address             addr;
    char                path[MAX_NAMELENGTH];
    char                subdir[MAX_NAMELENGTH];
    Error               err;

    // We make entries for all the children, recurring on children directories 
    // as necessary.

    for(i = 0; i < root.getNumChildren(); i++) {
        err = root.getChild(i,node);
        if(err != NSR::error_None) {
            continue;
        }
        if(node.address_.sector == ROOT_ICB_ADDRESS) {
            continue;
        }
        sprintf(subdir,"%s/%s",dir,(CHAR_PTR) node.filename_);
        if(node.attr_.fileType() == NSR::file_Directory) {
            if(strlen((CHAR_PTR) node.filename_) == 0) {        // Parent entry!
                continue;
            }

            printf("Restoring directory %s\n",subdir);
            if(mkdir(subdir,node.attr_.unixMode() | S_IRWXU)) {
                fprintf(stderr,"RESTORE FAILED:  could not make directory %s\n",
                        subdir);
                perror("mkdir");
                return -1;
            }
            rval = recreateDirTree(node,subdir);
            if(rval) {
                return rval;
            }
        }
        else {                          // Normal file; restore it
            printf("Restoring file %s,length = %d\n",subdir,node.elist_.totalBytes());
            rval = recreateFile(node,subdir);
            if(rval) {
                return rval;
            }
        }
    }

    // If we got here, all went well.  Success!

    return 0;
}

//-----------------------------------------------------------------------------
// FUNCTION:  recreateFile()
//
// Creates a file and/or a symbolic link.
//
// ARGUMENTS:
//
// TreeNode&       node      IN  Reference to treeNode object describing the 
//                               file to restore.
// char           *filepath  IN  Pathname of file to restore.
//
// RETURNS:
//
// Returns 0 on success, -1 on failure.
//
// PRECONDITIONS:
//
// --The file specified by node is a file or symbolic link (not a directory).
//
// POSTCONDITIONS:
//
// --The specified file will be created with 0 length and minimally user
//   RWX permissions.  Other attributes will not be restored.  If a
//   symbolic link, the link will be created.
//-----------------------------------------------------------------------------

static int
recreateFile(TreeNode& node,CHAR_PTR filepath)
{
    FILE        *fp;
    char        path[MAX_NAMELENGTH];

    // We have to handle symbolic links with a different system call

    if(node.attr_.fileType() == NSR::file_SymbolicLink) {  // Symbolic link!
        if(getSymlinkPath(path,node)) {
            return -1;
        }
        if(symlink(path,filepath)) {
            fprintf(stderr,"RESTORE FAILED:  could not restore symbolic \
link %s\n",filepath);
            perror("symlink");
            return -1;
        }
        return 0;
    }

    // It's a normal file; create it.

    if((fp = fopen(filepath,"w")) == NULL) { 
        fprintf(stderr,"RESTORE FAILED:  could not create file %s\n",
                filepath);
        perror("fopen");
        return -1;
    }
    fclose(fp);
    return 0;
}

//-----------------------------------------------------------------------------
// FUNCTION:  getSymlinkPath()
//
// Gets the Unix representation of the symbolic link path from the file
// information.  Parses the PathElementList.
//
// ARGUMENTS:
//
// char           *path   OUT  Buffer to hold the pathname representation of 
//                             the symbolic link.
// TreeNode&       node   IN   File information for the symbolic link.
//
// RETURNS:
//
// Returns 0 on success, -1 on failure.
//
// PRECONDTIONS:
//
// --Argument node object corresponds to a symbolic link with valid
//   PathElementList information.
//
// POSTCONDTIONS:
//
// --The Unix representation of the symbolic link (a pathname) will be
//   written into the argument path variable.
//-----------------------------------------------------------------------------

static int
getSymlinkPath( CHAR_PTR path,TreeNode& node )
{
    int                 i;
    int                 len;
    PathElementList     *plistp;

    // Get the PathElementList for the symbolic link

    plistp = &(node.plist_);
    if(plistp == NULL) {
        fprintf(stderr,"getSymlinkPath():  no PathElementList!\n");
        return -1;
    }

    // Now we have to parse through the PathElementList and construct a
    // standard Unix path from it.

    path[0] = 0;
    for(i = 0; i < plistp->getCount(); i++) {
        PathElement& pelt = (*plistp)[i];

        len = strlen(pelt.name);
        switch(pelt.type) {
        case NSR::path_Custom_Root:             // Root if nonzero length
            if(len == 0) {      // Reset
                path[0] = 0;
                strcat(path,"/");
            } 
            else {              // This is an unusual case!
                fprintf(stderr,"WARNING:  Symbolic link path component of \
type 1 with nonzero length!\n");
                path[0] = 0;
                strcat(path,"/");
            }
            break;
        case NSR::path_Standard_Root:           // Root
            path[0] = 0;
            strcat(path,"/");
            break;
        case NSR::path_Parent:                  // ..
            strcat(path,"../");
            break;
        case NSR::path_Same:                    // .
            strcat(path,"./");
            break;
        case NSR::path_Name:                    // Just a path element
            strcat(path,pelt.name);
            strcat(path,"/");
            break;
        default:
            fprintf(stderr,"WARNING:  Symbolic link contains unknown path \
component type!\n");
            strcat(path,"unknown_type/");
            break;
        }
    }
    len = strlen(path);
    if(path[len - 1] == '/') {          // Remove trailing /'s.
        path[len - 1] = 0;
    }
    return 0;
}

//-----------------------------------------------------------------------------
// FUNCTION:  restoreAttributes()
//
// Restores the attributes to the given file.
//
// We have to restore all of the following:
//
// User id
// Group id
// Mode bits
// Access Time
// Modify Time
//
// Note that we change the user id and group id last, since doing so may
// make it impossible for the current process to perform the other tasks.
//
// Symbolic links do not have any attributes to restore (besides mode bits,
// which are restored by the symlink() call).
//
// This function is recursive.  We use a postorder-traversal type of
// recursion:  we must modify the attributes of all children of the
// current directory first, because the permissions of the current directory
// may not allow for modification of children entries, and to preserve
// modify and access times of the current directory.
//
// ARGUMENTS:
//
// TreeNode&       root       IN  node representing a directory
// char           *dir        IN  directory path of current directory
//
// RETURNS:
//
// Returns 0 on success, -1 on failure.
//
// PRECONDITIONS:
//
// --The current nsrTreeNode is a directory
// --The directory path argument corresponds to the path of the directory
//   given by the nsrTreeNode argument.
//
// POSTCONDITIONS:
// 
// --The attributes of the children of the current directory will be 
//   restored.
// --The attributes of the current directory will be restored.
//-----------------------------------------------------------------------------

static int
restoreAttributes( TreeNode& root,CHAR_PTR dir )
{
    int                 i;
    int                 rval;
    struct utimbuf      tbuf;
    char                subdir[MAX_NAMELENGTH];
    TreeNode            tnode;
    Error               err;

    if(root.attr_.fileType() == NSR::file_SymbolicLink) {
        return 0;
    }

    // We now restore the attributes of the children of the current directory,
    // if any, recursively.

    for(i = 0; i < root.getNumChildren(); i++) {
        err = root.getChild(i,tnode);
        if(err != NSR::error_None) {            // Skip parent entries
            continue;
        }
        if(root.attr_.fileType() == NSR::file_Directory) {
            sprintf(subdir,"%s/%s",dir,(CHAR_PTR) root.filename_);
        }
        else {
            sprintf(subdir,"%s",dir);
        }
        rval = restoreAttributes(tnode,subdir);
        if(rval) {
            return rval;
        }
    }

    if(root.address_.sector == ROOT_ICB_ADDRESS) {
        return 0;
    }
    else {
        sprintf(subdir,"%s/%s",dir,(CHAR_PTR) root.filename_);
    }

    // chmod():  Restore permission and set-uid/set-gid bits

    if(chmod(subdir,root.attr_.unixMode())) {
        fprintf(stderr,"RESTORE FAILED, could not chmod file %s\n",
                subdir);
        perror("chmod");
        return -1;
    }

    // utime():  Restore access and modify times

    tbuf.actime = root.attr_.accessTime();
    tbuf.modtime = root.attr_.modificationTime();
    if(utime(subdir,&tbuf)) {
        fprintf(stderr,"RESTORE FAILED, could not utime file %s\n",
                subdir);
        perror("utime");
        return -1;
    }

    // chown():  Restore uid and gid

    if(chown(subdir,root.attr_.uid(),root.attr_.gid())) {
        fprintf(stderr,"RESTORE FAILED, could not chown file %s\n",
                subdir);
        perror("chown");
        return -1;
    }
    return 0;
}

//-----------------------------------------------------------------------------
// FUNCTION:  restoreData()
//
// Restores data from the specified surface to all files in the tree that
// have extents on the surface.
//
// This function is tail-recursive (make iterative?).
//
// In the general case, a file's extent list can contain extents from any
// and all surfaces in the NSR volume.  To minimize swapping, we restore all
// data from each surface to whatever files the data belongs to.  To do
// so, we make multiple passes through the directory tree.  Note that this
// may cause, for instance, the second half of a file's data to be written
// into the file before the first half.  The open(2), lseek(2), and write(2)
// system calls are used to perform this random-access file writing.
//
// ARGUMENTS:
//
// TreeNode&        node       IN  Node representing a file or directory.
// hpACDev&         acdev      IN  Autochanger device object.
// char            *dir        IN  Directory pathname of current file.
// Sid_t            sid        IN  Sid (surface) to recover data from.
//
// RETURNS:
//
// Returns 0 on success, -1 on failure.
//
// PRECONDITIONS:
//
// --Argument dir pathname corresponds to the pathname of the current file
//   described by the argument nsrImportTree *.
//
// POSTCONDITIONS:
//
// --All data in files in the given tree which is on the given Sid_t will
//   be restored.
//-----------------------------------------------------------------------------
 
static int      
restoreData(TreeNode& node,hpACDev& acdev,CHAR_PTR dir,Sid_t sid)
{
    int                 i;
    int                 fd;
    int                 offset;
    int                 rval;
    char                subdir[MAX_NAMELENGTH];
    TreeNode            tnode;
    Error               err;
    
    if(node.attr_.fileType() == NSR::file_SymbolicLink) {
        return 0;
    }

    // The algorithm for restoring data is as follows:  Go through the
    // file's extent list.  For each extent that is on the current sid,
    // read the data from optical and write it to the file system.

    if((node.attr_.fileType() != NSR::file_Directory)) {  // Regular file.
        sprintf(subdir,"%s/%s",dir,(CHAR_PTR) node.filename_);
        fd = open(subdir,O_WRONLY);
        if(fd < 0) {
            fprintf(stderr,"restoreData():  unable to open file %s\n",subdir);
            perror("open");
        }
        offset = 0;
//      printf("file %s, node.elist_.totalBytes() = %d\n",subdir,
//             node.elist_.totalBytes());
        for(i = 0; i < node.elist_.getCount(); i++) {
            Extent& ext = (node.elist_)[i];
            if(ext.start.sid != sid) {
                offset += ext.len;
                continue;
            }

            // Have to restore data.  

            rval = copyExtent(fd,acdev,sid,offset,ext);
            if(rval) {
                printf("copyExtent() failed, aborting!\n");
                close(fd);
                return rval;
            }
            offset += ext.len;
        }
        close(fd);
        return 0;
    }

    // If we got here, the current file is a directory.  Recur on its
    // children.

    printf(".");
    fflush(stdout);
    for(i = 0; i < node.getNumChildren(); i++) {
        err = node.getChild(i,tnode);
        if(err != NSR::error_None) {
            continue;
        }
        if(tnode.attr_.fileType() == NSR::file_Directory) {
            sprintf(subdir,"%s/%s",dir,(CHAR_PTR) tnode.filename_);
        }
        else {
            sprintf(subdir,"%s",dir);
        }
        rval = restoreData(tnode,acdev,subdir,sid);
        if(rval) {
            return rval;
        }
    }
    return 0;
} 

//-----------------------------------------------------------------------------
// FUNCTION:  copyExtent()
//
// Reads an extent and copies it to the appropriate offset in the file.
//-----------------------------------------------------------------------------

static int
copyExtent(int fd,hpACDev& acdev,Sid_t sid,int offset,Extent& ext)
{
    int         residue;
    int         len;
    int         rval;
    UINT32      sectorSize;
    UINT32      sector_offset;
    static BYTE buf[MAX_IOSIZE];        // Don't reallocate constantly

    sectorSize = acdev.sectorSize(sid);
    if(sectorSize == 0) {
        printf("copyExtent():  bad sectorSize returned!\n");
        return -1;
    }
    
    residue = ext.len;
    if(residue == 0) {
        printf("copyExtent: hmm, we have an extent of 0 length!\n");
    }
    sector_offset = ext.start.sector;
    while(residue > 0) {
        len = ((residue < MAX_IOSIZE) ? residue : MAX_IOSIZE);
        rval = acdev.read(sid,sector_offset,len,buf);
        if(rval < len) {        // Got a short count
            fprintf(stderr,"restoreData() WARNING:  got short count reading \
file data, %d should be %d bytes\n",rval,len);
            fprintf(stderr,"acdev.error is %s\n",(acdev.error()).message());
        }
        sector_offset += (len / sectorSize);

        // Now write the data to the file, seeking if necessary.

        if(lseek(fd,offset,SEEK_SET) < 0) {
            fprintf(stderr,"restoreData():  lseek() failed on file!\n");
            perror("lseek");
            return -1;
        }

        rval = write(fd,buf,len);
        if(rval < len) {
            fprintf(stderr,"restoreData() WARNING:  got short count \
writing file, %d should be %d bytes\n",rval,len);
            perror("write");
        }
        residue -= len;
        offset += len;
    }
    return 0;
}
    
//-----------------------------------------------------------------------------
// FUNCTION:  Usage
//
// Prints usage information to screen.
//-----------------------------------------------------------------------------

static void
Usage( void )
{
    static CHAR_PTR usage_str[] = {
"Usage:\n",
#ifndef _ARCBROWSE_
"arcRead [-a acrcFile] [-s jbstatFile] [-r directory ] [-f fsname] ",
"        [-c cmrFile] [-l logFile] [-I] [-o outfile] [-P]",
#else
"arcBrowse [-a acrcFile] [-s jbstatFile] [-r directory ] [-f fsname] ",
#endif  // _ARCBROWSE_
"        [-g volname] [-t val] [-w swapFile] [-W] -v vsname",
"",
"[-a acrcFile]:    Optional.  Pathname of the acrc configuration file.",
"[-s jbStatFile]:  Optional.  Pathname of the jukebox status file.",
"[-r directory]:   Optional.  If this option is given, the test will",
"                  restore the files to the specified directory.  Note",
"                  that you must ensure that the target directory has",
"                  sufficient free space.",
"[-R]:             Optional.  Specifies that the archive was created with",
"                  a synth file.  Compare data to the synth pattern, without",
"                  doing a restore.",
"[-f fsname]:      Optional.  The file set name to import, quoted if ",
"                  necessary.  If not specified, the most recent fileset is",
"                  imported from the given volume set.",
"[-g volname]:     Optional.  Grow the volume set by one volume.",
#ifndef _ARCBROWSE_
"[-c cmrFile]:     Optional.  Used the specified cmrfile to compare metadata",
"                  from the imported fileset.",
"[-C cmrFile]:     Optional.  Create a NEW CMR file, separate of the comparision",
"                  CMR file (opt -c).",
"[-l logFile]:     Optional.  Specify the name of the difference file output",
"                  by the CMR validation.  Default is arcRead.out",
"[-I]:             Optional.  Do not do import",
"[-o outFile]:     Optional.  Overrides all other options.  Inventories",
"                  the jukebox and generates a status file of the specified",
"                  name.",
"[-P]:             Optional.  Specifies that a list of the imported files is",
"                  not to be printed to screen following the import.",
#endif  // _ARCBROWSE_
"[-t val]:         Optional.  Set value of the Archive object algorithm trace var.",
"[-w swapFile]:    Optional.  Specifies the pathname of the swapfile.",
"                  This option is only useful if the paging tree option",
"                  has been compiled in (ROGUEWAVIFIED).  The swapfile,",
"                  and a swapfile named swapfile.md, will be created to",
"                  store metadata information.  Default:  /tmp/arcRead.swap",
"[-W]:             Optional.  If specified, the swap file and the",
"                  metadata file will NOT be removed at the completion",
"                  of the program.",
"-v vsname:        Mandatory.  The volume set name, quoted if necessary.",
"",
"Example:  ",
"",
#ifndef _ARCBROWSE_
"arcRead -a .acrc -s /tmp/jbstat.1 -v \"Volset 1\" -f \"Fileset \"",
#else
"arcBrowse -a .acrc -s /tmp/jbstat.1 -v \"Volset 1\" -f \"Fileset \"",
#endif  // _ARCBROWSE_
"",
"ENVIRONMENT VARIABLES:",
"",
"The acrc file and the jukebox status file can both be specified by ",
"environment variables:",
"",
"ACRC_FILE:    Environment variable containing the acrc file pathname.",
"JBSTAT_FILE:  Environment variable containing the jukebox status file ",
"              pathname.",
"ARCTRACE:     Environment variable containing the value of algorithmic",
"              trace variable.",
"",
"Note that the -a and -s options will override any value in the corresponding",
"environment variable, and the -a and -s options are mandatory if the",
"environment variables are not set.",
NULL
};
    int i = 0;

    while(usage_str[i] != NULL) {
        printf("%s\n",usage_str[i]);
        i++;
    }
}

void *startDirFunc(void *parent, char *name, Attributes &attr)
{
    cmrDiff->beginDir(parent, name, attr);
}
void endDirFunc(void *dir)
{
    cmrDiff->endDir(dir);
}
void fileFunc(char *name, Attributes& attr, ExtentList& elist)
{
    cmrDiff->file(name, attr, elist);
}
void symlinkFunc(char *name, Attributes& attr, PathElementList& plist)
{
    cmrDiff->symlink(name, attr, plist);
}

//-----------------------------------------------------------------------------
// rootAccess           Provide root directory information to the CmrDiff
//                      object.
// usage
//   This is one of two functions required for use with the CmrDiff class.
//   The function declaration must agree with the "RiderRootAccessFunc"
//   function typedef in "cmrDiff.h".
//
// parameters
//   rootName     OUT   Pointer to root directory name.
//   attr         OUT   Reference to CmrDiff root attributes structure.
//   rootPointer  OUT   Pointer to the GLOBAL root treenode structure.
//-----------------------------------------------------------------------------
void rootAccess (RWCString &rootName, Attributes &attr, void **rootPointer)
{
    *rootPointer = gRootptr;
    rootName = gRootptr->filename_;
    attr = gRootptr->attr_;
}

void endDirectory(void *dir)
{

    /*if (dir != (void*)gRootptr)
        if ((dir < (void*)0x40000000) ||
            (dir > (void*)0x50000000)) {
            printf ("pointer (%x) out of range at %d, %s\n", (int) dir,
                    __LINE__, __FILE__);
        }
    
    if ( (dir != (void*)gRootptr) &&
         (dir != NULL)) {
        TreeNode *node = (TreeNode*) dir;
        delete node;
    }
    */
}

INT32
PathToNode(CHAR_PTR path,TreeNode& node)
{
    TreeNode    cur,child;
    CHAR_PTR    cptr;
    char        curelt[256];
    INT32       index;

    cptr = path;
    cur = *gRootptr ;

    // Start consuming path elements.

    while(1) {
        index = 0;
        while((!isspace(*cptr)) && (*cptr != 0) && (*cptr != '/')) {
            curelt[index] = *cptr;
            cptr++;
            index++;
        }
        if(index == 0) {
            break;
        }
        curelt[index] = 0;              // null-terminate

        // We should now have a path element in curelt.

        if(cur.search(curelt,child) != NSR::error_None) {
            if(strcmp(curelt,".") == 0) {
                child = cur;
            }
            else {
                if(strcmp(curelt,"..") == 0) {
                    cur.getParent(child);
                }
                else {
                    return -1;
                }
            }
        }

        cur = child;
        if((*cptr == 0) || isspace(*cptr)) {            // We're done.
            break;
        }

        if(child.attr_.fileType() != NSR::file_Directory) {
            printf("Path component %s is not a directory!\n",curelt);
            return -1;
        }

        // We found a directory node, good.
        while(*cptr == '/') cptr++;             // Remove trailing /'s
    }

    node = cur;
    return 0;
}

//-----------------------------------------------------------------------------
// treeAccess           Access target directory per a pointer to the parent
//                      treenode, populate a list with child entries of the
//                      target directory, and return a pointer to the target
//                      directory tree node.
//
// usage
//   This is one of two functions required for use with the CmrDiff class.
//   The ValNode, FileNode and SymlinkNode classes defined in "cmrDiff.h"
//   are the required container for target entry child nodes.  The
//   function declaration must also agree with the "RiderAccessFunc"
//   typedef in "cmrDiff.h".
//
// parameters
//   dirName        OUT Name of the target directory.
//   dirEntries     OUT An ordered list of target directory child entries.
//   voidTargetDir  OUT Return pointer of target directory tree node.
//-----------------------------------------------------------------------------
void treeAccess (char *dirName,
                 RWBinaryTree *dirEntries,
                 void **voidTargetDir)
{
    *voidTargetDir = NULL;
    TreeNode targetDir;
    if (PathToNode(dirName, targetDir) != 0)
        return;

    // Iterate through each child entry of the provided parent
    // directory.
    TreeNode child;
    for(int i = 0; i < targetDir.getNumChildren(); i++)
    {
        Error err = targetDir.getChild(i, child);
        if(err != NSR::error_None) {            // Parent entry
            continue;
        }

        if (child.attr_.fileType() == NSR::file_Directory)
        {
            ValNode *node = new ValNode(nodetype_dir, child.filename_, child.attr_);
            dirEntries->insert (node);
        }
        else if (child.attr_.fileType() == NSR::file_SymbolicLink)
        {
            ValNode *node = new ValNode(nodetype_dir, child.filename_, child.attr_);
            *node << child.plist_;
            dirEntries->insert (node);
        }
        else if (child.attr_.fileType() == NSR::file_Raw)
        {
            ValNode *node = new ValNode(nodetype_dir, child.filename_, child.attr_);
            *node << child.elist_;
            dirEntries->insert (node);
        }
    }
}
    

//-----------------------------------------------------------------------------
// FUNCTION:  freeTree()
//
// Frees memory associated with a tree.
//-----------------------------------------------------------------------------

static void
freeTree( TreeNode& node )
{
    TreeNode    child;
    INT32       i;

    for(i = 0; i < node.getNumChildren(); i++) {
        node.getChild(i,child);
        freeTree(child);
    }
    (node.getDB())->freeNode(node);
}

//#ifdef _ARCBROWSE_
#include "bsh.cpp"
//#endif        // _ARCBROWSE_

#ifndef _ARCBROWSE_             


//-----------------------------------------------------------------------------
// writeCmr     Recursively write CMR from fileset information in tree.
//
// preconditions:
//      The tree is populated with a fileset.
//
// postconditions:
//      The output file has been populated with CMR refecting fileset info in
//      the tree, or an error has terminated processing.
//
// This code was stolen shamelessly from ../msVerify/udftocmr.C.
//-----------------------------------------------------------------------------

static int writeCmr ( TreeNode& root, CmrWriter& cmrWriter )
{
    int                 i;
    int                 rval;
    TreeNode            tnode;
    Error               err;

    // Tell the CmrWriter of this directory, file or symlink.
    switch (root.attr_.fileType()) {

        // currently supported file types.
    case NSR::file_Directory:
        cmrWriter.beginDir(root.filename_, root.attr_);

        // recurr on children
        for(i = 0; i < root.getNumChildren(); i++) {

            err = root.getChild(i,tnode);
            if(err != NSR::error_None) {                // Parent entry
                continue;
            }
            rval = writeCmr(tnode, cmrWriter);
            if(rval) {
                return rval;
            }
        }

        // Close the current directory.
        cmrWriter.endDir();
        break;
    case NSR::file_Raw:
        cmrWriter.file (root.filename_, root.attr_, root.elist_);
        break;
    case NSR::file_SymbolicLink:
        cmrWriter.symlink(root.filename_, root.attr_, root.plist_);
        break;

        // currently unsupported file types.
    case NSR::file_SpecialBlockDevice:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_SpecialBlockDevice);
        break;
    case NSR::file_SpecialCharacterDevice:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_SpecialCharacterDevice);
        break;
    case NSR::file_ExtendedAttributes:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_ExtendedAttributes);
        break;
    case NSR::file_Fifo:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_Fifo);
        break;
    case NSR::file_Sock:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_Sock);
        break;
    case NSR::file_Unspecified:
    default:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_Unspecified);
        break;
    }

    return 0;
}

static int 
writeCpio(TreeNode& node,cpioWriter& writer,hpACDev& acdev,char *parentName)
{
    INT32               i;
    INT32               count,rval,len;
    INT32               offset,sector_offset,residue;
    UINT32              sectorSize;
    cpioHeader          hdr;
    TreeNode            child;
    char                path[MAX_NAMELENGTH];
    Extent *            extp;
    unsigned char       buf[MAX_IOSIZE];

    // Write the current node first, then recur on the children if any.

    nodeToCpio(node,hdr,parentName,path);

    // Write the header out

    if(writer.writeHeader(hdr)) {
        fprintf(stderr,"writeCpio:  fatal error writing cpio header!\n");
        return -1;
    }

    // If the file was a symbolic link, write the path as data.

    if(node.attr_.fileType() == NSR::file_SymbolicLink) {
        assert((strlen(path) + 1) == hdr.fileSize());
        if(writer.writeData(path,hdr.fileSize()) != hdr.fileSize()) {
            fprintf(stderr,"writeCpio:  fatal error in symlink!\n");
            return -1;
        }
        assert(writer.dataExpected() == 0);
    }

    // If the file was a regular file, write its data.

    if(node.attr_.fileType() == NSR::file_Raw) {
        for(i = 0; i < node.elist_.entries(); i++) {
            extp = node.elist_.at(i);
            residue = extp->len;
            sector_offset = extp->start.sector;
            sectorSize = acdev.sectorSize(extp->start.sid);
            while(residue > 0) {
                len = ((residue < MAX_IOSIZE) ? residue : MAX_IOSIZE);
                rval = acdev.read(extp->start.sid,sector_offset,len,buf);
                if(rval < len) {        // Got a short count
                    fprintf(stderr,"cpioWrite() WARNING:  got short count \
reading file data, %d should be %d bytes\n",rval,len);
                    fprintf(stderr,"acdev.error is %s\n",
                            (acdev.error()).message());
                }
                sector_offset += (len / sectorSize);

                // Now write the data to the cpio archive

                rval = writer.writeData((char *) buf,len);
                if(rval < len) {
                    fprintf(stderr,"cpioWrite() WARNING:  got short count \
writing file, %d should be %d bytes\n",rval,len);
                }
                residue -= len;
                offset += len;
            }
        }
        assert(writer.dataExpected() == 0);
    }

    // Now recur on any children.

    for(i = 0; i < node.getNumChildren(); i++) {
        node.getChild(i,child);
        if(strcmp((CHAR_PTR) node.filename_,"/") == 0) {
            sprintf(path,".");
        }
        else {
            sprintf(path,"%s/%s",parentName,(CHAR_PTR) node.filename_);
        }
        if(writeCpio(child,writer,acdev,path)) {
            return -1;
        }
    }
    return 0;
}

static void 
nodeToCpio(TreeNode& node,cpioHeader& hdr,char *parentName,char *path)
{
    static int  count = 0;

    if(strcmp((CHAR_PTR) node.filename_,"/") == 0) {
        sprintf(hdr.c_filename,"%s",parentName);
    }
    else {
        sprintf(hdr.c_filename,"%s/%s",parentName,(CHAR_PTR) node.filename_);
    }
    hdr.c_namesize = strlen(hdr.c_filename) + 1;
    hdr.c_dev = 0x01;
    hdr.c_ino = count++;
    hdr.c_mode = node.attr_.unixMode();

    switch(node.attr_.fileType()) {
    case NSR::file_Raw:
        hdr.fileSize(node.elist_.totalBytes());
        break;
    case NSR::file_Directory:
        hdr.c_mode |= S_IFDIR;
        hdr.fileSize(0);
        break;
    case NSR::file_SymbolicLink:
        hdr.c_mode |= S_IFLNK;
        getSymlinkPath(path,node);
        hdr.fileSize(strlen(path) + 1);
        break;
    }

    hdr.c_uid = node.attr_.uid();
    hdr.c_gid = node.attr_.gid();
    hdr.c_nlink = node.attr_.linkCount();
    if(node.attr_.fileType() == NSR::file_Directory) {
        hdr.c_nlink++;
    }
    hdr.c_rdev = 0x00;
    hdr.modifyTime(node.attr_.modificationTime());
}

#endif // _ARCBROWSE_
