//-*-c++-*-
//-----------------------------------------------------------------------------
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//-----------------------------------------------------------------------------

#ifdef ROGUEWAVIFIED
#include <rw/disktree.h>
#include <rw/filemgr.h>
#include <rw/collect.h>
#include <rw/cstring.h>
#include <rw/rwfile.h>
#endif
#include <unistd.h>             // For access(2)
#include "arcTree.h"

//-----------------------------------------------------------------------------
// The following macro is necessary for the treeNode, because it is
// derived from the RogueWave RWCollectable class.  It doesn't need to be
// exactly here, but it needs to be somewhere in a .cpp file somewhere.
//-----------------------------------------------------------------------------

#ifdef ROGUEWAVIFIED
RWDEFINE_COLLECTABLE(TreeNode,NSRIMPORTTREE_ID)
#endif

//-----------------------------------------------------------------------------
// METHOD:  TreeNode::TreeNode
//
// Default constructor for the TreeNode class.
//-----------------------------------------------------------------------------

TreeNode::TreeNode(void) : 
#ifndef ROGUEWAVIFIED
                           parent_(NULL),
                           realNode_(NULL),
#endif
                           nchildren_(0),
                           maxChildren_(0),
                           db_(NULL),
                           child_(NULL),
                           data_(NULL)
{
    // No need to do anything.
}

TreeNode::TreeNode( TreeDB *db ) :
#ifndef ROGUEWAVIFIED
                           parent_(NULL),
                           realNode_(NULL),
#endif
                           nchildren_(0),
                           maxChildren_(0),
                           db_(db),
                           child_(NULL),
                           data_(NULL)
{
    // No need to do anything
}

TreeNode::TreeNode( TreeDB *db, const IntegerField& key ) :
#ifndef ROGUEWAVIFIED
                           parent_(NULL),
                           realNode_(NULL),
#endif
                           nchildren_(0),
                           maxChildren_(0),
                           db_(db),
                           child_(NULL),
                           data_(NULL)
{
    key_ = key;
}

//-----------------------------------------------------------------------------
// METHOD:  TreeNode::TreeNode
//
// Copy constructor for the TreeNode class.
//-----------------------------------------------------------------------------

TreeNode::TreeNode(const TreeNode& node) : 
#ifndef ROGUEWAVIFIED
                                           parent_(NULL),
                                           realNode_(NULL),
#endif
                                           nchildren_(0),
                                           maxChildren_(0),
                                           db_(NULL),
                                           child_(NULL),
                                           data_(NULL)
{
    operator=(node);
}

//-----------------------------------------------------------------------------
// METHOD:  TreeNode::~TreeNode
//
// Destructor.
//-----------------------------------------------------------------------------

TreeNode::~TreeNode() 
{
    delete [] child_;
}

//-----------------------------------------------------------------------------
// METHOD:  TreeNode::resize()
//
// Resizes the child_ array.  Primarily used for preallocation.
//-----------------------------------------------------------------------------

NSR::Error
TreeNode::resize(UINT32 numChildren) 
{
#ifdef ROGUEWAVIFIED

    int            i,n;
    IntegerField   *newarr;

    newarr = new IntegerField[numChildren];
    if(newarr == NULL) {
        return Error(NSR::error_No_free_store);
    }
    n = ((numChildren < nchildren_) ? numChildren : nchildren_);
    for(i = 0; i < n; i++) {
        newarr[i] = child_[i];
    }
    delete [] child_;
    child_ = newarr;
    maxChildren_ = numChildren;

    return Error(NSR::error_None);
   
#else

    int         i,n;
    TreeNode   **newarr;

    newarr = new (TreeNode *)[numChildren];
    if(newarr == NULL) {
        return Error(NSR::error_No_free_store);
    }
    n = ((numChildren < nchildren_) ? numChildren : nchildren_);
    for(i = 0; i < n; i++) {
        newarr[i] = child_[i];
    }
    delete [] child_;
    child_ = newarr;
    maxChildren_ = numChildren;

    return Error(NSR::error_None);

#endif
}

//-----------------------------------------------------------------------------
// METHOD:  TreeNode::addChild()
//
// Adds a child to the node.
//-----------------------------------------------------------------------------

NSR::Error 
TreeNode::addChild(const TreeNode& child)
{
#ifdef ROGUEWAVIFIED

    Error       err;

    if(nchildren_ < maxChildren_) {
        child_[nchildren_] = child.key_;
    }
    else {
        // we have to increase the size of the child_ array.  We only increase
        // it by 1.  The user should really use resize() to preallocate space
        // in the child_ array.

        err = resize(getNumChildren() + 1);
        if(err != NSR::error_None) {
            return err;
        }
        child_[nchildren_] = child.key_;
    }
    nchildren_++;
    return Error(NSR::error_None);

#else

    Error       err;

    if(nchildren_ < maxChildren_) {
        child_[nchildren_] = child.realNode_;
    }
    else {

        // We have to increase the size of the child_ array.  We only increase
        // it by 1.  The user should really use resize() to preallocate
        // space in the child_ array.

        err = resize(nchildren_ + 1);
        if(err != NSR::error_None) {
            return err;
        }
        child_[nchildren_] = child.realNode_;
    }
    nchildren_++;
    return Error(NSR::error_None);

#endif
}

//-----------------------------------------------------------------------------
// METHOD:  TreeNode::removeChild()
//
// Removes a child from the node.
//-----------------------------------------------------------------------------

NSR::Error
TreeNode::removeChild(const TreeNode& child)
{
#ifdef ROGUEWAVIFIED

    int         i,index = -1;
    TreeNode    child;
    Error       err;

    for(i = 0; i < getNumChildren(); i++) {
        if(child_[i] == child.key_) {
            index = i;
            break;
        }
    }
    if(index < 0) {
        return Error(NSR::error_Argument_not_found);
    }

    err = getChild(index,child);
    assert(err == NSR::error_None);

    if(child.getNumChildren() != 0) {
        return Error(NSR::error_Directory_dirty);
    }

    // Delete the child

    err = db_->freeNode(child);
    if(err != NSR::error_None) {
        return err;
    }

    // Have to move all the children to the right of the deleted child one
    // place to the left.

    for(i = index; i < (nchildren_ - 1); i++) {
        child_[i] = child_[i + 1];
    }
    nchildren_--;

    return Error(NSR::error_None);

#else

    int         i,index = -1;
    TreeNode    *cur;

    for(i = 0; i < nchildren_; i++) {
        cur = child_[i];
        if((child.address_.sid == cur->address_.sid) &&
           (child.address_.sector == cur->address_.sector)) {
            index = i;
            break;
        }
    }
    if(index < 0) {
        return Error(NSR::error_Argument_not_found);
    }

    // We have to verify that the child doesn't have any children.  If so,
    // we'll have to fail the call.

    if(child_[index]->nchildren_ != 0) {
//      return Error(NSR::error_BLAH_Directory_not_empty);
        return Error(NSR::error_Directory_dirty);
    }

    // Delete the child.

    delete child_[index];

    // Have to move all the children to the right of the deleted child one
    // place to the left.

    for(i = index; i < (nchildren_ - 1); i++) {
        child_[i] = child_[i + 1];
    }
    nchildren_--;

    return Error(NSR::error_None);
#endif
}

//-----------------------------------------------------------------------------
// METHOD:  TreeNode::setParent()
//
// Sets the parent entry.
//-----------------------------------------------------------------------------

void
TreeNode::setParent(const TreeNode& parent) 
{
#ifdef ROGUEWAVIFIED

    parent_ = parent.key_;
#else

    parent_ = parent.realNode_;

#endif
}

//-----------------------------------------------------------------------------
// METHOD:  TreeNode::getParent()
//
// Fills in a TreeNode struct with the parent node information.
//-----------------------------------------------------------------------------

NSR::Error
TreeNode::getParent(TreeNode& parent) const
{
#ifdef ROGUEWAVIFIED

    Error       err;

    if(db_ == NULL) {
        return Error(NSR::error_Argument_not_found);
    }
    err = db_->getNode(parent_,parent);

    return err;

#else

    if(parent_ == NULL) {
        return Error(NSR::error_Argument_not_found);
    }

    parent = *parent_;
    parent.realNode_ = parent_;
    return Error(NSR::error_None);
#endif
}

//-----------------------------------------------------------------------------
// METHOD:  TreeNode::getChild()
//
// Fills in a TreeNode struct with the data from the desired child.
//-----------------------------------------------------------------------------

NSR::Error
TreeNode::getChild(UINT32 which,TreeNode& child) const
{
#ifdef ROGUEWAVIFIED

    Error       err;

    if(db_ == NULL) {
        return Error(NSR::error_Argument_not_found);
    }
    if(which >= nchildren_) {
        return Error(NSR::error_Argument_not_found);
    }
    err = db_->getNode(child_[which],child);
    return err;

#else

    if(which >= nchildren_) {
        return Error(NSR::error_Argument_not_found);
    }
    if(child_[which] == NULL) {
        return Error(NSR::error_Argument_not_found);
    }
    child = *(child_[which]);
    child.realNode_ = child_[which];

    return Error(NSR::error_None);

#endif
}

//-----------------------------------------------------------------------------
// METHOD:  TreeNode::sync()
//
// Syncs the node with its on-disk representation:  copies the user's version
// of the node's information into the on-disk representation (ToDisk).
//-----------------------------------------------------------------------------

NSR::Error
TreeNode::sync(void) 
{
#ifdef ROGUEWAVIFIED

    if(db_ == NULL) {
        return Error(NSR::error_Argument_not_found);
    }
    return db_->putNode(*this);

#else

    if(db_ == NULL) {
        return Error(NSR::error_Argument_not_found);
    }
    db_->putNode(*this);
    return Error(NSR::error_None);

#endif
}

//-----------------------------------------------------------------------------
// METHOD:  TreeNode::refresh()
//
// Copies the data from the node's on-disk representation into the user's local
// copy (FromDisk).  This operation is the inverse of a sync().
//-----------------------------------------------------------------------------

Error
TreeNode::refresh(void)
{
#ifdef ROGUEWAVIFIED

    if(db_ == NULL) {
        return Error(NSR::error_Argument_not_found);
    }
    return db_->getNode(this->key_,*this);

#else

    if(db_ == NULL) {
        return Error(NSR::error_Argument_not_found);
    }
    db_->getNode(this->key_,*this);
    return Error(NSR::error_None);
#endif
}

//-----------------------------------------------------------------------------
// METHOD:  TreeNode::search()
//
// Searches node for file with specified name.
//-----------------------------------------------------------------------------

Error 
TreeNode::search(CHAR_PTR filename,TreeNode& node) const
{
    INT32       i;
    TreeNode    child;
    Error       err;

    for(i = 0; i < nchildren_; i++) {
        err = getChild(i,child);
        if(err != NSR::error_None) {
            return err;
        }
        if(strcmp((CHAR_PTR) child.filename_,filename) == 0) {
            node = child;
            return Error(NSR::error_None);
        }
    }
    return Error(NSR::error_Argument_not_found);
}

//-----------------------------------------------------------------------------
// METHOD:  TreeNode::operator =()
//
// Assignment operator between two TreeNodes.
//-----------------------------------------------------------------------------

TreeNode&
TreeNode::operator =(const TreeNode& node)
{
#ifdef ROGUEWAVIFIED

    int i;

    nchildren_ = node.nchildren_;
    maxChildren_ = node.maxChildren_;
    delete [] child_;
    child_ = new IntegerField[maxChildren_];
    if(child_ == NULL) {
        return *this;
    }
    for(i = 0; i < nchildren_; i++) {
        child_[i] = node.child_[i];
    }
    parent_ = node.parent_;
    key_ = node.key_;
    address_ = node.address_;
    attr_ = node.attr_;
    db_ = node.db_;
    filename_ = node.filename_;
    plist_ = node.plist_;
    elist_ = node.elist_;
    data_ = node.data_;

    return *this;

#else

    int         i;

    nchildren_ = node.nchildren_;
    maxChildren_ = node.maxChildren_;
    delete [] child_;
    child_ = new (TreeNode *)[maxChildren_];
    if(child_ == NULL) {
        return *this;
    }
    for(i = 0; i < nchildren_; i++) {
        child_[i] = node.child_[i];
    }
    parent_ = node.parent_;
    realNode_ = node.realNode_;

    address_ = node.address_;
    attr_ = node.attr_;
    db_ = node.db_;
    filename_ = node.filename_;
    plist_ = node.plist_;
    elist_ = node.elist_;
    data_ = node.data_;

    return *this;
#endif
}

//-----------------------------------------------------------------------------
// METHOD:  TreeNode::getKey()
//
// Retrieve's a node's key.
//-----------------------------------------------------------------------------

void
TreeNode::getKey(IntegerField& key) const
{
    key = key_;
}

//-----------------------------------------------------------------------------
// METHOD:  TreeNode::setKey()
//
// Sets a node's key.
//-----------------------------------------------------------------------------

void
TreeNode::setKey(const IntegerField& key) 
{
#ifdef ROGUEWAVIFIED

    // NOTE:  If this node was allocated by the database, the user must
    // blow away the old node and resync.

    key_ = key;

#else

    key_ = key;

#endif
}

void TreeNode::print(const char *msg)
{
    int         i;
    RWCString   attributes;

    if(msg != NULL) {
        printf("%s",msg);
    }
    printf("----------------- Node ------------------\n");
    printf("Filename:      %s\n",(CONST_CHAR_PTR) filename_);
    printf("Key:           %s\n",key_.field);
#ifdef ROGUEWAVIFIED
    printf("Parent:        %s\n",parent_.field);
#endif  // ROGUEWAVIFIED
    printf("ICB Address:   (%d,%d)\n",address_.sid,address_.sector);
    printf("Max children:  %d\n",maxChildren_);
    printf("Num children:  %d\n",nchildren_);
    attr_.appendAttributes(attributes);
    printf("Attributes:    %s\n",(CONST_CHAR_PTR) attributes);
    printf("Extent List:   ( ");
    for(i = 0; i < elist_.getCount(); i++) {
        Extent  *extp = elist_.at(i);
        printf("(%d,%d,%d) ",extp->start.sid,extp->start.sector,extp->len);
    }
    printf(")\n");
    fflush(stdout);
}

#ifdef ROGUEWAVIFIED
RWspace
TreeNode::binaryStoreSize() const
{
    int                 namelen,childlen,attrlen,linklen,extlen;
    int                 totalLen;
    RWCString           attributes;
    char                path[512];

//    namelen = filename_.length() + 1;         // Add 1 for null-termination
// BLAH
    namelen = strlen((CONST_CHAR_PTR) filename_) + 1;
    childlen = key_.size()*getNumChildren();
    attr_.appendAttributes(attributes);
    attrlen = attributes.length() + 1;          // Add 1 for null-termination
    if(attr_.fileType() != NSR::file_SymbolicLink) {
        linklen = 0;
    }
    else {
        plistToPath(path);
        linklen = strlen(path) + 1;             // Add 1 for null-termination
    }
    extlen = 3*key_.size()*elist_.getCount();

    // The 160 is 10 * integer field size

#ifdef OLD
    totalLen = 120 + namelen + childlen + attrlen + linklen + extlen;
#else
    totalLen = 160 + namelen + childlen + attrlen + linklen + extlen;
#endif
    totalLen += RWCollectable::binaryStoreSize();
    return totalLen;
}
#endif

#ifdef ROGUEWAVIFIED
void
TreeNode::restoreGuts(RWFile& file)
{
    char *              buffer;
    IntegerField        field;
    unsigned            len;

    RWCollectable::restoreGuts(file);
    if(file.Read(field.field,field.size()) == FALSE) {
        return;
    }
       
    len = field.fromAscii();

    buffer = new char[len];
    if(buffer == NULL) {
        return;
    }
    memcpy(buffer,field.field,field.size());
    if(file.Read(buffer+field.size(),len - field.size()) == FALSE) {
        delete [] buffer;
        return;
    }
    fromAscii(buffer);
    delete [] buffer;
}
#endif

#ifdef ROGUEWAVIFIED
void
TreeNode::saveGuts(RWFile& file) const
{
    char *              buffer;
    unsigned            len;

    RWCollectable::saveGuts(file);
    len = binaryStoreSize();
    buffer = new char[len];
   // printf("saveGuts:  len = %d\n",len);
//    fflush(stdout);
    if(buffer == NULL) {
        return;
    }
    memset(buffer,0,len);
    toAscii(buffer);
    if(file.Write(buffer,len) == FALSE) {
        delete [] buffer;
        return;
    }

    delete [] buffer;
    return;
}
#endif

#ifdef ROGUEWAVIFIED

int
TreeNode::compareTo(const RWCollectable *ptr) const
{
    TreeNode    *tptr = (TreeNode *) ptr;

    printf("c");
    fflush(stdout);
    return strcmp(key_.field,tptr->key_.field);
}

#endif

#ifdef ROGUEWAVIFIED

RWBoolean
TreeNode::isEqual(const RWCollectable *ptr) const
{
    TreeNode    *tptr = (TreeNode *) ptr;

    printf("i");
    fflush(stdout);
    return (strcmp(key_.field,tptr->key_.field) == 0);
}

#endif 

//-----------------------------------------------------------------------------
// METHOD:  TreeNode::toAscii()
//
// Copies the tree node's data to an ASCII representation.
//
// The format is as follows:
//
// offset          length  description                             format
// ------          ------  -----------                             ------
// 0               12      record length                     r0    %011o\000
// 12              12      node key                          f0    %011o\000
// 24              12      filename length (+ null byte)     f1*   %011o\000
// 36              12      parent key                        f2    %011o\000
// 48              12      # of children                     f3*   %011o\000
// 60              12      length of attributes              f4*   %011o\000
// 72              12      length of link path               f5*   %011o\000
// 84              12      # of extents                      f6*   %011o\000
// 96              12      ICB address sid                   f7    %011o\000
// 108             12      ICB address sector                f8    %011o\000
// 120             f1      null-terminated filename                string
// 120+f1          f3*12   child keys                        l3    %011o\000
// 120+f1+l3       f4      attributes                              string
// 120+f1+l3+f4    f5      link path (if necessary)                string
// 120+f1+l3+f4+f5 f6*36   extents                           l6      
//
// Total length:  120 + f1 + f3*12 + f4 + f5 + f6*36
//-----------------------------------------------------------------------------

#ifdef ROGUEWAVIFIED
int
TreeNode::toAscii(char *buffer) const
{
    int                 i;
    int                 namelen,childlen,attrlen,linklen,extlen;
    int                 totalLen;
    RWCString           attributes;
    char                path[512];
    IntegerField        field;
    char                *cptr = buffer;

    if(buffer == NULL) {
        return 0;
    }

    // Calculate the total length.

    totalLen = binaryStoreSize();
//    namelen = filename_.length() + 1;         // Add 1 for null-termination
//  BLAH
    namelen = strlen((CONST_CHAR_PTR) filename_) + 1;
    childlen = key_.size()*getNumChildren();
    attr_.appendAttributes(attributes);
    attrlen = attributes.length() + 1;          // Add 1 for null-termination
    if(attr_.fileType() != NSR::file_SymbolicLink) {
        linklen = 0;
    }
    else {
        plistToPath(path);
        linklen = strlen(path) + 1;             // Add 1 for null-termination
    }
    extlen = 3*key_.size()*elist_.getCount();

    field.toAscii(totalLen);                    // Total record length
    cptr = field.toMem(cptr);
    int len=strlen(key_.field);
    memset(&(key_.field[len]),0,key_.size() - len);
    cptr = key_.toMem(cptr);                    // File key
    field.toAscii(namelen);                     // Filename length
    cptr = field.toMem(cptr);
    cptr = parent_.toMem(cptr);
    field.toAscii(nchildren_);                  // Number of children
    cptr = field.toMem(cptr);
    field.toAscii(attrlen);                     // Attributes length
    cptr = field.toMem(cptr);
    field.toAscii(linklen);                     // Length of link path
    cptr = field.toMem(cptr);
    field.toAscii(elist_.getCount());           // Number of extents;
    cptr = field.toMem(cptr);
    field.toAscii(address_.sid);                // ICB Address sid
    cptr = field.toMem(cptr);
    field.toAscii(address_.sector);             // ICB Address sector
    cptr = field.toMem(cptr);
    strcpy(cptr,(CHAR_PTR) filename_);          // Filename
    cptr += namelen;
    for(i = 0; i < getNumChildren(); i++) {     // Child keys
        cptr = child_[i].toMem(cptr);
    }
    strcpy(cptr,(CHAR_PTR) attributes);         // Attributes
    cptr += attrlen;
    if(attr_.fileType() == NSR::file_SymbolicLink) {    // Link path
        strcpy(cptr,path);
        cptr += linklen;
    }
    for(i = 0; i < elist_.getCount(); i++) {            // Extent list
        const Extent *extp = elist_.at(i);

        field.toAscii((unsigned int) extp->start.sid);  // Sid
        cptr = field.toMem(cptr);
        field.toAscii((unsigned int) extp->start.sector); // Start sector
        cptr = field.toMem(cptr);
        field.toAscii((unsigned int) extp->len);        // Length
        cptr = field.toMem(cptr);
    }
    //assert((cptr - buffer) == totalLen);
    return totalLen;
}

#endif  // ROGUEWAVIFIED

//-----------------------------------------------------------------------------
// METHOD:  TreeNode::fromAscii()
//
// Fills in the tree node's data to from an ASCII representation (does not 
// include the extent list!).
//
// The format is as follows:
//
// offset          length  description                             format
// ------          ------  -----------                             ------
// 0               12      record length                     r0    %011o\000
// 12              12      node key                          f0    %011o\000
// 24              12      filename length (+ null byte)     f1*   %011o\000
// 36              12      parent key                        f2    %011o\000
// 48              12      # of children                     f3*   %011o\000
// 60              12      length of attributes              f4*   %011o\000
// 72              12      length of link path               f5*   %011o\000
// 84              12      # of extents                      f6*   %011o\000
// 96              12      ICB address sid                   f7    %011o\000
// 108             12      ICB address sector                f8    %011o\000
// 120             f1      null-terminated filename                string
// 120+f1          f3*12   child keys                        l3    %011o\000
// 120+f1+l3       f4      attributes                              string
// 120+f1+l3+f4    f5      link path (if necessary)                string
// 120+f1+l3+f4+f5 f6*36   extents                           l6      
//
// Total length:  120 + f1 + f3*12 + f4 + f5 + f6*36
//-----------------------------------------------------------------------------

#ifdef ROGUEWAVIFIED
void
TreeNode::fromAscii(char *buffer)
{
    int                 i;
    int                 namelen,attrlen,pathlen,totalLen;
    int                 numChildren;
    int                 numExtents;
    RWCString           attributes;
    char                path[512];
    IntegerField        field;
    char                *cptr = buffer;

    cptr += key_.size();                                // Skip length field
    int len=strlen(key_.field);
    memset(&(key_.field[len]),0,key_.size() - len);
    key_ = cptr;                                                // Node key
    cptr += key_.size();
    namelen = field.fromAscii(cptr);                            // Name length
    cptr += key_.size();
    parent_ = cptr;                                             // parent key
    cptr += key_.size();
    numChildren = field.fromAscii(cptr);                        // # children
    cptr += key_.size();
    attrlen = field.fromAscii(cptr);                            // attr len
    cptr += key_.size();
    pathlen = field.fromAscii(cptr);                            // link len
    cptr += key_.size();
    numExtents = field.fromAscii(cptr);                         // # extents
    cptr += key_.size();
    address_.sid = field.fromAscii(cptr);                       // ICB addr
    cptr += key_.size();
    address_.sector = field.fromAscii(cptr);                    // ICB addr
    cptr += key_.size();
    filename_ = cptr;                                           // Filename
    cptr += namelen;
    resize(numChildren);
    for(i = 0; i < numChildren; i++) {                          // Children
        child_[i] = cptr;
        cptr += key_.size();
    }
    nchildren_ = numChildren;
    attributes = cptr;                                          // Attributes
    attr_.setAttributes(attributes);
    cptr += attrlen;
    if(attr_.fileType() == NSR::file_SymbolicLink) {            // Link path
        pathToPlist(cptr);
    }
    cptr += pathlen;
    elist_.clear();
    for(i = 0; i < numExtents; i++) {                           // ExtentList
        Extent  ext;

        ext.start.sid = (Sid_t) field.fromAscii(cptr);
        cptr += key_.size();
        ext.start.sector = field.fromAscii(cptr);
        cptr += key_.size();
        ext.len = field.fromAscii(cptr);
        cptr += key_.size();

        elist_.append(ext);
    }

    return;
}
#endif

//-----------------------------------------------------------------------------
// METHOD:  TreeNode::plistToPath()
//
// Converts a PathElementList to a Unix-style path.
//-----------------------------------------------------------------------------

#ifdef ROGUEWAVIFIED
void
TreeNode::plistToPath(char *path) const
{
    int                    i;
    int                    len;
    const PathElementList *plistp;

    // Get the PathElementList for the symbolic link

    plistp = &plist_;

    // Now we have to parse through the PathElementList and construct a
    // standard Unix path from it.

    path[0] = 0;
    for(i = 0; i < plistp->getCount(); i++) {
        const PathElement& pelt = (*plistp)[i];

        len = strlen(pelt.name);
        switch(pelt.type) {
        case NSR::path_Custom_Root:             // Root if nonzero length
            if(len == 0) {      // Reset
                path[0] = 0;
                strcat(path,"/");
            } 
            else {              // This is an unusual case!
                fprintf(stderr,"WARNING:  Symbolic link path component of \
type 1 with nonzero length!\n");
                path[0] = 0;
                strcat(path,"/");
            }
            break;
        case NSR::path_Standard_Root:           // Root
            path[0] = 0;
            strcat(path,"/");
            break;
        case NSR::path_Parent:                  // ..
            strcat(path,"../");
            break;
        case NSR::path_Same:                    // .
            strcat(path,"./");
            break;
        case NSR::path_Name:                    // Just a path element
            strcat(path,pelt.name);
            strcat(path,"/");
            break;
        default:
            fprintf(stderr,"WARNING:  Symbolic link contains unknown path \
component type!\n");
            strcat(path,"unknown_type/");
            break;
        }
    }
    len = strlen(path);
    if(path[len - 1] == '/') {          // Remove trailing /'s.
        path[len - 1] = 0;
    }
    return;
}
#endif

//-----------------------------------------------------------------------------
// METHOD:  TreeNode::pathToPlist()
//
// Converts a Unix-style path to a PathElementList
//-----------------------------------------------------------------------------

#ifdef ROGUEWAVIFIED
void
TreeNode::pathToPlist(const char *buf)
{
    int         i,index,count;
    char        element[512];
    PathElement pelt;

    i = 0;
    plist_.clear();
    count = strlen(buf);
    while(i < count) {
        index = 0;
        if(buf[i] == '/') {                     // Rooted path
            pelt.type = NSR::path_Standard_Root;
            pelt.name = "";
            plist_.append(pelt);
            i++;
            continue;
        }

        while((buf[i] != '/') && (buf[i] != 0)) {
            element[index] = buf[i];
            index++,i++;
        }
        i++;
        element[index] = 0;
        if(!element[0]) {       // Empty path element; assume we're done
            break;
        }
        if(strcmp(element,".") == 0) {          // Current dir path element
            pelt.type = NSR::path_Same;
            pelt.name = "";
            plist_.append(pelt);
            continue;
        }
        if(strcmp(element,"..") == 0) {         // Parent dir path element
            pelt.type = NSR::path_Parent;
            pelt.name = "";
            plist_.append(pelt);
            continue;
        }
        // If we got here, it's a regular path element.

        pelt.type = NSR::path_Name;
        pelt.name = element;
        plist_.append(pelt);
    }
    return;
}
#endif

//-----------------------------------------------------------------------------
// METHOD:  TreeDB::TreeDB
//
// Constructor for the TreeDB class.
//-----------------------------------------------------------------------------

TreeDB::TreeDB(const RWCString& swapFile,int rmFile) : rmFile_(rmFile)
{
#ifdef ROGUEWAVIFIED

    curKey_.toAscii(1);                         // Reserve 0. . .

    fileManager_ = NULL;
    bTree_ = NULL;
    swapFile_ = swapFile;
    mdFile_ = swapFile;
    mdFile_.append(".md");

    // create the RWFileManager

    fileManager_ = new RWFileManager((char *) swapFile);
    if(fileManager_ == NULL) {
        fprintf(stderr,"FATAL:  treeDB, fileManager create failed!\n");
        exit(1);
    }

    // If this is a new database file, allocate some space for the root node's
    // key.  The first allocation to an RWFileManager is special in that it
    // can never be deallocated; the location of the first allocation is also
    // available by the RWFileManager::start() method.  We use it to store
    // the root node's key, so that we can get started again later.

    if(::access((CONST_CHAR_PTR) fileManager_,F_OK)) {
        fileManager_->allocate(curKey_.size());
    }
    mdManager_ = new RWFileManager((char *) mdFile_);
    if(mdManager_ == NULL) {
        fprintf(stderr,"FATAL:  treeDB, mdManager_ create failed!\n");
        exit(1);
    }

    // Create the RWBTreeOnDisk.

#ifdef OLD
    bTree_ = new RWBTreeOnDisk(*mdManager_,     // Metadata manager
                               10,              // Number of buffers
                               RWBTreeOnDisk::autoCreate,   // Create tree
                                                            // if necessary
                               curKey_.size(),  // Length of keys
                               FALSE,           // Ignore nulls
                               RWNIL,           // Use RWFM for start node
                               RWBTreeOnDisk::V6Style,      // New Btrees
                               10,              // half order of the BTree
                               10               // Minimum entries in a node
                              );
#else
    bTree_ = new RWBTreeOnDisk(*mdManager_);
#endif

    if(bTree_ == NULL) {
        fprintf(stderr,"FATAL:  treeDB, BTree create failed!\n");
        exit(1);
    }

    return;

#else

    // DO NOTHING!

    return;

#endif
}

//-----------------------------------------------------------------------------
// METHOD:  TreeDB::~TreeDB
//
// Destructor for the TreeDB class.
//-----------------------------------------------------------------------------

TreeDB::~TreeDB()
{
#ifdef ROGUEWAVIFIED

    delete bTree_;
    delete fileManager_;
    delete mdManager_;
    if(rmFile_) {
        if(!swapFile_.isNull()) {
            unlink((CHAR_PTR) swapFile_);
        }
        if(!mdFile_.isNull()) {
            unlink((CHAR_PTR) mdFile_);
        }
    }
    return;

#else

    // DO NOTHING!

    return;

#endif
}

//-----------------------------------------------------------------------------
// METHOD:  TreeDB::allocNode()
//
// Allocate a node.
//-----------------------------------------------------------------------------

NSR::Error
TreeDB::allocNode(TreeNode& node)
{
#ifdef ROGUEWAVIFIED

    Error       err;

    node.setDB(this);
    node.setKey(curKey_);
    node.elist_.clear();
    node.plist_.clear();
    node.filename_ = "";
    node.nchildren_ = 0;
    node.maxChildren_ = 0;
    delete [] node.child_;
    node.child_ = NULL;
    node.data_ = NULL;
    ++curKey_;

    err = putNode(node);
    return err;

#else

    TreeNode    *newnode;

    newnode = new TreeNode(this);
    if(newnode == NULL) {
        return Error(NSR::error_No_free_store);
    }

    newnode->realNode_ = newnode;
    newnode->setKey(curKey_);
    ++curKey_;

    node = *newnode;

    return Error(NSR::error_None);

#endif
}

//-----------------------------------------------------------------------------
// METHOD:  TreeDB::allocNode()
//
// Allocate a node.
//-----------------------------------------------------------------------------

NSR::Error
TreeDB::allocNode(TreeNode& node,const Address& addr)
{
#ifdef ROGUEWAVIFIED

    IntegerField        key;
    Error       err;

    node.elist_.clear();
    node.plist_.clear();
    node.filename_ = "";
    node.nchildren_ = 0;
    node.maxChildren_ = 0;
    delete [] node.child_;
    node.child_ = NULL;
    node.data_ = NULL;
    node.setDB(this);
    addrToKey(addr,key);
    node.setKey(key);

    err = putNode(node);
    return err;

#else

    TreeNode    *newnode;
    IntegerField key;

    addrToKey(addr,key);

    newnode = new TreeNode(this,key);
    if(newnode == NULL) {
        return Error(NSR::error_No_free_store);
    }

    newnode->realNode_ = newnode;
    node = *newnode;

    return Error(NSR::error_None);

#endif
}

//-----------------------------------------------------------------------------
// METHOD:  TreeDB::freeNode()
//
// Frees a node.
//-----------------------------------------------------------------------------

NSR::Error
TreeDB::freeNode(TreeNode& node)
{
#ifdef ROGUEWAVIFIED

    IntegerField        key;

    node.getKey(key);
    RWoffset loc = bTree_->findValue(key.field);

    if(loc != RWNIL) {
        bTree_->remove(key.field);
        fileManager_->deallocate(loc);
    }
    else {
        return Error(NSR::error_Argument_not_found);
    }
    return Error(NSR::error_None);

#else

    delete node.realNode_;
    node.realNode_ = NULL;
    return Error(NSR::error_None);

#endif

}

//-----------------------------------------------------------------------------
// METHOD:  TreeDB::getNode()
//
// Gets a node.
//-----------------------------------------------------------------------------

NSR::Error
TreeDB::getNode(const IntegerField& key,TreeNode& node)
{
#ifdef ROGUEWAVIFIED

    RWoffset loc = bTree_->findValue(key.field);
    if(loc == RWNIL) {
        return Error(NSR::error_Argument_not_found);
    }
    else {
        fileManager_->SeekTo(loc);
//        *fileManager_ >> node;
        node.restoreGuts(*fileManager_);
    }
    node.setDB(this);
    return Error(NSR::error_None);

#else

    // This can't really "get" anything at all, since the key isn't meaningful
    // in the pointer representation.  This method shouldn't be called anyhow
    // by the user.

    if(node.realNode_ != NULL) {
        node = *(node.realNode_);
    }
    return Error(NSR::error_None);

#endif

}

//-----------------------------------------------------------------------------
// METHOD:  TreeDB::getRoot()
//
// Gets root.
//-----------------------------------------------------------------------------

#ifdef ROGUEWAVIFIED

NSR::Error
TreeDB::getRoot(TreeNode& node)
{
    IntegerField        key;

    RWoffset loc = fileManager_->start();
    if(loc == RWNIL) {
        return Error(NSR::error_Argument_not_found);
    }
    else {
        fileManager_->SeekTo(loc);
        if(fileManager_->Read(key.field,key.size()) == FALSE) {
           return Error(NSR::error_Argument_not_found);
        }
    }
    return getNode(key,node);
}

#endif  // ROGUEWAVIFIED

//-----------------------------------------------------------------------------
// METHOD:  TreeDB::setRoot()
//
// Sets root.
//-----------------------------------------------------------------------------

#ifdef ROGUEWAVIFIED

NSR::Error
TreeDB::setRoot(TreeNode& root)
{
    IntegerField        key;
    RWoffset loc = fileManager_->start();
    if(loc == RWNIL) {
        return Error(NSR::error_Argument_not_found);
    }
    else {
        root.getKey(key);
        fileManager_->SeekTo(loc);
        if(fileManager_->Write(key.field,key.size()) == FALSE) {
           return Error(NSR::error_Argument_not_found);
        }
    }
    return Error(NSR::error_None);
}
#endif



//-----------------------------------------------------------------------------
// METHOD:  TreeDB::putNode()
//
// Writes a node to the database.
//-----------------------------------------------------------------------------

NSR::Error
TreeDB::putNode(const TreeNode& node)
{
#ifdef ROGUEWAVIFIED

    IntegerField     key;

    node.getKey(key);
    RWoffset loc = bTree_->findValue(key.field);
    if(loc != RWNIL) {                  // Nuke it
        bTree_->remove(key.field);
        fileManager_->deallocate(loc);
    }

    // Now (re)insert.

    loc = fileManager_->allocate(node.binaryStoreSize());
    fileManager_->SeekTo(loc);
//    *fileManager_ << node;
    node.saveGuts(*fileManager_);
    bTree_->insertKeyAndValue(key.field,loc);

    return Error(NSR::error_None);

#else

    // Synchronizes the user's copy with the "database" copy.

    if(node.realNode_ == NULL) {
        return Error(NSR::error_Argument_not_found);
    }
    *(node.realNode_) = node; 

    return Error(NSR::error_None);

#endif

}

//-----------------------------------------------------------------------------
// METHOD:  TreeDB::addrToKey()
//
// Create a IntegerField string out of an Address.
//-----------------------------------------------------------------------------

void
TreeDB::addrToKey(const Address& addr,IntegerField& key)
{
    char        buffer[key.size()];     // This decl. is not portable

    sprintf(buffer,"%u.%d",addr.sid,addr.sector);
    key = buffer;
}
