//-*-c++-*-
//-----------------------------------------------------------------------------
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//-----------------------------------------------------------------------------

#ifndef _ARCTREE_H_
#define _ARCTREE_H_

#include <string.h>
#include <rw/cstring.h>
#include "nsrLib.h"

#ifdef ROGUEWAVIFIED
#include <rw/collect.h>
#include <rw/disktree.h>
#include <rw/filemgr.h>
#endif

#ifdef OLD
const int INT_FIELD_WIDTH = 12;
#else
const int INT_FIELD_WIDTH = 16;
#endif

//-----------------------------------------------------------------------------
// CLASS:  IntegerField
//
// This class, implemented as a struct, is a representation of a 32-bit 
// integer.
//
// PUBLIC DATA MEMBERS:
//
// char field[INT_FIELD_WIDTH]  ASCII (octal) representation of an integer.
//
// PUBLIC METHODS:
//
// IntegerField()
//      Constructor and copy constructor.
// ~IntegerField()
//      Destructor.
// toAscii()
//      Copies an integer into ASCII representation.
// fromAscii()
//      Returns the integer stored in the field.
// operator==
//      Equality operator between two IntegerFields and between an IntegerField
//      and a regular integer.
// operator=
//      Assignment operators between IntegerFields and ints.
// operator++
//      Increments value by one.  (prefix and postfix syntax)
// operator--
//      Decrements value by one.  (prefix and postfix syntax)
// size
//      Returns the size of the integer field, in bytes
// 
// USAGE NOTES:
//
//      None.
//-----------------------------------------------------------------------------
// COLLABORATORS:
//
//      None.
//
// PRIVATE DATA MEMBERS:
//
//      None.
//
// PRIVATE METHODS:
//
//      None.
//-----------------------------------------------------------------------------

struct IntegerField
{
   char         field[INT_FIELD_WIDTH];

   //--------------------------------------------------------------------------
   // C++isms:  Constructors, destructors, assignment operators
   //--------------------------------------------------------------------------

                        IntegerField() { toAscii(0); }
                        IntegerField(unsigned int initval) { toAscii(initval); }
                        IntegerField(const char *str) { *this = str; }
                        ~IntegerField() { }
                        IntegerField(const IntegerField& ifield) {
                            operator=(ifield);
                        }
    IntegerField&       operator=(const IntegerField& ifield) {
        memcpy(field,ifield.field,INT_FIELD_WIDTH);
        return *this;
    }
    IntegerField&       operator=(const char *str) {
        if(str != NULL) {
            memcpy(field,str,INT_FIELD_WIDTH);
        }
        return *this;
    }
    int                 operator==(const IntegerField& ifield) {
        unsigned int val1 = fromAscii();
        unsigned int val2 = ifield.fromAscii();
        return (val1 == val2);
    }
    int                 operator==(unsigned int val) {
        unsigned int val1 = fromAscii();
        return (val1 == val);
    }
    int                 operator!=(const IntegerField& ifield) {
        return !(*this == ifield);
    }
    int                 operator!=(unsigned int val) {
        return !(*this == val);
    }
    IntegerField&       operator++() {
        unsigned int    val = fromAscii();
        toAscii(val + 1);
    }
    IntegerField&       operator++(int dummy) {
        return ++(*this);
    }
    IntegerField&       operator--() {
        unsigned int    val = fromAscii();
        toAscii(val - 1);
    }
    IntegerField&       operator--(int dummy) {
        return --(*this);
    }

    //--------------------------------------------------------------------------
    // Other methods
    //--------------------------------------------------------------------------

    void                toAscii(unsigned int newVal) {
        sprintf(field,"%011o",newVal);
    }
    unsigned int        fromAscii(void) const {
        unsigned int val;
        sscanf(field,"%011o",&val);
        return val;
    }
    unsigned int        fromAscii(const char *buffer) {
        unsigned int val;
        sscanf(buffer,"%011o",&val);
        return val;
    }
    char *              toMem(char *buffer) const {
        memcpy(buffer,field,INT_FIELD_WIDTH);
        return (buffer + INT_FIELD_WIDTH);
    }
    int                 size() const { return INT_FIELD_WIDTH; }
};


#ifdef ROGUEWAVIFIED
#define NSRIMPORTTREE_ID        0x0744          // Arbitrary
#endif

//-----------------------------------------------------------------------------
// CLASS:  TreeNode
//
// This class is used to represent an n-ary tree whose purpose is to be an
// in-core representation of the directory tree being imported and restored
// from or exported to NSR format.  Each node, which represents a file,
// contains the ICB information for the file.
//
// PUBLIC DATA MEMBERS:
//
// Address              address_        ICB address
// Attributes           attr_           Attributes
// RWCString            filename_       Filename
// PathElementList      plist_          For symbolic links
// ExtentList           elist_          For reg. files and directories
// void *               data_           Application specific data
//
// Non-paging version only:
//
// TreeNode *           realNode_       Pointer to the "real" node
//
// PUBLIC METHODS:
//
// TreeNode()
//      Constructor and copy constructor.
// ~TreeNode()
//      Destructor.
// resize()
//      Resizes the child array to have the specified number of entries.
// addChild()
//      Adds the specified TreeNode as a child of the current TreeNode.
// removeChild()
//      Removes the TreeNode at the specified index from the child array.
// setParent()
//      Sets the parent of the current TreeNode.
// getParent()
//      Gets the parent of the current TreeNode.
// getChild()
//      Gets the child at the specified index in the child array.
// getNumChildren()
//      Returns the number of children (NOT the capacity of the child array!)
// sync()
//      Synchronizes the TreeNode with the on-disk/core representation (write)
// refresh()
//      Reads the TreeNode in from disk/core (read)
// search()
//      Searches the TreeNode's immediate children for a node with the given
//      filename.
// operator=()
//      Assigns between two TreeNodes.
// getKey()
//      Gets the TreeNode's key (an IntegerField).
// setKey()
//      Sets the TreeNode's key (an IntegerField).
// getDB()
//      Returns a pointer to the TreeNode's TreeDB object.
// setDB()
//      Sets the TreeNode's TreeDB object.
// print()
//      Print some information about the TreeNode to screen.
//
// Paging Version Only (inherited from RWCollectable):
//
// binaryStoreSize()
//      Returns the number of bytes required to store the object to disk.
// restoreGuts()
//      Restores a TreeNode from an RWFile.
// saveGuts()
//      Saves a TreeNode to an RWFile.
// hash()
//      Returns a hash value for the TreeNode.
// compareTo()
//      Compares two TreeNodes logically.
// isEqual()
//      Compares two TreeNodes logically.
//
// USAGE NOTES:
//
// The TreeNode, in either its in-core or paging representations, must be
// used cautiously.  It uses value-based semantics in that the local 
// TreeNode object which the user uses is not the "real" TreeNode.  TreeNodes
// should be thought of as local handles to chunks of information that are
// stored in a database (the TreeDB class).  If a TreeNode is allocated from
// a TreeDB, any changes that are made to the TreeNode object will not be
// persistent unless a sync() call is used, which causes the information from
// the user's local TreeNode object to be flushed to the database (on disk
// in the paging version, in a bit of memory on the heap otherwise).  
// Similarly, local copies must be refresh()'ed if the TreeNode object has
// been changed by code elsewhere.
//
//-----------------------------------------------------------------------------
// COLLABORATORS:
//
//      TreeDB
//      IntegerField
//      Address
//      Attributes
//      ExtentList
//      PathElementList
//      RWCollectable
//      RWCString
//
// PRIVATE DATA MEMBERS:
//
// UINT32       nchildren_      Number of children in node
// UINT32       maxChildren_    Size of child_ array
// TreeDB       *db_            Database pointer
// IntegerField key_            Node's key
//
// Paging Version:
//
// IntegerField *child_         Array of child keys
// IntegerField parent_         For parent pointer
//
// Non-Paging Version:
//
// TreeNode     **child_        Array of child keys
// TreeNode     *parent_        For parent pointer
//
// PRIVATE METHODS:
//
// Paging Version Only:
//
//
// toAscii()
//      Writes an ASCII representation of the TreeNode's data to a buffer.
// fromAscii()
//      Fills in a TreeNode's data from an ASCII representation.
// plistToPath()
//      Converts a PathElementList into a Unix-style path.
// pathToPlist()
//      Converts a Unix-style path into a PathElementList.
//-----------------------------------------------------------------------------

class TreeDB;                   // Forward declaration

#ifdef ROGUEWAVIFIED

class TreeNode : public RWCollectable {

#else

class TreeNode {

#endif

    //-------------------------------------------------------------------------
    // Private data
    //-------------------------------------------------------------------------

    private:

#ifdef ROGUEWAVIFIED
RWDECLARE_COLLECTABLE(TreeNode)         // For RogueWave RWCollectable Class
#endif

    IntegerField        key_;           // Node's key
    UINT32              nchildren_;     // Number of children in node
    UINT32              maxChildren_;   // Size of child_ array
    TreeDB              *db_;           // Database pointer

#ifdef ROGUEWAVIFIED
    IntegerField        *child_;        // Array of child keys
    IntegerField        parent_;        // For parent pointer
#else
    TreeNode            **child_;       // Array of children pointers
    TreeNode            *parent_;       // For parent pointer
#endif

    //-------------------------------------------------------------------------
    // Public data 
    //-------------------------------------------------------------------------

    public:

    Address             address_;       // ICB address
    Attributes          attr_;          // Attributes
    RWCString           filename_;      // Filename
    PathElementList     plist_;         // For symbolic links
    ExtentList          elist_;         // For reg. files and directories
    void                *data_;         // Application specific data

#ifndef ROGUEWAVIFIED
    TreeNode            *realNode_;     // INTERNAL USE ONLY, don't use it!
#endif

    //-------------------------------------------------------------------------
    // Private methods
    //-------------------------------------------------------------------------

    private:

#ifdef ROGUEWAVIFIED
    int                 toAscii(char *buffer) const;
    void                fromAscii(char *buffer);
    void                plistToPath(char *path) const;
    void                pathToPlist(const char *buf);
#endif

    //-------------------------------------------------------------------------
    // Public methods
    //-------------------------------------------------------------------------

    public:

                        TreeNode();
                        TreeNode(const TreeNode& node);
                        TreeNode(TreeDB *db);
                        TreeNode(TreeDB *db,const IntegerField& key);
                        ~TreeNode();
    Error               resize(UINT32 numChildren);
    Error               addChild(const TreeNode& child);
    Error               removeChild(const TreeNode& child);
    void                setParent(const TreeNode& parent);
    Error               getChild(UINT32 which,TreeNode& child) const;
    Error               getParent(TreeNode& parent) const;
    UINT32              getNumChildren(void) const {
                            return nchildren_;
                        }
    Error               sync(void);
    Error               refresh(void);
    Error               search(CHAR_PTR filename,TreeNode& node) const;
    TreeNode&           operator=(const TreeNode& node);
    void                getKey(IntegerField& key) const;
    void                setKey(const IntegerField& key);
    TreeDB *            getDB(void) const { return db_; }
    void                setDB(TreeDB *newDB) { db_ = newDB; }
    void                print(const char *msg);

    //-------------------------------------------------------------------------
    // These methods must be overloaded from RWCollectable
    //-------------------------------------------------------------------------

    public:

#ifdef ROGUEWAVIFIED
    RWspace             binaryStoreSize() const;
    int                 compareTo(const RWCollectable *) const;
    RWBoolean           isEqual(const RWCollectable *) const;
    unsigned            hash() const { return address_.sector; }
    void                restoreGuts(RWFile&);
    void                saveGuts(RWFile&) const;
#endif
};
    
//-----------------------------------------------------------------------------
// CLASS:  TreeDB
//
// This class is intended to provide a pagable n-ary tree representation of
// a directory tree.  In its current implementation, it simply keeps every
// node in memory (in the form of TreeNode objects).  It is provided
// so that future implementations of a true paging tree will not 
// affect interfaces.
//
// PUBLIC DATA MEMBERS:
//
//      None.
//
// PUBLIC METHODS:
//
// TreeDB()
//      Constructor and copy constructor.
// ~TreeDB()
//      Destructor.
// allocNode()
//      Allocates a node in the database (or in core).
// freeNode()
//      Removes a node from the database (or from core).
// getNode()
//      Fetches the node with the given key.
// putNode()
//      Writes/updates the node with the given key into the database.
// addrToKey()
//      Converts from an Address to a key (IntegerField)
//
//-----------------------------------------------------------------------------
// COLLABORATORS:
//
// IntegerField
// TreeNode
//
// Paging Version:
//
// RWFileManager
// RWBTreeOnDisk
// RWCString
//
// PRIVATE DATA MEMBERS:
//
// IntegerField         curKey_         // Current key to be assigned
//
// Paging Version Only:
//
// RWFileManager *      fileManager_    // For data
// RWFileManager *      mdManager_      // For btree
// RWBTreeOnDisk *      bTree_          // The btree itself
// RWCString            swapFile_       // Data file name
// RWCString            mdFile_         // Btree file name
//
// PRIVATE METHODS:
//
//      None.
//-----------------------------------------------------------------------------

class TreeDB {

private:

    IntegerField        curKey_;
    int                 rmFile_;

#ifdef ROGUEWAVIFIED
    RWFileManager       *fileManager_;          // For data
    RWFileManager       *mdManager_;            // For btree
    RWBTreeOnDisk       *bTree_;
    RWCString           swapFile_;
    RWCString           mdFile_;
#endif

public:

                        TreeDB(const RWCString& swapFile,int rmFile = 1);
                        ~TreeDB();
    Error               allocNode(TreeNode& node);
    Error               allocNode(TreeNode& node,const Address& addr);
    Error               freeNode(TreeNode& node);
#ifdef ROGUEWAVIFIED
    Error               setRoot(TreeNode& node);
    Error               getRoot(TreeNode& node);
#endif
    Error               getNode(const IntegerField& key,TreeNode& node);
    Error               putNode(const TreeNode& node);
    void                addrToKey(const Address& addr,IntegerField &key);
};

#endif
