// arcWrite.cpp -*-c++-*-
//-----------------------------------------------------------------------------
// $Source: /projects/HMS/nsr/udfAncil/msTest/arcWrite.cpp,v $
// $Revision: 1.39 $    $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 96/05/09 10:52:24 $
//-----------------------------------------------------------------------------
// (c) Copyright 1995  Hewlett-Packard Company.
// The contents of this software are proprietary and confidential to
// the Hewlett-Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the Government  is
//      subject to restrictions as set forth in subdivision (b)
//      (3)  (ii)  of the Rights in Technical Data and Computer
//      Software clause at 52.227-7013.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304
//
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <assert.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <rw/gslist.h>
#include <rw/cstring.h>
#include <rw/ctoken.h>
#include "nsrLib.h"
#include <devs/nsrRSSdv.h>
#include <devs/nsrRMSdv.h>
#include "dlist.h"                      // Doubly-linked list toolkit
#include "dqueue.h"                     // Queue macro package based on dlist
#include "arcImEx.h"
#include "fsalloc.h"
#include "cmr.h"
#include "cpio.h"
#include "arcSynth.h"

//-----------------------------------------------------------------------------
// Some macros
//-----------------------------------------------------------------------------

#define MAX_FILESET_NAMELENGTH  32              // File set name maximum
#define MAX_VOLUME_NAMELENGTH   32              // Volume name maximum

#define DATA_START_SECTOR       300

#ifndef MAX_LINE
#define MAX_LINE                512
#endif

const char *defaultDBFile = "/tmp/arcWrite.swap";

//-----------------------------------------------------------------------------
// Function prototypes
//-----------------------------------------------------------------------------

static int              getFiles(TreeDB& tree,TreeNode& root,
                                 hpACDev *acdev,CHAR_PTR srcdir,
                                 FreeSpaceAllocator& fsa,
                                 UINT32 sectorSize);
static int              readCpio(TreeDB& tree,TreeNode& root,
                                 hpACDev *acdev,FreeSpaceAllocator& fsa,
                                 UINT32 sectorSize);
static int              readSynthFile(TreeDB& tree,TreeNode& root,
                                 hpACDev* acdev,FreeSpaceAllocator& fsa,
                                 UINT32 sectorSize, CHAR_PTR synthScriptName);
static int              getAttributes(TreeNode& node,CHAR_PTR pathname,
                                      UINT32 *fileSize = NULL);
static int              recursiveGetFiles(TreeDB& tree,TreeNode& node,
                                          hpACDev *acdev,CHAR_PTR path,
                                          FreeSpaceAllocator& fsa,UINT32 ss);
static int              copyData(CHAR_PTR path,int size,hpACDev *acdev,
                                 FreeSpaceAllocator& fsa,ExtentList& elist,
                                 UINT32 ss);
static int              getPlist(CHAR_PTR path,PathElementList& plist);
static void             formPlist(CHAR_PTR buf,PathElementList& plist);
static void             Usage(void);
static int              writeCmr ( TreeNode& root, CmrWriter& cmrWriter );
void                    cpioToNode(cpioHeader& cpioHdr,TreeNode& node);

//-----------------------------------------------------------------------------
// Global variables
//-----------------------------------------------------------------------------

// The trace global variable seems to be needed for the Archive object.

int trace = 0;

//-----------------------------------------------------------------------------
// arcWrite.cpp:  This file contains source for a simple test to use the NSR
//              MSLIB Archive object in conjunction with the hpACDev autochanger
//              device object to write an nsr volume.  
//
// Usage:
//
// arcWrite [-a acrcFile] [-s jbStatFile] [-g volname] [-m mtype] [-z size] 
//          [-c cmrFile] [-E] [-k] [-w swapFile] [-W] -v vsname -f fsname 
//          { -d srcdir | -x }
//
// [-a acrcFile]:    Optional.  Pathname of the acrc configuration file.
// [-s jbStatFile]:  Optional.  Pathname of the jukebox status file.
// [-g volname]:     Optional if -v is specified.  Grow the volume set
//                   by one volume, with the specified name.
// [-m mtype]:       Optional if vsname exists.  Media type of grow volumes.
//                   Valid values are "mo" and "worm".
// [-z size]:        Optional if vsname exists.  Sector size of grow volumes.
//                   Valid values are 512 and 1024.
// [-c cmrFile]:     Optional.  Specifies name of output CMR file.  Defaults
//                   to export.cmr.
// [-E]:             Optional.  No export; label only.
// [-k]:             Optional.  OVERRIDES ALL OTHER OPTIONS.  Destroys the
//                   volume set specified by the -v option and frees its 
//                   volumes to the available list.  USE WITH CARE.
// [-w swapFile]:    Optional.  Specifies the pathname of the swapfile.
//                   This option is only useful if the paging tree option
//                   has been compiled in (ROGUEWAVIFIED).  The swapfile,
//                   and a swapfile named swapfile.md, will be created to
//                   store metadata information.  Default:  /tmp/arcWrite.swap
// [-W]:             Optional.  If specified, the swap file and the
//                   metadata file will NOT be removed at the completion
//                   of the program.
// -v vsname:        Mandatory.  Export to the specified volume set.
// -f fsname:        Mandatory.  Specifies the name of the exported file set.
// -d srcdir:        Specifies the source directory for export.
// -D synthfile:     Specifies a synth file script.
// -x:               Read fileset from cpio archive on stdin.
//
// Example:  
//
// arcWrite -a .acrc -s /tmp/jbstat.1 -v "Volset 1" -f "Fileset 2" -d /users
//
// ENVIRONMENT VARIABLES:
//
// The acrc file and the jukebox status file can both be specified by
// environment variables:
//
// ACRC_FILE:    Contains the acrc file pathname.
// JBSTAT_FILE:  Contains the jukebox status file pathname.
// ARCTRACE:     Environment variable containing the value of the trace var.
//
// Note that the -a and -s options will override any value in the corresponding
// environment variable, and the -a and -s options are mandatory if the
// environment variables are not set.
//-----------------------------------------------------------------------------

declare(RWGSlist,RWCString)

main (int argc,CHAR_PTR argv[])
{
    int                 c;
    CHAR_PTR            cptr;
    INT32               i;
    int                 rval;
    int                 usage = 0;
    int                 killVS = 0;
    int                 noExport = 0;
    int                 readFromCpio = 0;
    int                 rmFile = 1;
    MediaType           mtype;
    UINT32              ssize;
    RWCString           acrcFile;
    RWCString           jbStatFile;
    RWCString           vsname;
    RWCString           fsname;
    RWCString           srcDir;
    RWCString           cmrFile(DEFAULT_CMRFILE);
    RWCString           *vnptr;
    RWCString           synthScriptName;
    RWCString           dbFile(defaultDBFile);
    RWCString           freeFileName;
    RWGSlist(RWCString) growVols;
    FreeSpaceAllocator  fsa;

    extern CHAR_PTR optarg;             // For getopt()
    extern int  optind,opterr;          // For getopt()

    // Get environment variables, if any.

    cptr = getenv("ACRC_FILE");
    if(cptr != NULL) {
        acrcFile = cptr;
    }
    cptr = getenv("JBSTAT_FILE");
    if(cptr != NULL) {
        jbStatFile = cptr;
    }
    cptr = getenv("ARCTRACE");
    if(cptr != NULL) {
        trace = atoi(cptr);
    }

    // Parse command-line arguments.

    while((c = getopt(argc,argv,"a:s:v:m:z:f:d:D:xg:e:c:kt:Ew:W")) != EOF) {
        switch(c) {
        case 'a':
            acrcFile = optarg;
            break;
        case 's':
            jbStatFile = optarg;
            break;
        case 'v':
            vsname = optarg;
            break;
        case 'm':
            if( (strcmp(optarg,"worm") == 0) || (strcmp(optarg,"WORM") == 0) ) {
                mtype = NSR::media_Worm;
                break;
            }
            if( (strcmp(optarg,"mo") == 0) || (strcmp(optarg,"MO") == 0) ) {
                mtype = NSR::media_Rewritable;
                break;
            }
            fprintf (stderr, "Invalid media type (must be \"mo\" or \"worm\")\n");
            usage = 1;
            break;
        case 'z':
            ssize = atoi(optarg);
            if((ssize != 512) && (ssize != 1024)) {
                fprintf(stderr,"Invalid sector size %s!\n",ssize);
                usage = 1;
            }
            break;
        case 'd':
            srcDir = optarg;
            break;
        case 'D':
            synthScriptName = optarg;
            break;
        case 'x':
            readFromCpio = 1;
            break;
        case 'f':
            fsname = optarg;
            break;
        case 'g':
            vnptr = new RWCString(optarg);
            if(vnptr == NULL) {
                fprintf(stderr,"No free store, abort!\n");
                return (2);
            }
            growVols.append(vnptr);
            break;
        case 'c':
            cmrFile = optarg;
            break;
        case 'e':
            freeFileName = optarg;
            break;
        case 'k':
            killVS = 1;
            break;
        case 't':
            trace = atoi(optarg);
            break;
        case 'E':
            noExport = 1;
            break;
        case 'w':
            dbFile = optarg;
            break;
        case 'W':
            rmFile = 0;
            break;
        default:
            fprintf (stderr, "Invalid option: %c\n", c);
            usage = 1;
            break;
        }
    }

    // Make sure the user has specified everything that is mandatory.

    if(acrcFile.isNull() || jbStatFile.isNull() || vsname.isNull()) {
        fprintf(stderr,"You must specify acrc file, status file, and vsname!\n");
        usage = 1;
    }
    else {
        if(!killVS) {
            if(!noExport) {
                if(/*srcDir.isNull() ||*/ fsname.isNull()) {
                    fprintf (stderr, "You must specify a fileset name unless not exporting\n");
                    usage = 1;
                }
            }
        }
    } 

    // Make sure that the argument directory for the export exists and can be 
    // accessed.

    if(!killVS && !noExport && synthScriptName.isNull()) {
        if(!srcDir.isNull()) {
            if(access((CHAR_PTR) srcDir,R_OK | X_OK | F_OK)) {
                fprintf(stderr,"Source directory %s not accessible!\n",
                        (CHAR_PTR) srcDir);
                usage = 1;
            }
        }
    }
    if(srcDir.isNull() && !readFromCpio && !killVS && synthScriptName.isNull()) {
        fprintf(stderr,"You must specify either -d or -x or -D!\n");
        usage = 1;
    }

    // Make sure that the synth script, if specified, is accessible.  This
    // check is useful because the synth parser is instantiated AFTER volume
    // labeling occurs, and if the script isn't accessible the program just
    // bails out, leaving a labeled volume set.

    if(!synthScriptName.isNull())
    {
        if(access((CONST_CHAR_PTR) synthScriptName,R_OK))
        {
            fprintf(stderr,"Synth file %s is not accessible!\n",
                    (CONST_CHAR_PTR) synthScriptName);
            usage = 1;
        }
    }

    if(usage) {
        Usage();
        return (TERM_ERROR);
    }

    // If we got here, we're ready to start testing.
    TreeDB              tree(dbFile,rmFile);
    NSRJukeBox          jb(acrcFile);
    TreeNode            root;

    if(jb.error() != NSR::error_None) {
        fprintf(stderr,"Construction of NSRJukeBox failed, error is %s\n",
                (jb.error()).message());
        return (TERM_ERROR);
    }

    // Read the status file.

    jb.readStatusFile(jbStatFile);
    if(jb.error() != NSR::error_None) {
        fprintf(stderr,"Read of status file %s failed, error is %s\n",
                (CHAR_PTR) jbStatFile,(jb.error()).message());
        return (TERM_ERROR);
    }

    // Set the working volume set.  

    VolsetNameMapList   vsnmList;
    VolsetNameMap       *vsnmptr;

    vsnmList = jb.volsetNameMapList();
    vsnmptr = vsnmList.search(vsname);
    if(vsnmptr == NULL) {
        jb.setNewWVS(vsname,ssize,mtype);
    }
    else {
        jb.setExistingWVS(vsname);
    }
    if(jb.error() != NSR::error_None) {
        fprintf(stderr,"Unable to set working volume set, error is %s\n",
                (CHAR_PTR) (jb.error()).message());
        return (TERM_NOERROR_TEST_FAILED);
    }
     

    // If the -k option was specified, we're killing the volume set, so do
    // it now and get out.

    if(killVS) {
        jb.freeVolumeSet(vsname);
        if(jb.error() != NSR::error_None) {
            fprintf(stderr,"freeVolumeSet() failed, error is %s\n",
                    (jb.error()).message());
        }
        else {
            printf("Freed volume set %s\n",(CHAR_PTR) vsname);
        }
        return TERM_NOERROR_TEST_PASSED;
    }

    // If we have any grow volumes, grow the volume set now.

    if(growVols.entries() > 0) {
        jb.beginGrow();
        if(jb.error() != NSR::error_None) {
            fprintf(stderr,"beginGrow() failed, error is %s\n",
                    (jb.error()).message());
            return TERM_NOERROR_TEST_FAILED;
        }
    }

    for(i = 0; i < growVols.entries(); i++) {
        jb.growWVS(*(growVols.at(i)),DATA_START_SECTOR);
        if(jb.error() != NSR::error_None) {
            fprintf(stderr,"Unable to grow working volume set, error is %s\n",
                    (CHAR_PTR) (jb.error()).message());
            return (TERM_NOERROR_TEST_FAILED);
        }
        printf("Added volume \"%s\" to volume set \"%s\"\n",
               (CHAR_PTR) *(growVols.at(i)),(CHAR_PTR) vsname);
    }

    if(growVols.entries() > 0) {
        jb.completeGrow();
        if(jb.error() != NSR::error_None) {
            fprintf(stderr,"completeGrow() failed, error is %s\n",
                (jb.error()).message());
            return TERM_NOERROR_TEST_FAILED;
        }
    }

    // If we're just labeling, we're done.  I think.

    if(noExport) {
        return TERM_NOERROR_TEST_PASSED;
    }

    // We have to prepare the archive for modification.

    jb.prepareArchive();
    if(jb.error() != NSR::error_None) {
        fprintf(stderr,"Unable to prepare archive, error is %s\n",
                (CHAR_PTR) (jb.error()).message());
        return (TERM_NOERROR_TEST_FAILED);
    }
    
    // We need to get freespace information for the volumes.  For right
    // now, we have a slight problem with grown volume sets--they won't have
    // any freespace information at all.  The user will have to know which
    // volumes are new, and set the freespace accordingly.  This is a bit of
    // a pain.

    VolumeFreeSpaceList         vfsl;
    VolumeFreeSpace             *vfsptr;

    jb.wvsFreeSpace(vfsl);
    if(jb.wvsMediaType() == NSR::media_Rewritable) {
        // For each of the growVols(), we need to set its free space.  We're
        // just going to say the whole disk is available.
        
        for(i = 0; i < vfsl.entries(); i++) {
            Extent      extent;
            UINT32      len;
          
            vfsptr = vfsl.at(i);
            vfsptr->dataStart_ = DATA_START_SECTOR;
            vfsptr->freeSpace_.clear();
            vfsptr->erasedSpace_.clear();
            extent.start.sid = vfsptr->sid_;
            extent.start.sector = DATA_START_SECTOR;
            len = ((jb.acdev())->numSectors(vfsptr->sid_) - 
                   DATA_START_SECTOR - 2) * jb.wvsSectorSize();
            extent.len = len;
            vfsptr->freeSpace_.append(extent);
        }
    }

    // Now take the freespace information and feed it to the FreeSpaceAllocator.

    for(i = 0; i < vfsl.entries(); i++) {
        vfsptr = vfsl.at(i);
        if(jb.wvsMediaType() == NSR::media_Rewritable) {
            fsa.dealloc(jb.wvsSectorSize(),vfsptr->freeSpace_);
            if (!synthScriptName.isNull())
                fsa.erasedDealloc(jb.wvsSectorSize(),vfsptr->erasedSpace_);
        }
        else {
            fsa.dealloc(jb.wvsSectorSize(),vfsptr->freeSpace_);
            if (!synthScriptName.isNull())
                fsa.erasedDealloc(jb.wvsSectorSize(),vfsptr->erasedSpace_);
        }
    }

    if (!synthScriptName.isNull()) {
        if (jb.wvsSectorSize() == 0) {
            fprintf (stderr, "Error in arcWrite, about to read the synth file (%s, %d)\n",
                     __FILE__, __LINE__);
            fprintf (stderr, "The jukebox class thinks sector size is zero.  This may mean\n");
            fprintf (stderr, "that you are attempting to export to an unlabeld volume set.\n\n");
            return (TERM_ERROR);
        }
        
        printf("Reading synthetic fileset script...\n");
        if(readSynthFile(tree,root,jb.acdev(),fsa,jb.wvsSectorSize(),
                         (char*)synthScriptName)) {
            // fprintf(stderr,"readSynthFile() failed!\n");
            return(TERM_NOERROR_TEST_FAILED);
        }
    }
    else if(srcDir.isNull()) {
        printf("Reading cpio archive from stdin. . .\n");
        if(readCpio(tree,root,jb.acdev(),fsa,jb.wvsSectorSize())) {
            fprintf(stderr,"readCpio() failed!\n");
            return(TERM_NOERROR_TEST_FAILED);
        }
    }
    else {
        printf("Reading source directory. . .\n");
        if(getFiles(tree,root,jb.acdev(),(CHAR_PTR) srcDir,fsa,
           jb.wvsSectorSize())) {
            fprintf(stderr,"getFiles() failed!\n");
            return (TERM_NOERROR_TEST_FAILED);
        }
    }
    
    //
    // HERE IS WHERE TREE MANIPULATION CODE SHOULD GO!!!
    //
  

    // Unfortunately, we have to convert from the FreeSpaceAllocator back to
    // a VolumeFreeSpaceList.  

    for(i = 0; i < vfsl.entries(); i++) {
        vfsptr = vfsl.at(i);
        if(jb.wvsMediaType() == NSR::media_Rewritable) {
            vfsptr->freeSpace_.clear();
            fsa.alloc(vfsptr->sid_,vfsptr->freeSpace_);
            if (!synthScriptName.isNull())
                fsa.erasedAlloc(vfsptr->sid_,vfsptr->erasedSpace_);
        }
        else {
            vfsptr->freeSpace_.clear();
            vfsptr->erasedSpace_.clear();
            fsa.alloc(vfsptr->sid_,vfsptr->freeSpace_);
            if (!synthScriptName.isNull())
                fsa.erasedAlloc(vfsptr->sid_,vfsptr->erasedSpace_);
        }
    }    

    CmrWriter   cmrWriter((CHAR_PTR) cmrFile);
    cmrWriter.useShortReservedWords();
    cmrWriter.setTabSize(2);

    printf("Writing CMR file %s\n",(CHAR_PTR) cmrFile);
    writeCmr(root,cmrWriter);

    printf("Exporting fileset metadata. . .\n");
    jb.exportArchive(fsname,root,vfsl);

    // Write a file containing free space.
    if (freeFileName != NULL) {
        FILE *free = fopen ((char*)freeFileName, "w");
        String volBuf;
        jb.writeAsText (volBuf);
        fwrite((char*)volBuf, sizeof(char), volBuf.length(), free);
        
//        fprintf (free,
//                 "%s\n",
//                 (char*)volBuf);
        fclose(free);
    }
    
    if(jb.error() != NSR::error_None) {
        fprintf(stderr,"jb.exportArchive() failed, error is %s\n",
                (jb.error()).message());
        rval = TERM_NOERROR_TEST_FAILED;
    }
    else {
        printf("Export was successful\n");
        rval = TERM_NOERROR_TEST_PASSED;
    }
#ifdef ROGUEWAVIFIED
    tree.setRoot(root);
#endif
    return rval;
}

//-----------------------------------------------------------------------------
// FUNCTION:  getFiles()
//
// This function traverses the entire source directory tree and recursively
// copies data to optical.  It also builds an nsrArchiveTree containing
// all of the file metadata.
//
// IMPORTANT NOTE:  This function assumes that all surfaces in the srfcList
// are completely unallocated and free (no WORM support right now!).  
//
// ARGUMENTS:
//
// TreeDB&          tree     IN/OUT  Archive tree to be filled up
// TreeNode&        root     OUT     Root node of tree
// hpACDev         *acdev    IN      Autochanger device object
// char            *srcdir   IN      Source directory
// FreeSpaceAllocator& fsa   IN/OUT  Free space information
// UINT32           ssize    IN      Sector size
//-----------------------------------------------------------------------------

int
getFiles(TreeDB& tree,TreeNode& root,hpACDev *acdev,CHAR_PTR srcdir,
         FreeSpaceAllocator& fsa,UINT32 ssize)
{
    Error               err;

    // We get the root directory's attributes first.  Since it's a directory,
    // we can't fill in its ExtentList now.  Fill in the filename ("/").

    err = tree.allocNode(root);
    if(err != NSR::error_None) {
        fprintf(stderr,"tree.allocNode() failed, error is %s\n",
                err.message());
        return -1;
    }
    if(getAttributes(root,srcdir)) {
        fprintf(stderr,"getFile():  getAttributes failed on root, aborting\n");
        return -1;
    }
    root.filename_ = "/";
    root.sync();

    // Now that our tree is rooted, we can start recurring on root's children.  

    if(recursiveGetFiles(tree,root,acdev,srcdir,fsa,ssize)) {
        return -1;
    }
    else {
        return 0;
    }
}

//-----------------------------------------------------------------------------
// FUNCTION:  getAttributes()
//
// Stats the specified file (given by pathname) and fills in the Attributes
// of the given nsrTreeNode.
//
// ARGUMENTS:
//
// TreeNode&        node      OUT  Node to be filled in
// char            *pathname  IN   File to stat
// UINT32          *fileSize  OUT  Size of file in bytes
//-----------------------------------------------------------------------------

static int
getAttributes( TreeNode& node,CHAR_PTR pathname,UINT32 *fileSize )
{
    struct stat statbuf;
    Attributes  attr;

    if(lstat(pathname,&statbuf)) {
        fprintf(stderr,"lstat() failed on file %s\n",pathname);
        perror("lstat");
        return -1;
    }
    if(S_ISCHR(statbuf.st_mode) || S_ISBLK(statbuf.st_mode) || 
       S_ISFIFO(statbuf.st_mode)) {
        printf("File %s is not a directory, regular file, or symlink, skip!\n",
               pathname);
        return -1;
    }
//    printf("%s\n",pathname);

    // Now just fill in the Attributes struct from the appropriate stat fields.

    if(S_ISDIR(statbuf.st_mode)) {
        node.attr_.fileType(NSR::file_Directory);
    }
    else {
        if(S_ISLNK(statbuf.st_mode)) {
            node.attr_.fileType(NSR::file_SymbolicLink);
        }
        else {
            node.attr_.fileType(NSR::file_Raw);
        }
    }
    node.attr_.uid(statbuf.st_uid);
    node.attr_.gid(statbuf.st_gid);
    node.attr_.unixMode(statbuf.st_basemode);

    // We need to set the link count.  We'll use the link count from the
    // stat struct, and adjust it by 1 for directories (since there are no
    // parent entries); the link count will be 1 for all other files.

    if(node.attr_.fileType() == NSR::file_Directory) {
        node.attr_.linkCount(statbuf.st_nlink - 1);
    }
    else {
        node.attr_.linkCount(1);
    }

    node.attr_.accessTime(statbuf.st_atime);
    node.attr_.modificationTime(statbuf.st_mtime);
    node.attr_.attributeTime(statbuf.st_ctime);
    if(fileSize != NULL) {
        *fileSize = statbuf.st_size;
    }
    node.sync();

    return 0;
}

//-----------------------------------------------------------------------------
// FUNCTION:  recursiveGetFiles()
//
// Recursively adds children files to the current node and copies their
// data to optical.
//
// ARGUMENTS:
//
// TreeDB&          tree     IN      Archive tree
// TreeNode&        node     IN/OUT  Node to add children to
// hpACDev *        acdev    IN      Autochanger device object
// char *           dirpath  IN      Pathname of current directory 
// FreeSpaceAllocator& fsa   IN/OUT  Free space information.
// UINT32           ssize    IN      Sector size
//
//-----------------------------------------------------------------------------

static int
recursiveGetFiles(TreeDB& tree,TreeNode& node,hpACDev *acdev,CHAR_PTR dirpath,
                  FreeSpaceAllocator& fsa,UINT32 ssize)
{
    int                 i;
    int                 rval;
    int                 numChildren;
    DIR                 *curdir;
    UINT32              fileSize;
    struct dirent       *dirent;
    TreeNode            tnode;
    char                curfile[MAX_NAMELENGTH];
    Error               err;

    curdir = opendir(dirpath);
    if(curdir == NULL) {
        fprintf(stderr,"Unable to open directory %s, aborting!\n",dirpath);
        return -1;
    }

    // Unfortunately, we have to go through each directory twice.  The first
    // time we merely count how many non-parent non-self entries there are,
    // so that we can resize the parent node's child pointer array.

    numChildren = 0;
    while((dirent = readdir(curdir)) != NULL) {
        // Kick out "." or ".." entries
        if(strcmp(dirent->d_name,".") == 0) {
            continue;
        }
        if(strcmp(dirent->d_name,"..") == 0) {
            continue;
        }
        numChildren++;
    }

    // Rewind the directory

    rewinddir(curdir);

    // Resize the node

    node.resize(numChildren);
    node.sync();

    while((dirent = readdir(curdir)) != NULL) {
        // Kick out "." or ".." entries
        if(strcmp(dirent->d_name,".") == 0) {
            continue;
        }
        if(strcmp(dirent->d_name,"..") == 0) {
            continue;
        }
        if(strcmp(dirpath,"/") == 0) {
            sprintf(curfile,"%s%s",dirpath,dirent->d_name);
        }
        else {
            sprintf(curfile,"%s/%s",dirpath,dirent->d_name);
        }
        err = tree.allocNode(tnode);
        if(err != NSR::error_None) {
            fprintf(stderr,"Unable to allocate TreeNode, aborting!\n");
            return -1;
        }
        if(getAttributes(tnode,curfile,&fileSize)) {
            fprintf(stderr,"WARNING:  Not exporting file %s!!!!\n",curfile);
            tree.freeNode(tnode);
            continue;
        }
        tnode.filename_ = dirent->d_name;
        tnode.setParent(node);

        // If it's a regular file, copy its data to optical.
        // If it's a symbolic link, get its PathElementList.

        if(tnode.attr_.fileType() != NSR::file_Directory) {
            if(tnode.attr_.fileType() != NSR::file_SymbolicLink) {
                tnode.elist_.clear();
                tnode.plist_.clear();
                if(copyData(curfile,fileSize,acdev,fsa,
                            tnode.elist_,ssize)) {
                    // Don't do anything. . . data wasn't successfully
                    // copied but we'll plug on nonetheless.
                    fprintf(stderr,"WARNING:  data copy failed for file %s\n",
                            (CHAR_PTR) tnode.filename_);
                }
            }
            else {                                      // Symbolic link
                if(getPlist(curfile,tnode.plist_)) {
                    fprintf(stderr,"Aborting export!\n");
                    return -1;
                }
            }
        }

        // Connect the child to the parent.

        node.addChild(tnode);
        tnode.sync();
        node.sync();
    }
    closedir(curdir);

    // Now we recur through the children directories.

    for(i = 0; i < node.getNumChildren(); i++) {
        err = node.getChild(i,tnode);
        if(err != NSR::error_None) {
            fprintf(stderr,"node.getChild(%d) failed, err is %s\n",i,
                    err.message());
            continue;
        }
        if(tnode.attr_.fileType() != NSR::file_Directory) {
            continue;
        }
        if(strcmp((CHAR_PTR) tnode.filename_,"/") == 0) {
            sprintf(curfile,"%s%s",dirpath,(CHAR_PTR) tnode.filename_);
        }
        else {
            sprintf(curfile,"%s/%s",dirpath,(CHAR_PTR) tnode.filename_);
        }
        tnode.sync();
        rval = recursiveGetFiles(tree,tnode,acdev,curfile,fsa,ssize);
        tnode.sync();
//      if(rval) {
//          return rval;
//      }
    }
    return 0;
}

//-----------------------------------------------------------------------------
// FUNCTION:  copyData()
//
// Copies data from a file to optical.
//
// ARGUMENTS:
//
// char       *pathname     IN      File to supply data
// int         size         IN      Size of the file, in bytes
// hpACDev    *acdev        IN      Autochanger device object
// FreeSpaceAllocator& fsa  IN/OUT  Free space
// ExtentList& elist        OUT     Extent List for file on optical.
// UINT32      ssize        IN      Sector size
//-----------------------------------------------------------------------------

static int
copyData(CHAR_PTR pathname,int size,hpACDev *acdev,FreeSpaceAllocator& fsa,
         ExtentList& elist,UINT32 ssize)
{
    int         fd;
    int         rval;
    int         nsectors;
    int         residue;
    int         sector_offset;
    int         len;
    Sid_t       sid;
    Extent      extent;
    static BYTE buf[MAX_IOSIZE];        // Don't reallocate constantly

    if(fsa.totalBytes() == 0) {
        fprintf(stderr,"copyData:  Out of space!!!\n");
        return -1;
    }

    // Open the file.

    fd = open(pathname,O_RDONLY);
    if(fd < 0) {
        fprintf(stderr,"Unable to open file %s for read!\n",pathname);
        return -1;
    }
    residue = size;
    while(residue > 0) {
        len = ((residue < MAX_IOSIZE) ? residue : MAX_IOSIZE);
        if(fsa.alloc(ssize,len,extent)) {
            len = fsa.largestFreeExtent();
            // MAKE SURE THIS ASSERTION DOESN'T GET COMPILED OUT
            assert(fsa.alloc(ssize,len,extent) == 0);
        }

        // Read from the native file system

        if(read(fd,buf,len) < len) {
            fprintf(stderr,"Got a short count reading file %s\n",pathname);
            close(fd);
            return -1;
        }

        sid = extent.start.sid;

        // Write to optical.

        if(acdev->write(sid,extent.start.sector,len,buf) < len) {
            fprintf(stderr,"Got a short count writing file %s\n",pathname);
            fprintf(stderr,"acdev error is %s, aborting data copy\n",
                    (acdev->error()).message());
            close(fd);
            return -1;
        }

        // Update variables, create an extent

        elist.append(extent);
        residue -= len;
    }
    close(fd);
    return 0;
}

//-----------------------------------------------------------------------------
// FUNCTION:  getPlist()
//
// Forms a PathElementList for a symbolic link.
//
// ARGUMENTS:
//
// char *            curfile   IN  Symbolic link pathname
// PathElementList&  plist     OUT PathElementList to fill in.
//-----------------------------------------------------------------------------

static int
getPlist(CHAR_PTR curfile,PathElementList& plist)
{
    int         i;
    int         index;
    int         count;
    char        buf[MAX_NAMELENGTH];
    char        element[MAX_NAMELENGTH];

    count = readlink(curfile,buf,MAX_NAMELENGTH);
    if(count < 0) {
        fprintf(stderr,"readlink() failed on file %s\n",curfile);
        perror("readlink");
        return -1;
    }
    buf[count] = 0;             // Null-terminate

    // Now we iterate through the path and form a PathElementList.  Yuck.

    PathElement pelt;

    i = 0;
    while(i < count) {
        index = 0;
        if(buf[i] == '/') {                     // Rooted path
            pelt.type = NSR::path_Standard_Root;
            pelt.name = "";
            plist.append(pelt);
            i++;
            continue;
        }

        while((buf[i] != '/') && (buf[i] != 0)) {
            element[index] = buf[i];
            index++,i++;
        }
        i++;
        element[index] = 0;
        if(!element[0]) {       // Empty path element; assume we're done
            break;
        }
        if(strcmp(element,".") == 0) {          // Current dir path element
            pelt.type = NSR::path_Same;
            pelt.name = "";
            plist.append(pelt);
            continue;
        }
        if(strcmp(element,"..") == 0) {         // Parent dir path element
            pelt.type = NSR::path_Parent;
            pelt.name = "";
            plist.append(pelt);
            continue;
        }
        // If we got here, it's a regular path element.

        pelt.type = NSR::path_Name;
        pelt.name = element;
        plist.append(pelt);
    }
    return 0;
}

static void
formPlist(CHAR_PTR buf,PathElementList& plist)
{
    int         i,index,count;
    char        element[MAX_NAMELENGTH];
    PathElement pelt;

    i = 0;
    count = strlen(buf);
    while(i < count) {
        index = 0;
        if(buf[i] == '/') {                     // Rooted path
            pelt.type = NSR::path_Standard_Root;
            pelt.name = "";
            plist.append(pelt);
            i++;
            continue;
        }

        while((buf[i] != '/') && (buf[i] != 0)) {
            element[index] = buf[i];
            index++,i++;
        }
        i++;
        element[index] = 0;
        if(!element[0]) {       // Empty path element; assume we're done
            break;
        }
        if(strcmp(element,".") == 0) {          // Current dir path element
            pelt.type = NSR::path_Same;
            pelt.name = "";
            plist.append(pelt);
            continue;
        }
        if(strcmp(element,"..") == 0) {         // Parent dir path element
            pelt.type = NSR::path_Parent;
            pelt.name = "";
            plist.append(pelt);
            continue;
        }
        // If we got here, it's a regular path element.

        pelt.type = NSR::path_Name;
        pelt.name = element;
        plist.append(pelt);
    }
}
        
//-----------------------------------------------------------------------------
// FUNCTION:  Usage()
//
// Prints usage information to screen.
//-----------------------------------------------------------------------------

static void
Usage( void )
{
    static CHAR_PTR usage_str[] = {
"Usage:",
"",
"arcWrite [-a acrcFile] [-s jbStatFile] [-g volname] [-m mtype] [-z size]",
"         [-c cmrFile] [-E] [-k] [-t val] -v vsname -f fsname ",
"         [-w swapFile] [-W] { -x | -d srcdir } ",
"",
"[-a acrcFile]:    Optional.  Pathname of the acrc configuration file.",
"[-s jbStatFile]:  Optional.  Pathname of the jukebox status file.",
"[-g volname]:     Optional if -v is specified.  Grow the volume set",
"                  by one volume, with the specified name.",
"[-m mtype]:       Optional if vsname exists.  Media type of grow volumes.",
"                  Valid values are \"mo\" and \"worm\".",
"[-z size]:        Optional if vsname exists.  Sector size of grow volumes.",
"                  Valid values are 512 and 1024.",
"[-c cmrFile]:     Optional.  Overrides default CMR file name (export.out)",
"[-E]:             Optional.  Specifies that no fileset export is to occur;",
"                  volumes in the volume set are only labeled.",
"[-k]:             Optional.  OVERRIDES ALL OTHER OPTIONS.  Destroys the",
"                  volume set specified by the -v option and frees its",
"                  volumes to the available list.  USE WITH CARE.",
"[-t val]:         Optional.  Set the value of the Archive trace variable.",
"-v vsname:        Mandatory.  Export to the specified volume set.",
"-f fsname:        Mandatory.  Specifies the name of the exported file set.",
"[-w swapFile]:    Optional.  Specifies the pathname of the swapfile.",
"                  This option is only useful if the paging tree option",
"                  has been compiled in (ROGUEWAVIFIED).  The swapfile,",
"                  and a swapfile named swapfile.md, will be created to",
"                  store metadata information.  Default:  /tmp/arcWrite.swap",
"[-W]:             Optional.  If specified, the swap file and the",
"                  metadata file will NOT be removed at the completion",
"                  of the program.",
"-d srcdir:        Specifies the source directory for export.",
"-D synthFile:     Specifies a synthetic fileset script.",
"-x:               Specifies that the exported fileset is to be read from",
"                  a cpio archive from standard input.  Exactly one of the",
"                  -d and -x options must be specified.",
"",
"Example:",
"",
"arcWrite -a .acrc -s /tmp/jbstat.1 -v \"Volset 1\" -f \"Fileset 2\" -d /users",
"",
"ENVIRONMENT VARIABLES:",
"",
"The acrc file and the jukebox status file can both be specified by",
"environment variables:",
"",
"ACRC_FILE:    Contains the acrc file pathname.",
"JBSTAT_FILE:  Contains the jukebox status file pathname.",
"ARCTRACE:     Environment variable containing the value of the trace var.",
"",
"Note that the -a and -s options will override any value in the corresponding",
"environment variable, and the -a and -s options are mandatory if the",
"environment variables are not set.",
NULL
};
    int i = 0;

    while(usage_str[i] != NULL) {
        printf("%s\n",usage_str[i]);
        i++;
    }
}

//-----------------------------------------------------------------------------
// writeCmr     Recursively write CMR from fileset information in tree.
//
// preconditions:
//      The tree is populated with a fileset.
//
// postconditions:
//      The output file has been populated with CMR refecting fileset info in
//      the tree, or an error has terminated processing.
//
// This code was stolen shamelessly from ../msVerify/udftocmr.C.
//-----------------------------------------------------------------------------

static int writeCmr ( TreeNode& root, CmrWriter& cmrWriter )
{
    int                 i;
    int                 rval;
    TreeNode            tnode;
    Error               err;

    // Tell the CmrWriter of this directory, file or symlink.
    switch (root.attr_.fileType()) {

        // currently supported file types.
    case NSR::file_Directory:
        cmrWriter.beginDir(root.filename_, root.attr_);

        // recurr on children
        for(i = 0; i < root.getNumChildren(); i++) {

            err = root.getChild(i,tnode);
            if(err != NSR::error_None) {                // Parent entry
                continue;
            }
            rval = writeCmr(tnode, cmrWriter);
            if(rval) {
                return rval;
            }
        }

        // Close the current directory.
        cmrWriter.endDir();
        break;
    case NSR::file_Raw:
        cmrWriter.file (root.filename_, root.attr_, root.elist_);
        break;
    case NSR::file_SymbolicLink:
        cmrWriter.symlink(root.filename_, root.attr_, root.plist_);
        break;

        // currently unsupported file types.
    case NSR::file_SpecialBlockDevice:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_SpecialBlockDevice);
        break;
    case NSR::file_SpecialCharacterDevice:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_SpecialCharacterDevice);
        break;
    case NSR::file_ExtendedAttributes:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_ExtendedAttributes);
        break;
    case NSR::file_Fifo:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_Fifo);
        break;
    case NSR::file_Sock:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_Sock);
        break;
    case NSR::file_Unspecified:
    default:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_Unspecified);
        break;
    }

    return 0;
}

static int
readCpio(TreeDB& tree,TreeNode& root,hpACDev *acdev,FreeSpaceAllocator& fsa, 
         UINT32 sectorSize)
{
    int         rval;
    int         i;
    cpioHeader  cpioHdr;
    cpioReader  cpioRdr;
    TreeNode    node,current,newdir;
    static BYTE buf[MAX_IOSIZE];        // Don't reallocate constantly


    if(!cpioRdr.isValid()) {            // Make sure construction worked
        return -1;
    }

    // We must make root first.

    if(tree.allocNode(root) != NSR::error_None) {
        fprintf(stderr,"Tree allocation error, aborting!\n");
        return -1;
    }
    root.attr_.fileType(NSR::file_Directory);
    root.attr_.unixMode(S_IRWXU);
    root.attr_.uid(getuid());
    root.attr_.gid(getgid());
    root.attr_.linkCount(1);
    root.attr_.accessTime(time((time_t *) NULL));
    root.attr_.attributeTime(time((time_t *) NULL));
    root.attr_.modificationTime(time((time_t *) NULL));
    root.filename_ = "/";
    root.setParent(root);
    root.sync();

    // Now iterate until the archive is empty

    while(1) {
        rval = cpioRdr.nextFile(cpioHdr);
        if(rval == CPIO_BAD_HEADER) {
            fprintf(stderr,"Read a bad record, quitting cpio read!\n");
            break;
        }
        if(rval == CPIO_ARCHIVE_EMPTY) {
            fprintf(stdout,"Read last archive record!!!\n\n");
            fflush(stdout);
            break;
        }
        if(!(S_ISDIR(cpioHdr.c_mode) || S_ISREG(cpioHdr.c_mode) ||
             S_ISLNK(cpioHdr.c_mode))) {
            fprintf(stderr,"File %s is not a regular file, directory, or \
symbolic link!\n",cpioHdr.c_filename);
            fprintf(stderr,"\tNOT EXPORTING file %s!\n",cpioHdr.c_filename);
            continue;
        }

        // Now connect the child to its parent.  We do so by parsing its
        // path one component at a time, searching for each component in
        // turn.  We start at root and traverse down the tree one component
        // at a time.

        RWCString       path = cpioHdr.c_filename;
        RWCString       component;
        RWCString       lookahead;

        // We have to get rid of any leading /'s, or else the tokenizer will
        // get confused.

        for(i = 0; i < path.length(); i++) {
            if(path[i] == '/') {
                path.remove(i,1);
                i--;
            }
            else {
                break;
            }
        }

        RWCTokenizer   tokenizer(path);

        current = root;
        current.refresh();
        lookahead = tokenizer("/");
        while(!lookahead.isNull()) {
            TreeNode    tempnode;

            component = lookahead;
            lookahead = tokenizer("/");
            if(lookahead.isNull()) {    // We're to the file itself
                // Add the file to the tree.

                if(tree.allocNode(node) != NSR::error_None) {
                    fprintf(stderr,"Tree allocation error, aborting!\n");
                    return -1;
                }

                cpioToNode(cpioHdr,node);

                // If it's a regular file, we have to copy the data.

                Extent  ext;
                int             count = 0;
                int             len;

                if(node.attr_.fileType() == NSR::file_Raw) {
                    while(1) {
                        len = ((cpioRdr.dataLeft() < MAX_IOSIZE) ? 
                               cpioRdr.dataLeft() : MAX_IOSIZE);
                        if(len == 0) {
                            break;
                        }
                        if(len > fsa.largestFreeExtent()) {
                            fprintf(stderr,"OUT OF SPACE ON OPTICAL, ABORT!!!\n");
                            return -1;          
                        }
                        rval = cpioRdr.readData((char *) buf,len);
                        if(rval <= 0) {
                            fprintf(stderr,"File read ended abnormally, ABORT!!!\n");
                            return -1;
                        }
                        if(fsa.alloc(sectorSize,rval,ext)) {
                            fprintf(stderr,"OUT OF SPACE ON OPTICAL, ABORT!!!\n");
                            return -1;          
                        }

                        // Write to optical.

                        if(acdev->write(ext.start.sid,ext.start.sector,rval,buf) < rval) {
                            fprintf(stderr,"Got a short count writing file %s\n",
                                    cpioHdr.c_filename);
                            fprintf(stderr,"acdev error is %s, aborting data copy\n",
                                    (acdev->error()).message());
                            return -1;
                        }

                        node.elist_.append(ext);
                    }
                }

                // If it's a symbolic link, we read the data and generate a 
                // PathElementList.

                if(node.attr_.fileType() == NSR::file_SymbolicLink) {
                    rval = cpioRdr.readData((char *) buf,MAX_IOSIZE);
                    buf[rval] = 0;        // Null-terminate
                    formPlist((char *) buf,node.plist_);
                }
                node.setParent(current);
                if(node.attr_.fileType() == NSR::file_Directory) {
                    current.attr_.linkCount(current.attr_.linkCount() + 1);
                    current.sync();
                }

                current.addChild(node);
                current.sync();
                node.sync();
                continue;
            }
            if(component == "/") {        // shouldn't happen. . .
                current = root;
                continue;
            }
            if(component == "..") {
                current.getParent(tempnode);
                current = tempnode;
                continue;
            }
            if(component == ".") {
                continue;
            }
            if(current.search((CHAR_PTR) component,tempnode) != 
               NSR::error_None)
            {
                if(tree.allocNode(newdir) != NSR::error_None) {
                    fprintf(stderr,"Tree allocation error, ABORTING!\n");
                    return -1;
                }
                newdir.attr_.fileType(NSR::file_Directory);
                newdir.attr_.unixMode(S_IRWXU);
                newdir.attr_.uid(getuid());
                newdir.attr_.gid(getgid());
                newdir.attr_.linkCount(1);
                newdir.attr_.accessTime(time((time_t *) NULL));
                newdir.attr_.attributeTime(time((time_t *) NULL));
                newdir.attr_.modificationTime(time((time_t *) NULL));
                newdir.filename_ = component;
                newdir.setParent(current);
                newdir.sync();
                current.addChild(newdir);
                current.attr_.linkCount(current.attr_.linkCount() + 1);
                current.sync();
                current = newdir;
            }
            else {
                current = tempnode;
            }
        }

    }        // End of big while loop

    root.refresh();
    return 0;
}

void
cpioToNode(cpioHeader& cpioHdr,TreeNode& node)
{
    int         i;

    for(i = strlen(cpioHdr.c_filename); i >= 0; i--) {
        if(cpioHdr.c_filename[i] == '/') {
            break;
        }
    }

    if(i < 0) {
        node.filename_ = cpioHdr.c_filename;
    }
    else {
        node.filename_ = &(cpioHdr.c_filename[i + 1]);
    }

    if(S_ISDIR(cpioHdr.c_mode)) {
        node.attr_.fileType(NSR::file_Directory);
    }
    if(S_ISREG(cpioHdr.c_mode)) {
        node.attr_.fileType(NSR::file_Raw);
    }
    if(S_ISLNK(cpioHdr.c_mode)) {
        node.attr_.fileType(NSR::file_SymbolicLink);
    }
    node.attr_.unixMode(cpioHdr.c_mode);
    node.attr_.uid(cpioHdr.c_uid);
    node.attr_.gid(cpioHdr.c_gid);
    node.attr_.linkCount(cpioHdr.c_nlink);
    node.attr_.accessTime(time((time_t *) NULL));
    node.attr_.attributeTime(time((time_t *) NULL));
    node.attr_.modificationTime(cpioHdr.modifyTime());

    node.sync();
}

static int
readSynthFile(TreeDB& tree, TreeNode& root,
              hpACDev *acdev,
              FreeSpaceAllocator& fsa, UINT32 sectorSize,
              CHAR_PTR synthScriptName)
{
    SynthFile* synth;

    synth = new SynthFile (tree, root, acdev, fsa, sectorSize,
                     synthScriptName);
    synth->parse();
    synth->returnUnusedSpace();
    if (synth->objectIsValid())
        return 0;
    else
        return 1;
}
