//-----------------------------------------------------------------------------
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//-----------------------------------------------------------------------------

#include <rw/regexp.h>

//-----------------------------------------------------------------------------
// METHOD:  browseShell::browseShell()
//
// Constructor for the browseShell class.
//
// ARGUMENTS:
//
// TreeNode& root   IN   Root of the filesystem tree
//
// RETURNS:  none
//
// PRECONDITIONS:
//
// --The argument root node is the root node of a tree structure describing
//   a fileset.
// --The tree has parent nodes set accurately.
//
// POSTCONDITIONS:
//
// --The browseShell's current working directory will be set to the root.
//-----------------------------------------------------------------------------

browseShell::browseShell(hpACDev *acdev,NSRJukeBox *jb,TreeNode& root) : 
    acdev_(acdev), jbp_(jb), cwd_(root), root_(root)
{

}

//-----------------------------------------------------------------------------
// METHOD:  browseShell::start()
//
// Main entry point to the shell.  Starts up the command interpreter.
//
// ARGUMENTS:  none
// 
// RETURN VALUES:  
//
// Returns 0 upon normal completion, -1 on some error condition.
// 
// PRECONDITIONS:
//
// --Construction was successful (valid tree node root was passed as argument)
//
// POSTCONDITIONS:
//
// --None in particular
//-----------------------------------------------------------------------------

INT32
browseShell::start(void)
{
    INT32       quit = 0;
    INT32       error = 0;
    INT32       argc;
    char        **argv;
    char        buffer[MAX_CMDLINE];

    while(!quit) {
        printf("bsh> ");                                // Display prompt
        if(fgets(buffer,MAX_LINE,stdin) == NULL) {
            printf("Type exit to quit\n");
            continue;
        }
        fflush(stdin);
        if(expandRegExps(buffer,MAX_CMDLINE)) {
            quit = 1;
            continue;
        }
        if(getArgs(buffer,&argc,&argv)) {
            quit = 1;
            continue;
        }
        if(argc == 0) {
            continue;
        }
        if(strcmp(argv[0],"dir") == 0) {
            error = do_dirlist(argc,argv);
            freeArgs(argc,argv);
            if(error) {
                quit = 1;
            }
            continue;
        }
        if(strcmp(argv[0],"cd") == 0) {
            error = do_cd(argc,argv);
            freeArgs(argc,argv);
            if(error) {
                quit = 1;
            }
            continue;
        }
        if(strcmp(argv[0],"pwd") == 0) {
            error = do_pwd();
            freeArgs(argc,argv);
            if(error) {
                quit = 1;
            } 
            continue;
        }
        if(strcmp(argv[0],"vi") == 0) {
            error = do_vi(argc,argv);
            freeArgs(argc,argv);
            if(error) {
                quit = 1;
            }
            continue;
        }
        if(strcmp(argv[0],"restore") == 0) {
            error = do_restore(argc,argv);
            freeArgs(argc,argv);
            if(error) {
                quit = 1;
            }
            continue;
        }
        if(strcmp(argv[0],"info") == 0) {
            error = do_fileInfo(argc,argv);
            freeArgs(argc,argv);
            if(error) {
                quit = 1;
            }
            continue;
        }
        if(strcmp(argv[0],"help") == 0) {
            do_help(argc,argv);
            freeArgs(argc,argv);
            continue;
        }
        if(strcmp(argv[0],"exit") == 0) {
            freeArgs(argc,argv);
            quit = 1;
            continue;
        }
        printf("Huh?\n");
    }

    return error;
}
        
//-----------------------------------------------------------------------------
// METHOD:  browseShell::getArgs()
//
// Gets a command string from stdin and parses into an argc/argv argument list.
//
// My apologies about the triple pointer.  It's a bit heinous.
//
// ARGUMENTS:
//
// INT32 *argc    OUT   Number of arguments, including command
// char  ***argv  OUT   Pointer to array of pointers to argument strings.
//
// RETURN VALUES:
//
// Returns 0 on success, -1 on failure or on EOF.
//
// PRECONDITIONS:
//
// --Both pointers are non-NULL
//
// POSTCONDITIONS:
//
// --argc will be filled in to the number of arguments, including the command
//   itself.
// --memory will be allocated for *argv as well as for the strings containing
//   the individual arguments.
//-----------------------------------------------------------------------------

INT32 browseShell::getArgs(char *buffer,INT32 *argc,char ***argv) 
{ 
    INT32       i,len,inword,index,start,which;
    char        *word;
    

    // We do two passes here.  The first pass is merely to count the number
    // of words so that we can allocated the argv array.  

    len = strlen(buffer);

    // First pass:  count words

    *argc = 0;
    inword = 0;
    for(i = 0; i < len; i++) {
        if(inword) {
            if(isspace(buffer[i])) {
                inword = 0;
                continue;
            }
        }
        else {
            if(!isspace(buffer[i])) {
                inword = 1;
                *argc += 1;
                continue;
            }
        }
    }

    // Allocate argv

    *argv = new (char *)[*argc];
    if(*argv == NULL) {
        return -1;
    }
    for(i = 0; i < *argc; i++) {
        (*argv)[i] = NULL;
    }
        
    // Second pass:  copy arguments into allocated strings and put them into
    // argv array.

    inword = 0;
    which = 0;
    for(i = 0; i < len; i++) {
        if(inword) {
            if(isspace(buffer[i])) {
                word = new char[i - start + 1];
                if(word == NULL) {
                    return -1;
                }
                strncpy(word,&(buffer[start]),i - start);
                word[i - start] = 0;
                (*argv)[which] = word;
                which++;
                assert(*argc >= which);
                inword = 0;
            }
        }
        else {
            if(!isspace(buffer[i])) {
                start = i;
                inword = 1;
            }
        }
    }

    // We're done.

    return 0;
}
                
//-----------------------------------------------------------------------------
// METHOD:  browseShell::freeArgs()
//
// Frees command line arguments.
//-----------------------------------------------------------------------------

void
browseShell::freeArgs(INT32 argc,char **argv)
{
    INT32       i;

    for(i = 0; i < argc; i++) {
        delete [] argv[i];
    }
    delete [] argv;
}

//-----------------------------------------------------------------------------
// METHOD:  browseShell::expandRegExps()
//
// Expands regular expressions into filenames.  This is fairly nontrivial.
//-----------------------------------------------------------------------------

INT32
browseShell::expandRegExps(char *buffer,INT32 buflen)
{
    INT32       i,j,argc,len,match;
    char        **argv;
    char        regexp[MAX_LINE];

    if(getArgs(buffer,&argc,&argv)) {
        return -1;
    }
    if(argc <= 1) {
        return 0;
    }
    sprintf(buffer,"%s ",argv[0]);

    // Each argument might be a regular expression.  If it doesn't match
    // anything at all, we copy the argument unchanged; if it does, we
    // only copy over the matches.  Yuck!  This doesn't take into account
    // quoted arguments and all that kind of stuff.  Shells are a pain.
    // This one is free; you get what you pay for!

    // Right now I'm doing absolutely nothing in way of checking to make
    // sure that all of the expansions fit in the buffer.  This is bad.
    // I'm hoping I'll have rectified the situation before someone else
    // reads this comment and figures (out?) I'm a complete idiot.
    
    for(i = 1; i < argc; i++) {
        match = 0;
        convertRegExp(argv[i],regexp);
        len = strlen(regexp);
        for(j = len; j >= 0; j--) {             // Remove trailing \n's
            if(regexp[j] != '\n') {
                break;
            }
            regexp[j] = 0;
        }
        if(len == 0) {
            strcpy(regexp,".*");
        }
        RWCRegexp       re(regexp);
        RWCString       arg(argv[i]);
        
        Error           err;
        TreeNode        child;

        // This is a performance hit:  we're going through the directory looking
        // for names every time.  We really should read the directory once,
        // tucking away the names somewhere, and go through the somewhere for
        // each regular expression.

        for(j = 0; j < cwd_.getNumChildren(); j++) {
            err = cwd_.getChild(j,child);
            if(err != NSR::error_None) {
                printf("ERROR:  cwd_.getChild(%d,child) failure, error is %s\n",
                       j,err.message());
                return -1;
            }
            if(!child.filename_(re).isNull()) {
                match++;
                strcat(buffer,(CHAR_PTR) child.filename_);
                strcat(buffer," ");
            }
        }
        if(match == 0) {
            strcat(buffer,argv[i]);
            strcat(buffer," ");
        }
    }

    freeArgs(argc,argv);
    return 0;
}
    

//-----------------------------------------------------------------------------
// METHOD:  browseShell::do_dirlist()
//
// Does an ls -l on the current directory.
//
// ARGUMENTS:
//
// INT32  argc   IN   Number of arguments
// char **argv   IN   Array of pointers to arguments
//
// RETURN VALUES:
//
// Returns 0 on success, -1 on failure
//
// PRECONDITIONS:
//
// POSTCONDITIONS:
//-----------------------------------------------------------------------------

INT32
browseShell::do_dirlist(int argc,char **argv)
{
    INT32       i;
    TreeNode    file,child;
    Error       err;

    if(argc == 1) {             // special case. . .
        for(i = 0; i < cwd_.getNumChildren(); i++) {
            err = cwd_.getChild(i,child);
            if(err != NSR::error_None) {
                printf("ERROR:  file.getChild(%d,child) failure, error is %s\n",
                       i,err.message());
                return -1;
            }
            printFile((CHAR_PTR) child.filename_,child.attr_,
                      child.elist_.totalBytes());
        }
        printf("\n");
        return 0;
    }
    for(i = 1; i < argc; i++) {
        if(pathToNode(argv[i],file)) {
            printf("%s:  not found\n",argv[i]);
            continue;
        }
        if(file.attr_.fileType() == NSR::file_Directory) {
            printf("%s:\n\n",argv[i]);
            
            // List all of the files in this directory.

            INT32       j;

            for(j = 0; j < file.getNumChildren(); j++) {
                err = file.getChild(j,child);
                if(err != NSR::error_None) {
                    printf("ERROR:  file.getChild(%d,child) failure, error is %s\n",
                           j,err.message());
                    return -1;
                }
                printFile((CHAR_PTR) child.filename_,child.attr_,
                          child.elist_.totalBytes());
            }
        }
        else {
            printFile((CHAR_PTR) file.filename_,file.attr_,file.elist_.totalBytes());
        }
    }
    printf("\n");
    return 0;
}

//-----------------------------------------------------------------------------
// METHOD:  browseShell::do_cd()
//
// Change current working directory.
//
// ARGUMENTS:
//
// INT32    argc  IN  Number of arguments
// char   **argv  IN  Command-line arguments
//
// RETURN VALUES:
//
// Returns -1 on some tree inconsistency problem, 0 otherwise.
//
// PRECONDITIONS:
// POSTCONDITIONS:
//-----------------------------------------------------------------------------

INT32
browseShell::do_cd(int argc,char **argv)
{
    TreeNode    newNode;

    if(argc > 2) {
        printf("Huh?  Try:  cd [destination_directory]\n");
        return 0;
    }
        
    if(argc == 1) {
        cwd_ = root_;
        do_pwd();
        return 0;
    }

    // We have the destination path.  We have to parse the thing.  Yuck.

    if(pathToNode(argv[1],newNode)) {
        printf("Directory %s not found.\n",argv[1]);
        return 0;
    }
   if(newNode.attr_.fileType() != NSR::file_Directory) {
        printf("File %s is not a directory\n",argv[1]);
        return 0;
    }

    cwd_ = newNode;
    do_pwd();
    return 0;
}

//-----------------------------------------------------------------------------
// METHOD:  browseShell::pathToNode()
//
// Parses a path string and finds the node corresponding to it.
//
// ****** This routine does NOT follow symbolic links! ******
//
// ARGUMENTS:
//
// CHAR_PTR  path  IN   The path to parse
// TreeNode& node  OUT  Node corresponding to last element of path
//
// RETURNS:
//
// Returns 0 on success, -1 if a component of the path does not exist or is
// not a directory.
//
// PRECONDITIONS:
// POSTCONDITIONS:
//-----------------------------------------------------------------------------

INT32
browseShell::pathToNode(CHAR_PTR path,TreeNode& node)
{
    TreeNode    cur,child;
    CHAR_PTR    cptr;
    char        curelt[MAX_LINE];
    INT32       index;

    cptr = path;
    if(*cptr == '/') {          // Absolute path
        cur = root_;
        while(*cptr == '/') cptr++;             // Removing preceding /'s
    }
    else {
        cur = cwd_;
    }

    // Start consuming path elements.  

    while(1) {
        index = 0;
        while((!isspace(*cptr)) && (*cptr != 0) && (*cptr != '/')) {
            curelt[index] = *cptr;
            cptr++;
            index++;
        }
        if(index == 0) {
            break;
        }
        curelt[index] = 0;              // null-terminate

        // We should now have a path element in curelt.

        if(cur.search(curelt,child) != NSR::error_None) {
            if(strcmp(curelt,".") == 0) {
                child = cur;
            }
            else {
                if(strcmp(curelt,"..") == 0) {
                    cur.getParent(child);
                }
                else {
                    return -1;
                }
            }
        }

        cur = child;
        if((*cptr == 0) || isspace(*cptr)) {            // We're done.
            break;
        }

        if(child.attr_.fileType() != NSR::file_Directory) {
            printf("Path component %s is not a directory!\n",curelt);
            return -1;
        }

        // We found a directory node, good.


        while(*cptr == '/') cptr++;             // Remove trailing /'s
    }

    node = cur;
    return 0;
}

//-----------------------------------------------------------------------------
// METHOD:  browseShell::do_vi()
//
// Vi's the specified file by restoring it to a temporary file.  This is a
// virtual vi, as the file won't actually be modified on optical.
//-----------------------------------------------------------------------------

INT32
browseShell::do_vi(int argc,char **argv)
{
    INT32       i,j,rval,fd,nfiles = 0;
    sid_t       sid;
    char        command[MAX_CMDLINE];
    CHAR_PTR    *tmpname = new (char *)[argc - 1];
    TreeNode    node;


    command[0] = 0;
    strcat(command,argv[0]);
    strcat(command," ");
    
    // We accept any vi options.  This is a bit tricky.  Since I don't offhand
    // actually _know_ what all the vi options are, any argument that doesn't
    // match to a filename I assume will be an option argument to vi.  

    // In this for-loop, we find any files in the argument list and
    // restore them to /tmp somewhere.

    for(i = 1; i < argc; i++) {
    
        // Get the file's node.

        if(pathToNode(argv[i],node)) {
            strcat(command,argv[i]);
            strcat(command," ");
            continue;
        }

        // Restore the file from optical to /tmp somewhere.

        tmpname[nfiles] = tempnam("/tmp","bsh");
        if(tmpname[nfiles] == NULL) {
            printf("tempnam() failed!\n");
            return 0;
        }
        fd = creat(tmpname[nfiles],0777);
        if(fd < 0) {
            printf("open failed\n");
            delete [] tmpname;
            return 0;
        }

        // Iterate through the extents and copy the data.

        INT32 offset = 0;
        for(j = 0; j < node.elist_.getCount(); j++) {
            Extent& ext = (node.elist_)[j];

            // Have to restore data.  

            rval = copyExtent(fd,*acdev_,ext.start.sid,offset,ext);
            if(rval) {
                printf("copyExtent() failed, aborting!\n");
                close(fd);
                return 0;
            }
            offset += ext.len;
        }
        close(fd);              // File is now restored.
        strcat(command,tmpname[nfiles]);
        strcat(command," ");
        nfiles++;
    }

    // Vi the file

    system(command);

    // Nuke the files

    for(i = 0; i < nfiles; i++) {
        unlink(tmpname[i]);
        free(tmpname[i]);
    }
    delete [] tmpname;
    return 0;
}

//-----------------------------------------------------------------------------
// METHOD:  browseShell::do_fileInfo()
//
// Prints UDF-related information about a file.
//-----------------------------------------------------------------------------

INT32
browseShell::do_fileInfo(int argc,char **argv)
{
    int         i,j;
    TreeNode    curnode;
    Extent      *extp;

    if(argc < 2)
    {
        printf("Usage:  info [filelist]\n");
        return 0;
    }
    for(i = 1; i < argc; i++)
    {
        if(pathToNode(argv[i],curnode))
        {
            printf("%s:  not found\n",argv[i]);
            continue;
        }

        // Okay, we have the file's node.  Print out the node.

        curnode.print(NULL);
    }
    return 0;
}
        
//-----------------------------------------------------------------------------
// METHOD:  browseShell::do_pwd()
//
// Print current working directory.
//-----------------------------------------------------------------------------

INT32
browseShell::do_pwd(void)
{
    INT32       i = 0,j,len;
    char        path[MAX_LINE];
    TreeNode    parent;
    TreeNode    nodePath[MAX_NODES];

    if(!equalNodes(cwd_,root_)) {
        nodePath[i++] = cwd_;
        while(1) {
            nodePath[i-1].getParent(parent);
            if(equalNodes(parent,root_)) {
                break;
            }                
            nodePath[i++] = parent;
        }
    }

    strcpy(path,"/");
    for(j = i - 1; j >= 0; j--) {
        strcat(path,(char *) nodePath[j].filename_);
        strcat(path,"/");
    }
    len = strlen(path);
    if(len > 1) {
        path[len - 1] = 0;
    }
    printf("The current working directory is %s\n",path);
    return 0;
}

//-----------------------------------------------------------------------------
// METHOD:  browseShell::help()
//
// Prints some help information.
//-----------------------------------------------------------------------------

void
browseShell::do_help(int argc,char **argv)
{
    static char *help_str[] = {
"Browse shell commands:\n",
"dir                           Directory (ls -l) listing of current directory",
"pwd                           Print current working directory",
"cd [dir]                      Change current working directory",
"vi [file]                     Browse file using vi",
"restore [srcdir] [hpux_dir]   Restore subtree to native file system dir.",
"info [file]                   Prints UDF-related information about the file.",
"                              hpux_dir takes the form:  hpux:/dir1/dir2",
"exit                          Exit the browse shell",
"help                          This help list\n",
NULL };
    int i = 0;

    while(help_str[i] != NULL) {
        printf("%s\n",help_str[i]);
        i++;
    }
}

//-----------------------------------------------------------------------------
// METHOD:  browseShell::do_restore()
//
// Restores a subtree to a directory in the native file system.  
//
// Usage:  restore [srcdir] [hpux_dir]
//
// [srcdir] is the source directory in the virtual browseShell filesystem;
// [hpux_dir] is the directory on the native filesystem, and must take the
// form:
//
// hpux:[pathname]
//-----------------------------------------------------------------------------

INT32
browseShell::do_restore(int argc,char **argv)
{
    TreeNode    src;
    char        hpuxdir[MAX_LINE];

    if(argc != 3) {
        printf("restore usage:\n\n");
        printf("restore [srcdir] [hpux_dir]\n\n");
        printf("[srcdir]:  source dir in virtual browseShell filesystem\n");
        printf("[hpux_dir]:  destination dir in native FS\n");
        printf("             Must take the form:  hpux:pathname\n");
        return 0;
    }

    // Try to get the source directory's node.

    if(pathToNode(argv[1],src)) {
        printf("%s:  not found\n",argv[1]);
        return 0;
    }
    src.filename_ = "/";
    src.address_.sector = ROOT_ICB_ADDRESS;

    // Now verify that the user specified an HP-UX directory, and if so,
    // strip off the "hpux:" and verify that we have access.

    if(strncmp(argv[2],"hpux:",5) != 0) {
        printf("[hpux_dir] arg must take the form:  hpux:pathname!\n");
        return 0;
    }
    strcpy(hpuxdir,argv[2] + 5);        // This is pretty kludgy. . .

    if(access(hpuxdir,R_OK | W_OK | X_OK | F_OK)) {
        printf("[hpux_dir] %s is not accessible, sorry!\n",argv[2]);
        return 0;
    }

    // Okay, we should be ready, in theory, to do the restore.

    VolsetNameMapList   vsnmList;
    VolsetNameMap *     vsnmptr;
    RWCString           vsname(jbp_->wvsName());
    SidList             sidList;
    
    vsnmList = jbp_->volsetNameMapList();
    vsnmptr = vsnmList.search(vsname);
    sidList = vsnmptr->sidList_;
    
    restoreFiles(src,*acdev_,hpuxdir,sidList);
    return 0;
}

//-----------------------------------------------------------------------------
// METHOD:  browseShell::equalNodes()
//
// Determine whether two nodes are equal.  
//
// This function is necessary because there currently is no operator== defined
// for the TreeNode class.  We consider two nodes equal if they have the same
// ICB address.  Note that this will only work for trees that have been
// constructed by an import from NSR, since trees constructed from the native
// FS do not use the ICB address field at all.
//
// ARGUMENTS:
//
// TreeNode& n1  IN  First node to compare
// TreeNode& n2  IN  Second node to compare
//
// RETURNS VALUES:
//
// Returns 1 if the nodes are equal, 0 otherwise.
//
// PRECONDITIONS:
//
// --Both nodes have been initialized; see above note about the definition
//   of node equality.
//
// POSTCONDITIONS:  see return values
//-----------------------------------------------------------------------------

INT32
browseShell::equalNodes(TreeNode& n1,TreeNode& n2)
{
    if((n1.address_.sid == n2.address_.sid) &&
       (n1.address_.sector == n2.address_.sector)) {
        return 1;
    }
    else {
        return 0;
    }
}

//-----------------------------------------------------------------------------
// METHOD:  browseShell::convertRegExp()
//
// This method converts from a standard shell-like regular expression to the
// form used by the RogueWave RWCRegExp class.  
//
// The primary difference is that shell regexp's are implicitly anchored
// and that the * syntax means 'match zero or more of any character.'.
//
// There is a lot of icky C-style char pointer arithmetic stuff going on here.
// I was in a hurry.
//
// ARGUMENTS:
//
// char *src     IN  Source string containing the original regexp
// char *regexp  OUT Character buffer to contain the converted regexp
//
// RETURN VALUES:  none
//
// PRECONDITIONS:
//
// --Both strings are non-NULL
// --The destination buffer regexp must be large enough to hold the converted
//   string. 
//
// POSTCONDITIONS:
//
// --The converted string will be copied into the buffer pointed to by
//   regexp.
//-----------------------------------------------------------------------------

void
browseShell::convertRegExp(char *src,char *regexp)
{
    char *s,*re;

    s = src;
    re = regexp;
    assert((s != NULL) && (re != NULL));
    
    // ".." is a special case.  We have to convert it to \.\. 

    if(strcmp(src,"..") == 0) {
        strcpy(regexp,"\\.\\.");
        return;
    }
   
    // Basically, prepend a '^' to anchor the expression, then convert any
    // '*' to .*.

    *re++ = '^';

    while(*s) {
        if(*s == '*') {
            *re++ = '.';
            *re++ = '*';
            s++;
        }
        else {
            *re++ = *s++;
        }
    }
    *re++ = '$';     // Anchor expression
    *re = 0;         // Null terminate
}

//-----------------------------------------------------------------------------
// FUNCTION:  setParent()
//
// Recursive function which sets the parent pointers for all the nodes in a 
// tree.  This is necessary because currently the importArchive() function does
// not set the parent pointers when building the tree.
//
// ARGUMENTS:
//
// TreeNode&  parent  IN   Parent of current node
// TreeNode&  node    OUT  Current node
//
// RETURN VALUES:  none
//
// PRECONDITIONS:
//
// --both nodes are valid, allocated nodes.
//
// POSTCONDITIONS:
//
// --The node will have its parent value set to parent, and all of its 
//   children will (recursively) have their parent values set appropriately.
//-----------------------------------------------------------------------------

void    setParent(TreeNode& parent,TreeNode& node)
{
    INT32       i;
    TreeNode    child;

    node.setParent(parent);
    node.sync();
    if(node.attr_.fileType() == NSR::file_Directory) {
        for(i = 0; i < node.getNumChildren(); i++) {
            node.getChild(i,child);
            setParent(node,child);
            
        }
    }
}
