//-----------------------------------------------------------------------------
// cpioHdr.cpp                                                  --JKH
//
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//
//  CLASSES:
//   cpioHeader 
//-----------------------------------------------------------------------------
//
// $Source: /projects/HMS/nsr/udfAncil/msTest/cpioHdr.cpp,v $
// $Revision: 1.3 $     $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 95/10/25 17:27:13 $
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include "cpio.h"

//-----------------------------------------------------------------------------
// METHOD:  cpioHeader::operator=()                             --JKH
//
// Assignment operator between two cpioHeader objects.
//
// ARGUMENTS:
//
// const cpioHeader& hdr  IN  object to copy.
//
// RETURN VALUES:   None.
//
// PRECONDITIONS:  None.
// POSTCONDITIONS:
//
// --The current object will be assigned equal to the argument.
//-----------------------------------------------------------------------------

cpioHeader&
cpioHeader::operator=(const cpioHeader& hdr)
{
    c_magic = hdr.c_magic;
    c_dev = hdr.c_dev;
    c_ino = hdr.c_ino;
    c_mode = hdr.c_mode;
    c_uid = hdr.c_uid;
    c_gid = hdr.c_gid;
    c_nlink = hdr.c_nlink;
    c_rdev = hdr.c_rdev;
    c_mtime[0] = hdr.c_mtime[0];
    c_mtime[1] = hdr.c_mtime[1];
    c_namesize = hdr.c_namesize;
    c_filesize[0] = hdr.c_filesize[0];
    c_filesize[1] = hdr.c_filesize[1];
    memcpy(c_filename,hdr.c_filename,CPIO_MAXNAME);

    return *this;
}

//-----------------------------------------------------------------------------
// METHOD:  cpioHeader::modifyTime()                            --JKH
//
// Sets and gets the modify time.
//
// ARGUMENTS:
//
// None
//
// OR
//
// int  newMtime  IN  The new modify time.
//
// RETURN VALUES:  Returns the modify time OR void
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:  None.
//-----------------------------------------------------------------------------

int
cpioHeader::modifyTime(void) const
{
    return ((unsigned) c_mtime[1] + (unsigned) (c_mtime[0] << 16));
}

void
cpioHeader::modifyTime(int newMtime)
{
    c_mtime[0] = ((newMtime >> 16) & 0xFFFF);
    c_mtime[1] = (newMtime & 0xFFFF);
}

//-----------------------------------------------------------------------------
// METHOD:  cpioHeader::filesize()                              --JKH
//
// Sets and gets the filesize
//
// ARGUMENTS:
//
// None
//
// OR
//
// int  newfilesize  IN  The new filesize
//
// RETURN VALUES:  Returns the filesize OR void
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:  None.
//-----------------------------------------------------------------------------

int
cpioHeader::fileSize(void) const
{
    return ((unsigned) c_filesize[1] + (unsigned) (c_filesize[0] << 16));
}

void
cpioHeader::fileSize(int newfilesize)
{
    c_filesize[0] = ((newfilesize >> 16) & 0xFFFF);
    c_filesize[1] = (newfilesize & 0xFFFF);
}
