//-----------------------------------------------------------------------------
// cpioWrit.cpp                                                 --JKH
//
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//
//  CLASSES:
//   cpioWriter
//-----------------------------------------------------------------------------
//
// $Source: /projects/HMS/nsr/udfAncil/msTest/cpioWrit.cpp,v $
// $Revision: 1.2 $     $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 95/10/25 17:27:15 $
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include "cpio.h"

#define PIPESIZE        4096

//-----------------------------------------------------------------------------
// METHOD:  cpioWriter::cpioWriter()                            --JKH
//
// Constructor for the cpioWriter class.
//
// ARGUMENTS:
//
// int type              IN  Type of archive (ascii or binary)
// const char *cpioFile  IN  Pathname of cpio archive to read
//
// RETURN VALUES:   None.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:
//
// --The argument archive file will be opened for write.
//-----------------------------------------------------------------------------

cpioWriter::cpioWriter(int type,const char *cpioFile) : fd_(-1),
                                                        residue_(0),
                                                        odd_(0),
                                                        isbinary_(0),
                                                        open_(0),
                                                        count_(0)
{
    if(cpioFile == NULL) {                      // Write to stdout
        fd_ = fileno(stdout);
    }
    else {
        fd_ = creat(cpioFile,S_IRWXU | S_IRWXG | S_IRWXO);
    }
    if(fd_ >= 0) {
        open_ = 1;
    }
    if(type != CPIO_ASCII) {
        isbinary_ = 1;
    }
}

//-----------------------------------------------------------------------------
// METHOD:  cpioWriter::~cpioWriter()                           --JKH
//
// Destructor for the cpioReader class.
//
// ARGUMENTS:  None.
//
// RETURN VALUES:   None.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:
//
// --The archive file descriptor, if open, will be closed.
//-----------------------------------------------------------------------------

cpioWriter::~cpioWriter()
{
    if(fd_ < 0) {
        return;
    }
    if(open_) {
        close(fd_);
    }
}

//-----------------------------------------------------------------------------
// METHOD:  cpioWriter::writeHeader()                           --JKH
//
// Writes the user's cpioHeader to the output stream.
//
// ARGUMENTS:
//
// cpioHeader&  hdr  IN   User's struct to be written
//
// RETURN VALUES: 
//
// Returns an integer with value 0 upon success, -1 on error 
//
// PRECONDITIONS:
//
// --The cpioReader object was successfully opened.
// --The previous file's data was completely written.
//
// POSTCONDITIONS:
//-----------------------------------------------------------------------------

int
cpioWriter::writeHeader(const cpioHeader& hdr)
{
    int         rval;
    int         len;
    int         index = 0;
    char        buffer[CPIO_MAXNAME];
    short       *sptr;
    char        *cptr;

    if(fd_ < 0) {
        return -1;
    }

    if(residue_ > 0) {
        return -1;
    }

    odd_ = 0;

    if(isbinary_) {                                     // Write binary header
        sptr = (short *) buffer;
        *sptr++ = CPIO_MAGIC;
        *sptr++ = hdr.c_dev;
        *sptr++ = hdr.c_ino;
        *sptr++ = hdr.c_mode;
        *sptr++ = hdr.c_uid;
        *sptr++ = hdr.c_gid;
        *sptr++ = hdr.c_nlink;
        *sptr++ = hdr.c_rdev;
        *sptr++ = hdr.c_mtime[0];
        *sptr++ = hdr.c_mtime[1];
        *sptr++ = hdr.c_namesize;
        *sptr++ = hdr.c_filesize[0];
        *sptr++ = hdr.c_filesize[1];

        // Now write the filename.

        cptr = (char *) sptr;
        memcpy(cptr,hdr.c_filename,hdr.c_namesize);
        cptr += hdr.c_namesize;
        if(hdr.c_namesize % 2) {
            *cptr++ = 0x00;                     // Pad by one byte
        }
        len = ((unsigned) cptr) - ((unsigned) buffer);
    }
    else {                                              // Write ASCII header
        if(sprintf(buffer,"%06ho%06ho%06ho%06ho%06ho%06ho%06ho%06ho%011lo%06ho%011lo",
               CPIO_MAGIC,hdr.c_dev,hdr.c_ino,hdr.c_mode,
               hdr.c_uid,hdr.c_gid,hdr.c_nlink,hdr.c_rdev,
               hdr.modifyTime(),hdr.c_namesize,hdr.fileSize()) < 0)
        {
            return -1;
        }

        if(sprintf(&(buffer[ASCII_HEADER_LENGTH]),"%s%c",hdr.c_filename,0)<0) {
            return -1;
        }

        len = ASCII_HEADER_LENGTH + hdr.c_namesize;
    }

    // At this point, we've filled in the header, and are ready to receive
    // file data, so write the header out to the stream.

    rval = Write(fd_,buffer,len);
    if(rval < len) {
        return -1;
    }

    residue_ = hdr.fileSize();
    odd_ = (hdr.fileSize() % 2);

    return 0;
}

//-----------------------------------------------------------------------------
// METHOD:  cpioWriter::writeData()                             --JKH
//
// Writes file data to the output stream.
//
// ARGUMENTS:
//
// const char * buf     OUT  User's buffer.
// int          nbytes  IN   Number of bytes to write.
//
// RETURN VALUES: 
//
// Returns the number of bytes written.
//
// PRECONDITIONS:
//
// --The cpioWritten object was successfully opened.
// --The file header has already been written by a previous call to 
//   writeHeader().
// --All of the file data has not already been written.
//
// POSTCONDITIONS:
//
// --The user's buffer will be written to the output stream.
// --The amount of data written will be nbytes or residue, whichever is less.
//-----------------------------------------------------------------------------

int
cpioWriter::writeData(const char *buf,int nbytes)
{
    int         len,rval;

    if(fd_ < 0) {
        return -1;
    }
    if(residue_ == 0) {
        return 0;
    }
    if(nbytes < residue_) {
        len = nbytes;
    }
    else {
        len = residue_;
    }

    rval = Write(fd_,buf,len);
    if(rval > 0) {
        residue_ -= rval;
    }
    if(residue_ == 0) {
        if(isbinary_ && odd_) {         // Pad the file data
            Write(fd_,buf,1);
        }
    }
    return rval;
}

//-----------------------------------------------------------------------------
// METHOD:  cpioWriter::reInit()                                --JKH
//
// Reinitializes the writer to connect to a different file.
//
// ARGUMENTS:
//
// const char * cpioFile  IN  New file to open.
//
// RETURN VALUES: 
//
// Returns 0 on success, -1 on failure.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:
//
// --If the new file is opened successfully, the old one will be closed.
// --If the new file is not opened successfully, the old one will be left open.
//-----------------------------------------------------------------------------

int
cpioWriter::reInit(int type,const char *cpioFile)
{
    int         fd;

    if(cpioFile == NULL) {
        fd = fileno(stdout);
    }
    else {
        fd = open(cpioFile,O_RDONLY);
    }
    if(fd < 0) {
        return -1;
    }

    close(fd_);
    fd_ = fd;
    residue_ = 0;
    odd_ = 0;
    isbinary_ = (type ? 1 : 0);
    open_ = 1;
    count_ = 0;

    return 0;
}

//-----------------------------------------------------------------------------
// METHOD:  cpioWriter::closeArchive()                          --JKH
//
// Writes the trailer record and closes the output file.
//
// ARGUMENTS:  None.
// RETURN VALUES:  0 on success, -1 on failure.
// PRECONDITIONS:
//
// --The cpioWriter object was successfully opened.
//
// POSTCONDITIONS:
//
// --See above.
//-----------------------------------------------------------------------------

int
cpioWriter::closeArchive()
{
    cpioHeader  cpioHdr;
    char        buffer[512];

    if(open_ == 0) {                    // Archive is already closed.
        return 0;
    }

    // Set up a trailer record.

    cpioHdr.c_magic = CPIO_MAGIC;
    cpioHdr.c_dev = 0;
    cpioHdr.c_ino = 0;
    cpioHdr.c_mode = 0;
    cpioHdr.c_uid = 0;
    cpioHdr.c_gid = 0;
    cpioHdr.c_nlink = 0;
    cpioHdr.c_rdev = 0;
    cpioHdr.c_mtime[0] = 0;
    cpioHdr.c_mtime[1] = 0;
    cpioHdr.c_namesize = strlen(LAST_RECORD_NAME) + 1;
    cpioHdr.c_filesize[0] = 0;
    cpioHdr.c_filesize[1] = 0;
    strcpy(cpioHdr.c_filename,LAST_RECORD_NAME);

    // Write it.

    if(writeHeader(cpioHdr)) {
        return -1;
    }

    // Pad to a 512 byte boundary.

    if(count_ % 512) {
        memset(buffer,0,512);
        Write(fd_,buffer,512 - (count_ % 512));
    }

    // Close the output file if it's not stdout.

    if(fd_ != fileno(stdout)) {
        close(fd_);
    }
    fd_ = -1;
    odd_ = 0;
    residue_ = 0;
    isbinary_ = 0;
    open_ = 0;
    count_ = 0;

    return 0;
}

//-----------------------------------------------------------------------------
// METHOD:  cpioReader::Write()                                 --JKH
//
// Supplements the write(2) system call for reading from the stdin stream.
// If writing to stdout, this call iterates until all desired bytes have
// been written.  Note that this can lead to a blocked process if the stdout
// stream does not read all the bytes.
//
// ARGUMENTS:
//
// int          fd      IN   File descriptor.
// const char * buf     OUT  User's buffer.
// int          nbytes  IN   Number of bytes to write.
//
// RETURN VALUES: 
//
// Returns the number of bytes written.
//
// PRECONDITIONS:
//
// --The cpioReader object was successfully opened.
//
// POSTCONDITIONS:
//
// --The user's buffer will be written to the current output file.
//-----------------------------------------------------------------------------

int
cpioWriter::Write(int fd,const char *buf,int nbytes)
{
    int    count = 0;
    int    rval;
    int    len;

    if(fd == fileno(stdout)) {
        while(count < nbytes) {
            if(count > PIPESIZE) {
                len = PIPESIZE;
            }
            else {
                len = nbytes;
            }
            rval = ::write(fd,buf + count,len);
            if(rval > 0) {
                count += rval;
            }
        }
        if(count > 0) {
            count_ += rval;
        }
        return count;
    }
    else {
        rval = ::write(fd,buf,nbytes);
        if(rval > 0) {
            count_ += rval;
        }
        return rval;
    }
}
