/****************************************************************************
** hash.h:  header file for doubly-linked list routines.
**
**      Unpublished Confidential Information of Hewlett-Packard
**      Company. Do not disclose.  Copyright (c) Hewlett-Packard
**      Company 1995. All rights reserved
** 
** The contents of this software are proprietary and confidential to
** the Hewlett- Packard Company, and are limited in distribution to
** those with a direct need to know.  Individuals having access to
** this software are responsible for maintaining the confidentiality
** of the content and for keeping the software secure when not in use.
** Transfer to any party is strictly forbidden other than as expressly
** permitted in writing by Hewlett-Packard Company.  Unauthorized
** transfer to or possession by any unauthorized party may be a
** criminal offense.
**
**                  RESTRICTED RIGHTS LEGEND
**
**      Use,  duplication,  or disclosure by the U. S. Government is
**      subject to restrictions as set forth in subparagraph (c) (1)
**      (ii) in the Rights in Technical Data and Computer Software
**      clause in DFARS 252.227-7013 or any other successor clause.
**
**                      HEWLETT-PACKARD COMPANY
**                      3000 Hanover St.
**                      Palo Alto, CA  94304 U.S.A.
**
**      Rights for non-DOD U.S. Government Departments and Agencies
**      are as set forth in FAR 52.227-19 (c) (1,2) or any other
**      successor clause.
**
*****************************************************************************/

#ifndef _HASH_H_
#define _HASH_H_

#include "dlist.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Error codes:  hash_err */

#define HERR_NOERR      DLERR_NOERR     /* No error */
#define HERR_NULL       DLERR_NULL      /* Invalid NULL arg. to function */
#define HERR_NOMEM      DLERR_NOMEM     /* Malloc failed */
#define HERR_NOTFND     DLERR_NOTFND    /* Data not found */
#define HERR_APPLF      DLERR_APPLF     /* Apply function failed */
#define HERR_BADARG     100             /* Bad argument */

#ifdef __STDC__
typedef struct hash_table
{
        int                             capacity;
        dlist                           **table;
        int                             count;
        int                             (*hash_func)(void *);
        int                             hash_err;
} hash_table;
#else
typedef struct hash_table
{
        int                             capacity;
        dlist                           **table;
        int                             count;
        int                             (*hash_func)();
        int                             hash_err;
} hash_table;
#endif

/*
** The hash_table function declarations.
**
*/

#ifdef __STDC__

hash_table *    htableCreate(int size,int (*cmp_func)(void *,void *), 
                             int (*hash_func)(void *));
int             htableHash(hash_table *table, void *datap);
void *          htableFind(hash_table *table, void *datap);
void *          htableDelete(hash_table *table, void *datap);
int             htableDestroy(hash_table *table);
int             htableApply(hash_table *table, int (*apply_func)(void *),
                            int failsafe);

#else

hash_table *    htableCreate();
int             htableHash();
void *          htableFind();
void *          htableDelete();
int             htableDestroy();
int             htableApply();

#endif 

/*
** The hash table macro function(s)
*/

#define         htableCount(table)      (((table) == (hash_table *) NULL) ? \
                                            0 : ((table)->count))

#define         htableError(table)      (((table) == (hash_table *) NULL) ? \
                                            HERR_NULL : ((table)->hash_err))

#ifdef __cplusplus
}
#endif

#endif          /* _HASH_H_ */
