//-*-c++-*-
//-----------------------------------------------------------------------------
// nsrJB.cpp
//
// This file contains implementations for the nsrJukeBox class and its
// associated ancillary classes.
//
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//
// CLASSES:
//
// SurfaceMapList
// SidList
// AvailSurfacesList
// VolsetNameMapList
// FilesetNameMapList
// VolumeFreeSpaceList
// NSRJukeBox
//
//-----------------------------------------------------------------------------
// $Source: /projects/HMS/nsr/udfAncil/msTest/nsrJB.cpp,v $
// $Revision: 1.27 $    $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 96/05/01 00:03:27 $
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/stat.h>
#include <rw/cstring.h>
#include <rw/ctoken.h>
#include "arcImEx.h"
#include "nsrJB.h"
#include "hash.h"
#include "dqueue.h"

#define MAX_LINE        512
#define MAX_DRIVES      12      // Good enough for SolarFlare with 4x. . .
#define IS_VALID_SECTOR_SIZE(ss) (((ss) == 512) || ((ss) == 1024))

extern int trace;
static int      TreeNode_cmp(void *d1,void *d2);
static int      TreeNode_hash(void *d1);

//-----------------------------------------------------------------------------
// METHOD:  SurfaceMapList::search()                            --JKH
//
// This method searches the SurfaceMapList for an entry with the given
// slot/side or sid information.
//
// ARGS:
//
// UINT16 slot  IN      Which slot in jukebox (starting with 1)
// UINT16 side  IN      Which side (0 = 'A', 1 = 'B')
//
// OR
//
// Sid_t sid    IN      Which Sid_t
//
// OR
//
// const RWCString volname IN  Which volume name 
//
// RETURNS:
//
// Returns a pointer to the SurfaceMap object with the given slot/side 
// or Sid_t information on success, NULL on failure.
//
// PRECONDITIONS:
//
// None.
//
// POSTCONDITIONS:
//
// See return values
//-----------------------------------------------------------------------------

SurfaceMap *
SurfaceMapList::search(UINT16 slot,UINT16 side)
{
    INT32       i;

    for(i = 0; i < theList_.entries(); i++) {
        SurfaceMap *smptr = theList_.at(i);
        if((smptr->slot_ == slot) && (smptr->side_ == side)) {
            return smptr;
        }
    }
    return NULL;
}

SurfaceMap *
SurfaceMapList::search(Sid_t sid)
{
    INT32       i;

    for(i = 0; i < theList_.entries(); i++) {
        SurfaceMap *smptr = theList_.at(i);
        if(smptr->sid_ == sid) {
            return smptr;
        }
    }
    return NULL;
}

SurfaceMap *
SurfaceMapList::search(const RWCString& volname)
{
    INT32       i;

    for(i = 0; i < theList_.entries(); i++) {
        SurfaceMap *smptr = theList_.at(i);
        if(smptr->volname_ == volname) {
            return smptr;
        }
    }
    return NULL;
}

//-----------------------------------------------------------------------------
// METHOD:  SurfaceMapList::at()                                --JKH
//
// Returns a pointer to the specified item given by the argument index
// (beginning at 0).
//
// ARGUMENTS:
//
// UINT32       index   IN  Index of desired list item.
//
// RETURNS:  
//
// Returns a pointer to the desired item on success, NULL on error (such as
// an index out of range).
// 
// PRECONDITIONS: 
//
// --Index is in the range [0,# of items in list)
//
// POSTCONDITIONS:
//
// --See return values
//-----------------------------------------------------------------------------

SurfaceMap *
SurfaceMapList::at(UINT32 index)
{
    if(index >= entries()) {
        return NULL;
    }
    return theList_.at(index);
}

const SurfaceMap *
SurfaceMapList::at(UINT32 index) const
{
    if(index >= entries()) {
        return NULL;
    }
    return theList_.at(index);
}

//-----------------------------------------------------------------------------
// METHOD:  SurfaceMapList::append()                            --JKH
//
// Adds a copy of the argument item to the list.
//
// ARGUMENTS:
//
// const SurfaceMap&    smap   IN  Item to add to the list
//
// RETURNS:  None.
// 
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:
//
// --A copy of the item will be added to the list.
//-----------------------------------------------------------------------------

void
SurfaceMapList::append(const SurfaceMap& smap)
{
    SurfaceMap  *smptr = new SurfaceMap;

    if(smptr == NULL) {
        return;
    }

    *smptr = smap;
    theList_.append(smptr);
}

//-----------------------------------------------------------------------------
// METHOD:  SurfaceMapList::remove()                            --JKH
//
// Removes the item specified by the given index from the list.
//
// ARGUMENTS:
//
// UINT32  index   IN   Index of item to remove.
//
// RETURNS:  None.
// 
// PRECONDITIONS:  
//
// --index must be in the range [0,# of items in list)
//
// POSTCONDITIONS:
//
// --The specified item will be removed and its memory freed.
//-----------------------------------------------------------------------------

void
SurfaceMapList::remove(UINT32 index)
{
    if(index >= entries()) {
        return;
    }
    SurfaceMap *smptr = theList_.removeReference(theList_.at(index));
    delete smptr;
}

//-----------------------------------------------------------------------------
// METHOD:  SurfaceMapList::copy()                              --JKH
//
// Performs a deep copy of one SurfaceMapList to another.
//
// ARGUMENTS:
//
// const SurfaceMapList&  obj   IN  Source list for copy.
//
// RETURNS:  None.
// 
// PRECONDITIONS:  None
//
// POSTCONDITIONS:
//
// --The current list will be set equal to the source list.
//-----------------------------------------------------------------------------

void
SurfaceMapList::copy(const SurfaceMapList& obj)
{
    if(&obj == this) {          // Trying to overwrite myself!
        return;
    }

    clear();                    // Clear myself out.

    INT32       i;

    for(i = 0; i < obj.entries(); i++) {
        const SurfaceMap        *smptr = obj.at(i);

        if(smptr == NULL) {     // This is an error case.  Bail out.
            clear();
            return;
        }
        append(*smptr);
    }
}

//-----------------------------------------------------------------------------
// METHOD:  SurfaceMapList::free()                              --JKH
//
// Clears out the contents of a SurfaceMapList.
//
// ARGUMENTS:  None.
//
// RETURNS:  None.
// 
// PRECONDITIONS:  None
//
// POSTCONDITIONS:
//
// --The current list will be freed.
//-----------------------------------------------------------------------------

void
SurfaceMapList::free()
{
    SurfaceMap  *smptr;
    INT32       i;

    for(i = 0; i < entries(); i++) {
        smptr = at(i);
        delete smptr;
    }
    theList_.clear();
}

//-----------------------------------------------------------------------------
// METHOD:  SidList::search()                                   --JKH
//
// This method searches the SidList for an entry with the given sid 
// information.
//
// ARGS:
//
// Sid_t sid    IN      Which Sid_t
//
// RETURNS:
//
// Returns a pointer to the Sid_t object with the given value, NULL on failure.
//
// PRECONDITIONS:
//
// None.
//
// POSTCONDITIONS:
//
// See return values
//-----------------------------------------------------------------------------

Sid_t *
SidList::search(Sid_t sid)
{
    INT32       i;

    for(i = 0; i < entries(); i++) {
        Sid_t   *sptr = theList_.at(i);

        if(sptr == NULL) {              // Problem, NULL entry in list!
            return NULL;
        }

        if(*sptr == sid) {
            return sptr;
        }
    }
    return NULL;
}

//-----------------------------------------------------------------------------
// METHOD:  SidList::operator==()                               --JKH
//
// Equality operator for the SidList class.
//
// In this definition, two SidLists are considered equal if each contains
// the same list of sids--that is, every sid in the right-hand list is in
// the left-hand list.  NOTE:  order is NOT a criterion.  Since SidLists are
// not ordered, the operator== is therefore an O(mn) operation, where n
// is the number of items in the right-hand list, and m is the number of
// items in the left-hand list, and thus should be used with care.
//
// ARGS:
//
// const SidList& slist   IN   List to check against
//
// RETURN VALUES:
//
// Returns 1 if the lists are equal, 0 otherwise.
//
// PRECONDITIONS:   None.
//
// POSTCONDITIONS:  None.
//-----------------------------------------------------------------------------

INT32
SidList::operator==(const SidList& slist)
{
    INT32       i,j,found;

    for(i = 0; i < slist.entries(); i++) {
        const Sid_t     *sptr = slist.at(i);
        found = 0;
        if(sptr == NULL) {              // Problem, NULL entry in list!
            return 0;
        }
        for(j = 0; j < entries(); j++) {
            Sid_t       *sptr2 = theList_.at(j);
            if(*sptr == *sptr2) {
                found = 1;
                break;
            }
        }
        if(!found) {
            return 0;
        }
    }
    return 1;
}

//-----------------------------------------------------------------------------
// METHOD:  SidList::at()                               --JKH
//
// Returns a pointer to the specified item given by the argument index
// (beginning at 0).
//
// ARGUMENTS:
//
// UINT32       index   IN  Index of desired list item.
//
// RETURNS:  
//
// Returns a pointer to the desired item on success, NULL on error (such as
// an index out of range).
// 
// PRECONDITIONS: 
//
// --Index is in the range [0,# of items in list)
//
// POSTCONDITIONS:
//
// --See return values
//-----------------------------------------------------------------------------

Sid_t *
SidList::at(UINT32 index)
{
    if(index >= entries()) {
        return NULL;
    }
    return theList_.at(index);
}

const Sid_t *
SidList::at(UINT32 index) const
{
    if(index >= entries()) {
        return NULL;
    }
    return theList_.at(index);
}

//-----------------------------------------------------------------------------
// METHOD:  SidList::append()                           --JKH
//
// Appends an item to the list.
//
// ARGUMENTS:
//
// const Sid_t  sid   IN  Sid to append to the list.
//
// RETURNS:  None.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:
//
// --An sid will be allocated and appended to the list.
//-----------------------------------------------------------------------------

void
SidList::append(const Sid_t& sid)
{
    Sid_t       *sptr = new Sid_t;

    if(sptr == NULL) {
        return;
    }
    *sptr = sid;
    theList_.append(sptr);
}

//-----------------------------------------------------------------------------
// METHOD:  SidList::remove()                                   --JKH
//
// Removes an item from the list.
//
// ARGUMENTS:
//
// UINT32       index   IN  Index of item to remove.
//
// RETURNS:  None.
//
// PRECONDITIONS:  
//
// --The index is in the range [0,nentries)
//
// POSTCONDITIONS:
//
// --The item will be removed and its memory freed.
//-----------------------------------------------------------------------------

void
SidList::remove(UINT32 index)
{
    if(index >= entries()) {
        return;
    }
    Sid_t *sptr = this->theList_.removeReference(this->theList_.at(index));
    delete sptr;
}

//-----------------------------------------------------------------------------
// METHOD:  SidList::copy()                                     --JKH
//
// Performs a deep copy of one SurfaceMapList to another.
//
// ARGUMENTS:
//
// const SidList&  obj   IN  Source list for copy.
//
// RETURNS:  None.
// 
// PRECONDITIONS:  None
//
// POSTCONDITIONS:
//
// --The current list will be set equal to the source list.
//-----------------------------------------------------------------------------

void
SidList::copy(const SidList& obj)
{
    if(&obj == this) {          // Trying to overwrite myself!
        return;
    }

    clear();                    // Clear myself out.

    INT32       i;

    for(i = 0; i < obj.entries(); i++) {
        const Sid_t     *sidptr = obj.at(i);

        if(sidptr == NULL) {    // This is an error case.  Bail out.
            clear();
            return;
        }
        append(*sidptr);
    }
}

//-----------------------------------------------------------------------------
// METHOD:  SidList::free()                                     --JKH
//
// Clears out the contents of a SidList.
//
// ARGUMENTS:  None.
//
// RETURNS:  None.
// 
// PRECONDITIONS:  None
//
// POSTCONDITIONS:
//
// --The current list will be freed.
//-----------------------------------------------------------------------------

void
SidList::free()
{
    Sid_t       *sptr;
    INT32       i;

    for(i = 0; i < entries(); i++) {
        sptr = at(i);
        delete sptr;
    }
    theList_.clear();
}

//-----------------------------------------------------------------------------
// METHOD:  AvailSurfacesList::search()                         --JKH
//
// This method searches the AvailSurfacesList for an entry with the given
// sectorsize and media type information.
//
// ARGS:
//
// UINT32     sectorSize  IN  Sector size of media, in bytes
// MediaType  mediatype   IN  Type of media (MO or WORM)
//
// RETURNS:
//
// Returns a pointer to the AvailSurfaces object with the given sector size
// and media type information on success, NULL on failure.
//
// PRECONDITIONS:
//
// None.
//
// POSTCONDITIONS:
//
// See return values
//-----------------------------------------------------------------------------

AvailSurfaces *
AvailSurfacesList::search(UINT32 sectorSize,MediaType mediatype)
{
    INT32       i;

    for(i = 0; i < theList_.entries(); i++) {
        AvailSurfaces *asptr = theList_.at(i);
        if((asptr->sectorSize_ == sectorSize) && 
           (asptr->mtype_ == mediatype)) {
            return asptr;
        }
    }
    return NULL;
}

//-----------------------------------------------------------------------------
// METHOD:  AvailSurfacesList::at()                             --JKH
//
// Returns a pointer to the specified item given by the argument index
// (beginning at 0).
//
// ARGUMENTS:
//
// UINT32       index   IN  Index of desired list item.
//
// RETURNS:  
//
// Returns a pointer to the desired item on success, NULL on error (such as
// an index out of range).
// 
// PRECONDITIONS: 
//
// --Index is in the range [0,# of items in list)
//
// POSTCONDITIONS:
//
// --See return values
//-----------------------------------------------------------------------------

AvailSurfaces *
AvailSurfacesList::at(UINT32 index)
{
    if(index >= entries()) {
        return NULL;
    }
    return theList_.at(index);
}

const AvailSurfaces *
AvailSurfacesList::at(UINT32 index) const
{
    if(index >= entries()) {
        return NULL;
    }
    return theList_.at(index);
}

//-----------------------------------------------------------------------------
// METHOD:  AvailSurfacesList::append()                         --JKH
//
// Appends an item to the list.
//
// ARGUMENTS:
//
// const AvailSurfaces& as   IN  Sid to append to the list.
//
// RETURNS:  None.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:
//
// --An sid will be allocated and appended to the list.
//-----------------------------------------------------------------------------

void
AvailSurfacesList::append(const AvailSurfaces& as)
{
    AvailSurfaces       *asptr = new AvailSurfaces;

    if(asptr == NULL) {
        return;
    }
    *asptr = as;
    theList_.append(asptr);
}

//-----------------------------------------------------------------------------
// METHOD:  AvailSurfacesList::remove()                         --JKH
//
// Removes an item from the list.
//
// ARGUMENTS:
//
// UINT32       index   IN  Index of item to remove.
//
// RETURNS:  None.
//
// PRECONDITIONS:  
//
// --The index is in the range [0,nentries)
//
// POSTCONDITIONS:
//
// --The item will be removed and its memory freed.
//-----------------------------------------------------------------------------

void
AvailSurfacesList::remove(UINT32 index)
{
    if(index >= entries()) {
        return;
    }
    AvailSurfaces *asptr = theList_.removeReference(theList_.at(index));
    delete asptr;
}

//-----------------------------------------------------------------------------
// METHOD:  AvailSurfacesList::copy()                           --JKH
//
// Performs a deep copy of one AvailSurfaceList to another.
//
// ARGUMENTS:
//
// const AvailSurfacesList&  obj   IN  Source list for copy.
//
// RETURNS:  None.
// 
// PRECONDITIONS:  None
//
// POSTCONDITIONS:
//
// --The current list will be set equal to the source list.
//-----------------------------------------------------------------------------

void
AvailSurfacesList::copy(const AvailSurfacesList& obj)
{
    if(&obj == this) {          // Trying to overwrite myself!
        return;
    }

    clear();                    // Clear myself out.

    INT32       i;

    for(i = 0; i < obj.entries(); i++) {
        const AvailSurfaces     *asptr = obj.at(i);

        if(asptr == NULL) {     // This is an error case.  Bail out.
            clear();
            return;
        }
        append(*asptr);
    }
}

//-----------------------------------------------------------------------------
// METHOD:  AvailSurfacesList::free()                           --JKH
//
// Clears out the contents of a AvailSurfacesList.
//
// ARGUMENTS:  None.
//
// RETURNS:  None.
// 
// PRECONDITIONS:  None
//
// POSTCONDITIONS:
//
// --The current list will be freed.
//-----------------------------------------------------------------------------

void
AvailSurfacesList::free()
{
    AvailSurfaces       *asptr;
    INT32               i;

    for(i = 0; i < entries(); i++) {
        asptr = at(i);
        delete asptr;
    }
    theList_.clear();
}

//-----------------------------------------------------------------------------
// METHOD:  VolsetNameMapList::search()                         --JKH
//
// This method searches the VolsetNameMapList for an entry with the given
// volume set name.
//
// ARGS:
//
// const RWCString&  vsname IN  Volume set name to search for
//
// RETURNS:
//
// Returns a pointer to the VolsetNameMap object with the given volume set
// name on success, NULL on failure.
//
// PRECONDITIONS:
//
// None.
//
// POSTCONDITIONS:
//
// See return values
//-----------------------------------------------------------------------------

VolsetNameMap *
VolsetNameMapList::search(const RWCString& vsname)
{
    INT32       i;

    for(i = 0; i < theList_.entries(); i++) {
        VolsetNameMap *vnmptr = theList_.at(i);
        if(vnmptr->vsname_ == vsname) {
            return vnmptr;
        }
    }
    return NULL;
}

//-----------------------------------------------------------------------------
// METHOD:  VolsetNameMapList::at()                             --JKH
//
// Returns a pointer to the specified item given by the argument index
// (beginning at 0).
//
// ARGUMENTS:
//
// UINT32       index   IN  Index of desired list item.
//
// RETURNS:  
//
// Returns a pointer to the desired item on success, NULL on error (such as
// an index out of range).
// 
// PRECONDITIONS: 
//
// --Index is in the range [0,# of items in list)
//
// POSTCONDITIONS:
//
// --See return values
//-----------------------------------------------------------------------------

VolsetNameMap *
VolsetNameMapList::at(UINT32 index)
{
    if(index >= entries()) {
        return NULL;
    }
    return theList_.at(index);
}

const VolsetNameMap *
VolsetNameMapList::at(UINT32 index) const
{
    if(index >= entries()) {
        return NULL;
    }
    return theList_.at(index);
}

//-----------------------------------------------------------------------------
// METHOD:  VolsetNameMapList::append()                         --JKH
//
// Appends an item to the list.
//
// ARGUMENTS:
//
// const VolsetNameMap& as   IN  Sid to append to the list.
//
// RETURNS:  None.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:
//
// --A VolsetNameMap will be allocated and appended to the list.
//-----------------------------------------------------------------------------

void
VolsetNameMapList::append(const VolsetNameMap& vnmap)
{
    VolsetNameMap       *vnmptr = new VolsetNameMap(vnmap.vsname_);

    if(vnmptr == NULL) {
        return;
    }
    *vnmptr = vnmap;
    theList_.append(vnmptr);
}

//-----------------------------------------------------------------------------
// METHOD:  VolsetNameMapList::remove()                         --JKH
//
// Removes an item from the list.
//
// ARGUMENTS:
//
// UINT32       index   IN  Index of item to remove.
//
// RETURNS:  None.
//
// PRECONDITIONS:  
//
// --The index is in the range [0,nentries)
//
// POSTCONDITIONS:
//
// --The item will be removed and its memory freed.
//-----------------------------------------------------------------------------

void
VolsetNameMapList::remove(UINT32 index)
{
    if(index >= entries()) {
        return;
    }
    VolsetNameMap *vsnmptr = theList_.removeReference(theList_.at(index));
    delete vsnmptr;
}

//-----------------------------------------------------------------------------
// METHOD:  VolsetNameMapList::copy()                           --JKH
//
// Performs a deep copy of one VolsetNameMapList to another.
//
// ARGUMENTS:
//
// const VolsetNameMapList&  obj   IN  Source list for copy.
//
// RETURNS:  None.
// 
// PRECONDITIONS:  None
//
// POSTCONDITIONS:
//
// --The current list will be set equal to the source list.
//-----------------------------------------------------------------------------

void
VolsetNameMapList::copy(const VolsetNameMapList& obj)
{
    if(&obj == this) {          // Trying to overwrite myself!
        return;
    }

    clear();                    // Clear myself out.

    INT32       i;

    for(i = 0; i < obj.entries(); i++) {
        const VolsetNameMap     *vnmptr = obj.at(i);

        if(vnmptr == NULL) {    // This is an error case.  Bail out.
            clear();
            return;
        }
        append(*vnmptr);
    }
}

//-----------------------------------------------------------------------------
// METHOD:  VolsetNameMapList::free()                           --JKH
//
// Clears out the contents of a VolsetNameMapList.
//
// ARGUMENTS:  None.
//
// RETURNS:  None.
// 
// PRECONDITIONS:  None
//
// POSTCONDITIONS:
//
// --The current list will be freed.
//-----------------------------------------------------------------------------

void
VolsetNameMapList::free()
{
    VolsetNameMap       *vnmptr;
    INT32               i;

    for(i = 0; i < entries(); i++) {
        vnmptr = at(i);
        delete vnmptr;
    }
    theList_.clear();
}


//-----------------------------------------------------------------------------
// METHOD:  FilesetNameMapList::search()                                --JKH
//                              searchByVSName()
//
// This method searches the FilesetNameMapList for an entry with the given
// fileset name, or by volume set name.
//
// NOTES:
//
// Since there is a many-to-one relationship between file set names and
// volume sets (i.e., one volume set can have many filesets), searchByVSName()
// finds only the FIRST occurence of the specified volume set name.
//
// ARGS:
//
// const RWCString&  fsname IN  Fileset name to search for
//                   vsname IN  Volume set name to search for.
//
// RETURNS:
//
// Returns a pointer to the FilesetNameMap object with the given fileset
// or volume set name on success, NULL on failure.
//
// PRECONDITIONS:
//
// None.
//
// POSTCONDITIONS:
//
// See return values
//-----------------------------------------------------------------------------

FilesetNameMap *
FilesetNameMapList::search(const RWCString& fsname)
{
    INT32       i;

    for(i = 0; i < theList_.entries(); i++) {
        FilesetNameMap *fnmptr = theList_.at(i);
        if(fnmptr->fsname_ == fsname) {
            return fnmptr;
        }
    }
    return NULL;
}

FilesetNameMap *
FilesetNameMapList::searchByVSName(const RWCString& vsname)
{
    INT32       i;

    for(i = 0; i < theList_.entries(); i++) {
        FilesetNameMap *fnmptr = theList_.at(i);
        if(fnmptr->vsname_ == vsname) {
            return fnmptr;
        }
    }
    return NULL;
}

//-----------------------------------------------------------------------------
// METHOD:  FilesetNameMapList::at()                            --JKH
//
// Returns a pointer to the specified item given by the argument index
// (beginning at 0).
//
// ARGUMENTS:
//
// UINT32       index   IN  Index of desired list item.
//
// RETURNS:  
//
// Returns a pointer to the desired item on success, NULL on error (such as
// an index out of range).
// 
// PRECONDITIONS: 
//
// --Index is in the range [0,# of items in list)
//
// POSTCONDITIONS:
//
// --See return values
//-----------------------------------------------------------------------------

FilesetNameMap *
FilesetNameMapList::at(UINT32 index)
{
    if(index >= entries()) {
        return NULL;
    }
    return theList_.at(index);
}

const FilesetNameMap *
FilesetNameMapList::at(UINT32 index) const
{
    if(index >= entries()) {
        return NULL;
    }
    return theList_.at(index);
}

//-----------------------------------------------------------------------------
// METHOD:  FilesetNameMapList::append()                        --JKH
//
// Appends an item to the list.
//
// ARGUMENTS:
//
// const FilesetNameMap&  fsnmap   IN  Item to append to the list
//
// RETURNS:  None.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:
//
// --A FilesetNameMap will be allocated and appended to the list.
//-----------------------------------------------------------------------------

void
FilesetNameMapList::append(const FilesetNameMap& fsnmap)
{
    FilesetNameMap      *fsnmptr = new FilesetNameMap(fsnmap);

    if(fsnmptr == NULL) {
        return;
    }
    *fsnmptr = fsnmap;
    theList_.append(fsnmptr);
}

//-----------------------------------------------------------------------------
// METHOD:  FilesetNameMapList::remove()                        --JKH
//
// Removes an item from the list.
//
// ARGUMENTS:
//
// UINT32       index   IN  Index of item to remove.
//
// RETURNS:  None.
//
// PRECONDITIONS:  
//
// --The index is in the range [0,nentries)
//
// POSTCONDITIONS:
//
// --The item will be removed and its memory freed.
//-----------------------------------------------------------------------------

void
FilesetNameMapList::remove(UINT32 index)
{
    if(index >= entries()) {
        return;
    }
    FilesetNameMap *fsnmptr = theList_.removeReference(theList_.at(index));
    delete fsnmptr;
}

//-----------------------------------------------------------------------------
// METHOD:  FilesetNameMapList::copy()                          --JKH
//
// Performs a deep copy of one FilesetNameMapList to another.
//
// ARGUMENTS:
//
// const FilesetNameMapList&  obj   IN  Source list for copy.
//
// RETURNS:  None.
// 
// PRECONDITIONS:  None
//
// POSTCONDITIONS:
//
// --The current list will be set equal to the source list.
//-----------------------------------------------------------------------------

void
FilesetNameMapList::copy(const FilesetNameMapList& obj)
{
    if(&obj == this) {          // Trying to overwrite myself!
        return;
    }

    clear();                    // Clear myself out.

    INT32       i;

    for(i = 0; i < obj.entries(); i++) {
        const FilesetNameMap    *fsnmptr = obj.at(i);

        if(fsnmptr == NULL) {   // This is an error case.  Bail out.
            clear();
            return;
        }
        append(*fsnmptr);
    }
}

//-----------------------------------------------------------------------------
// METHOD:  FilesetNameMapList::free()                          --JKH
//
// Clears out the contents of a FilesetNameMapList.
//
// ARGUMENTS:  None.
//
// RETURNS:  None.
// 
// PRECONDITIONS:  None
//
// POSTCONDITIONS:
//
// --The current list will be freed.
//-----------------------------------------------------------------------------

void
FilesetNameMapList::free()
{
    FilesetNameMap      *fsnmptr;
    INT32               i;

    for(i = 0; i < entries(); i++) {
        fsnmptr = at(i);
        delete fsnmptr;
    }
    theList_.clear();
}

//-----------------------------------------------------------------------------
// METHOD:  VolumeFreeSpaceList::search()                               --JKH
//
// This method searches the VolumeFreeSpaceList for an entry with the given
// sid.
//
// ARGS:
//
// Sid_t  sid IN  sid to search for
//
// RETURNS:
//
// Returns a pointer to the VolumeFreeSpace object with the given Sid_t
// on success, NULL on failure.
//
// PRECONDITIONS:
//
// None.
//
// POSTCONDITIONS:
//
// See return values
//-----------------------------------------------------------------------------

VolumeFreeSpace *
VolumeFreeSpaceList::search(Sid_t sid)
{
    INT32       i;

    for(i = 0; i < theList_.entries(); i++) {
        VolumeFreeSpace *vfsptr = theList_.at(i);
        if(vfsptr->sid_ == sid) {
            return vfsptr;
        }
    }
    return NULL;
}

//-----------------------------------------------------------------------------
// METHOD:  VolumeFreeSpaceList::at()                           --JKH
//
// Returns a pointer to the specified item given by the argument index
// (beginning at 0).
//
// ARGUMENTS:
//
// UINT32       index   IN  Index of desired list item.
//
// RETURNS:  
//
// Returns a pointer to the desired item on success, NULL on error (such as
// an index out of range).
// 
// PRECONDITIONS: 
//
// --Index is in the range [0,# of items in list)
//
// POSTCONDITIONS:
//
// --See return values
//-----------------------------------------------------------------------------

VolumeFreeSpace *
VolumeFreeSpaceList::at(UINT32 index)
{
    if(index >= entries()) {
        return NULL;
    }
    return theList_.at(index);
}

const VolumeFreeSpace *
VolumeFreeSpaceList::at(UINT32 index) const
{
    if(index >= entries()) {
        return NULL;
    }
    return theList_.at(index);
}

//-----------------------------------------------------------------------------
// METHOD:  VolumeFreeSpaceList::append()                       --JKH
//
// Appends an item to the list.
//
// ARGUMENTS:
//
// const VolumeFreeSpace&  vfs   IN  Item to append to the list
//
// RETURNS:  None.
//
// PRECONDITIONS:  None.
//
// POSTCONDITIONS:
//
// --A VolumeFreeSpace object will be allocated and appended to the list.
//-----------------------------------------------------------------------------

void
VolumeFreeSpaceList::append(const VolumeFreeSpace& vfs)
{
    VolumeFreeSpace     *vfsptr = new VolumeFreeSpace;

    if(vfsptr == NULL) {
        return;
    }
    *vfsptr = vfs;
    theList_.append(vfsptr);
}

//-----------------------------------------------------------------------------
// METHOD:  VolumeFreeSpaceList::remove()                       --JKH
//
// Removes an item from the list.
//
// ARGUMENTS:
//
// UINT32       index   IN  Index of item to remove.
//
// RETURNS:  None.
//
// PRECONDITIONS:  
//
// --The index is in the range [0,nentries)
//
// POSTCONDITIONS:
//
// --The item will be removed and its memory freed.
//-----------------------------------------------------------------------------

void
VolumeFreeSpaceList::remove(UINT32 index)
{
    if(index >= entries()) {
        return;
    }
    VolumeFreeSpace *vfsptr = theList_.removeReference(theList_.at(index));
    delete vfsptr;
}

void
VolumeFreeSpaceList::print(void)
{
    INT32               i;
    VolumeFreeSpace     *vfsptr;

    for(i = 0; i < theList_.entries(); i++) {
        vfsptr = theList_.at(i);
        vfsptr->print();
    }
}

//-----------------------------------------------------------------------------
// METHOD:  VolumeFreeSpaceList::copy()                         --JKH
//
// Performs a deep copy of one VolumeFreeSpaceList to another.
//
// ARGUMENTS:
//
// const VolumeFreeSpaceList&  obj   IN  Source list for copy.
//
// RETURNS:  None.
// 
// PRECONDITIONS:  None
//
// POSTCONDITIONS:
//
// --The current list will be set equal to the source list.
//-----------------------------------------------------------------------------

void
VolumeFreeSpaceList::copy(const VolumeFreeSpaceList& obj)
{
    if(&obj == this) {          // Trying to overwrite myself!
        return;
    }

    clear();                    // Clear myself out.

    INT32       i;

    for(i = 0; i < obj.entries(); i++) {
        const VolumeFreeSpace   *vfsptr = obj.at(i);

        if(vfsptr == NULL) {    // This is an error case.  Bail out.
            clear();
            return;
        }
        append(*vfsptr);
    }
}

//-----------------------------------------------------------------------------
// METHOD:  VolumeFreeSpaceList::free()                         --JKH
//
// Clears out the contents of a VolumeFreeSpaceList.
//
// ARGUMENTS:  None.
//
// RETURNS:  None.
// 
// PRECONDITIONS:  None
//
// POSTCONDITIONS:
//
// --The current list will be freed.
//-----------------------------------------------------------------------------

void
VolumeFreeSpaceList::free()
{
    VolumeFreeSpace     *vfsptr;
    INT32               i;

    for(i = 0; i < entries(); i++) {
        vfsptr = at(i);
        delete vfsptr;
    }
    theList_.clear();
}

//-----------------------------------------------------------------------------
// METHOD:  NSRJukeBox::NSRJukeBox()                            --JKH
//
// Constructor for the NSRJukebox class.
//
// This method takes as argument an RWCString containing the pathname of an
// acrc file, which is expected to contain lines of the following syntax:
//
// acdev = <pathname>
// drive <uint> = <pathname>
//
// where <pathname> is the pathname of a character special device file (in
// the case of the acdev rule, that of a passthru device file associated with
// the jukebox's controller; in the case of the drives, that of a disk
// device file).
//
// ARGUMENTS:
//
// const RWCString& acrcFile    IN  Filename of acrc configuration file.
//
// RETURNS:  none
//
// PRECONDITIONS:
//
// --The specified acrcFile exists and is accessible (readable)
// --The device files specified in the acrc file exist and are accessible
//
// POSTCONDITIONS:
//
// --An hpACDev object will be created bound to the autochanger device 
//   specified in the acrc file.
// --Element addresses and element status information will be acquired from
//   the autochanger device and all surfaces mapped to Sid_t's, and a
//   SurfaceMapList created.
//
//-----------------------------------------------------------------------------

NSRJukeBox::NSRJukeBox(const RWCString& acrcFile) : arc_(NULL),
                                                    acdev_(NULL),
                                                    openIntegrity_(NSR::False),
                                                    newvs_(NSR::False),
                                                    err_(NSR::error_None)
{
    if(makeACDev(acrcFile)) {
        return;
    }

    // acdev_ should now be valid.  Make the map table.

    if(mapSurfaces()) {
        return;
    }

    // We now have a map table.  We're done.
}

NSRJukeBox::~NSRJukeBox()
{
    delete acdev_;
    delete arc_;
}

//-----------------------------------------------------------------------------
// METHOD:  NSRJukeBox::readStatusFile()                        --JKH
//
// Reads and parses the specified status file, and generates mappings lists
// to keep track of volume and volumeset information.
//
// ARGUMENTS:  None.
//
// RETURNS:
//
// Returns an error object (see error values below).
//
// PRECONDITIONS:
//
// --The nsrJukeBox has been successfully constructed.
//
// POSTCONDITIONS:
//
// --The various lists will be filled out according to the information given
//   in the status file.
//
// SIDE EFFECTS:
//
// --Upon error, this method sets the err_ variable.
//
// ERRORS:
//
// error_None                           Operation completed successfully.
//-----------------------------------------------------------------------------

Error
NSRJukeBox::readStatusFile(RWCString& statusFile)
{
    INT32               i;
    FILE *              infile;
    FilesetNameMapList  fsnmList;
    VolsetNameMapList   vsnmList;
    AvailSurfacesList   asList;

    err_ = NSR::error_None;
    if(acdev_ == NULL) {                // Construction failed
        return err_;
    }

    if(statusFile.isNull()) {
        err_ = NSR::error_Null_status_file;
        return err_;
    }

    // REPLACE THIS CODE WITH REAL PARSER CODE.
    // This code is pretty much a horrendous kludge, just to get something
    // in place for right now.  

    infile = fopen((CHAR_PTR) statusFile,"r");
    if(infile == NULL) {
        err_ = NSR::error_Cannot_access_status_file;
        return err_;
    }

    char        buffer[MAX_LINE];
    char        word[MAX_LINE];
    CHAR_PTR    cptr;
    UINT32      which;

    while(1) {
        if(fgets(buffer,MAX_LINE,infile) == NULL) {
            break;
        }
        if(sscanf(buffer,"%s",word) != 1) {
            continue;
        }
        if(strcmp(word,"#") == 0) {
            continue;
        }

        RWCString       line(buffer);
        MyTokenizer     next(line);
        RWCString       token;

        // Get the first token.

        token = next();
        if(token.isNull()) {
            continue;
        }

        // The syntax we're looking for here is:
        //
        // volset <vsname> = <surface_list>

        if(strcmp((CHAR_PTR) token,"volset") == 0) {
            token = next();     // token is now the volset name
            if(token.isNull()) {
                err_ = NSR::error_Status_file_syntax_error;
                fclose(infile);
                return err_;
            }
            processIdentifier(token,IN);
            VolsetNameMap       vsnmap(token);

            // Verify that the next token is '=' and skip it.

            token = next();
            if(strcmp((CHAR_PTR) token,"=") != 0) {
                err_ = NSR::error_Status_file_syntax_error;
                fclose(infile);
                return err_;
            }

            // We should now just have a list of surfaces.

            char        c;
            UINT32      slot;
            UINT16      side;

            while(!(token = next()).isNull()) {
                if(sscanf((CHAR_PTR)token,"%u%c",&slot,&c) != 2) {
                    err_ = NSR::error_Status_file_syntax_error;
                    fclose(infile);
                    return err_;
                }
                if((c == 'a') || (c == 'A')) {
                    side = 0;
                }
                else {
                    side = 1;
                }

                // We got a valid surface specifier.  Find its sid, if 
                // possible, and add it to the VolsetNameMapList.

                SurfaceMap      *smptr = smList_.search(slot,side);
                if(smptr == NULL) {
                    err_ = NSR::error_Invalid_surface;
                    fclose(infile);
                    return err_;
                }

                vsnmap.sidList_.append(smptr->sid_);
            }
            if(vsnmap.sidList_.entries() == 0) {
                err_ = NSR::error_Status_file_syntax_error;
                fclose(infile);
                return err_;
            }

            // Add the vsnmap to the temporary vsnmList.

            vsnmList.append(vsnmap);
            continue;
        }

        // The syntax we're looking for here is:
        //
        // fileset <fsname> in <vsname>

        if(strcmp((CHAR_PTR) token,"fileset") == 0) {
            if((token = next()).isNull()) {
                err_ = NSR::error_Status_file_syntax_error;
                fclose(infile);
                return err_;
            }

            // token contains <fsname>.  Make a FilesetNameMap.

            processIdentifier(token,IN);
            FilesetNameMap      fsnmap(token);

            // Get the "in" keyword

            if((token = next()).isNull()) {
                err_ = NSR::error_Status_file_syntax_error;
                fclose(infile);
                return err_;
            }

            if(strcmp((CHAR_PTR) token,"in") != 0) {
                err_ = NSR::error_Status_file_syntax_error;
                fclose(infile);
                return err_;
            }

            // Get the <vsname>

            if((token = next()).isNull()) {
                err_ = NSR::error_Status_file_syntax_error;
                fclose(infile);
                return err_;
            }

            processIdentifier(token,IN);
            fsnmap.vsname_ = token;

            // Add it to the temporary FilesetNameMapList.

            
            fsnmList.append(fsnmap);
            continue;
        }

        // The syntax we're looking for here is:
        //
        // avail <sector_size> <media_type> = <surface_list>

        if(strcmp((CHAR_PTR) token,"avail") == 0) {
            AvailSurfaces       avail;

            if((token = next()).isNull()) {
                err_ = NSR::error_Status_file_syntax_error;
                fclose(infile);
                return err_;
            }

            avail.sectorSize_ = atoi((CHAR_PTR) token);
            if(!IS_VALID_SECTOR_SIZE(avail.sectorSize_)) {
                err_ = NSR::error_Status_file_syntax_error;
                fclose(infile);
                return err_;
            }
                 
            if((token = next()).isNull()) {
                err_ = NSR::error_Status_file_syntax_error;
                fclose(infile);
                return err_;
            }

            if(strcmp((CHAR_PTR) token,"worm") == 0) {
                avail.mtype_ = NSR::media_Worm;
            }
            else {
                if(strcmp((CHAR_PTR) token,"mo") == 0) {
                    avail.mtype_ = NSR::media_Rewritable;
                }
                else {
                    err_ = NSR::error_Status_file_syntax_error;
                    fclose(infile);
                    return err_;
                }
            }

            if((token = next()).isNull()) {
                err_ = NSR::error_Status_file_syntax_error;
                fclose(infile);
                return err_;
            }

            if(strcmp((CHAR_PTR) token,"=") != 0) {
                err_ = NSR::error_Status_file_syntax_error;
                fclose(infile);
                return err_;
            }

            // We should now just have a list of surfaces.

            char        c;
            UINT32      slot;
            UINT16      side;

            while(!(token = next()).isNull()) {
                if(sscanf((CHAR_PTR) token,"%u%c",&slot,&c) != 2) {
                    err_ = NSR::error_Status_file_syntax_error;
                    fclose(infile);
                    return err_;
                }
                if((c == 'a') || (c == 'A')) {
                    side = 0;
                }
                else {
                    side = 1;
                }

                // We got a valid surface specifier.  Find its sid, if 
                // possible, and add it to the VolsetNameMapList.

                SurfaceMap      *smptr = smList_.search(slot,side);
                if(smptr == NULL) {
                    err_ = NSR::error_Invalid_surface;
                    fclose(infile);
                    return err_;
                }

                avail.sidList_.append(smptr->sid_);
            }
            if(avail.sidList_.entries() == 0) {
                err_ = NSR::error_Status_file_syntax_error;
                fclose(infile);
                return err_;
            }

            // Append the AvailSurfaces to the temporary list.

            asList.append(avail);
            continue;
        }
    }
    fclose(infile);

    // At this point, in theory, we will have read the entire status file
    // and parsed it, and filled in some lists.  We have to do some 
    // consistency checks to make sure the information we read from the
    // file isn't ridiculous.  Namely, we want to make sure that the
    // items in the FilesetNameMap correspond to volume sets listed in
    // the VolsetNameMapList.  We also verify that the sid's in the
    // available lists aren't in the VolsetNameMapList.

    // CHECKS NOT YET IMPLEMENTED
    
    // Everything's consistent.  Save off the jukebox status file name
    // and reset the lists.

    jbstatFile_ = statusFile;
    asList_ = asList;
    vnmList_ = vsnmList;
    fnmList_ = fsnmList;

    return err_;
}

void 
NSRJukeBox::processIdentifier(RWCString& id,INT32 which)
{
    UINT32      i;

    // We're trying to strip out any leading and trailing quotation marks.

    if(which == IN) {
        for(i = 0; i < id.length(); i++) {
            if(id[i] == '\"') {
                if(i > 0) {
                    if(id[i - 1] == '\\') {
                        id.remove(i - 1,1);
                        i--;
                        continue;
                    }
                }
                id.remove(i,1);
                i--;
                continue;
            }
            if(id[i] == '\\') {
                if((id[i + 1] == '\"') || (id[i + 1] == '\\')) {
                    continue;
                }
                if(i > 0) {
                    if(id[i - 1] == '\\') {
                        id.remove(i - 1,1);
                        i--;
                        continue;
                    }
                }
                id.remove(i,1);
                i--;
                continue;
            }
        }
    }
    else {
        for(i = 0; i < id.length(); i++) {
            if(id[i] == '\"') {
                id.insert(i,"\\");
                i++;
            }
            if(id[i] == '\\') {
                id.insert(i,"\\");
                i++;
            }
        }
    }
}

//-----------------------------------------------------------------------------
// METHOD:  NSRJukeBox::genStatusFile()                         --JKH
//
// Catalogs the contents of the jukebox, determining which volumes belong
// to which volumesets, and generates a status file describing the state of
// the jukebox.
//
// ARGUMENTS:
//
// RCWString&  statusFile   IN  Name of the status file to be written.
//
// RETURNS:
//
// Returns an error object (see Errors below).
//
// PRECONDITIONS:
//
// --If statusFile is null, then jbStatFile_ must be non-Null.
//
// POSTCONDITIONS:   See above.
//
// ERRORS:
//
// error_None                   Operation completed successfully.
// error_Null_volume_set_name   Null volume set name passed.
// error_Name_not_unique        Specified name is not unique.
//
// ALGORITHM NOTES:
//
// This method performs a getVolumeInfo() call on every surface in the
// jukebox.  It adds any non-NSR volume to the appropriate available list.
//-----------------------------------------------------------------------------

Error
NSRJukeBox::genStatusFile(RWCString& statusFile)
{
    INT32       i,j;
    SurfaceMap  *smptr;
    Archive     *arcp;

    // First clear all of the database lists.  

    asList_.clear();
    vnmList_.clear();
    fnmList_.clear();
    vfsList_.clear();

    arcp = new Archive;
    if(arcp == NULL) {
        err_ = NSR::error_No_free_store;
        return err_;
    }

    // Set the Archive object to do appropriate tracing.
    if (trace == 0) {
        arcp->logFile(NULL);
        arcp->logErrors(NSR::False);
        arcp->logCalls(NSR::False);
        arcp->logAlgorithms(NSR::False);
    } else {
        arcp->logFile(stderr);
        arcp->logErrors(NSR::True);
#ifdef MULT_TRACE_LEVELS
        if (trace > 1) {
            arcp->logAlgorithms(NSR::True);
        }
        if (trace > 2) {
            arcp->logCalls(NSR::True);
        }
#else
        arcp->logAlgorithms(NSR::True);
        arcp->logCalls(NSR::True);
#endif // MULT_TRACE_LEVELS
    }

    arcp->setDevice(acdev_);
    if((err_ = arcp->error()) != NSR::error_None) {
        return err_;
    }

    UINT16      nvols,vid;
    UINT32      dataStart;
    char        vsname[MAX_LINE];
    char        volname[MAX_LINE];
    ExtentList  freeSpace;
    ExtentList  erasedSpace;
    UINT32      sectorSize;
    MediaType   mtype;

    for(i = 0; i < smList_.entries(); i++) {
        smptr = smList_.at(i);
        strcpy(vsname,"");
        strcpy(volname,"");
        if(arcp->getVolumeInfo(smptr->sid_,&nvols,vsname,volname,&vid,
                               &freeSpace,&erasedSpace,&dataStart) !=
                               NSR::True) {
            // Add the surface to the correct available list.  

            sectorSize = acdev_->sectorSize(smptr->sid_);
            mtype = acdev_->mediaType(smptr->sid_);

            AvailSurfaces       *asp = asList_.search(sectorSize,mtype);
            if(asp == NULL) {
                AvailSurfaces   as;

                as.sectorSize_ = sectorSize;
                as.mtype_ = mtype;
                as.sidList_.append(smptr->sid_);

                asList_.append(as);
            }
            else {
                asp->sidList_.append(smptr->sid_);
            }
        }
        else {                  // Surface is an NSR volume
            // We need to add the surface to its volume set list.

            VolsetNameMap       *vsnmptr = vnmList_.search(vsname);

            if(vsnmptr == NULL) { 
                VolsetNameMap   vsnmap(vsname);

                vsnmap.sidList_.append(smptr->sid_);

                vnmList_.append(vsnmap);
            }
            else {
                vsnmptr->sidList_.append(smptr->sid_);
            }

            // Now update freespace information.

            VolumeFreeSpace     *vfsptr = vfsList_.search(smptr->sid_);

            if(vfsptr == NULL) {
                VolumeFreeSpace vfs;

                vfs.sid_ = smptr->sid_;
                vfs.volname_ = volname;
                vfs.freeSpace_ = freeSpace;
                vfs.erasedSpace_ = erasedSpace;
                vfs.dataStart_ = dataStart;

                vfsList_.append(vfs);
            }
            else {
                // THIS SHOULD NEVER HAPPEN

                assert(1 == 0);
            }
        }
    }

    // We now have all of the volumes in available lists or in volume set
    // lists.  For each volumeset, we have to add them to the archive object
    // and get fileset information out so that we can produce the fileset
    // mapping list.

    delete arcp;

    for(i = 0; i < vnmList_.entries(); i++) {
        VolsetNameMap   *vsnmptr = vnmList_.at(i);

        if(vsnmptr->sidList_.entries() == 0) {
            // This is a serious error. . .
            assert(1 == 0);
            continue;
        }
        arcp = new Archive;
        if(arcp == NULL) {
            err_ = NSR::error_No_free_store;
            return err_;
        }
        // Set the Archive object to do appropriate tracing.
        if (trace == 0) {
            arcp->logFile(NULL);
            arcp->logErrors(NSR::False);
            arcp->logCalls(NSR::False);
            arcp->logAlgorithms(NSR::False);
        } else {
            arcp->logFile(stderr);
            arcp->logErrors(NSR::True);
#ifdef MULT_TRACE_LEVELS           
            if (trace > 1) {
                arcp->logAlgorithms(NSR::True);
            }
            if (trace > 2) {
                arcp->logCalls(NSR::True);
            }
#else
            arcp->logAlgorithms(NSR::True);
            arcp->logCalls(NSR::True);
#endif // MULT_TRACE_LEVELS
        }

        arcp->setDevice(acdev_);
        if((err_ = arcp->error()) != NSR::error_None) {
            delete arcp;
            return err_;
        }

        arcp->openArchive(vsnmptr->sidList_.entries(),
                          (CHAR_PTR) vsnmptr->vsname_,
                          acdev_->sectorSize(*(vsnmptr->sidList_.at(0))),
                          acdev_->mediaType(*(vsnmptr->sidList_.at(0))));

        if((err_ = arcp->error()) != NSR::error_None) {
            printf("genStatusFile():  openArchive failed, error is %s\n",
                   err_.message());
            delete arcp;
            return err_;
        }

        for(j = 0; j < vsnmptr->sidList_.entries(); j++) {
            arcp->addLabeledVolume(*(vsnmptr->sidList_.at(j)));
            if((err_ = arcp->error()) != NSR::error_None) {
                delete arcp;
                return err_;
            }
        }

        NSR::VolumeSetInfo      vsi;
        arcp->getVolSetInfo(&vsi);
        if(arcp->error() != NSR::error_None) {
            printf("genStatusFile():  getVolSetInfo failed, error is %s\n",
                   (arcp->error()).message());
            delete arcp;
            continue;
        }
        delete arcp;

        // Go through the FileSetList and make fileset mappings

        for(j = 0; j < vsi.fileSetList.entries(); j++) {
            NSR::FileSetInfo    *fsiptr = vsi.fileSetList.at(j);

            FilesetNameMap      fsnmap(fsiptr->name,(CHAR_PTR) vsi.volSetName);

            fnmList_.append(fsnmap);
        }
    }

    // Okay, at this point we should have valid internal lists.  Write the
    // actual file out.

    if(!statusFile.isNull()) {
        jbstatFile_ = statusFile;
    }

    writeStatusFile();
    return err_;
}

//-----------------------------------------------------------------------------
// METHOD:  NSRJukeBox::setNewWVS()                             --JKH
//
// Sets the volume set name for a new (to be created) working volume set.  
// Clears the old one.
//
// ARGUMENTS:
//
// const RCWString&  newvs   IN  Name of new working volume set
//
// RETURNS:
//
// Returns an error object (see Errors below).
//
// PRECONDITIONS:
//
// --The specified name does not correspond to a pre-existing volume set.
//
// POSTCONDITIONS:
//
// --The current working volume set name will be set to the new volume set.
//
// ERRORS:
//
// error_None                   Operation completed successfully.
// error_Null_volume_set_name   Null volume set name passed.
// error_Name_not_unique        Specified name is not unique.
//-----------------------------------------------------------------------------

Error
NSRJukeBox::setNewWVS(const RWCString& newvs,UINT32 sectorSize,MediaType mtype)
{
    err_ = NSR::error_None;

    if(newvs.isNull()) {
        err_ = NSR::error_Null_volume_set_name;
        return err_;
    }

    // We want to make sure that the new volume set name isn't already in use.

    if(vnmList_.search(newvs) != NULL) {
        err_ = NSR::error_Name_not_unique;
        return err_;
    }

    wvs_ = newvs; 
    newvs_ = NSR::True;

    sectorSize_ = sectorSize;
    mtype_ = mtype;

    // We need to make an entry for the volumeset in the VolsetNameMapList.

    VolsetNameMap       vsnmap(wvs_);
    vnmList_.append(vsnmap);
    openIntegrity_ = NSR::False;

    // Attempt to open the archive.

    initArchive();
    return err_;
}

//-----------------------------------------------------------------------------
// METHOD:  NSRJukeBox::setExistingWVS()                        --JKH
//
// Sets the volume set name for an existing  working volume set.  
// Clears the old one.
//
// ARGUMENTS:
//
// const RCWString&  newvs   IN  Name of working volume set
//
// RETURNS:
//
// Returns an error object (see Errors below).
//
// PRECONDITIONS:
//
// --The specified name corresponds to a pre-existing volume set.
//
// POSTCONDITIONS:
//
// --The current working volume set name will be set to the new volume set.
//
// ERRORS:
//
// error_None                   Operation completed successfully.
// error_Null_volume_set_name   Null volume set name passed.
// error_Name_not_found         Specified name is not found.
//-----------------------------------------------------------------------------

Error
NSRJukeBox::setExistingWVS(const RWCString& newvs)
{
    err_ = NSR::error_None;

    if(newvs.isNull()) {
        err_ = NSR::error_Null_volume_set_name;
        return err_;
    }

    // We want to make sure that the volume set name exists.

    VolsetNameMap       *vsnmptr;

    if((vsnmptr = vnmList_.search(newvs)) == NULL) {
        err_ = NSR::error_Name_not_found;
        return err_;
    }

    // Unfortunately, we need to cause a swap here so that we can find out
    // the sector size and media type.

    assert(vsnmptr->sidList_.entries() > 0);
    sectorSize_ = acdev_->sectorSize(*(vsnmptr->sidList_.at(0)));
    mtype_ = acdev_->mediaType(*(vsnmptr->sidList_.at(0)));

    wvs_ = newvs; 
    newvs_ = NSR::False;
    openIntegrity_ = NSR::False;

    // Attempt to open the archive.

    initArchive();
    return err_;
}

//-----------------------------------------------------------------------------
// METHOD:  NSRJukeBox::beginGrow()                             --JKH
//
// Notifies the archive object that the current working volume set is going
// to be grown.
//
// ARGUMENTS:   None.
//
// RETURN VALUES:
//
// Returns an Error object.
//
// PRECONDITIONS:
//
// --There is a currently defined working volume set.
//
// POSTCONDITIONS:
//
// --Archive::beginGrow() will be called.
//-----------------------------------------------------------------------------

Error
NSRJukeBox::beginGrow(void)
{
    if(wvs_.isNull()) {
        err_ = NSR::error_No_working_volume_set;
        return err_;
    }

#ifdef BLAH
    if(newvs_ != NSR::True) {
        arc_->markOpenIntegrity();
        if((err_ = arc_->error()) != NSR::error_None) {
            return err_;
        }
    }
#endif

    arc_->beginGrow();
    err_ = arc_->error();
    return err_;
}
    
//-----------------------------------------------------------------------------
// METHOD:  NSRJukeBox::completeGrow()                          --JKH
//
// Notifies the archive object that the current working volume set has been
// grown and that growing is complete.
//
// ARGUMENTS:   None.
//
// RETURN VALUES:
//
// Returns an Error object.
//
// PRECONDITIONS:
//
// --There is a currently defined working volume set.
// --beginGrow() has previously been called.
//
// POSTCONDITIONS:
//
// --Archive::completeGrow() will be called.
//-----------------------------------------------------------------------------

Error
NSRJukeBox::completeGrow(void)
{
    if(wvs_.isNull()) {
        err_ = NSR::error_No_working_volume_set;
        return err_;
    }

    arc_->completeGrow();
    err_ = arc_->error();
    return err_;
}

//-----------------------------------------------------------------------------
// METHOD:  NSRJukeBox::growWVS()                               --JKH
//
// Grows the current working volume set by one volume.
//
// ARGUMENTS:
//
// const RWCString&  volname   IN  Name of new volume
// Sid_t             sid       IN  Sid of volume to add
// UINT32            dataStart IN  Data start sector
//
// OR
//
// const RWCString&  volname   IN  Name of new volume
// UINT32            dataStart IN  Data start sector
//
// OR
//
// const RWCString&  volname   IN  Name of new volume
// UINT16            slot      IN  Slot location of volume to add
// UINT16            side      IN  Side of cartridge of volume to add
// UINT32            dataStart IN  Data start sector 
// 
// RETURNS:
//
// Returns an error object (see Errors below).
//
// PRECONDITIONS:
//
// --There is at least one available volume of the necessary sector size and
//   media type.
//
// POSTCONDITIONS:
//
// --The current working volume set will be grown and the new volume labeled.
//
// ERRORS:
//
// error_None                   Operation completed successfully.
// error_Null_volume_name       Null volume name
// error_No_working_volume_set  No current working volume set is defined.
// error_No_free_volume         No free volumes of the required type.
//-----------------------------------------------------------------------------

Error
NSRJukeBox::growWVS(const RWCString& volname,Sid_t sid,UINT32 dataStart)
{
    INT32       i;

    // Make sure there is a working volume set.

    if(wvs_.isNull()) {
        err_ = NSR::error_No_working_volume_set;
        return err_;
    }

    // Make sure the volume name isn't null

    if(volname.isNull()) {
        err_ = NSR::error_Null_volume_name;
        return err_;
    }

    // Make sure that the sid is in the appropriate list.

    AvailSurfaces       *asptr;

    asptr = asList_.search(sectorSize_,mtype_);
    if(asptr == NULL) {
        err_ = NSR::error_No_free_volume;
        return err_;
    }

    for(i = 0; i < asptr->sidList_.entries(); i++) {
        if(*(asptr->sidList_.at(i)) == sid) {
            break;
        }
    }
    if(i >= asptr->sidList_.entries()) {
        err_ = NSR::error_No_free_volume;
        return err_;
    }

    // Remove the sid from the avail list

    asptr->sidList_.remove(i);

    // We know have the sid of the volume to add to the volume set.  Add it
    // now.

    VolsetNameMap       *vsnmptr = vnmList_.search(wvs_);
    assert(vsnmptr != NULL);

    vsnmptr->sidList_.append(sid);

    // We need to add an entry to the VolumeFreeSpaceList.  By default, we'll
    // just make a free space extent starting from the data start sector to
    // the n-2th sector.  (where n is the number of sectors).

    VolumeFreeSpace     vfs;
    Extent              ext;

    vfs.sid_ = sid;
    vfs.volname_ = volname;
    ext.start.sector = dataStart;
    ext.start.sid = sid;
    ext.len = (sectorSize_ * (acdev_->numSectors(sid) - dataStart - 2));
    vfs.freeSpace_.append(ext);
    vfsList_.append(vfs);

    // Now we actually have to label the volume.  Do it now.

    arc_->addUnlabeledVolume(sid,(CHAR_PTR) volname,dataStart,vfs.freeSpace_,
                             vfs.erasedSpace_);
    err_ = arc_->error();

    // Update the jukebox status file.

    if(!jbstatFile_.isNull() && (err_ == NSR::error_None)) {
        writeStatusFile();
    }
    return err_;
}

Error
NSRJukeBox::growWVS(const RWCString& volname,UINT32 dataStart)
{
    // Just find a volume sid of the appropriate type from the avail list,
    // if one exists.

    AvailSurfaces       *asptr = asList_.search(sectorSize_,mtype_);
    if(asptr == NULL) {
        err_ = NSR::error_No_free_volume;
        return err_;
    }

    if(asptr->sidList_.entries() == 0) {
        err_ = NSR::error_No_free_volume;
        return err_;
    }

    return growWVS(volname,*(asptr->sidList_.at(0)),dataStart);
}

Error
NSRJukeBox::growWVS(const RWCString& volname,UINT16 slot,UINT16 side,
                    UINT32 dataStart)
{
    // Just find the sid corresponding to the slot/side pair, if one exists

    SurfaceMap  *smptr = smList_.search(slot,side);
    if(smptr == NULL) {
        err_ = NSR::error_No_free_volume;
        return err_;
    }

    return growWVS(volname,smptr->sid_,dataStart);
}

//-----------------------------------------------------------------------------
// METHOD:  NSRJukeBox::freeVolumeSet()                         --JKH
//
// This method frees an NSR volume set and returns its volumes to the 
// appropriate available list.  The label on each volume in the set is
// destroyed.
//
// ARGUMENTS:
//
// const RWCString& vsname   IN      Name of volume set to free.
//
// RETURN VALUES:
//
// Returns an error object (see Errors below).
//
// PRECONDITIONS:
//
// --The specified volume set exists.
//
// POSTCONDITIONS:  See above.
//
// ERRORS:
//
// Any errors that can result from Archive object methods:
//
// error_Unknown_fileset_name       Unknown fileset name specified.
//
// SIDE EFFECTS:
//
// --If there is a jukebox status file defined, this method will rewrite it
//   with the updated information.
//-----------------------------------------------------------------------------

Error
NSRJukeBox::freeVolumeSet(const RWCString& vsname)
{
    INT32       i;

    err_ = NSR::error_None;

    if(vsname.isNull()) {
        err_ = NSR::error_Null_volume_set_name;
        return err_;
    }

    VolsetNameMap       *vsnmptr;

    vsnmptr = vnmList_.search(vsname);
    if(vsnmptr == NULL) {
        err_ = NSR::error_Name_not_found;
        return err_;
    }

    if(arc_ == NULL) {
        arc_ = new Archive;
        if(arc_ == NULL) {
            err_ = NSR::error_No_free_store;
            return err_;
        }
        // Set the Archive object to do appropriate tracing.
        if (trace == 0) {
            arc_->logFile(NULL);
            arc_->logErrors(NSR::False);
            arc_->logCalls(NSR::False);
            arc_->logAlgorithms(NSR::False);
        } else {
            arc_->logFile(stderr);
            arc_->logErrors(NSR::True);
#ifdef MULT_TRACE_LEVELS           
            if (trace > 1) {
                arc_->logAlgorithms(NSR::True);
            }
            if (trace > 2) {
                arc_->logCalls(NSR::True);
            }
#else
            arc_->logAlgorithms(NSR::True);
            arc_->logCalls(NSR::True);
#endif // MULT_TRACE_LEVELS
        }
    }

    // Get the media type and sector size of the media from the first entry,
    // so we know into which available list to put the freed volumes.

    UINT32              sectorSize;
    MediaType           mtype;
    AvailSurfaces       *asptr;

    sectorSize = acdev_->sectorSize(*(vsnmptr->sidList_.at(0)));
    mtype = acdev_->mediaType(*(vsnmptr->sidList_.at(0)));

    if(mtype == NSR::media_Worm) {      // Can't free a WORM volume set!
        err_ = NSR::error_Invalid_operation;
        return err_;
    }
    asptr = asList_.search(sectorSize,mtype);
    if(asptr == NULL) {
        asptr = new AvailSurfaces;
        if(asptr == NULL) {
            err_ = NSR::error_No_free_store;
            return err_;
        }
        asptr->sectorSize_ = sectorSize;
        asptr->mtype_ = mtype;
        asList_.append(*asptr);
        delete asptr;
        asptr = asList_.search(sectorSize,mtype);
        assert(asptr != NULL);
    }

    // Nuke the label on each volume.  

    for(i = 0; i < vsnmptr->sidList_.entries(); i++) {
        arc_->destroyVolumeLabel(*(vsnmptr->sidList_.at(i)));
        if((err_ = arc_->error()) != NSR::error_None) { // This is trouble!
            return err_;
        }

        // Add the freed volume to the available free space list.

        asptr->sidList_.append(*(vsnmptr->sidList_.at(i)));
    }

    // Remove the volume set from the VolsetNameMapList.

    for(i = 0; i < vnmList_.entries(); i++) {
        if((vnmList_.at(i))->vsname_ == vsname) {
            vnmList_.remove(i);
            delete vsnmptr;
            break;
        }
    }

    // Remove any filesets associated with the volume set

    for(i = 0; i < fnmList_.entries(); i++) {
        if(fnmList_.at(i)->vsname_ == vsname) {
            fnmList_.remove(i);
            i--;
        }
    }

    // If we have a jukebox status file name defined, write the new file out.

    if(!jbstatFile_.isNull()) {
        writeStatusFile();
    }
    return err_;
}

//-----------------------------------------------------------------------------
// METHOD:  NSRJukeBox::wvsFreeSpace()                          --JKH
//
// This method fills in the user's VolumeFreeSpaceList with free space 
// information for the volumes in the current working volume set.
//
// ARGUMENTS:
//
// VolumeFreeSpaceList& vfsl  OUT  User's list
//
// RETURN VALUES:
//
// Returns an error object (see Errors below).
//
// PRECONDITIONS:
//
// --There is a current working volume set (set with setNewWVS() or 
//   setExistingWVS()).
//
// POSTCONDITIONS:  See above.
//
// ERRORS:
//
// error_No_working_volume_set      No working volume set
//-----------------------------------------------------------------------------

Error
NSRJukeBox::wvsFreeSpace(VolumeFreeSpaceList& vfsl)
{
    INT32       i;

    err_ = NSR::error_None;

    if(wvs_.isNull()) {
        return (err_ = NSR::error_No_working_volume_set);
    }

    // Get the SidList for the working volume set.  Go through the volumes
    // and copy their free space entry into the user's list.

    VolsetNameMap       *vsnmptr;
    VolumeFreeSpace     *vfsptr;

    vsnmptr = vnmList_.search(wvs_);
    assert(vsnmptr != NULL);

    vfsl.clear();
    for(i = 0; i < vsnmptr->sidList_.entries(); i++) {
        vfsptr = vfsList_.search(*(vsnmptr->sidList_.at(i)));
        assert(vfsptr != NULL);
        vfsl.append(*vfsptr);
    }
    return err_;
}

//-----------------------------------------------------------------------------
// METHOD:  NSRJukeBox::getImplementationInfo()                 --JKH
//
// This method gets the implementation info for a fileset in the current 
// working volume set.
//
// ARGUMENTS:
//
// const RWCString&    fsname  IN   Name of fileset
// ImplementationInfo& iinfo   OUT  User's struct
//
// RETURN VALUES:
//
// Returns an error object (see Errors below).
//
// PRECONDITIONS:
//
// --The current working volume set contains at least one volume.
// --This method must be called before a call to importArchive().
//
// POSTCONDITIONS:
//
// --The user's struct will be filled out with the implementation info.
//
// ERRORS:
//
// Any errors that can result from Archive object methods:
//
// error_Unknown_fileset_name       Unknown fileset name specified.
//-----------------------------------------------------------------------------

NSR::Error
NSRJukeBox::getImplementationInfo(const RWCString& fsname, 
                                  ImplementationInfo& iinfo)
{
    INT32               i;
    VolumeSetInfo       vsinfo;

    if(wvs_.isNull()) {
        err_ = NSR::error_No_working_volume_set;
        return err_;
    }
    if(arc_->isCompleteVolSet() != NSR::True) {
        err_ = NSR::error_No_working_volume_set;
        return err_;
    }

    arc_->getVolSetInfo(&vsinfo);
    if((err_ = arc_->error()) != NSR::error_None) {
        return err_;
    }
    FileSetInfo         *fsiptr = NULL;
    INT32               fsnum = -1;

    if(!fsname.isNull()) {
        for(i = 0; i < vsinfo.fileSetList.entries(); i++) {
            fsiptr = vsinfo.fileSetList.at(i);

            if(fsiptr->name == fsname) {
                fsnum = fsiptr->num;
                break;
            }
        }
        if(fsnum == -1) {               // User specified unknown fs name!
            err_ = NSR::error_Unknown_fileset_name;
            return err_;
        }
    }

    if(mtype_ == NSR::media_Rewritable) {
        arc_->getImplementInfo(&iinfo);
    }
    else {
        arc_->getImplementInfo(fsnum,&iinfo);
    }
    err_ = arc_->error();
    return err_;
}
    

//-----------------------------------------------------------------------------
// METHOD:  NSRJukeBox::importArchive()                         --JKH
//
// This method imports a fileset from an NSR volumeset and stores the
// metadata into a tree structure.
//
// ARGUMENTS:
//
// const RWCString& fsname   IN      Name of fileset to import.
// TreeDB&          db       IN/OUT  Tree database to allocate nodes.
// TreeNode&        root     OUT     Root node will be stored here.
//
// NOTE:  if the fsname parameter is an empty string (null), the imported
// fileset will default to the last fileset on the volume set (for worm);
// for mo, the fileset name is only used for a check, since mo can only
// contain a single fileset.
//
// RETURN VALUES:
//
// Returns an error object (see Errors below).
//
// PRECONDITIONS:
//
// --The current working volume set contains at least one volume.
//
// POSTCONDITIONS:
//
// --Freespace information will be stored internally.
// --The fileset will be imported and the metadata stored in a tree structure
//   with root stored in the user argument root.
//
// ERRORS:
//
// Any errors that can result from Archive object methods:
//
// error_Unknown_fileset_name       Unknown fileset name specified.
//-----------------------------------------------------------------------------

Error
NSRJukeBox::importArchive(const RWCString& fsname,TreeDB& db,TreeNode& root)
{
    INT32               i;
    NSR::VolumeSetInfo  vsinfo;
#ifdef TEST_EA
    Extent      EAExtent;
    Extent      EASymLinkExtent;

    EASymLinkExtent.start.sid = 42;
    EASymLinkExtent.start.sector = 4337;
    EASymLinkExtent.len = 5824;
#endif // TEST_EA

    // Make sure we actually have a volume set to work with.

    if(wvs_.isNull()) {
        err_ = NSR::error_No_working_volume_set;
        return err_;
    }
    if(arc_->isCompleteVolSet() != NSR::True) {
        err_ = NSR::error_No_working_volume_set;
        return err_;
    }

    // We run getVolsetInfo() to store away freespace information.
    // NOTE:  getVolsetInfo() isn't currently implemented, so we use a loop
    // of getVolumeInfo() calls to get freespace information.

    VolsetNameMap       *vsnmptr = vnmList_.search(wvs_);
    assert(vsnmptr != NULL);

    UINT16      nvols;                  // Ignored
    char        vsnameBuf[MAX_LINE];
    char        volnameBuf[MAX_LINE];
    UINT16      high_vid = 0,vid;
    Sid_t       root_sid;

    VolumeFreeSpace             vfs;
    for(i = 0; i < vsnmptr->sidList_.entries(); i++) {

        vfs.sid_ = *(vsnmptr->sidList_.at(i));
        arc_->getVolumeInfo(vfs.sid_,&nvols,vsnameBuf,volnameBuf,&vid,
                            &(vfs.freeSpace_),&(vfs.erasedSpace_),
                            &(vfs.dataStart_));
        if((err_ = arc_->error()) != NSR::error_None) {
            return err_;
        }
        if(vid >= high_vid) {
            high_vid = vid;
            root_sid = vfs.sid_;
        }

        vfs.volname_ = volnameBuf;
    }

    // We should have the root sid at this point.  We're ready to start the
    // import.  We need to get the fileset number, if the user specified
    // a fileset name.

    arc_->getVolSetInfo(&vsinfo);
    if((err_ = arc_->error()) != NSR::error_None) {
        return err_;
    }
    FileSetInfo         *fsiptr = NULL;
    INT32               fsnum = -1;

    if(!fsname.isNull()) {
        for(i = 0; i < vsinfo.fileSetList.entries(); i++) {
            fsiptr = vsinfo.fileSetList.at(i);

            if(fsiptr->name == fsname) {
                fsnum = fsiptr->num;
                break;
            }
        }
        if(fsnum == -1) {               // User specified unknown fs name!
            err_ = NSR::error_Unknown_fileset_name;
            return err_;
        }
    }

    // Call beginImport()

    ExtentList          metaDataSpace;

    if(mtype_ == NSR::media_Rewritable) {
        arc_->beginImport(&metaDataSpace);
    }
    else {
        arc_->beginImport(fsnum);
    }

    if((err_ = arc_->error()) != NSR::error_None) {
        return err_;
    }

    // Reclaim metadata space into the free list.

    if(metaDataSpace.getCount() > 0) {
        for(i = 0; i < metaDataSpace.getCount(); i++) {
            Extent      ext = metaDataSpace[i];
            VolumeFreeSpace     *vfsptr = vfsList_.search(vfs.sid_);

            if(vfsptr == NULL) {
                vfs.freeSpace_.append(ext);
            }
            else {
                *vfsptr = vfs;
            }
        }
    }

    // In theory, we're ready to start reading ICB's. 

    Attributes          root_attr;
    Extent              root_extent;
    Address             root_addr;

    arc_->readRootICB(&root_attr,&root_extent);
    if((err_ = arc_->error()) != NSR::error_None) {
        fprintf(stderr,"arc.readRootICB() failed, error is %s\n",
                (arc_->error()).message());
        return err_;
    }
    fprintf(stderr,"Successfully read root ICB\n");


    // Now we want to read the rest of the ICB's.  We'll create a hash table
    // to keep track of the ICB's.  Add the root ICB to the hash table as
    // well.

    TreeNode    tnode;
#ifndef ROGUEWAVIFIED
    hash_table  *icb_table;

    icb_table = htableCreate(HTABLE_SIZE,TreeNode_cmp,TreeNode_hash);
    if(icb_table == NULL) {
        err_ = NSR::error_No_free_store;
        return err_;
    }
#endif  // ROGUEWAVIFIED
    root_addr.sid = root_sid;
    root_addr.sector = ROOT_ICB_ADDRESS;

    err_ = db.allocNode(root,root_addr);
    if(err_ != NSR::error_None) {
        return err_;
    }
    root.attr_ = root_attr;
    root.elist_.append(root_extent);
    root.filename_ = "/";
    root.sync();

#ifndef ROGUEWAVIFIED
    if(htableHash(icb_table,(void *) root.realNode_)) {
        err_ = NSR::error_No_free_store;
        return err_;
    }
#endif  // ROGUEWAVIFIED

    // Read the rest of the ICB's.

    while(1) {
        Address         curAddr;
        Attributes      curAttr;
        ExtentList      curExtList;
        Extent          curEAExt;

        if(arc_->endOfICBTable() == NSR::True) {
            break;
        }

        curExtList.clear();
#ifdef TEST_EA
        arc_->readNextICB(&curAddr),&curAttr,
                        &curExtList,&curEAExt);
#else
        arc_->readNextICB(&curAddr,&curAttr,&curExtList);
                        
#endif // TEST_EA

        if(arc_->error() != NSR::error_None) {  // We're done
            db.freeNode(tnode);
            return err_;
        }
        err_ = db.allocNode(tnode,curAddr);
        if(err_ != NSR::error_None) {
            return err_;
        }
        tnode.address_ = curAddr;
        tnode.attr_ = curAttr;
        tnode.elist_ = curExtList;

        if(tnode.attr_.fileType() == NSR::file_Directory) {
            if(tnode.elist_.getCount() < 1) {
                printf("pass 1, directory has no extlist!\n");
                fflush(stdout);
            }
        }

#ifdef TEST_EA
        if(tnode.attr_.fileType() == NSR::file_SymbolicLink) {
            if(EAExtent != EASymLinkExtent) {
                fprintf(stderr,"EA FAILURE on symlink!!!\n");
                fprintf(stderr,"Read extent (sid: %d,start sector: %d, \
len: %d\n",EAExtent.start.sid,EAExtent.start.sector,EAExtent.len);
                fprintf(stderr,"Expected (sid: %d,start sector: %d, \
len: %d\n",EASymLinkExtent.start.sid,EASymLinkExtent.start.sector,
EASymLinkExtent.len);
                fprintf(stderr,"ICB address is (%d,%d)\n",tnode.address_.sid,
                        tnode.address_.sector);
            }
        }
        else {
            Extent      *EAextp = tnode.elist_.at(0);
            if(EAExtent != *EAextp) {
                fprintf(stderr,"EA FAILURE on file or dir!!!\n");
                fprintf(stderr,"Read extent (sid: %d,start sector: %d, \
len: %d\n",EAExtent.start.sid,EAExtent.start.sector,EAExtent.len);
                fprintf(stderr,"Expected (sid: %d,start sector: %d, \
len: %d\n",EAextp->start.sid,EAextp->start.sector,EAextp->len);
                fprintf(stderr,"ICB address is (%d,%d)\n",tnode.address_.sid,
                        tnode.address_.sector);
            }
        }
#endif // TEST_EA
        

        // If the file is a symbolic link, read in its PathElementList.

        if(tnode.attr_.fileType() == NSR::file_SymbolicLink) {

            arc_->readPath(tnode.elist_,&(tnode.plist_));
            if((err_ = arc_->error()) != NSR::error_None) {
                fprintf(stderr,"arc.readPath() failed, error is %s\n",
                        (arc_->error()).message());
                return err_;
            }
        }

        // Add the entry to the hash table

        tnode.sync();

#ifndef ROGUEWAVIFIED
        if(htableHash(icb_table,(void *) tnode.realNode_)) {
            err_ = NSR::error_No_free_store;
            return err_;
        }
#endif  // ROGUEWAVIFIED
    }

    // At this point, we should have the root ICB address, attributes, and
    // extent, as well as a hash table containing all of the rest of the
    // ICB's.  Now we enter a recursive function that gets the actual 
    // directory entries, so that we can associate ICB information with
    // file names. 

    printf("Import first pass complete\n");
    fflush(stdout);

#ifdef ROGUEWAVIFIED
    if(recursiveGetEntry(root_extent,db,root)) {
#else
    if(recursiveGetEntry(root_extent,icb_table,root)) {
#endif  // ROGUEWAVIFIED
        return err_;
    }

    // Clean up; we're done.

#ifndef ROGUEWAVIFIED
    htableDestroy(icb_table);
#endif  // ROGUEWAVIFIED

    arc_->completeImport();
    if((err_ = arc_->error()) != NSR::error_None) {
        fprintf(stderr,"arc_->completeImport() failed, error is %s\n",
                (arc_->error()).message());
        return err_;
    }
    arc_->closeArchive();
    if((err_ = arc_->error()) != NSR::error_None) {
        fprintf(stderr,"arc.closeArchive() failed, error is %s\n",
                (arc_->error()).message());
        return err_;
    }

    return err_;
}

//-----------------------------------------------------------------------------
// METHOD:  NSRJukeBox::prepareArchive()                        --JKH
//
// This method prepares the current working volume set to be modified.  This
// step is necessary if the CWVS is to be exported to.
//
// ARGUMENTS:  None.
//
// RETURN VALUES:
//
// Returns an error object (see Errors below).
//
// PRECONDITIONS:
//
// --The current working volume set contains at least one volume.
//
// POSTCONDITIONS:
//
// --The volume set will be marked open.
// --If the volume set is MO, the fileset will be destroyed.
//
// ERRORS:
//
// Any errors that can result from Archive object methods:
//
// error_Unknown_fileset_name       Unknown fileset name specified.
//-----------------------------------------------------------------------------

Error
NSRJukeBox::prepareArchive(void)
{
    err_ = NSR::error_None;

    if(wvs_.isNull()) {
        err_ = NSR::error_No_working_volume_set;
        return err_;
    }

    // We need to find out if there's actually a fileset associated with
    // this volume set.  If not, we NOP it for now until the Archive object
    // supports markOpenIntegrity() and destroyFileset() calls on empty
    // volume sets.

    if(newvs_ == NSR::True) {           // It's a new volume set
        openIntegrity_ = NSR::True;
        return err_;
    }

    assert(arc_ != NULL);
    arc_->markOpenIntegrity();
    if((err_ = arc_->error()) != NSR::error_None) {
        return err_;
    }

    // If it's rewritable MO, we destroy the fileset.  Also, we must 
    // remember to remove the destroyed fileset from the fileset mapping
    // list, and update the status file.

    if(mtype_ == NSR::media_Rewritable) {
        arc_->destroyFileSet();
        if((err_ = arc_->error()) != NSR::error_None) {
            return err_;
        }
        FilesetNameMap  *fsnmptr;
        for(INT32 i = 0; i < fnmList_.entries(); i++) {
            fsnmptr = fnmList_.at(i);
            if(fsnmptr->vsname_ == wvs_) {
                fnmList_.remove(i);
                i--;
            }
        }
        writeStatusFile();
    }
    openIntegrity_ = NSR::True;
    return err_;
}

//-----------------------------------------------------------------------------
// METHOD:  NSRJukeBox::exportArchive()                         --JKH
//
// This method exports a fileset from a tree structure to an NSR volume set.
//
// ARGUMENTS:
//
// const RWCString& fsname   IN      Name of fileset to export.
// TreeNode&        root     IN      Root node of tree to export.
// VolumeFreeSpaceList& vfsl IN      List of freespace information.
//
// RETURN VALUES:
//
// Returns an error object (see Errors below).
//
// PRECONDITIONS:
//
// --The current working volume set contains at least one volume.
// --The fsname parameter is non-NULL.
// --The free space information list has an entry for every surface.
//
// POSTCONDITIONS:
//
// --The fileset will be exported to the specified volume set.
//
// ERRORS:
//
// Any errors that can result from Archive object methods:
//
// error_Unknown_fileset_name       Unknown fileset name specified.
//-----------------------------------------------------------------------------

Error
NSRJukeBox::exportArchive(const RWCString& fsname,TreeNode& root,
                          VolumeFreeSpaceList& vfsl)
{
    INT32               i;
    VolsetNameMap       *vsnmptr;

    err_ = NSR::error_None;

    if(wvs_.isNull()) {
        err_ = NSR::error_No_working_volume_set;
        return err_;
    }

    if(openIntegrity_ != NSR::True) {
        err_ = NSR::error_Invalid_operation;
        return err_;
    }

    // Get the current working volume set's information

    vsnmptr = vnmList_.search(wvs_);
    if(vsnmptr == NULL) {
        err_ = NSR::error_No_working_volume_set;
        return err_;
    }

    // We must set the free space information.  Make sure that the user's
    // VolumeFreeSpaceList has an entry for every volume.

    VolumeFreeSpace     *vfsptr;

    for(i = 0; i < vsnmptr->sidList_.entries(); i++) {
        vfsptr = vfsl.search(*(vsnmptr->sidList_.at(i)));
        if(vfsptr == NULL) {
            err_ = NSR::error_No_freespace_information;
            return err_;
        }
    }

    // Okay, we're ready to start the actual export sequence.  Set available
    // space.

    for(i = 0; i < vsnmptr->sidList_.entries(); i++) {
        vfsptr = vfsl.search(*(vsnmptr->sidList_.at(i)));
        arc_->setAvailSpace(vfsptr->sid_,vfsptr->freeSpace_,
                            vfsptr->erasedSpace_);
        if((err_ = arc_->error()) != NSR::error_None) {
            return err_;
        }
    }

    // Allocate ICB Table

    UINT32 ndirs = 0,nfiles = 0,overflow = 0;
    countNodes(root,&ndirs,&nfiles,&overflow);

    arc_->allocateICBTable(ndirs,nfiles,overflow);
    if((err_ = arc_->error()) != NSR::error_None) {
        return err_;
    }

    // Begin Export

    arc_->beginExport(fsname);
    if((err_ = arc_->error()) != NSR::error_None) {
        return err_;
    }

    // Create directories.  First recursive pass.

    if(recursiveWriteDirectory(root,1)) {
        return err_;
    }

    // Write ICBs.  Second recursive pass.

    if(recursiveWriteICB(root,1)) {
        return err_;
    }

    // We're done.  Complete export and close the archive.

    arc_->completeExport();
    if((err_ = arc_->error()) != NSR::error_None) {
        return err_;
    }

    arc_->closeArchive();
    if((err_ = arc_->error()) != NSR::error_None) {
        return err_;
    }

    // Since we got here, we assume everything happened correctly.  Update
    // the internal data base, and if we have a jukebox status file, write
    // out the updated version.  We make a new fileset entry.

    FilesetNameMap      fsnmap(fsname);

    fsnmap.vsname_ = wvs_;
    fnmList_.append(fsnmap);

    if(!jbstatFile_.isNull()) {
        writeStatusFile();
    }
    openIntegrity_ = NSR::False;

    return err_;
}

//-----------------------------------------------------------------------------
// METHOD:  NSRJukeBox::makeACDev()                             --JKH
//
// Reads and parses the acrc file, and creates an hpACDev.
//
// ARGUMENTS:
//
// const RWCString& acrcFile    IN  Filename of acrc configuration file.
//
// RETURNS:
//
// Returns 0 on success, -1 on failure
//
// PRECONDITIONS:
//
// --The acrcFile is non-NULL and contains a valid path to an acrc file.
//
// POSTCONDITIONS:
//
// --An autochanger device object will be created with appropriate device
//   files mapped.
//
// SIDE EFFECTS:
//
// --Upon error, this method sets the err_ variable.
//
// ERRORS:
//
// error_Null_acrc_file         A NULL acrc filename was specified.
// error_Cannot_access_acrc_file        Cannot open file for read.
// error_Syntax_error_in_acrc_file Bad input in file
// error_Cannot_stat_acrc_file  stat(2) failed on acrc file
// error_Bad_AC_device_file     Not a special character device file
//-----------------------------------------------------------------------------

INT32
NSRJukeBox::makeACDev(const RWCString& acrcFile)
{
    INT32       i;
    FILE        *infile;
    struct stat  buf;

    if(acrcFile.isNull()) {
        err_ = NSR::error_Null_acrc_file; 
        return -1; 
    } 
    
    // Open the file.  
    
    infile = fopen((CHAR_PTR) acrcFile,"r"); 
    if(infile == NULL) { 
        err_ = NSR::error_Cannot_access_acrc_file;
        return -1;
    }

    // REPLACE THIS CODE WITH REAL PARSER CODE.
    // This code is pretty much a horrendous kludge, just to get something
    // in place for right now.

    char        buffer[MAX_LINE];
    char        word[MAX_LINE];
    UINT32      which;
    RWCString   acDevfile;
    struct {
        RWCString       driveDevfile;
        UINT32          driveNum;
    } driveInfo[MAX_DRIVES];

    while(1) {
        if(fgets(buffer,MAX_LINE,infile) == NULL) {
            break;
        }
        if(sscanf(buffer,"%s",word) != 1) {
            continue;
        }
        if(strcmp(word,"#") == 0) {
            continue;
        }
        if(strcmp(word,"acdev") == 0) {
            if(sscanf(buffer,"%*s = %s",word) != 1) {
                err_ = NSR::error_Syntax_error_in_acrc_file;
                return -1;
            }

            // word should contain the autochanger device file.  Stat the
            // file.

            if(stat(word,&buf)) {
                err_ = NSR::error_Cannot_stat_acrc_file;
                return -1;
            }

            // Verify that the file is a character special device file

            if(!S_ISCHR(buf.st_mode)) {
                err_ = NSR::error_Bad_AC_device_file;
                return -1;
            }
            if(!acDevfile.isNull()) {
                err_ = NSR::error_Bad_AC_device_file;
                return -1;
            }
            acDevfile = word;
            continue;
        }
        if(strcmp(word,"drive") == 0) {
            if(sscanf(buffer,"%*s %lu = %s",&which,word) != 2) {
                err_ = NSR::error_Syntax_error_in_acrc_file;
                return -1;
            }
            if(which > MAX_DRIVES) {
                err_ = NSR::error_Syntax_error_in_acrc_file;
                return -1;
            }
            which--;
            if(!driveInfo[which].driveDevfile.isNull()) {
                err_ = NSR::error_Syntax_error_in_acrc_file;
                return -1;
            }
            driveInfo[which].driveDevfile = word;
            driveInfo[which].driveNum = which;
            continue;
        }
        err_ = NSR::error_Syntax_error_in_acrc_file;
        return -1;
    }
    fclose(infile);

    // At this point, we've read in the file, gotten the acdrv file and
    // drive files.  We now create an hpACDev device, get element addresses
    // information, and make sure we have entries for every drive in the
    // the autochanger.

    acdev_ = new hpACDev((CHAR_PTR) acDevfile);
    if(acdev_ == NULL) {
        err_ = NSR::error_No_free_store;
        return -1;
    }
    if(acdev_->error() != NSR::error_None) {
        err_ = acdev_->error();
        return -1;
    }

    struct element_addresses    eltaddr;

    if(acdev_->elementAddresses(eltaddr)) {
        err_ = acdev_->error();
        return -1;
    }

    // Make sure we got an entry for every drive in the autochanger.

    for(i = 0; i < eltaddr.num_data_transfers; i++) {
        if(driveInfo[i].driveDevfile.isNull()) {
            err_ = NSR::error_Syntax_error_in_acrc_file;
            delete acdev_;
            acdev_ = NULL;
            return -1;
        }
    }

    // Now loop through and add the drive mappings.  We need to do an
    // element status on every drive to get the SCSI target and LUN ids.

    struct element_status       eltstat;

    for(i = 0; i < eltaddr.num_data_transfers; i++) {
        if(acdev_->elementStatus(eltaddr.first_data_transfer + i,eltstat)) {
            err_ = acdev_->error();
            delete acdev_;
            acdev_ = NULL;
            return -1;
        }
        if(acdev_->addDriveMap(eltstat.bus_address,eltstat.lun,
           (CHAR_PTR) driveInfo[i].driveDevfile)) {
            err_ = acdev_->error();
            delete acdev_;
            acdev_ = NULL;
            return -1;
        }
    }

    // We're done, everything went successfully.

    return 0;
}

//-----------------------------------------------------------------------------
// METHOD:  NSRJukeBox::mapSurfaces()                           --JKH
//
// Maps surfaces to Sid_t's and generates the map table.
//
// ARGUMENTS:  None.
//
// RETURNS:
//
// Returns 0 on success, -1 on failure
//
// PRECONDITIONS:
//
// --The nsrJukeBox has been successfully constructed.
//
// POSTCONDITIONS:
//
// --All surfaces in the jukebox will be mapped to an Sid_t.
//
// SIDE EFFECTS:
//
// --Upon error, this method sets the err_ variable.
//
// ERRORS:
//
// error_None                           Operation completed successfully.
//-----------------------------------------------------------------------------

INT32
NSRJukeBox::mapSurfaces()
{
    INT32                       i;
    struct element_status       eltstat;
    struct element_addresses    eltaddr;

    err_ = NSR::error_None;
    if(acdev_ == NULL) {                // Not constructed properly.
        return -1;
    }

    // Get element addresses information

    if(acdev_->elementAddresses(eltaddr)) {
        err_ = acdev_->error();
        return -1;
    }

    // Now get element status on every storage slot, and for each one that
    // is occupied, map it to an Sid_t and add an entry to the surface map
    // list.

    for(i = 0; i < eltaddr.num_storages; i++) {
        if(acdev_->elementStatus(eltaddr.first_storage + i,eltstat)) {
            err_ = acdev_->error();
            return -1;
        }
        if(eltstat.full) {
            SurfaceMap  smap;

            smap.slot_ = i + 1;
            smap.side_ = 0;             // side A first
            smap.sid_ = acdev_->addSurface(smap.slot_,smap.side_);
            if(smap.sid_ == (Sid_t) SID_T_INVALID) {
                err_ = acdev_->error();
                return -1;
            }

            // Add it to the SurfaceMapList.

            smList_.append(smap);

            // Now side B

            smap.side_ = 1;
            smap.sid_ = acdev_->addSurface(smap.slot_,smap.side_);
            if(smap.sid_ == (Sid_t) SID_T_INVALID) {
                err_ = acdev_->error();
                return -1;
            }

            smList_.append(smap);
        }
    }
    return 0;
}

//-----------------------------------------------------------------------------
// METHOD:  NSRJukeBox::writeStatusFile()                       --JKH
//
// Creates a jukebox status file from the current database information.
//
// ARGUMENTS:  None.
//
// RETURNS:
//
// Returns 0 on success, -1 on failure.
//
// PRECONDITIONS:
//
// --The nsrJukeBox has been successfully constructed.
//
// POSTCONDITIONS:
//
// --A status file will be created.
//
// SIDE EFFECTS:
//
// --Upon error, this method sets the err_ variable.
//
// ERRORS:
//
// error_None                           Operation completed successfully.
//-----------------------------------------------------------------------------

INT32
NSRJukeBox::writeStatusFile()
{
    INT32       i,j;
    FILE        *outfile;
    RWCString   str1,str2;

    if(jbstatFile_.isNull()) {
        return -1;
    }

    outfile = fopen((CHAR_PTR) jbstatFile_,"w");
    if(outfile == NULL) {
        return -1;
    }

    // Write volume set information first.
 
    for(i = 0; i < vnmList_.entries(); i++) {
        VolsetNameMap *vsnmptr = vnmList_.at(i);

        str1 = vsnmptr->vsname_;
        processIdentifier(str1,OUT);
        fprintf(outfile,"volset \"%s\" = ",(CHAR_PTR) str1);
        for(j = 0; j < vsnmptr->sidList_.entries(); j++) {
            SurfaceMap  *smptr = smList_.search(*(vsnmptr->sidList_.at(j)));

            fprintf(outfile,"%u%c ",smptr->slot_,(smptr->side_ == 0) ?
                    'a' : 'b');
        }
        fprintf(outfile,"\n");
    }

    // Now write file set map information

    for(i = 0; i < fnmList_.entries(); i++) {
        FilesetNameMap *fsnmptr = fnmList_.at(i);

        str1 = fsnmptr->fsname_;
        str2 = fsnmptr->vsname_;
        processIdentifier(str1,OUT);
        processIdentifier(str2,OUT);
        fprintf(outfile,"fileset \"%s\" in \"%s\"\n",(CHAR_PTR) str1,
                (CHAR_PTR) str2);
    }

    // Now write avail space information

    for(i = 0; i < asList_.entries(); i++) {
        AvailSurfaces *asptr = asList_.at(i);

        if(asptr->sidList_.entries() == 0) {
            continue;
        }
        fprintf(outfile,"avail %d %s = ",asptr->sectorSize_,
                (asptr->mtype_ == NSR::media_Worm) ? "worm" : "mo");
        
        for(j = 0; j < asptr->sidList_.entries(); j++) {
            SurfaceMap  *smptr = smList_.search(*(asptr->sidList_.at(j)));

            fprintf(outfile,"%u%c ",smptr->slot_,(smptr->side_ == 0) ?
                    'a' : 'b');
        }
        fprintf(outfile,"\n");
    }
    fclose(outfile);
    return 0;
}

//-----------------------------------------------------------------------------
// METHOD:  NSRJukeBox::print()
//
// Prints out the guts of the jukebox object.  Mainly for debugging purposes.
//-----------------------------------------------------------------------------

void
NSRJukeBox::print()
{
    INT32       i,j;

    // Surface list first

    printf("Jukebox surface list:\n\n");

    for(i = 0; i < smList_.entries(); i++) {
        SurfaceMap      *smptr = smList_.at(i);

        if(smptr == NULL) {
            printf("Oops, Null entry in list, index = %ld\n",i);
            continue;
        }
        printf("Surface %d%c mapped to Sid_t %u\n",smptr->slot_,
               (smptr->side_ ? 'B' : 'A'),smptr->sid_);
    }

    // Available Surfaces List

    printf("\nAvailable Surfaces Lists:\n\n");

    for(i = 0; i < asList_.entries(); i++) {
        AvailSurfaces   *asptr = asList_.at(i);

        if(asptr == NULL) {
            printf("Oops, Null entry in list, index = %ld\n",i);
            continue;
        }
        printf("Available %u byte/sector %s media (sid's):  ",
               asptr->sectorSize_,
               (asptr->mtype_ == NSR::media_Worm ? "WORM" : "MO"));
        
        for(j = 0; j < asptr->sidList_.entries(); j++) {
            printf("%u ",*(asptr->sidList_.at(j)));
        }
        printf("\n");
    }

    // Volume Set Name Map List

    printf("\nVolume Set Name Map Lists\n\n");

    for(i = 0; i < vnmList_.entries(); i++) {
        VolsetNameMap   *vsnmptr = vnmList_.at(i);

        if(vsnmptr == NULL) {
            printf("Oops, Null entry in list, index = %ld\n",i);
            continue;
        }
        printf("Volume set \"%s\" sid's:  ",(CHAR_PTR) vsnmptr->vsname_);
        
        for(j = 0; j < vsnmptr->sidList_.entries(); j++) {
            printf("%u ",*(vsnmptr->sidList_.at(j)));
        }
        printf("\n");
    }

    // File Set Name Map List

    printf("\nFile Set Name Map Lists:\n\n");

    for(i = 0; i < fnmList_.entries(); i++) {
        FilesetNameMap          *fsnmptr = fnmList_.at(i);

        if(fsnmptr == NULL) {
            printf("Oops, Null entry in list, index = %ld\n",i);
            continue;
        }
        printf("Fileset \"%s\" belongs to volume set \"%s\"\n",
               (CHAR_PTR) fsnmptr->fsname_,(CHAR_PTR) fsnmptr->vsname_);
    }
            
}

//-----------------------------------------------------------------------------
// METHOD:  NSRJukeBox::initArchive()                           --JKH
//
// Creates an archive object, sets the device, and opens the archive, adding
// volumes as necessary from the current working volume set.
//
// ARGUMENTS:  None.
//
// RETURN VALUES:  None.
//
// PRECONDITIONS:
//
// --There is a current working volume set.
//
// POSTCONDITIONS:
//
// --An archive object will be created and openArchive() called, as well as
//   addLabeledVolume for every volume in the current working volume set.
//-----------------------------------------------------------------------------

void
NSRJukeBox::initArchive()
{
    if(wvs_.isNull()) {
        return;
    }

    if(arc_ != NULL) {
        delete arc_;
    }

    // Create a new Archive object

    arc_ = new Archive;

    if(arc_ == NULL) {
        err_ = NSR::error_No_free_store;
        return;
    }

    // Set the Archive object to do appropriate tracing.
    if (trace == 0) {
        arc_->logFile(NULL);
        arc_->logErrors(NSR::False);
        arc_->logCalls(NSR::False);
        arc_->logAlgorithms(NSR::False);
    } else {
        arc_->logFile(stderr);
        arc_->logErrors(NSR::True);
#ifdef MULT_TRACE_LEVELS           
        if (trace > 1) {
            arc_->logAlgorithms(NSR::True);
        }
        if (trace > 2) {
            arc_->logCalls(NSR::True);
        }
#else
        arc_->logAlgorithms(NSR::True);
        arc_->logCalls(NSR::True);
#endif // MULT_TRACE_LEVELS
    }

    // Set the device.
    arc_->setDevice(acdev_);
    if((err_ = arc_->error()) != NSR::error_None) {
        return;
    }

    // Open the archive object. 

    VolsetNameMap       *vsnmptr = vnmList_.search(wvs_);
    if(vsnmptr == NULL) {
        err_ = NSR::error_Null_volume_name;
        return;
    }
    arc_->openArchive(vsnmptr->sidList_.entries(),(CHAR_PTR) wvs_,sectorSize_,
                     mtype_);
    if((err_ = arc_->error()) != NSR::error_None) {
        return;
    }

    // Add any pre-existing labeled volumes.  Also, do a getVolInfo on each
    // volume so that we can get its freespace information.

    INT32       i;

    for(i = 0; i < vsnmptr->sidList_.entries(); i++) {
        VolumeFreeSpace         vfs;
        UINT16                  nvols,vid;
        char                    fakeVSname[128];
        char                    volname[128];
        ExtentList              fs,es;
        vfs.sid_ = *(vsnmptr->sidList_.at(i));

        arc_->addLabeledVolume(vfs.sid_);
        if((err_ = arc_->error()) != NSR::error_None) {
            return;
        }
        arc_->getVolumeInfo(vfs.sid_,
                            &nvols,
                            fakeVSname,
                            volname,
                            &vid,
                            &fs,
                            &es,
                            &(vfs.dataStart_));
        if((err_ = arc_->error()) != NSR::error_None) {
            return;
        }
        vfs.volname_ = volname;

        // Set freespace.  The erased space list is only for MO using
        // pre-erasure; the NSRJukeBox doesn't support that, so we just
        // set the vfs.freeSpace_ equal to the free space list.

        vfs.freeSpace_ = fs;

        vfsList_.append(vfs);
    }

    // We're done.
}

//-----------------------------------------------------------------------------
// METHOD:  NSRJukeBox::recursiveGetEntry()
//
// Recursively traverses through the directory tree (preorder traversal) and
// displays filenames and file lengths.
//
// ALGORITHM:
//
// It is important, to minimize swapping, that the entries be read from the
// Archive object in a recursive preorder traversal.  For the current directory,
// nodes are made for all children.  If a child is a directory, it is placed
// into a queue for later recursion.  File entries are retrieved from the
// Archive object for each child, and the filenames stored in the hash table.
// Then each subdirectory is recurred upon, in queue (FIFO) order.  For
// maximum efficiency, directories in the Archive object must be traversed
// in the same order that they are retrieved by it.
//
// ARGUMENTS:
//
// Extent&         ext       IN      Extent of current directory
// hash_table     *icb_table IN      Hash table containing file ICB information
// TreeNode&       tnode     IN/OUT  Node of current directory
//
// PRECONDITIONS:
//
// --The hash table contains an entry for the current directory
//
// POSTCONDITIONS:
//
// --All file entries for the current directory will be read from the Archive
//   object and added to the n-ary tree.  
// --The file names will be added to the appropriate entries in the hash table.
//-----------------------------------------------------------------------------

INT32
#ifdef ROGUEWAVIFIED
NSRJukeBox::recursiveGetEntry(Extent& ext,TreeDB& db,TreeNode& tnode)
#else
NSRJukeBox::recursiveGetEntry(Extent& ext,hash_table *icb_table,TreeNode& tnode)
#endif  // ROGUEWAVIFIED
{
    int                 i,nentries;
    Boolean             isdir;
    char                name[MAX_NAMELENGTH];
    dqueue              *que;
    TreeNode            finfo;
#ifdef ROGUEWAVIFIED
    TreeNode            RealNode;
    TreeNode *          newnode = &RealNode;
#else
    TreeNode *          newnode;
#endif  // ROGUEWAVIFIED

    // Create the queue for keeping track of subdirectories

    que = dqueueCreate();
    if(que == NULL) {
        err_ = NSR::error_No_free_store;
        return -1;
    }

    // locateDirectory():  Sets current working directory

    arc_->locateDirectory(ext);
    if((err_ = arc_->error()) != NSR::error_None) {
        return -1;
    }
    
    nentries = arc_->numDirEntries();
    if((err_ = arc_->error()) != NSR::error_None) {
        return -1;
    }
    // Resize the tree node

    err_ = tnode.resize(nentries); 
    if(err_ != NSR::error_None) {
        return -1;
    }
    tnode.sync();

    // Now loop and get all of the entries in the directory.  If a
    // given entry is a directory, put it into our queue.

    for(i = 0; i < nentries; i++) {
        arc_->getDirEntry(i,&isdir,name,&(finfo.address_));
        if((err_ = arc_->error()) != NSR::error_None) {
            return -1;
        }

        // If it's a parent directory entry (NULL name), just continue:  we
        // don't put parent entries in the hash table.  This also means we
        // don't link parent entries into the tree.

        if(strlen(name) == 0) {
            continue;
        }

        // Find the entry in the hash table so that we can add its name
        // to the hash table entry.

#ifdef ROGUEWAVIFIED
        Error           err;
        IntegerField    key;
        db.addrToKey(finfo.address_,key);
        if((err = db.getNode(key,*newnode)) != NSR::error_None) {
            if(strlen(name) > 0) {
                fprintf(stderr,
                        "ERROR, dir entry for file \"%s\" not in database!  ",
                        name);
                fprintf(stderr,"Error is:  %s\n",err.message());
                continue;
            }
        }
#else
        newnode = (TreeNode *) htableFind(icb_table,(void *) &finfo);
        if(newnode == NULL) {
            if(strlen(name) > 0) {
                fprintf(stderr,"ERROR, dir entry for file \"%s\" not in hash table!\n", name);
            }
            continue;
        }
#endif  // ROGUEWAVIFIED
        newnode->filename_ = name;
        newnode->setParent(tnode);
        newnode->sync();

        // Link the parent to the child.

        err_ = tnode.addChild(*newnode);
        tnode.sync();
        if(err_ != NSR::error_None) {
            return -1;
        }

        // if it's a directory, put it into the queue.

        if((newnode->attr_.fileType() == NSR::file_Directory) && 
           (strlen((CHAR_PTR) newnode->filename_) > 0)) {
#ifdef ROGUEWAVIFIED
            TreeNode *nodep = new TreeNode;
            *nodep = *newnode;
#endif  // ROGUEWAVIFIED

#ifdef ROGUEWAVIFIED
            if(dqueueEnqueue(que,nodep)) {
#else   // ROGUEWAVIFIED
            if(dqueueEnqueue(que,newnode)) {
#endif
                fprintf(stderr,"Queue enqueue failed, aborting!\n");
                return 2;
            }
        }
    }                   // End of for loop

    // Now recur on the entries in the queue.

    int rval;
    while( (newnode = (TreeNode *) dqueueDequeue(que)) != NULL) {
        finfo = *newnode;
        if(finfo.elist_.getCount() < 1) {
            fprintf(stderr,"Problem, directory <%s> has no extent list!\n",
                   (const char *) finfo.filename_);
            fflush(stderr);
        }
#ifdef ROGUEWAVIFIED
        rval = recursiveGetEntry(finfo.elist_[0],db,finfo);
#else
        rval = recursiveGetEntry(finfo.elist_[0],icb_table,finfo);
#endif  // ROGUEWAVIFIED
        if(rval) {
            fprintf(stderr,"Recursive call to recursiveGetEntry() failed!\n");
            return rval;
        }
#ifdef ROGUEWAVIFIED
        delete newnode;
#endif  // ROGUEWAVIFIED
    }

    // Cleanup

    dqueueDestroy(que);
    return 0;
}

//-----------------------------------------------------------------------------
// METHOD:  NSRJukeBox::countNodes()
//
// Recursively counts all of the files and directories in the tree, and
// calculates overflow sectors.
//
// ARGUMENTS
//
// TreeNode&        node      IN     Current location in tree
// UINT32          *ndirs     IN/OUT Number of directories
// UINT32          *nfiles    IN/OUT Number of files
// UINT32          *overflow  IN/OUT Overflow sectors.
//-----------------------------------------------------------------------------

void
NSRJukeBox::countNodes(TreeNode& root,UINT32 *ndirs,UINT32 *nfiles,
                       UINT32 *overflow)
{
    int         i;
    TreeNode    tnode;
    Error       err;

    *ndirs += 1;                // Count the current directory!

    // Now count the children.  If the child is directory, recur on it.

    for(i = 0; i < root.getNumChildren(); i++) {
        err = root.getChild(i,tnode);
        if(err != NSR::error_None) {
            fprintf(stderr,"countNodes:  getChild(%d) failed\n",i);
            continue;
        }
        if(tnode.attr_.fileType() == NSR::file_Directory) {
            countNodes(tnode,ndirs,nfiles,overflow);
        }
        else {
            *nfiles += 1;       // It's a file.
            if(tnode.attr_.fileType() == NSR::file_SymbolicLink) {
                *overflow += arc_->overflowSectors(tnode.plist_);
            }
            else {
                *overflow += arc_->overflowSectors(tnode.elist_.getCount());
            }
        }
    }
}

//-----------------------------------------------------------------------------
// METHOD:  NSRJukeBox::recursiveWriteDirectory()
//
// Recursively writes directories to the export archive.
//
// ARGUMENTS:
//
// TreeNode         node      IN  Current location in the tree
// int              isroot    IN  nonzero if root directory
//-----------------------------------------------------------------------------

INT32
NSRJukeBox::recursiveWriteDirectory(TreeNode& node,INT32 isroot)
{
    INT32       i,rval;
    TreeNode    tnode;
    Error       err;

    // Call createDirectory()

    if(isroot) {
        arc_->createDirectory(NULL);
        if((err_ = arc_->error()) != NSR::error_None) {
            return -1;
        }
    }
    else {
        arc_->createDirectory((CHAR_PTR) node.filename_);
        if((err_ = arc_->error()) != NSR::error_None) {
            return -1;
        }
    }

    // Add file and directory entries for all children.  

    for(i = 0; i < node.getNumChildren(); i++) {
        err = node.getChild(i,tnode);
        if(err != NSR::error_None) {
            continue;
        }
        if(tnode.attr_.fileType() == NSR::file_Directory) {
            arc_->addDirEntry(NSR::True,(CHAR_PTR) tnode.filename_,0);
            if((err_ = arc_->error()) != NSR::error_None) {
                return -1;
            }
        }
        else {          
            UINT32  ovf;
            if(tnode.attr_.fileType() == NSR::file_SymbolicLink) {
                ovf = arc_->overflowSectors(tnode.plist_);
            }
            else {
                ovf = arc_->overflowSectors(tnode.elist_.getCount());
            }
            arc_->addDirEntry(NSR::False,(CHAR_PTR) tnode.filename_,ovf);
            if((err_ = arc_->error()) != NSR::error_None) {
                return -1;
            }
        }
    }

    // Write the directory.

    Extent      extent;
    arc_->writeDirectory(&extent);
    if((err_ = arc_->error()) != NSR::error_None) {
        return -1;
    }
    node.elist_.clear();
    node.elist_.append(extent);
    node.sync();

    // Now recur on subdirectories, if any.

    for(i = 0; i < node.getNumChildren(); i++) {
        err = node.getChild(i,tnode);
        if(err != NSR::error_None) {
            continue;
        }
        if(tnode.attr_.fileType() != NSR::file_Directory) {
            continue;
        }
        rval = recursiveWriteDirectory(tnode,0);
        if(rval) {
            return rval;
        }
    }

    // Close the directory

    arc_->closeDirectory();
    if((err_ = arc_->error()) != NSR::error_None) {
        return -1;
    }

    return 0;
}

//-----------------------------------------------------------------------------
// METHOD:  NSRJukeBox::recursiveWriteICB()                     --JKH
//
// Recursively writes ICB's to the export archive.
//
// ARGUMENTS:
//
// TreeNode         node      IN  Current location in the tree
// int              isroot    IN  nonzero if root directory
//-----------------------------------------------------------------------------

INT32
NSRJukeBox::recursiveWriteICB(TreeNode& node,INT32 isroot)
{
    INT32       i,rval;
    TreeNode    tnode;
    Error       err;
    UINT32      nsubdirs;
#ifdef TEST_EA
    Extent      EAExtent;
    Extent      EASymLinkExtent;

    EASymLinkExtent.start.sid = 42;
    EASymLinkExtent.start.sector = 4337;
    EASymLinkExtent.len = 5824;
#endif // TEST_EA

    node.refresh();

    // Call writeRootICB if necessary.

    if(isroot) {
        TreeNode        child;

        nsubdirs = 0;
        for(i = 0; i < node.getNumChildren(); i++) {
            err = node.getChild(i,child);
            if(err != NSR::error_None) {
                continue;
            }
            if(child.attr_.fileType() == NSR::file_Directory) {
                nsubdirs++;
            }
        }
        arc_->writeRootICB(node.attr_,node.elist_[0]);
        if((err_ = arc_->error()) != NSR::error_None) {
            return -1;
        }
    }

    // Write ICB's for children.
    for(i = 0; i < node.getNumChildren(); i++) {
        err = node.getChild(i,tnode);
        if(err != NSR::error_None) {
            printf("recursiveWriteICB(), node.getChild(%d) failed\n",i);
            continue;
        }
        if(tnode.attr_.fileType() == NSR::file_SymbolicLink) {
#ifdef TEST_EA
            arc_->writeNextICB(tnode.attr_,tnode.plist_,EASymLinkExtent);
#else
            arc_->writeNextICB(tnode.attr_,tnode.plist_);
#endif // TEST_EA
            if((err_ = arc_->error()) != NSR::error_None) {
                return -1;
            }
        }
        else {
            // We need to determine the number of subdirectories for this
            // child, if it's a directory.

            UINT32      nsubdirs = 0;
            if(tnode.attr_.fileType() == NSR::file_Directory) {
                int j;
                TreeNode        child;

                for(j = 0; j < tnode.getNumChildren(); j++) {
                    err = tnode.getChild(j,child);
                    if(err != NSR::error_None) {
                        continue;
                    }
                    if(child.attr_.fileType() == NSR::file_Directory) {
                        nsubdirs++;
                    }
                }
            }
#ifdef TEST_EA
            EAExtent = *(tnode.elist_.at(0));
            arc_->writeNextICB(tnode.attr_,tnode.elist_,EAExtent);
#else
            arc_->writeNextICB(tnode.attr_,tnode.elist_);
#endif // TEST_EA
            if((err_ = arc_->error()) != NSR::error_None) {
                return -1;
            }
        }
    }

    // Now recur on subdirectories, if any.

    for(i = 0; i < node.getNumChildren(); i++) {
        err = node.getChild(i,tnode);
        if(err != NSR::error_None) {
            continue;
        }
        if(tnode.attr_.fileType() != NSR::file_Directory) {
            continue;
        }
        rval = recursiveWriteICB(tnode,0);
        if(rval) {
            return rval;
        }
    }

    return 0;
}

// We need a little more sophisticated tokenizing than the RWCTokenizer seems
// to give us.  In particular, we need to allow quoted strings to have 
// whitespace in them.

RWCSubString
MyTokenizer::operator()(void)
{
    UINT32      i;
    UINT32      start = 0;
    UINT32      inword = 0;
    UINT32      quote = 0;

    if(curpos >= orig.length()) {
        orig.resize(0);
        return orig(0,0);
    }
    for(i = curpos; i < orig.length(); i++) {
        if(!inword) {
            if(isspace(orig[i])) {
                continue;
            }
            start = i;
            inword = 1;
            if(orig[i] == '\"') {
                if(i > 0) {
                    if(orig[i - 1] != '\\') {
                        quote++;
                        if(quote == 2) {
                            break;
                        }
                    }
                }
            }
        }
        else {
            if(isspace(orig[i]) && !quote) {
                break;
            }
            if(orig[i] == '\"') {
                if(i > 0) {
                    if(orig[i - 1] != '\\') {
                        quote++;
                        if(quote == 2) {
                            break;
                        }
                    }
                }
            }
        }
    }
    curpos = i + 1;
    if(inword) {
        return orig(start,i - start);
    }
    else {
        return orig(0,0);
    }
}

//-----------------------------------------------------------------------------
// FUNCTION:  TreeNode_cmp
//
// Comparison function for the TreeNode class.  This method is for use with
// the various dlist and htable C toolkit routines.
//
// ARGUMENTS:
//
// void         *d1  IN   First data item to compare
// void         *d2  IN   Second data item to compare
//
// RETURN VALUES:
//
// Returns a value less than 0 if d1 is less than d2, 0 if d1 is equal to d2,
// and a value greater than 0 if d1 is greater than d2.  TreeNodes are 
// compared based upon their ICB addresses.
//
// PRECONDITIONS:
//
// --d1 and d2 are non-NULL
// --d1 and d2 are pointers to TreeNode objects
//
// POSTCONDITIONS:
//
// --see return values
//-----------------------------------------------------------------------------

static int
TreeNode_cmp(void *d1,void *d2)
{
    TreeNode *f1 = (TreeNode *) d1;
    TreeNode *f2 = (TreeNode *) d2;

    if((f1 == NULL) || (f2 == NULL)) {
        return 0;
    }
    if(f1->address_.sid != f2->address_.sid) {
        return (f1->address_.sid - f2->address_.sid);
    }
    return (f1->address_.sector - f2->address_.sector);
}

//-----------------------------------------------------------------------------
// FUNCTION:  TreeNode_hash
//
// Hash function for the nsrTreeNode struct.  This method is for use with
// the various dlist and htable C toolkit routines.
//
// ARGUMENTS:
//
// void         *d1  IN   Data item to hash.
//
// RETURN VALUES:
//
// Returns a hash value for the argument TreeNode.
//
// PRECONDITIONS:
//
// --d1 is non-NULL
// --d1 is a pointer to TreeNode objects
//
// POSTCONDITIONS:  See return values
//-----------------------------------------------------------------------------

static int
TreeNode_hash(void *d1)
{
    TreeNode  *f1 = (TreeNode *) d1;

    if(f1 == NULL) {
        return 0;
    }
    return f1->address_.sector;
}
