//-*-c++-*-
//-----------------------------------------------------------------------------
// NSRJukeBox.h:  
// 
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//
// CLASSES:
//
// SurfaceMap
// SurfaceMapList
// SidList
// AvailSurfaces
// AvailSurfacesList
// VolsetNameMap
// VolsetNameMapList
// FilesetNameMap
// FilesetNameMapList
// VolumeFreeSpace
// VolumeFreeSpaceList
// NSRJukeBox
//
//-----------------------------------------------------------------------------
// $Source: /projects/HMS/nsr/udfAncil/msTest/nsrJB.h,v $
// $Revision: 1.17 $    $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 96/02/02 09:59:10 $
//-----------------------------------------------------------------------------

#ifndef _NSRJB_H_                       // Allow multiple inclusion
#define _NSRJB_H_

#include <rw/cstring.h>                 // RogueWave string class
#include <rw/gslist.h>                  // RogueWave generic singly-linked list
#include <devs/nsrRSSdv.h>
#include <devs/nsrRMSdv.h>
#include <nsrLib.h>

#include "hash.h"

#define         IN              0
#define         OUT             1

//-----------------------------------------------------------------------------
// CLASS:  SurfaceMap                                           --JKH
//
// The SurfaceMap class is basically just a data structure used to map
// slot and side information to an Sid_t.
//
// PUBLIC DATA MEMBERS:
//
// UINT16       slot_   Slot of surface within the autochanger
// UINT16       side_   Which side of the cartridge (0 = 'A', 1 = 'B')
// Sid_t        sid_    Sid_t assigned to the surface
// RWCString    volname_ Volume name, if known
//
// PUBLIC METHODS:
//
// SurfaceMap()
//      Constructor.  Initializes data members.
// ~SurfaceMap()
//      Destructor.  Does nothing.
// operator==()
//      Equality operator between two SurfaceMap objects.
// operator!=()
//      Inequality operator between two SurfaceMap objects.
// operator=()
//      Assignment operator between two SurfaceMap objects.
//
// USAGE NOTES:
//-----------------------------------------------------------------------------
// COLLABORATORS:               None.
//
// PRIVATE DATA MEMBERS:        None.
//
// PRIVATE METHODS:             
//
// copy()
//      Copies one SurfaceMap object to another.
//
// IMPLEMENTATION NOTES:        None.
//-----------------------------------------------------------------------------

class SurfaceMap 
{
    public:

    UINT16              slot_;
    UINT16              side_;
    Sid_t               sid_;
    RWCString           volname_;

    SurfaceMap() : slot_(0), side_(0), sid_(SID_T_INVALID) {}
    SurfaceMap(const SurfaceMap& smap) {
        copy(smap);
    }
    ~SurfaceMap() {}

    operator==(const SurfaceMap& smap) {
        return ((slot_ == smap.slot_) && 
                (side_ == smap.side_) &&
                (sid_ == smap.sid_));
    }
    operator!=(const SurfaceMap& smap) {
        return !(*this == smap);
    }
    SurfaceMap& operator=(const SurfaceMap& smap) {
        copy(smap);
        return *this;
    }

    private:

    void        copy(const SurfaceMap& smap) {
        slot_ = smap.slot_;
        side_ = smap.side_;
        sid_ = smap.sid_;
        volname_ = smap.volname_;
    }
};

//-----------------------------------------------------------------------------
// CLASS:  SurfaceMapList                                       --JKH
//
// This class is simply a list of SurfaceMaps.
// 
// PUBLIC DATA MEMBERS:
//
// RWGSlist(SurfaceMap) smList_         List of SurfaceMaps
//
// PUBLIC METHODS:
//
// SurfaceMapList()
//      Constructor.  Does nothing.
// ~SurfaceMapList()
//      Destructor.  Does nothing.
// search()
//      Overloaded function:  searches for an entry given a slot/side pair,
//      or searches given an Sid_t.  In either case, it returns a pointer to
//      the found entry on success.
// clear()
//      Clears the contents of the SurfaceMapList
// operator=()
//      Assignment operator (deep copy)
// at()
//      Returns a pointer to the item given by the specified index.
// append()
//      Appends a new SurfaceMap to the list.
// remove()
//      Removes an item given by the specified index from the SurfaceMapList.
// entries()
//      Returns the number of entries in the list.
//
// USAGE NOTES:
//-----------------------------------------------------------------------------
// COLLABORATORS:
//
// SurfaceMap
// RWGSlist
//
// PRIVATE DATA MEMBERS:        None.
//
// PRIVATE METHODS:             
//
// copy()
//      Does a deep copy of one SurfaceMapList to another.
//
// free()
//      Clears the contents of the SurfaceMapList and frees dynamic memory.
//
// IMPLEMENTATION NOTES:        None.
//-----------------------------------------------------------------------------

declare(RWGSlist,SurfaceMap)

class SurfaceMapList
{
    public:

    RWGSlist(SurfaceMap)        theList_;

    SurfaceMapList() {}
    SurfaceMapList(const SurfaceMapList& obj) { copy(obj); }
    ~SurfaceMapList() { free(); }

    SurfaceMap *        search(UINT16 slot,UINT16 side);
    SurfaceMap *        search(Sid_t sid);
    SurfaceMap *        search(const RWCString& volname);
    void                clear() { free(); }
    SurfaceMapList&     operator=(const SurfaceMapList& obj) {
        copy(obj);
        return *this;
    }
    SurfaceMap *        at(UINT32 i);
    const SurfaceMap *  at(UINT32 i) const;
    void                append(const SurfaceMap& newSMap);
    void                remove(UINT32 i);
    UINT32              entries() const { return theList_.entries(); }

    private:

    void copy(const SurfaceMapList& smList);
    void free();
};


//-----------------------------------------------------------------------------
// CLASS:  SidList                                              --JKH
//
// This class is simply a list of Sid_t's.
// 
// PUBLIC DATA MEMBERS:
//
// RWGSlist(Sid_t)      theList_        List of available sid's
//
// PUBLIC METHODS:
//
// SidList()
//      Constructor.  Does nothing.
// ~SidList()
//      Destructor.  Does nothing.
//
// USAGE NOTES:
//-----------------------------------------------------------------------------
// COLLABORATORS:               
//
// RWGSlist(Sid_t)
//
// PRIVATE DATA MEMBERS:        None.
//
// PRIVATE METHODS:             None.
//
// IMPLEMENTATION NOTES:        None.
//-----------------------------------------------------------------------------

declare(RWGSlist,Sid_t)

class SidList 
{
    public:

    RWGSlist(Sid_t)             theList_;

    SidList() {}
    SidList(const SidList& sidlist) { copy(sidlist); }
    ~SidList() {}

    Sid_t *             search(Sid_t sid);
    void                clear() { free(); }
    SidList&            operator=(const SidList& sidlist) {
        copy(sidlist);
        return *this;
    }
    INT32               operator==(const SidList& sidlist);
    Sid_t *             at(UINT32 i);
    const Sid_t *       at(UINT32 i) const;
    void                append(const Sid_t& newSid);
    void                remove(UINT32 i);
    UINT32              entries() const { return theList_.entries(); }

    private:

    void                copy(const SidList& sidlist);
    void                free();
};

//-----------------------------------------------------------------------------
// CLASS:  AvailSurfaces                                        --JKH
//
// This class is a simple data structure which keeps a list of Sid_t's 
// associated with a particular type of media as specified by sector size
// and type (MO/WORM).
// 
// PUBLIC DATA MEMBERS:
//
// UINT32               sectorSize_     Sector size, in bytes
// MediaType            mtype_          MO or WORM media
// RWGSlist(Sid_t)      sidList_        List of available sid's
//
// PUBLIC METHODS:
//
// AvailSurfaces()
//      Constructor.  Does nothing.
// ~AvailSurfaces()
//      Destructor.  Does nothing.
//
// USAGE NOTES:
//-----------------------------------------------------------------------------
// COLLABORATORS:               
//
// RWGSlist(Sid_t)
//
// PRIVATE DATA MEMBERS:        None.
//
// PRIVATE METHODS:             None.
//
// IMPLEMENTATION NOTES:        None.
//-----------------------------------------------------------------------------


class AvailSurfaces
{
    public:

    UINT32              sectorSize_;
    MediaType           mtype_;
    SidList             sidList_;

    AvailSurfaces() : sectorSize_(0),mtype_(NSR::media_Rewritable) {}
    ~AvailSurfaces() {}

    operator==(const AvailSurfaces& as) { 
        return ((sectorSize_ == as.sectorSize_) &&
                (mtype_ == as.mtype_) &&
                (sidList_ == as.sidList_));
    }
    operator!=(const AvailSurfaces& as) {
        return !(*this == as);
    }
    AvailSurfaces& operator=(const AvailSurfaces& as) {
        copy(as);
        return *this;
    }

    private:

    copy(const AvailSurfaces& as) {
        sectorSize_ = as.sectorSize_;
        mtype_ = as.mtype_;
        sidList_ = as.sidList_;
    }
        
};

//-----------------------------------------------------------------------------
// CLASS:  AvailSurfacesList                                    --JKH
//
// This class is a simple list of AvailSurfaces objects.
// 
// PUBLIC DATA MEMBERS:
//
// RWGSlist(AvailSurfaces)      theList_        Available surface lists
//
// PUBLIC METHODS:
//
// AvailSurfacesList()
//      Constructor.  Does nothing.
//
// ~AvailSurfacesList()
//      Destructor.  Does nothing.
//
// USAGE NOTES:
//-----------------------------------------------------------------------------
// COLLABORATORS:
//
// AvailSurfaces
// RWGSlist(AvailSurfaces)
//
// PRIVATE DATA MEMBERS:        None.
//
// PRIVATE METHODS:             None.
//
// IMPLEMENTATION NOTES:        None.
//-----------------------------------------------------------------------------

declare(RWGSlist,AvailSurfaces)

class AvailSurfacesList
{
    public:

    RWGSlist(AvailSurfaces)     theList_;

    AvailSurfacesList() {}
    AvailSurfacesList(const AvailSurfacesList& obj) { copy(obj); }
    ~AvailSurfacesList() { free(); }
    AvailSurfaces *     search(UINT32 ssize,MediaType mtype);
    void                clear() { free(); }
    AvailSurfacesList&  operator=(const AvailSurfacesList& asList) {
        copy(asList);
        return *this;
    }
    AvailSurfaces *     at(UINT32 i);
    const AvailSurfaces *at(UINT32 i) const;
    void                append(const AvailSurfaces& newAS);
    void                remove(UINT32 i);
    UINT32              entries() const { return theList_.entries(); }

    private:

    void                copy(const AvailSurfacesList& asList);
    void                free();
};

//-----------------------------------------------------------------------------
// CLASS:  VolsetNameMap                                        --JKH
//
// This class is a simple data structure providing a mapping between a
// volume set name and a list of Sid_t's.
// 
// PUBLIC DATA MEMBERS:
//
// RWCString            vsname_         Volume set name
// RWGSlist(Sid_t)      sidList_        List of sid's
//
// PUBLIC METHODS:
//
// VolsetNameMap()
//      Constructor.  Initializes the volume set name.
// ~VolsetNameMap()
//      Destructor.  Does nothing.
// operator==()
//      Equality operator between two VolsetNameMap objects.
//
// USAGE NOTES:
//-----------------------------------------------------------------------------
// COLLABORATORS:
//
// RWCString
// RWGSlist
//
// PRIVATE DATA MEMBERS:        None.
//
// PRIVATE METHODS:             None.
//
// IMPLEMENTATION NOTES:        None.
//-----------------------------------------------------------------------------

class VolsetNameMap
{
    public:

    RWCString           vsname_;                // Volume set name
    SidList             sidList_;               // List of sid's

    VolsetNameMap(const RWCString& vsname) : vsname_(vsname) {}
    VolsetNameMap(const VolsetNameMap& vsnmap) { copy(vsnmap); }
    ~VolsetNameMap() { free(); }

    VolsetNameMap& operator=(const VolsetNameMap& vsnmap) { 
        copy(vsnmap); 
        return *this;
    }
    operator==(const VolsetNameMap& vsnmap) {
        return (vsname_ == vsnmap.vsname_);
    }
    operator!=(const VolsetNameMap& vsnmap) {
        return !(*this == vsnmap);
    }

    private:

    void copy(const VolsetNameMap& vsnmap) {
        vsname_ = vsnmap.vsname_;
        sidList_ = vsnmap.sidList_;
    }
    void free() { sidList_.clear(); }   
};

//-----------------------------------------------------------------------------
// CLASS:  VolsetNameMapList                                    --JKH
//
// This class is a simple list of VolsetNameMap objects.
// 
// PUBLIC DATA MEMBERS:
//
// RWGSlist(VolsetNameMap)      vsnmList_       List of VolsetNameMaps
//
// PUBLIC METHODS:
//
// USAGE NOTES:
//-----------------------------------------------------------------------------
// COLLABORATORS:
//
// VolsetNameMap
// RWGSlist(VolsetNameMap)
//
// PRIVATE DATA MEMBERS:
//
// PRIVATE METHODS:
//
// IMPLEMENTATION NOTES: 
//-----------------------------------------------------------------------------

declare(RWGSlist,VolsetNameMap)

class VolsetNameMapList
{
    public:

    RWGSlist(VolsetNameMap)     theList_;

    VolsetNameMapList() {}
    VolsetNameMapList(const VolsetNameMapList& vnmList) { copy(vnmList); }
    ~VolsetNameMapList() {}

    VolsetNameMap       *search(const RWCString& vsname);
    void                clear() { free(); }
    VolsetNameMapList&  operator=(const VolsetNameMapList& obj) {
        copy(obj);
        return *this;
    }
    VolsetNameMap *     at(UINT32 i);
    const VolsetNameMap *at(UINT32 i) const;
    void                append(const VolsetNameMap& newVNMap);
    void                remove(UINT32 i);
    UINT32              entries() const { return theList_.entries(); }

    private:

    void                copy(const VolsetNameMapList& vnmList);
    void                free();
};

//-----------------------------------------------------------------------------
// CLASS:  FilesetNameMap                                       --JKH
//
// This class is a simple data structure providing a mapping between a
// fileset name and a volume set name.
// 
// PUBLIC DATA MEMBERS:
//
// RWGString            fsname_         Fileset name
// RWCString            vsname_         Volume set name
//
// PUBLIC METHODS:
//
// FilesetNameMap()
//      Constructor.  Initializes the fileset name.
// ~FilesetNameMap()
//      Destructor.  Does nothing.
// operator==()
//      Equality operator between two FilesetNameMap objects.
//
// USAGE NOTES:
//-----------------------------------------------------------------------------
// COLLABORATORS:
//
// RWCString
// RWGSlist
//
// PRIVATE DATA MEMBERS:        None.
//
// PRIVATE METHODS:             None.
//
// IMPLEMENTATION NOTES:        None.
//-----------------------------------------------------------------------------

class FilesetNameMap
{
    public:

    RWCString           fsname_;
    RWCString           vsname_;

    FilesetNameMap(const RWCString& fsname) : fsname_(fsname) {}
    FilesetNameMap(const RWCString& fsname,const RWCString& vsname) :
                   fsname_(fsname), vsname_(vsname) {}
    FilesetNameMap(const FilesetNameMap& map) { copy(map); }
    ~FilesetNameMap() {}
    operator==(const FilesetNameMap& map) {
        return (fsname_ == map.fsname_);
    }
    operator!=(const FilesetNameMap& map) {
        return !(*this == map);
    }
    FilesetNameMap      operator=(const FilesetNameMap& map) {
        copy(map);
        return *this;
    }

    private:

    void                copy(const FilesetNameMap& map) {
        fsname_ = map.fsname_;
        vsname_ = map.vsname_;
    }
};

//-----------------------------------------------------------------------------
// CLASS:  FilesetNameMapList                                   --JKH
//
// This class is a simple list of FilesetNameMap objects.
// 
// PUBLIC DATA MEMBERS:
//
// RWGSlist             theList_        List of FilesetNameMaps
//
// PUBLIC METHODS:
//
// FilesetNameMapList()
//      Constructor.  
// ~FilesetNameMapList()
//      Destructor.  Does nothing.
// search()
//      Searches for a FilesetNameMap with the given file set name, and
//      returns a pointer to the found object on succes.
//
// USAGE NOTES:
//-----------------------------------------------------------------------------
// COLLABORATORS:
//
// FilesetNameMap
// RWGSlist
//
// PRIVATE DATA MEMBERS:        None.
//
// PRIVATE METHODS:             None.
//
// IMPLEMENTATION NOTES:        None.
//-----------------------------------------------------------------------------

declare(RWGSlist,FilesetNameMap)

class FilesetNameMapList
{
    public:

    RWGSlist(FilesetNameMap)    theList_;

    FilesetNameMapList() {}
    FilesetNameMapList(const FilesetNameMapList& fnmList) { copy(fnmList); }
    ~FilesetNameMapList() {}
    FilesetNameMap      *search(const RWCString& fsname);
    FilesetNameMap      *searchByVSName(const RWCString& vsname);
    void                clear() { free(); }
    FilesetNameMapList& operator=(const FilesetNameMapList& list) {
        copy(list);
        return *this;
    }
    FilesetNameMap *    at(UINT32 i);
    const FilesetNameMap *at(UINT32 i) const;
    void                append(const FilesetNameMap& newMap);
    void                remove(UINT32 i);
    UINT32              entries() const { return theList_.entries(); }

    private:

    void                copy(const FilesetNameMapList& list);
    void                free();
};

//-----------------------------------------------------------------------------
// CLASS:  VolumeFreeSpace                                      --JKH
//
// This class is a simple data structure to store free space information
// for a volume.
// 
// PUBLIC DATA MEMBERS:
//
// Sid_t                sid_            Volume's sid
// RWCString            volname_        Volume's name
// ExtentList           freeSpace_      Volume's free space
// ExtentList           erasedSpace_    Volume's erased space
// UINT32               dataStart_      Volume's data start sector
//
// PUBLIC METHODS:
//
// VolumeFreeSpace()
//      Constructor.  
// ~VolumeFreeSpace()
//      Destructor.  Does nothing.
// operator==()
//      Equality operator between two VolumeFreeSpace objects.
// operator=()
//      Assignment operator between two VolumeFreeSpace objects.
//
// USAGE NOTES:
//-----------------------------------------------------------------------------
// COLLABORATORS:               None.
//
// PRIVATE DATA MEMBERS:        None.
//
// PRIVATE METHODS:             None.
//
// IMPLEMENTATION NOTES:        None.
//-----------------------------------------------------------------------------

class VolumeFreeSpace
{
    public:

    Sid_t               sid_;
    RWCString           volname_;
    ExtentList          freeSpace_;
    ExtentList          erasedSpace_;
    UINT32              dataStart_;

    VolumeFreeSpace() : sid_(SID_T_INVALID), dataStart_(0) {}
    VolumeFreeSpace(const VolumeFreeSpace& vfs) {
        copy(vfs);
    }
    ~VolumeFreeSpace() {}

    operator==(const VolumeFreeSpace& vfs) { 
        return ((sid_ == vfs.sid_) && (dataStart_ == vfs.dataStart_) &&
                (volname_ == vfs.volname_));
    }

    VolumeFreeSpace&    operator=(const VolumeFreeSpace& vfs) {
        copy(vfs);
        return *this;
    }

    void print(void) {
        printf("Volume Free Space:  sid_ = %d, volname = %s, data start = %u\n",
               sid_,(CHAR_PTR) volname_,dataStart_);
        printf("\tFree space:\n");
        freeSpace_.print();
        printf("\tErased space:\n");
        erasedSpace_.print();
    }

    private:

    void        copy(const VolumeFreeSpace& vfs) {
        sid_ = vfs.sid_;
        volname_ = vfs.volname_;
        freeSpace_ = vfs.freeSpace_;
        erasedSpace_ = vfs.erasedSpace_;
        dataStart_ = vfs.dataStart_;
    }
};

//-----------------------------------------------------------------------------
// CLASS:  VolumeFreeSpaceList                                  --JKH
//
// This class is a simple list of VolumeFreeSpaceList objects.
// 
// PUBLIC DATA MEMBERS:
//
// RWGSlist             theList_        List of VolumeFreeSpaceList
//
// PUBLIC METHODS:
//
// VolumeFreeSpaceList()
//      Constructor.  
// ~VolumeFreeSpaceList()
//      Destructor.  Does nothing.
// search()
//      Searches for a VolumeFreeSpaceList with the given sid, and
//      returns a pointer to the found object on succes.
//
// USAGE NOTES:
//-----------------------------------------------------------------------------
// COLLABORATORS:
//
// VolumeFreeSpace
// RWGSlist
//
// PRIVATE DATA MEMBERS:        None.
//
// PRIVATE METHODS:             None.
//
// IMPLEMENTATION NOTES:        None.
//-----------------------------------------------------------------------------

declare(RWGSlist,VolumeFreeSpace)

class VolumeFreeSpaceList
{
    public:

    RWGSlist(VolumeFreeSpace)   theList_;

    VolumeFreeSpaceList() {}
    VolumeFreeSpaceList(const VolumeFreeSpaceList& vfsList) { copy(vfsList); }
    ~VolumeFreeSpaceList() {}
    VolumeFreeSpace     *search(Sid_t sid);
    void                clear() { free(); }
    VolumeFreeSpaceList& operator=(const VolumeFreeSpaceList& list) {
        copy(list);
        return *this;
    }
    VolumeFreeSpace *    at(UINT32 i);
    const VolumeFreeSpace *at(UINT32 i) const;
    void                append(const VolumeFreeSpace& newvfs);
    void                remove(UINT32 i);
    UINT32              entries() const { return theList_.entries(); }

    void                print(void); 

    private:

    void                copy(const VolumeFreeSpaceList& list);
    void                free();
};

//-----------------------------------------------------------------------------
// CLASS:  NSRJukeBox                                           --JKH
//
// The NSRJukeBox class is an abstraction of an NSR-format cognizant jukebox
// device.  It combines the raw device functionality of the MSDevice class
// with NSR volume, volume set, and fileset management capability.
// 
// PUBLIC DATA MEMBERS:
// 
// None.
//
// PUBLIC METHODS:
//
// USAGE NOTES:
//-----------------------------------------------------------------------------
// COLLABORATORS:
//
// hpACDev
// arcIO
// SurfaceMap
// SurfaceMapList
// AvailSurfaces
// AvailSurfacesList
// VolsetNameMap
// VolsetNameMapList
// FilesetNameMap
// FilesetNameMapList
// RWCString
// RWGSlist
//
// PRIVATE DATA MEMBERS:
//
// hpACDev              *acdev_         Autochanger device object
// RWCString            jbstatFile_     Name of status file
// RWCString            wvs_            Name of working volume set
// SurfaceMapList       smList_         List of surface mappings
// AvailSurfacesList    asList_         List of available media
// VolsetNameMapList    vnmList_        List of volume set name mappings
// FilesetNameMapList   fnmList_        List of file set name mappings
// VolumeFreeSpaceList  vfsList_        List of freespace info for volumes
// Error                err_            Error from previous operation
//
// PRIVATE METHODS:
//
// makeACDev()
//      Reads and parses the acrc file, and constructs an hpACDev.
// mapSurfaces()
//      Gets element addresses and element status information, and generates
//      a SurfaceMapList for all surfaces in the jukebox.
//
//
// IMPLEMENTATION NOTES: 
//-----------------------------------------------------------------------------

class NSRJukeBox
{
    //-------------------------------------------------------------------------
    // Private Data
    //-------------------------------------------------------------------------

    private:

    Archive             *arc_;
    hpACDev             *acdev_;                
    RWCString           jbstatFile_;
    RWCString           wvs_;           
    Boolean             newvs_;
    Boolean             openIntegrity_;
    UINT32              sectorSize_;
    MediaType           mtype_;
    SurfaceMapList      smList_;
    AvailSurfacesList   asList_;
    VolsetNameMapList   vnmList_;
    FilesetNameMapList  fnmList_;
    VolumeFreeSpaceList vfsList_;
    Error               err_;

    //-------------------------------------------------------------------------
    // Public Member Functions
    //-------------------------------------------------------------------------

    public:

    // C++isms:  constructors, destructors

    NSRJukeBox(const RWCString& acrcFile);      // Constructor
    ~NSRJukeBox();

    // Status file methods

    Error               readStatusFile(RWCString& statusFile);
    Error               genStatusFile(RWCString& statusFile);

    // Volume set methods

    Error               setNewWVS(const RWCString& vsname,UINT32 sectorSize,
                                  MediaType mtype);
    Error               setExistingWVS(const RWCString& vsname);
    Error               beginGrow(void);
    Error               growWVS(const RWCString& volname,UINT32 dataStart);
    Error               growWVS(const RWCString& volname,Sid_t sid,
                                UINT32 dataStart);
    Error               growWVS(const RWCString& volname,UINT16 slot,
                                UINT16 side,UINT32 dataStart);
    Error               completeGrow(void);
    Error               freeVolumeSet(const RWCString& vsname);
    Error               getImplementationInfo(const RWCString& fsname,
                                              ImplementationInfo& iinfo);
    Error               importArchive(const RWCString& fsname,
                                      TreeDB& treeDB,
                                      TreeNode& root);
    Error               prepareArchive(void);
    Error               exportArchive(const RWCString& fsname,
                                      TreeNode& root,
                                      VolumeFreeSpaceList& vfsList);

    // Database methods
    
    const SurfaceMapList&      surfaceMapList(void) { return smList_; }
    const AvailSurfacesList&   availSurfacesList(void) { return asList_; }
    const VolsetNameMapList&   volsetNameMapList(void) { return vnmList_; }
    const FilesetNameMapList&  filesetNameMapList(void) { return fnmList_; }
    const VolumeFreeSpaceList& volumeFreeSpaceList(void) { return vfsList_; }

    Error               wvsFreeSpace(VolumeFreeSpaceList& vfsl);
    const RWCString&    wvsName() { return wvs_; }
    UINT32              wvsSectorSize(void) { return sectorSize_; }
    MediaType           wvsMediaType(void) { return mtype_; }

    // Miscellaneous methods

    hpACDev *           acdev(void) { return acdev_; }
    void                processIdentifier(RWCString& id,INT32 which);
    void                print();
    Error               error() { return err_; }

    /**/void writeAsText (String& text) 
            {
                NSR::VolumeSetInfo      vsi;
                arc_->getVolSetInfo(&vsi);
                vsi.writeAsText (text);
            }
            
                

    //-------------------------------------------------------------------------
    // Private Member Functions
    //-------------------------------------------------------------------------

    private:

    INT32               makeACDev(const RWCString& acrcFile);
    INT32               mapSurfaces();
    INT32               writeStatusFile();
    void                initArchive();
#ifdef ROGUEWAVIFIED
    INT32               recursiveGetEntry(Extent& extent,TreeDB& db,
                                          TreeNode& node);
#else
    INT32               recursiveGetEntry(Extent& extent,hash_table *icb_table,
                                          TreeNode& node);
#endif  // ROGUEWAVIFIED
    void                countNodes(TreeNode& node,UINT32 *ndirs,UINT32 *nfiles,
                                   UINT32 *overflow);
    INT32               recursiveWriteDirectory(TreeNode& node,INT32 isroot);
    INT32               recursiveWriteICB(TreeNode& node,INT32 isroot);

};

class MyTokenizer 
{
    UINT32      curpos;
    RWCString   orig;

    public:

    MyTokenizer(const RWCString& str) : curpos(0),orig(str) {}

    RWCSubString operator()(void);
};

#endif  // _NSRJB_H_
