//-*-c++-*-
//-----------------------------------------------------------------------------
// syndata.cpp
//
// This file contains implementations for the SynthData class.
//
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//
// CLASSES:
//
// SynthData
//
//-----------------------------------------------------------------------------
// $Source: /projects/HMS/nsr/udfAncil/msTest/syndata.cpp,v $
// $Revision: 1.2 $     $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 95/12/01 10:07:50 $
//-----------------------------------------------------------------------------

#include "syndata.h"

SynthData::SynthData(const char* seed, const long offset, const long len)
{
    makeData(seed, offset, len);
}

void SynthData::makeData(const char* seed, const long offset, const long len)
{
    if (len < 1) {
        data_[0]='\0';
        fprintf (stderr, "Error in SynthData::SynthData: ");
        fprintf (stderr, "Zero length pattern requested.\n");
        return;
    }
    if (len > MAX_IOSIZE) {
        data_[0]='\0';
        fprintf (stderr, "Error in SynthData::SynthData: ");
        fprintf (stderr, "pattern > 256K requested.\n");
        return;
    }
    data_ = new char [len*2];
    sOffset = new char[512];

    long seedLen = strlen(seed);
    if (seedLen >= (MAX_IOSIZE - OFFSET_STRLEN)) {
        data_[0]='\0';
        fprintf (stderr, "Error in SynthData::SynthData: ");
        fprintf (stderr, "Attempt to pattern with over-sized seed.\n");
        return;
    }

    long i = 0;
    data_[i++] = '[';
    while (i < len) {
        sprintf (sOffset, "<^%10lx>", (long) (i + offset+2));
        memmove (data_+i, sOffset, OFFSET_STRLEN);
        i+= OFFSET_STRLEN;

        if (i < len) {
            memmove(data_+i, seed, seedLen);
            i += seedLen;
        }
    }
    data_[len-2]=']';
    data_[len-1]='\0';

    needToMake = FALSE;
}

const unsigned char* SynthData::data()
{
    if (needToMake) {
        fprintf (stderr, "Error in SynthData::data().  Data has not been built.");
        return NULL;
    }
    return (unsigned char*)(char*)data_;
}

SynthData::~SynthData() {
    delete [] data_;
    delete [] sOffset;
}














