%{
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "arcSynth.h"
#include "synthY.h"

#ifdef TRACE
#define LEXTRACE(A) printf ("lex: %s\n", (A))
#else
#define LEXTRACE(A)
#endif

int Line = 1;

%}

ws                      [ \t]+
comment                 #.*
qstring                 \"[^\"\n]*[\"\n]
token                   [a-zA-Z][a-zA-Z0-9]*
path                    ([a-zA-z0-9/_.]*)|(\"[^\"\n][a-zA-Z/_\.][a-zA-Z0-9/_\.]*[\"\n])
nl                      \n
integer                 [0-9]+
Kint                    [0-9]+K
Mint                    [0-9]+M


%x sALLOC

%%

<sALLOC,INITIAL>\n      { Line++; };
<sALLOC,INITIAL>{ws}      ;
<sALLOC,INITIAL>{comment} ;

<INITIAL>{integer}      {LEXTRACE("<INITIAL>integer");
                         synthlval.string = strdup(synthtext);
                         return INTEGER;}
<INITIAL>{Kint}         {LEXTRACE("<INITIAL>Kint");
                         synthlval.string = strdup(synthtext);
                         return KINT;}
<INITIAL>{Mint}         {LEXTRACE("<INITIAL>Mint");
                         synthlval.string = strdup(synthtext);
                         return MINT;}
<sALLOC>{integer}       {LEXTRACE("<sAlloc>integer");
                         synthlval.string = strdup(synthtext);
                         return ALLOC_INTEGER;}
<sALLOC>,               {LEXTRACE("<sALLOC>,");return ALLOC_SEPARATOR; }

<INITIAL>{qstring}      {LEXTRACE("qstring");
                        synthlval.string = strdup(synthtext+1);
                        if(synthlval.string[synthleng-2] != '"')
                                fprintf(stderr, "Unterminated character string",(char *)0);
                        else
                                synthlval.string[synthleng-2] = '\0'; /* remove close quote */
                        return QSTRING;
                        }

<sALLOC>[\{]            { LEXTRACE("open alloc"); 
                          return ALLOC_BEGIN;
                        }
<sALLOC>[\}]            { LEXTRACE("close alloc");
                          BEGIN(INITIAL); 
                          return ALLOC_END;
                        }

<INITIAL>mkdir          { LEXTRACE("mkdir"); return SYNTH_MKDIR;}
<INITIAL>cd             { LEXTRACE("cd"); return SYNTH_CD;}
<INITIAL>mkfile         { LEXTRACE("mkfile"); return SYNTH_FILE;}
<INITIAL>mklink         { LEXTRACE("lns"); return SYNTH_SLINK;}
<INITIAL>attr           { LEXTRACE("attr"); return SYNTH_ATTR;}
<INITIAL>alloc          { LEXTRACE("alloc"); BEGIN(sALLOC); return SYNTH_ALLOC;}

<sALLOC>F               { LEXTRACE("F"); return ALLOC_FIRST; }
<sALLOC>L               { LEXTRACE("L"); return ALLOC_LAST; }
<sALLOC>e               { LEXTRACE("e"); return ALLOC_ERASED; }
<sALLOC>u               { LEXTRACE("u"); return ALLOC_UNERASED; }
<sALLOC>[\!]            { LEXTRACE("!"); return ALLOC_ALL; }
<sALLOC>[\~]            { LEXTRACE("~"); return ALLOC_EVERYTHING;}
<sALLOC>[\+]            { LEXTRACE("+"); return ALLOC_INCREMENT; }
<sALLOC>[\*]            { LEXTRACE("*"); return ALLOC_REPEAT; }

<INITIAL>{path}         {LEXTRACE("path"); LEXTRACE(synthtext);
                         synthlval.string = strdup(synthtext);
                         return PATH;}                  

<INITIAL>{token}        {LEXTRACE("token");LEXTRACE(synthtext);
                         synthlval.string = strdup(synthtext);
                         return TOKEN;}                 
%%

void 
syntherror (char *s)  
{
        printf ("\n%s on line %d, at %s.\n", s, Line, synthtext);
        printf ("This is probably a syntax error in the synth script file.\n\n");
        SynthFile::invalidateObject();
        exit(1);
}

int synthwrap()
{
    return 1; // tell parser to stop at eof
}


