%{
#include "arcSynth.h"

extern int yylex();
extern void yyerror(char *s);

#ifdef TRACE
#define YACCTRACE(A) printf ("yacc: %s\n", (A))
#else
#define YACCTRACE(A)
#endif

%}

%union {
        char *string;
}

%token SYNTH_MKDIR SYNTH_CD SYNTH_FILE SYNTH_SLINK SYNTH_ALLOC SYNTH_ATTR
%token ALLOC_FIRST ALLOC_LAST ALLOC_INCREMENT ALLOC_ERASED ALLOC_UNERASED
%token ALLOC_REPEAT ALLOC_ALL ALLOC_EVERYTHING ALLOC_END ALLOC_SEPARATOR
%token ALLOC_BEGIN

%token <string> QSTRING TOKEN INTEGER ALLOC_INTEGER PATH KINT MINT
%type <string> qstring token alloc_integer integer path Kint Mint

%%

synth:                  synth_dir
                        | synth_file
                        | synth_slink
                        | synth_cd
                        | synth_attr
                        | synth_alloc
                        | synth synth_dir
                        | synth synth_file
                        | synth synth_slink
                        | synth synth_cd
                        | synth synth_attr
                        | synth synth_alloc
                        ;

synth_dir:              SYNTH_MKDIR PATH {
                                YACCTRACE("synth_dir");
                                synth_dir ($2);
                        }
                        | SYNTH_MKDIR qstring {
                                YACCTRACE("synth_dir");
                                synth_dir ($2);
                                free($2);
                        };


synth_slink:            SYNTH_SLINK PATH PATH {
                                YACCTRACE("synth_slink");
                                synth_slink ($3, $2);
                                free($2);
                                free($3);
                        }
                        | SYNTH_SLINK qstring qstring {
                                YACCTRACE("synth_slink");
                                synth_slink ($3, $2);
                                free($2);
                                free($3);
                        }
                        | SYNTH_SLINK PATH qstring {
                                YACCTRACE("synth_slink");
                                synth_slink ($3, $2);
                                free($2);
                                free($3);
                        }
                        | SYNTH_SLINK qstring PATH {
                                YACCTRACE("synth_slink");
                                synth_slink ($3, $2);
                                free($2);
                                free($3);
                        };

synth_cd:               SYNTH_CD PATH {
                                YACCTRACE("synth_cd");
                                synth_cd ($2);
                        }
                        | SYNTH_CD qstring {
                                YACCTRACE("synth_cd");
                                synth_cd ($2);
                                free($2);
                        };

synth_file:             SYNTH_FILE PATH integer {
                                YACCTRACE("synth_file");
                                synth_file($2, $3);
                                free ($3);
                        }
                        | SYNTH_FILE qstring integer {
                                YACCTRACE("synth_file");
                                synth_file($2, $3);
                                free ($2);
                                free ($3);
                        }
                        | SYNTH_FILE qstring Kint {
                                YACCTRACE("synth_file");
                                synth_file_k($2, $3);
                                free($2);
                                free ($3);
                        }
                        | SYNTH_FILE qstring Mint {
                                YACCTRACE("synth_file");
                                synth_file_m($2, $3);
                                free ($2);
                                free ($3);
                        }
                        | SYNTH_FILE PATH Kint {
                                YACCTRACE("synth_file");
                                synth_file_k($2, $3);
                                free ($2);
                                free ($3);
                        }
                        | SYNTH_FILE PATH Mint {
                                YACCTRACE("synth_file");
                                synth_file_m($2, $3);
                                free ($2);
                                free ($3);
                        }
                        | SYNTH_FILE PATH {
                                YACCTRACE("synth_file");
                                synth_file($2);
                        }
                        | SYNTH_FILE qstring {
                                YACCTRACE("synth_file");
                                synth_file($2);
                        };

synth_attr:             SYNTH_ATTR qstring {
                                YACCTRACE("synth_attributes");
                                synth_attr($2);
                        };

synth_alloc:            alloc_begin alloc_contents alloc_end {
                                YACCTRACE("synth_allocation");
                        };


    /* allocation */

alloc_begin:            SYNTH_ALLOC ALLOC_BEGIN{
                            YACCTRACE("synth_alloc");
                            synth_begin_alloc();    
                        };

alloc_end:              ALLOC_END {
                            YACCTRACE("synth_end");
                            synth_end_alloc();
                        };

alloc_contents:         alloc_item
                        | alloc_contents separator alloc_item;

separator:              ALLOC_SEPARATOR {
                                YACCTRACE("alloc_separator");
                                YACCTRACE("alloc_separator");
                                synth_alloc_separator();
                        };
                                
alloc_item:             first
                        | last
                        | erased
                        | unerased
                        | increment
                        | repeat
                        | all
                        | everything
                        | sectors
                        | alloc_item first
                        | alloc_item last
                        | alloc_item erased
                        | alloc_item unerased
                        | alloc_item increment
                        | alloc_item repeat
                        | alloc_item all
                        | alloc_item everything
                        | alloc_item sectors
                        ;

first:                  ALLOC_FIRST {
                                YACCTRACE("alloc_first");
                                alloc_first();
                        };
last:                   ALLOC_LAST {
                                YACCTRACE("alloc_last");
                                alloc_last();
                        };
erased:                 ALLOC_ERASED {
                                YACCTRACE("alloc_erased");
                                alloc_erased();
                        };
unerased:               ALLOC_UNERASED {
                                YACCTRACE("alloc_unerased");
                                alloc_unerased();
                        };
increment:              ALLOC_INCREMENT {
                                YACCTRACE("alloc_increment");
                                alloc_increment();
                        };
repeat:                 ALLOC_REPEAT alloc_integer {
                                YACCTRACE("alloc_repeat");
                                alloc_repeat($2);
                                free ($2);
                        };
all:                    ALLOC_ALL {
                                YACCTRACE("alloc_ALL");
                                alloc_all();
                        };
everything:             ALLOC_EVERYTHING {
                                YACCTRACE("alloc_EVERYTHING");
                                alloc_everything();
                        };
sectors:                alloc_integer {
                                YACCTRACE("alloc_integer");
                                alloc_sectors($1);
                                free ($1);
                        };

    /* atomics */

qstring:                QSTRING {$$ = $1;};
path:                   PATH {$$ = $1;};
token:                  TOKEN {$$ = $1;};
alloc_integer:          ALLOC_INTEGER {$$ = $1;};
integer:                INTEGER {$$ = $1;};
Kint:                   KINT {$$ = $1;};
Mint:                   MINT {$$ = $1;};
