//-*-c++-*--------------------------------------------------------------
//  udf2cmr.cpp
//
// WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING
//
// This program is given as an example.  It has a bug in the tree database
// code that causes that list file in the fileset to be improperly processed.
// The important part of the program is just to follow the sequence of calls
// to the Archive object to understand how to do an import.  For a fully
// functional, albeit more complicated, example, please see arcRead and 
// arcWrite.
//
// WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING
//
//   Program to read in udf and spit out CMR
//
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//
//----------------------------------------------------------------------
//
// $Source: /projects/HMS/nsr/udfAncil/msTest/udf2cmr.cpp,v $
// $Revision: 1.3 $     $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 96/06/14 13:57:25 $
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <rw/cstring.h>
#include "nsrLib.h"
#include "dlist.h"                      // Doubly-linked list toolkit
#include "dqueue.h"                     // Queue macro package based on dlist
#include "hash.h"                       // Hash table based on dlist
#include "arcImEx.h"
#include "cmr.h"

//-----------------------------------------------------------------------------
// Some data types
//-----------------------------------------------------------------------------

struct srfcEntry {
    int         slot;
    int         side;
    sid_t       sid;
    ExtentList  elist;          // Used primarily in export
    char        *volname;       // Used primarily in export

    srfcEntry(void) : slot(0),
                      side(0),
                      sid(SID_T_INVALID),
                      volname(NULL) {}
    ~srfcEntry(void) {
        if(volname != NULL) {
            delete [] volname;
        }
    }
};

//-----------------------------------------------------------------------------
// Function prototypes
//-----------------------------------------------------------------------------

static void     Usage(void);
static hpACDev *configDevice(RWCString *devf,int ask,dlist *srfcList);
static int      getDriveMaps(hpACDev& acdev);
static int      srfc_cmp(void *d1,void *d2);
static int      srfc_free(void *d1);
static int      importArchive(hpACDev *acdev,dlist *srfcList,TreeDB& treeDB,
                              TreeNode& root);
#ifdef ROGUEWAVIFIED
static int      recursiveGetEntry(Extent& ext,Archive& arc,
                                  TreeDB& treeDB,
                                  TreeNode& tnode);
#else
static int      recursiveGetEntry(Extent& ext,Archive& arc,
                                  hash_table *icb_table,TreeDB& treeDB,
                                  TreeNode& tnode);
#endif
static int      TreeNode_hash(void *d1);
static int      TreeNode_cmp(void *d1,void *d2);
static void     freeTree(TreeNode& node);
static int      writeCmr(TreeDB& tree,TreeNode& root, CmrWriter& cmrWriter);

int trace = 0;

//-----------------------------------------------------------------------------
// udf2cmr.C:  This file contains source for a simple test to use the NSR
//             MSLIB Archive object in conjunction with the hpACDev autochanger
//             device object to read an nsr volume.  By default, it will simply 
//             echo to screen a list of filenames and some simple information 
//             about each file.  If the [ -r directory ] option is specified,
//             the files will actually be restored to the given directory.
//
// Usage:
//
// udf2cmr -d devfile -c cmrFileName [-y] surface_list
//
// -d devfile:      Devfile is the passthru device file for the autochanger
// -c cmrFileName:  Mandatory.  Name of the output CMR file to be generated.
// [-y]:            Optional.  If this option is given, the test will ask the 
//                  user to provide device file mappings for drives in the 
//                  autochanger.  Otherwise, the test will attempt to create 
//                  device files for the drives (you must be root for this to 
//                  succeed).
// surface_list:    A list of one or more entries of the form [slot_number][A|B]
//
// Example:  
//
// udf2cmr -d /dev/ac1 -y -c cmr.out 1a 1b 2a 2b 4a 4b
//-----------------------------------------------------------------------------

int
main(int argc,CHAR_PTR argv[])
{
    int         c;
    int         i;
    int         rval;
    int         slot;
    char        side;
    int         usage = 0;
    int         argAsk = 0;             // Ask user for drive mappings?
    RWCString   argDevfile;             // Device file for autochanger
    RWCString   cmrFileName;            // Name of output CMR file
    dlist       *srfcList;              // Store surface information
    srfcEntry   *sptr;
    srfcEntry   srfc;
    dlist_node  *dnode;

    extern CHAR_PTR optarg;             // For getopt()
    extern int  optind,opterr;          // For getopt()

    // Parse command-line arguments.

    while((c = getopt(argc,argv,"d:c:y")) != EOF) {
        switch(c) {
        case 'd':
            argDevfile = optarg;
            break;
        case 'c':
            cmrFileName = optarg;
            break;
        case 'y':
            argAsk = 1;
            break;
        default:
            usage = 1;
            break;
        }
    }

    if(argDevfile.isNull()) {
        usage = 1;
    } 

    // Create a dlist to keep track of our surfaces.

    srfcList = dlistCreate(srfc_cmp);
    if(srfcList == NULL) {
        fprintf(stderr,"%s:  memory allocation failure!\n",argv[0]);
        exit(2);
    }

    // Now process any remaining arguments and interpret them as surfaces.
    // Add them to the list of surfaces.

    for(i = optind; i < argc; i++) {
        if(sscanf(argv[i],"%d%c",&slot,&side) != 2) {
            usage = 1;
            break;
        }
        side = toupper(side);
        if((side != 'A') && (side != 'B')) {
            usage = 1;
            break;
        }

        side = ((side == 'A') ? 0 : 1);

        // Check to see if the surface is already in our list.  If so, don't
        // add it.

        srfc.slot = slot;
        srfc.side = side;

        dnode = dlistFind(srfcList,&srfc);
        if(dnode != NULL) {     // Already in the list
            continue;
        }
        sptr = new srfcEntry;
        if(sptr == NULL) {
            fprintf(stderr,"%s:  Memory allocation failure!\n",argv[0]);
            exit(2);
        }
        sptr->slot = slot;
        sptr->side = side;
        if(dlistInsert(srfcList,sptr)) {
            fprintf(stderr,"%s:  Memory allocation failure!\n",argv[0]);
            exit(2);
        }
    }

    if(dlistCount(srfcList) == 0) {
        usage = 1;
    }

    if(usage) {
        Usage();
        exit(2);
    }

    // If we got here, we're ready to start testing.  If the importArchive
    // function succeeds, the root nsrTreeNode will point to the 
    // root node of the imported directory structure, and the TreeDB
    // contain the tree structure and all file information.

    TreeDB              tree("/tmp/dummyfile");
    TreeNode            root;
    hpACDev             *acdev;

    acdev = configDevice(&argDevfile,argAsk,srfcList);
    if(acdev == NULL) {
        fprintf(stderr,"Unable to configure autochanger device object, \
aborting\n");
        return 2;
    }

    if(!importArchive(acdev,srfcList,tree,root)) {
        CmrWriter       cmrWriter((char*)cmrFileName);
        cmrWriter.useShortReservedWords();
        cmrWriter.setTabSize(2);
        writeCmr(tree, root, cmrWriter);
        printf("Wrote CMR file \"%s\"\n",(char *) cmrFileName);
    }
    else {
        fprintf(stderr,"Import archive failed, unable to write CMR file!\n");
    }

    dlistApply(srfcList,srfc_free,0);
    dlistDestroy(srfcList);

    freeTree(root);
    delete acdev;

    return rval;
}

//-----------------------------------------------------------------------------
// FUNCTION:  Usage
//
// Prints usage information to screen.
//-----------------------------------------------------------------------------

static void
Usage( void )
{
    static CHAR_PTR usage_str[] = {
"Usage:\n",
"udf2cmr -d devfile -c cmrFile [-y] surface_list\n",
"-d devfile:    Mandatory.  Devfile is the passthru device file for the ",
"               autochanger.",
"-c cmrFile:    Mandatory.  The output file name of the CMR file to be",
"               generated.",
"[-y]:          Optional.  If this option is given, the test will ask the",
"               user to provide device file mappings for drives in the",
"               autochanger.  Otherwise, the test will attempt to create",
"               device files for the drives (you must be root for this to",
"               succeed).",
"surface_list:  A list of one or more entries of the form [slot_number][A|B].",
"",
"Example:\n",
"./udf2cmr -d ac1 -c cmr.out -y 1A 1B 3A 3B\n",
NULL
};
    int i = 0;

    while(usage_str[i] != NULL) {
        printf("%s\n",usage_str[i]);
        i++;
    }
}

//-----------------------------------------------------------------------------
// FUNCTION:  configDevice
//
// Creates an hpACDev object, adds surface mappings, and adds drive mappings
// if necessary.  
//
// ARGUMENTS:
//
// RWCString *devf      IN      Pathname of passthru device file for the
//                            autochanger's controller.
// int      ask       IN      If nonzero, query the user for drive mappings
// dlist   *srfcList  IN/OUT  List of surfaces to map into autochanger.
//                            Sid_t's of mapped surfaces are written to
//                            the appropriate entry in the list.
//
// RETURNS:
//
// Returns a pointer to an hpACDev on success, -1 on failure.
//
// PRECONDITIONS:
//
// --The argument device file is valid.
//
// POSTCONDITIONS:
//
// --A hpACDev object will be created and returned
//-----------------------------------------------------------------------------

static hpACDev *
configDevice( RWCString *devf,int ask,dlist *srfcList )
{
    int         i;
    hpACDev     *acdev;

    assert(devf != NULL);
    assert(srfcList != NULL);

    // Try to create an hpACDev object.

    acdev = new hpACDev(String((CHAR_PTR) *devf));
    if(acdev == NULL) {
        fprintf(stderr,"Memory allocation failure in configDevice()!\n");
        return NULL;
    }
    if(acdev->error() != NSR::error_None) {     // Construction failed!
        fprintf(stderr,"Construction of hpACDev failed, error is %s\n",
                (acdev->error()).message());
        return NULL;
    }

    // Map surfaces.

    dlist_node  *nptr = dlistHead(srfcList);
    srfcEntry   *sptr;
    sid_t       sid;

    for(i = 0; i < dlistCount(srfcList); i++) {
        sptr = (srfcEntry *) nptr->data;
        sid = acdev->addSurface(sptr->slot,sptr->side);
        if(sid == SID_T_INVALID) {
            fprintf(stderr,"Surface %d%c is not in the autochanger!\n",
                    sptr->slot,(sptr->side) ? 'B' : 'A');
            dlistDelete(srfcList,nptr);
            delete sptr;
            continue;
        }
        nptr = nptr->next;
        printf("Mapped surface %d%c to sid_t %d\n",sptr->slot,
               (sptr->side) ? 'B' : 'A',sid);
        sptr->sid = sid;
    }

    if(dlistCount(srfcList) == 0) {
        fprintf(stderr,"No surfaces mapped, aborting!\n");
        delete acdev;
        return NULL;
    }

    // If necessary, ask user for drive mappings.

    if(ask) {
        if(getDriveMaps(*acdev)) {
            delete acdev;
            return NULL;
        }
    }

    return acdev;
}

//-----------------------------------------------------------------------------
// FUNCTION:  srfc_cmp
//
// Comparison function for the srfcEntry struct.
//-----------------------------------------------------------------------------

static int
srfc_cmp(void *d1,void *d2)
{
    srfcEntry   *s1 = (srfcEntry *) d1;
    srfcEntry   *s2 = (srfcEntry *) d2;

    if((s1 == NULL) || (s2 == NULL)) {
        return 0;
    }
    if(s1->slot != s2->slot) {
        return (s1->slot - s2->slot);
    }
    return (s1->side - s2->side);
}

//-----------------------------------------------------------------------------
// FUNCTION:  srfc_free
//
// Data free function for the srfcEntry struct.
//-----------------------------------------------------------------------------

static int
srfc_free(void *d1)
{
    srfcEntry   *s1 = (srfcEntry *) d1;

    if(s1 == NULL) {
        return -1;
    }
    delete s1;
    return 0;
}

//-----------------------------------------------------------------------------
// FUNCTION:  getDriveMaps
//
// Queries the user for drive map information and adds the drive mappings
// to the hpACDev.
//-----------------------------------------------------------------------------

static int
getDriveMaps( hpACDev& acdev )
{
    int                      i;
    char                     buffer[256];
    struct element_addresses eltaddr;
    struct element_status    eltstat;
    struct stat              statbuf;

    // We'll only ask about the drives that are in the autochanger.

    if(acdev.elementAddresses(eltaddr)) {
        fprintf(stderr,"Can't get element addresses information, aborting!\n");
        return -1;
    }

    for(i = 0; i < eltaddr.num_data_transfers; i++) {
        if(acdev.elementStatus(i + eltaddr.first_data_transfer,eltstat)) {
            fprintf(stderr,"Can't get element status information, aborting!\n");
            return -1;
        }
        while(1) {
            fprintf(stdout,"Please enter device file for drive %d, tgt=%d, \
lun=%d:  ",i+1,eltstat.bus_address,eltstat.lun);
            if(gets(buffer) == NULL) {
                fprintf(stdout,"Invalid input.\n");
                continue;
            }
            if(stat(buffer,&statbuf)) {
                fprintf(stdout,"Cannot stat specified file, please reenter\n");
                continue;
            }
            if(!S_ISCHR(statbuf.st_mode)) {
                fprintf(stdout,"Specified file is not a character special \
device file!\n");
                continue;
            }

            // Add the device file.

            acdev.addDriveMap(eltstat.bus_address,eltstat.lun,buffer);
            break;
        }
    }
    return 0;
}

//-----------------------------------------------------------------------------
// FUNCTION:  importArchive
//
// Creates an Archive object and attempts to read from the nsr volumes
// specified in the srfcList parameter.  The Archive object will import the
// volumes and print to screen the file names as it reads them.  If the
// restore parameter is nonzero, this function will attempt to restore
// files to argDir.
//
// ARGUMENTS:
//
// hpACDev        *acdev       IN      Pointer to autochanger device object
// dlist          *srfcList    IN      Linked-list containing surfaces to import
// nsrArchiveTree& tree        IN/OUT  Archive tree
//
// RETURNS:
//
// Returns a nsrTreeNode corresponding to the root node on success,
// NULL_NSRTREENODE on failure.
//
// PRECONDITIONS:
//
// --The argument acdev object is a valid autochanger device object
// --The argument srfcList is non-empty list containing sid's to import.
//
// POSTCONDITIONS:
//
// --The root argument will point to the root of an n-ary tree representing 
//   the tree structure of the imported directory tree and will contain
//   all file information for every file in the imported directory tree.
//-----------------------------------------------------------------------------

static int
importArchive( hpACDev *acdev,dlist *srfcList,TreeDB& treeDB,TreeNode& root )
{
    int                 i;
    int                 rval;
    UINT16              high_vid,vid;
    UINT16              dummy;
    UINT32              ldummy;
    UINT32              sector_size;
    MediaType           media_type;
    sid_t               root_sid;
#ifndef ROGUEWAVIFIED
    hash_table          *icb_table;
#endif
    char                vsname[MAX_NAMELENGTH];
    char                vname[MAX_NAMELENGTH];
    dlist_node          *nptr;
    srfcEntry           *sptr;
    Attributes          root_attr;
    Address             root_addr;
    Error               err;

    // We should be ready to create an Archive object and start the import.

    Archive     arc;

    arc.setDevice(acdev);
    if(arc.error() != NSR::error_None) {        
        fprintf(stderr,"Construction of Archive failed, error is %s\n",
                (arc.error()).message());
        return -1;
    }

    // Call getVolumeInfo():  iterate through the srfcList to determine 
    // which volume is root.  The only field that is important to 
    // getVolumeInfo() is the uvid (aka sid_t).  In the same loop we also
    // call addVolumeToImport() for every surface in the import group.

    nptr = dlistHead(srfcList);
    high_vid = 0;

    // Get the sector size and media type.

    sptr = (srfcEntry *) nptr->data;
    sector_size = acdev->sectorSize(sptr->sid);
    media_type = acdev->mediaType(sptr->sid);

    ExtentList  elist1,elist2;
    while(nptr != NULL) {
        sptr = (srfcEntry *) nptr->data;

        // We don't really need the free or erased space extent lists for
        // this application, though in the future we may need them.  

        arc.getVolumeInfo(sptr->sid,&dummy,vsname,vname,&vid,&elist1,&elist2,&(ldummy));
        if(arc.error() != NSR::error_None) {
            fprintf(stderr,"arc.getVolumeInfo() failed, error is %s\n",
                    (arc.error()).message());
            return -1;
        }
        printf("Read volume info for sid %d, volume name is %s\n",sptr->sid,
               vname);
        if(vid >= high_vid) {
            high_vid = vid;
            root_sid = sptr->sid;
        }
        sptr->volname = new char[strlen(vname) + 1];
        strcpy(sptr->volname,vname);

        nptr = nptr->next;
    }
    root_addr.sid = root_sid;
    root_addr.sector = ROOT_ICB_ADDRESS;

    // Open the archive

    arc.openArchive(dlistCount(srfcList),vsname,sector_size,media_type);

    if(arc.error() != NSR::error_None) {
        fprintf(stderr,"arc.openArchive() failed, error is %s\n",
                (arc.error()).message());
        return -1;
    }
    printf("Opened NSR Archive with %d volumes\n",dlistCount(srfcList));

    // Add the volumes

    nptr = dlistHead(srfcList);
    while(nptr != NULL) {
        sptr = (srfcEntry *) nptr->data;
        arc.addLabeledVolume(sptr->sid);
        if(arc.error() != NSR::error_None) {
            fprintf(stderr,"arc.addLabeledVolume(%d) failed, error is %s\n",
                    sptr->sid,
                    (arc.error()).message()); 
            return -1;
        }
        nptr = nptr->next;
    }

    // Make sure we have a complete volume set.

    if(arc.isCompleteVolSet() == NSR::False) {
        fprintf(stderr,"INCOMPLETE VOLUME SET, aborting import!\n");
        return -1;
    }

    // Begin Import.  We currently don't support WORM, and we don't care
    // about meta data space, so ignore the file meta data space ExtentList. 

    arc.beginImport(&elist1);
    if(arc.error() != NSR::error_None) {
        fprintf(stderr,"arc.beginImport() failed, error is %s\n",
               (arc.error()).message());
        return -1;
    }

    // readRootICB

    Extent              root_extent;


    arc.readRootICB(&root_attr,&root_extent);
    if(arc.error() != NSR::error_None) {
        fprintf(stderr,"arc.readRootICB() failed, error is %s\n",
                (arc.error()).message());
        return -1;
    }
    printf("Successfully read root ICB\n");


    // Now we want to read the rest of the ICB's.  We'll create a hash table
    // to keep track of the ICB's.  Add the root ICB to the hash table as
    // well.

    TreeNode    tnode;

#ifndef ROGUEWAVIFIED
    icb_table = htableCreate(HTABLE_SIZE,TreeNode_cmp,TreeNode_hash);
    if(icb_table == NULL) {
        fprintf(stderr,"Hash table creation failed!");
        return -1;
    }
#endif

    err = treeDB.allocNode(root,root_addr);
    if(err != NSR::error_None) {
        fprintf(stderr,"Memory allocation failure!\n");
        return -1;
    }
    root.attr_ = root_attr;
    root.elist_.append(root_extent);
    root.filename_ = "/";
    root.sync();

#ifndef ROGUEWAVIFIED
    if(htableHash(icb_table,(void *) root.realNode_)) {
        fprintf(stderr,"htableHash() call failed, aborting\n");
        return -1;
    }
#endif

    // Read the rest of the ICB's.

    Address     curAddr;
    Attributes  curAttr;
    ExtentList  curExtList;

    while(1) {
        curExtList.clear();
        arc.readNextICB(&curAddr,&curAttr,&curExtList);
        if(arc.error() != NSR::error_None) {    // We're done
            treeDB.freeNode(tnode);
            break;
        }
        err = treeDB.allocNode(tnode,curAddr);
        if(err != NSR::error_None) {
            fprintf(stderr,"Memory allocation failure!\n");
            return -1;
        }
        tnode.address_ = curAddr;
        tnode.attr_ = curAttr;
        tnode.elist_ = curExtList;

        // If the file is a symbolic link, read in its PathElementList.

        if(tnode.attr_.fileType() == NSR::file_SymbolicLink) {

            arc.readPath(tnode.elist_,&(tnode.plist_));
            if(arc.error() != NSR::error_None) {
                fprintf(stderr,"arc.readPath() failed, error is %s\n",
                        (arc.error()).message());
                return -1;
            }
        }

        // Add the entry to the hash table

        printf("Added key <%s> to database\n",tnode.key_.field);

        tnode.sync();

#ifndef ROGUEWAVIFIED
        if(htableHash(icb_table,(void *) tnode.realNode_)) {
            fprintf(stderr,"htableHash() call failed, aborting\n");
            return -1;
        }
#endif
    }

    // At this point, we should have the root ICB address, attributes, and
    // extent, as well as a hash table containing all of the rest of the
    // ICB's.  Now we enter a recursive function that gets the actual 
    // directory entries, so that we can associate ICB information with
    // file names. 

#ifdef ROGUEWAVIFIED
    if(recursiveGetEntry(root_extent,arc,treeDB,root)) 
#else
    if(recursiveGetEntry(root_extent,arc,icb_table,treeDB,root))
#endif
    {
        fprintf(stderr,"importArchive():  recursiveGetEntry() failed!\n");
        return -1;
    }

    // Clean up; we're done.

#ifndef ROGUEWAVIFIED
    htableDestroy(icb_table);
#endif

    arc.completeImport();
    if(arc.error() != NSR::error_None) {
        fprintf(stderr,"arc.completeImport() failed, error is %s\n",
                (arc.error()).message());
        return -1;
    }
    arc.closeArchive();
    if(arc.error() != NSR::error_None) {
        fprintf(stderr,"arc.closeArchive() failed, error is %s\n",
                (arc.error()).message());
        return -1;
    }

    return 0;
}

//-----------------------------------------------------------------------------
// FUNCTION:  recursiveGetEntry()
//
// Recursively traverses through the directory tree (preorder traversal) and
// displays filenames and file lengths.
//
// ALGORITHM:
//
// It is important, to minimize swapping, that the entries be read from the
// Archive object in a recursive preorder traversal.  For the current directory,
// nodes are made for all children.  If a child is a directory, it is placed
// into a queue for later recursion.  File entries are retrieved from the
// Archive object for each child, and the filenames stored in the hash table.
// Then each subdirectory is recurred upon, in queue (FIFO) order.  For
// maximum efficiency, directories in the Archive object must be traversed
// in the same order that they are retrieved by it.
//
// ARGUMENTS:
//
// Extent&         ext       IN      Extent of current directory
// Archive&        arc       IN      Archive object
// hash_table     *icb_table IN      Hash table containing file ICB information
// TreeDB&&        treeDB    IN/OUT  Archive Tree
// TreeNode&       tnode     IN/OUT  Node of current directory
//
// PRECONDITIONS:
//
// --The curdir parameter is the pathname of the current directory
// --The hash table contains an entry for the current directory
//
// POSTCONDITIONS:
//
// --All file entries for the current directory will be read from the Archive
//   object and added to the n-ary tree.  
// --The file names will be added to the appropriate entries in the hash table.
// --Every filename will be printed to screen with some attribute information
//   (rwx bits, file size).
//-----------------------------------------------------------------------------

static int
recursiveGetEntry
(
    Extent& ext,
    Archive& arc,
#ifndef ROGUEWAVIFIED
    hash_table *icb_table,
#endif
    TreeDB& treeDB,
    TreeNode& tnode
)
{
    int                 i,nentries;
    Boolean             isdir;
    char                name[MAX_NAMELENGTH];
    dqueue              *que;
    TreeNode            finfo;
#ifdef ROGUEWAVIFIED
    TreeNode            RealNode;
    TreeNode *          newnode = &RealNode;
#else
    TreeNode *          newnode;
#endif
    Error               err;

    // Create the queue for keeping track of subdirectories

    que = dqueueCreate();
    if(que == NULL) {
        fprintf(stderr,"Queue creation failed in recursiveGetEntry, abort\n");
        return -1;
    }

    // locateDirectory():  Sets current working directory

    arc.locateDirectory(ext);
    if(arc.error() != NSR::error_None) {
        fprintf(stderr,"arc.locateDirectory() failed, error is %s\n",
                (arc.error()).message());
        return -1;
    }
    
    nentries = arc.numDirEntries();
    if(arc.error() != NSR::error_None) {
        fprintf(stderr,"arc.numDirEntries() failed, error is %s\n",
                (arc.error()).message());
        return -1;
    }

    // Resize the tree node

    err = tnode.resize(nentries); 
    if(err != NSR::error_None) {
        fprintf(stderr,"tnode.resize() failed, no free store!\n");
        return -1;
    }
    tnode.sync();

    // Now loop and get all of the entries in the directory.  If a
    // given entry is a directory, put it into our queue.

    for(i = 0; i < nentries; i++) {
        arc.getDirEntry(i,&isdir,name,&(finfo.address_));
        if(arc.error() != NSR::error_None) {
            fprintf(stderr,"arc.getDirEntry(%d) failed, error is %s\n",i,
                    (arc.error()).message());
            return -1;
        }

        // If it's a parent directory entry (NULL name), just continue:  we
        // don't put parent entries in the hash table.  This also means we
        // don't link parent entries into the tree.

        if(strlen(name) == 0) {
            continue;
        }

        // Find the entry in the hash table so that we can add its name
        // to the hash table entry.

#ifdef ROGUEWAVIFIED
        Error           err;
        IntegerField    key;

        treeDB.addrToKey(finfo.address_,key);
        printf("Searching for file <%s>, key <%s>\n",
               (CONST_CHAR_PTR) name,key.field);
        if((err = treeDB.getNode(key,*newnode)) != NSR::error_None) {
            if(strlen(name) > 0) {
                fprintf(stderr,
                        "ERROR, dir entry for file \"%s\" not in database!  ",
                        name);
                fprintf(stderr,"Error is:  %s\n",err.message());
                continue;
            }
        }
#else
        newnode = (TreeNode *) htableFind(icb_table,(void *) &finfo);
        if(newnode == NULL) {
            if(strlen(name) > 0) {
                fprintf(stderr,"ERROR, dir entry for file \"%s\" not in hash table!\n", name);
            }
            continue;
        }
#endif
        newnode->filename_ = name;
        newnode->sync();

        // Link the parent to the child.

        err = tnode.addChild(*newnode);
        tnode.sync();
        if(err != NSR::error_None) {
            fprintf(stderr,"tnode.addChild() failed, no free store!\n");
            return -1;
        }

        // if it's a directory, put it into the queue.

        if((newnode->attr_.fileType() == NSR::file_Directory) && 
           (strlen((CHAR_PTR) newnode->filename_) > 0)) {
#ifdef ROGUEWAVIFIED
            TreeNode *nodep = new TreeNode;
            *nodep = *newnode;
#endif  // ROGUEWAVIFIED

#ifdef ROGUEWAVIFIED
            if(dqueueEnqueue(que,nodep)) {
#else
            if(dqueueEnqueue(que,newnode)) {
#endif
                fprintf(stderr,"Queue enqueue failed, aborting!\n");
                return 2;
            }
        }
    }                   // End of for loop

    // Now recur on the entries in the queue.

    int rval;
    while( (newnode = (TreeNode *) dqueueDequeue(que)) != NULL) {
        finfo = *newnode;
#ifdef ROGUEWAVIFIED
        rval = recursiveGetEntry((newnode->elist_)[0],arc,
                                 treeDB,finfo);
#else
        rval = recursiveGetEntry((newnode->elist_)[0],arc,icb_table,
                                 treeDB,finfo);
#endif
        if(rval) {
            fprintf(stderr,"Recursive call to recursiveGetEntry() failed!\n");
            return rval;
        }
    }

    // Cleanup

    dqueueDestroy(que);
    return 0;
}

//-----------------------------------------------------------------------------
// FUNCTION:  TreeNode_cmp
//
// Comparison function for the TreeNode class.
//-----------------------------------------------------------------------------

static int
TreeNode_cmp(void *d1,void *d2)
{
    TreeNode *f1 = (TreeNode *) d1;
    TreeNode *f2 = (TreeNode *) d2;

    if((f1 == NULL) || (f2 == NULL)) {
        return 0;
    }
    if(f1->address_.sid != f2->address_.sid) {
        return (f1->address_.sid - f2->address_.sid);
    }
    return (f1->address_.sector - f2->address_.sector);
}

//-----------------------------------------------------------------------------
// FUNCTION:  TreeNode_hash
//
// Hash function for the TreeNode struct.
//-----------------------------------------------------------------------------

static int
TreeNode_hash(void *d1)
{
    TreeNode  *f1 = (TreeNode *) d1;

    if(f1 == NULL) {
        return 0;
    }
    return f1->address_.sector;
}

//-----------------------------------------------------------------------------
// FUNCTION:  freeTree()
//
// Frees memory associated with a tree.
//-----------------------------------------------------------------------------

static void
freeTree( TreeNode& node )
{
    TreeNode    child;
    INT32       i;

    for(i = 0; i < node.getNumChildren(); i++) {
        node.getChild(i,child);
        freeTree(child);
    }
    (node.getDB())->freeNode(node);
}

//-----------------------------------------------------------------------------
// writeCmr     Recursively write CMR from fileset information in tree.
//
// preconditions:
//      The tree is populated with a fileset.
//
// postconditions:
//      The output file has been populated with CMR refecting fileset info in
//      the tree, or an error has terminated processing.
//-----------------------------------------------------------------------------

static int writeCmr ( TreeDB& tree,TreeNode& root, CmrWriter& cmrWriter )
{
    int                 i;
    int                 rval;
    TreeNode            tnode;
    Error               err;

    // Tell the CmrWriter of this directory, file or symlink.
    switch (root.attr_.fileType()) {

        // currently supported file types.
    case NSR::file_Directory:
        cmrWriter.beginDir(root.filename_, root.attr_);

        // recurr on children
        for(i = 0; i < root.getNumChildren(); i++) {

            err = root.getChild(i,tnode);
            if(err != NSR::error_None) {                // Parent entry
                continue;
            }
            rval = writeCmr(tree,tnode, cmrWriter);
            if(rval) {
                return rval;
            }
        }

        // Close the current directory.
        cmrWriter.endDir();    
        break;
    case NSR::file_Raw:
        cmrWriter.file (root.filename_, root.attr_, root.elist_);
        break;
    case NSR::file_SymbolicLink:
        cmrWriter.symlink(root.filename_, root.attr_, root.plist_);
        break;

        // currently unsupported file types.
    case NSR::file_SpecialBlockDevice:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_SpecialBlockDevice);
        break;
    case NSR::file_SpecialCharacterDevice:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_SpecialCharacterDevice);
        break;
    case NSR::file_ExtendedAttributes:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_ExtendedAttributes);
        break;
    case NSR::file_Fifo:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_Fifo);
        break;
    case NSR::file_Sock:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_Sock);
        break;
    case NSR::file_Unspecified:
    default:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_Unspecified);
        break;
    }

    return 0;
}
