//-*-c++-*-
// -----------------------------------------------------------------------------
//  acScript.cpp
//
// (c) Copyright 1995  Hewlett-Packard Company.
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the Government  is
//      subject to restrictions as set forth in subdivision (b)
//      (3)  (ii)  of the Rights in Technical Data and Computer
//      Software clause at 52.227-7013.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304
//
//  CLASSES:
//  TypedPath           Abstraction of a unix file path.  Includes type (file or
//                      directory - no links) and purpose, relative to Script.
//  TypedPathDictionary Keyed dictionary of TypedPath objects.
//  ScriptAction        An executable call, including command line options.
//  Script              Includes the path dictionary and a queue of script
//                      action objects.
//
// -----------------------------------------------------------------------------
// $Source: /projects/HMS/nsr/udfAncil/msVerify/acScript.cpp,v $
// $Revision: 1.11 $     $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 96/03/22 09:47:50 $
//-----------------------------------------------------------------------------

#include <signal.h>

#include "acScript.h"
ObjectStatus Script::object_status;
ActionStatus Script::lastActionStatus;

//----------------------------------------------------------------------
// METHOD: TypedPath:setPathType:  --trp
//
// ARGS:
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//      Sets expected pathtype, based on purpose.
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
virtual TypedPath::setPathType()
{
    switch (purpose_) {

    case purpose_native:
        pathtype_ = pathtype_directory;
        break;

    case purpose_cmr:
    case purpose_synth:
    case purpose_datacheck:
    case purpose_script:
    case purpose_config:
    case purpose_status:
    case purpose_free:
        pathtype_ = pathtype_file;
        break;
    
    case purpose_invalid:
    default:
        pathtype_ = pathtype_invalid;
        break;
    }
}

//----------------------------------------------------------------------
// METHOD: TypedPath::checkPath  --trp
//
// ARGS:
//  PathType&   actualType  OUT  Type of FOUND path resolution.  May
//                               be either target, or parent dir.
//  boolean&    foundTarget OUT  TRUE if target was found.
//  boolean&    foundParentWithoutTarget
//                               TRUE if the target was NOT found, but
//                               the parent directory of the target was.
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void TypedPath::checkPath(PathType& actualType, RWBoolean& foundTarget,
                          RWBoolean& foundParentWithoutTarget)
{
    // First, make a copy of the path without any trailing slashes.
    RWCString editedPath = path_;
    RWBoolean endsInSlash = (editedPath[editedPath.length()-1] == '/');
    if (endsInSlash) // remove trailing slash
        editedPath.remove((size_t)editedPath.length()-1);

        // Attempt to find the target.
    struct stat statbuf;
    if(lstat((char*)editedPath, &statbuf))
        foundTarget = FALSE;
    else {
        foundTarget = TRUE;
        foundParentWithoutTarget = FALSE;
    }
        

    // If the final target wasn't present, check for the parent directory.
    if (foundTarget == FALSE) {
        int index = editedPath.last('/');
        if (index == RW_NPOS) {
            foundParentWithoutTarget = FALSE;
        }
        else {
            editedPath.remove(index);
            if(lstat((char*)editedPath, &statbuf)) 
                foundParentWithoutTarget = FALSE;
            else
                foundParentWithoutTarget = TRUE;
        }
            
    }

    // If anything was found, indicate whether is was file, directory or
    // invalid.  Note that this type is of WHAT WAS FOUND, not necessarily
    // the type of the sought target.
    if (foundTarget || foundParentWithoutTarget) {
            
        if(S_ISDIR(statbuf.st_mode))
            actualType = pathtype_directory;
        else if (S_ISLNK(statbuf.st_mode)) {
            fprintf (stderr, "Error: symbolic links not supported: %s\n",
                     (char*) path_);
            actualType = pathtype_invalid;
        }
        else
            actualType = pathtype_file;
    }
    else
        actualType = pathtype_invalid;
}

//----------------------------------------------------------------------
// METHOD: TypedPath::appendPurposeDesc  --trp
//
// ARGS:
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
char* TypedPath::appendPurposeDesc(RWCString& s)
{
    switch(purpose_) {
    case purpose_datacheck:
        s += "datacheck";
        break;
    case  purpose_native:
        s += "native";
        break;
    case purpose_cmr:
        s += "cmr";
        break;
    case purpose_synth:
        s += "synth";
        break;
    case purpose_script:
        s += "script";
        break;
    case  purpose_config:
        s += "config";
        break;
    case purpose_status:
        s += "status";
    }
    return (char*)s;
}

//----------------------------------------------------------------------
// METHOD: TypedPath::appendTypeDesc  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
char* TypedPath::appendTypeDesc(RWCString& s)
{
    switch(pathtype_) {
    case pathtype_invalid:
        s += "invalid";
        break;
    case pathtype_file:
        s += "file";
        break;
    case pathtype_directory:
        s += "directory";
        break;
    }
    return (char*) s;
    
}

#ifdef VERIFY

//----------------------------------------------------------------------
// METHOD: TypedPath::dump  --trp
//      Prints out the typed path contents.
// ARGS:
// RETURNS:
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void TypedPath::dump(){
    RWCString dummy;
    printf ("path_: %s, purpose_: %s\n",
            (char*) path_, appendPurposeDesc(dummy=""));
}
//----------------------------------------------------------------------
// METHOD: TypedPath::validate  --trp
//
// ARGS:
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void TypedPath::validate(){
    if (!isValid()) {
        printf ("Class TypedPath is not Valid\n");
        dump();
    }
}
//----------------------------------------------------------------------
// METHOD: TypedPath::test  --trp
//      TypedPath self test.  Is highly context sensitive - in other
//      words, you have to run the self test in a directory that
//      contains the expected files.
// ARGS:
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void TypedPath::test(){
    printf ("beginning TypedPath test \n");
    printf ("(test assume present directory contents)\n");
    printf ("------------------------ \n");
        
    TypedPath t1;
    t1.purpose_ = purpose_script;
    t1.path_ = "~/nsr/udfAncil/msVerify/";
    t1.validate();

    TypedPath t2;
    t2.purpose_ = purpose_script;
    t2.path_ = "~/nsr/udfAncil/msVerify/";
    t2.validate();

    if (t1 != t2) {
        printf ("TypedPath 1 test failed\n");
    }

    t1.purpose_ = purpose_native;
    if (t1 == t2) {
        printf ("TypedPath test 2 failed\n");
    }

    t2.purpose_ = purpose_native;
    if (t1 != t2) {
        printf ("TypedPath test 3 failed\n");
    }
    t1.validate();
    t2.validate();

    printf ("testing checkPath\n");
    t1.purpose_ = purpose_script;
    RWBoolean foundTarget, foundParent;
    PathType type;
    t1.setPathType();
    printf ("testing absolute path to directory\n");
    t1.checkPath(type, foundTarget, foundParent);
    printf ("\tPathType: %d, foundTarget: %d, foundParent: %d\n",
            type, foundTarget, foundParent);

    t1.path_ += "acScript.cpp";
    printf ("testing absolute path to file\n");
    t1.checkPath(type, foundTarget, foundParent);
    printf ("\tPathType: %d, foundTarget: %d, foundParent: %d\n",
            type, foundTarget, foundParent);
    
    t1.path_ = "~/nsr/udfAncil/msVerify/newDir";
    printf ("testing absolute path to new file, with present parent dir.\n");
    t1.checkPath(type, foundTarget, foundParent);
    printf ("\tPathType: %d, foundTarget: %d, foundParent: %d\n",
            type, foundTarget, foundParent);

    printf ("testing relative path, with target file present\n");
    t1.path_ = "acScript.cpp";
    t1.checkPath(type, foundTarget, foundParent);
    printf ("\tPathType: %d, foundTarget: %d, foundParent: %d\n",
            type, foundTarget, foundParent);
    
    printf ("testing relative path, with target file not present.\n");
    t1.path_ = "newFile";
    t1.checkPath(type, foundTarget, foundParent);
    printf ("\tPathType: %d, foundTarget: %d, foundParent: %d\n",
            type, foundTarget, foundParent);
    
    printf ("testing relative path, with symlink.\n");
    t1.path_ = "linkme";
    t1.checkPath(type, foundTarget, foundParent);
    printf ("\tPathType: %d, foundTarget: %d, foundParent: %d\n",
            type, foundTarget, foundParent);
    
     
    t1.purpose_ = purpose_cmr;    t1.validate();
    t1.purpose_ = purpose_synth;  t1.validate();
    t1.purpose_ = purpose_script; t1.validate();
    t1.purpose_ = purpose_config; t1.validate();
    t1.purpose_ = purpose_status; t1.validate();

    printf ("Testing dump now, first t1 then t2.\n");
    t1.purpose_ = purpose_status;
    t1.path_= "t1/status";
    t1.dump();
    t2.purpose_ = purpose_native;
    t2.path_ = "t2/native";
    t2.dump();

    printf ("** Done testing TypedPath **\n\n");
}
#endif

//----------------------------------------------------------------------
// METHOD: TypedPathDictionary::insert  --trp
//
// ARGS:
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void TypedPathDictionary::insert(const RWCString& pathId, const TypedPath& path)
{
    RWCollectableString* insertPathId;
    insertPathId = new RWCollectableString(pathId);

    TypedPath* insertPath = new TypedPath;
    *insertPath = path;
        
    dictionary_.insertKeyAndValue((RWCollectable*) insertPathId,
                                  (RWCollectable*) insertPath);
}

//----------------------------------------------------------------------
// METHOD: TypedPathDictionary::retrieve  --trp
//
// ARGS:
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
RWBoolean TypedPathDictionary::retrieve(const RWCString& pathId, TypedPath& path)
{
    TypedPath* pathPtr;
    RWCollectableString key (pathId);
    pathPtr = (TypedPath*) dictionary_.findValue(&key);
    if (pathPtr == NULL) {
        return FALSE;
    }
    else {
        path = *pathPtr;
        return TRUE;
    }
}


#ifdef VERIFY
//----------------------------------------------------------------------
// METHOD: TypedPathDictionary::dump  --trp
//      Prints the typed 
// ARGS:
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void TypedPathDictionary::dump()
{
    printf ("\ntypedPathDictionary dump. %d items\n",
            (int)dictionary_.entries());
    printf ("key\ttype\tpath\n");
        
    RWHashDictionaryIterator iter(dictionary_);
    TypedPath* path;
    RWCString temp;
    RWCollectableString* str;

    iter.reset();
    str = (RWCollectableString*) iter();
    while (str != NULL) {
        path = (TypedPath*)iter.value();
        path->dump();
        
//        printf ("%s\t%s\t%s\n",
//                (char*) *((RWCString*)str),
//                path->appendPurposeDesc(temp=""),
//                (char*)path->path_);
//            
        str = (RWCollectableString*) iter();
    }
    printf ("\n");
}

//----------------------------------------------------------------------
// METHOD: TypedPathDictionary::test  --trp
//      Class self-test.
// ARGS:
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void TypedPathDictionary::test()
{
    printf ("beginning TypedPathDictionary test\n");
    printf ("------------------------------\n");
        
    TypedPathDictionary d1;
    TypedPath p;
    RWCollectableString key;

    p.purpose_ = purpose_native;
    p.path_ = "one/two";
    key ="native1";
    d1.insert(key, p);

    p.purpose_ = purpose_cmr;
    p.path_ = "one/two/cmr";
    key ="cmr1";
    d1.insert(key, p);
        
    p.purpose_ = purpose_script;
    p.path_ = "one/config";
    key ="config1";
    d1.insert(key, p);

    p.purpose_ = purpose_status;
    p.path_ = "one/status";
    key ="status1";
    d1.insert(key, p);
    d1.dump();

    if (d1.entries() != 4) {
        printf ("TypedDictionary Path count is wrong\n");
    }

    TypedPath findMe;
    key ="cmr1";
    printf ("looking up cmr1\n");
    if (d1.retrieve(key, findMe)) {
        printf ("found key cmr1: ");
        findMe.dump();
        printf ("\n");
    }
    else
        printf ("key cmr1 not found\n");

    d1.clearAndDestroy();
    printf ("** Done Testing TypedDictionary **\n");
}
    
#endif

//----------------------------------------------------------------------
// METHOD: ScriptAction::isEqual  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
virtual RWBoolean ScriptAction::isEqual (ScriptAction& otherAction)
{
    if (executableName_ != otherAction.executableName_)
        return FALSE;
    return TRUE;
}

//----------------------------------------------------------------------
// METHOD: ScriptAction::verifyFailure  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void ScriptAction::verifyFailure(char* reason)
{
    fprintf (stderr,
             "The test script failed valididation in action: \n");
    printf ("\t%s\n", reason);
        
    object_status = status_invalid;
}
//----------------------------------------------------------------------
// METHOD: showCommandLine  --trp
//      Non-class function used within apply.
// ARGS:
//  void*       parent  IN      driver dir parent, or NULL for root
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void showCommandLine(RWCollectable *c, void *dummy)
{
    ScriptAction* action = (ScriptAction*) c;
    printf ("%s\n", (char*)action->commandLine_);
}

//----------------------------------------------------------------------
// METHOD: Script::tallyAddedVolumes  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void tallyAddedVolumes(RWCollectable *c, void *volListObject)
{
    printf ("tallyAddedVolumes not implemented yet.\n");
}

//----------------------------------------------------------------------
// METHOD: executeActions   --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void executeActions (RWCollectable *c, void *scriptObject)
{
    // Validate the action.
    ScriptAction* action = (ScriptAction*) c;
    Script* script = (Script*) scriptObject;
    if (script->object_status != status_normal)
        return;
    if (!script->isValid())
        return;

    // If executing only export & import actions, skip others.
    // This is used for the repeat until fail option.
    if (script->executeExImOnly &&
        action->actionType != action_export &&
        action->actionType != action_import)
        return;
    
    // Return if the execute option doesn't match the last action
    // status.  Note that an onFail action is not executed on an
    // abort last action status.
    if ((action->execute_option == execute_ifOk) &&
        (script->lastActionStatus != actionstat_ok &&
         script->lastActionStatus != actionstat_pass))
        return;
    if ((action->execute_option == execute_ifNotOk) &&
        (script->lastActionStatus != actionstat_fail) &&
        (script->lastActionStatus != actionstat_abort) )
        return;
    
    // Check if this is a special action (continue or stop).
    if (action->specialAction == special_stop)
    {
        script->lastActionStatus = actionstat_fail;
        script->object_status = status_invalid;
    }
    if (action->specialAction == special_continue)
    {
        script->lastActionStatus = actionstat_pass;
    }

    // If executing only export & import (repeat until fail), increment
    // the fileset name each export cycle.
    if (script->executeExImOnly) {
        // update the fileset name
        if (action->actionType == action_export)
            script->repeatFileSetCounter++;
        
        sprintf ((char*)script->repeatFileSetName, " -f repeatFileSet%d ",
                     script->repeatFileSetCounter);

        // Use a rogue wave reg expression to replace the fileset name
        // in the command line.
        // Note that a file set name containing other than letters and
        // numbers will fail.  If you're reading this comment after
        // a long angry search, I hope I'm elsewhere...
        RWCRegexp fsNameSwitch(" -f [a-zA-Z0-9]* ");
        action->commandLine_(fsNameSwitch) = (char*)script->repeatFileSetName;
    }
    
    // Log the command line.
    if (script->redirectOutput && !script->outputFileName_.isNull()) {
        FILE* fOutput = fopen ((char*)script->outputFileName_, "a+");
        fprintf (fOutput,
                 "\n-----------------------------------------------------\
---------------------------\n");
        fprintf (fOutput, "%s\n", (char*)action->commandLine_);
        fclose(fOutput);
    }
    else {
        fprintf (stderr,
                 "\n-----------------------------------------------------\
---------------------------\n");
        fprintf (stderr, "%s\n", (char*)action->commandLine_);
    }
    

    // Optionally add stdout & stderr output to command line.
    RWCString editedCommandLine = action->commandLine_;
    if (script->redirectOutput && !script->outputFileName_.isNull()) {
        editedCommandLine += ">";
        editedCommandLine += script->outputFileName_;
        editedCommandLine += " 2>&1";
    }

    // Don't mark at test as passed if the last action is an erase that
    // is executed on both pass & fail.  Otherwise if the test fails
    // and an absolute (not ifPass) erase is called and succeeds,
    // the test will be marked as having passed.
    if (script->debug) {
//        fprintf (stderr,
//                 "DEBUG OPTION SELECTED, NO EXECUTION. Command Line:\n\n");
//        fprintf (stderr, "%s\n", (char*) editedCommandLine);
    }
    else {
        // Call the program with arguments.
        int actionReturn = system(editedCommandLine);
        
        int maskedReturn =
            script->analyzeSystemReturn(actionReturn, action->statusImpact);
        char retStatDesc[512];
        switch (maskedReturn) {
        case 0:
            sprintf (retStatDesc, "OK, NO PASS/FAIL STATUS");
            break;
            
        case 2:
            sprintf (retStatDesc, "FAIL");
            break;
            
        case 3:
            sprintf (retStatDesc, "PASS");
            break;
            
        case 1:
        default:
            sprintf (retStatDesc, "ABORT");
            break;
        }
        
        // Log the command line.
        if (script->redirectOutput && !script->outputFileName_.isNull()) {
            FILE* fOutput = fopen ((char*)script->outputFileName_, "a+");
            fprintf (fOutput, "Returned Status: %s\n\n", retStatDesc);
            fclose(fOutput);
        }
        else {
            fprintf (stderr, "Returned Status: %s\n\n", retStatDesc);
        }
    

        if (script->trace_)
            printf ("%s returned %d.\n", (char*)action->executableName_,
                    maskedReturn);
    }
}

//----------------------------------------------------------------------
// METHOD: Script::Script  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
Script::Script(char* scriptFileName)
{
    // Open the input file read only.
    inputFile_ = fopen (scriptFileName, "r");
    if (inputFile_ == NULL) {
        fprintf (stderr,
                 "Error: Input file %s cannot be opened for reading.\n",
                 scriptFileName);
        fprintf (stderr,
                 "Script object invalidated - no programs will be called.\n");
        object_status = status_invalid;
        return;
    }
    object_status = status_normal;
    redirectOutput = FALSE;
    trace_ = FALSE;
    arcLibTrace_ = FALSE;
    debug = FALSE;
    usePurify = FALSE;
    useCoverage = FALSE;

    // Specify the lex/yacc file.
    scriptin = inputFile_;

    // Get the arcTest working root.
   
    char* cptr = getenv("ARCTEST_ROOT");
    if(cptr != NULL) {
        arcTestRoot = cptr;
        arcTestRoot += "/";
    }
    else
        arcTestRoot = "/tmp";
}

void Script::scriptId(char* id)
{
    scriptId_ = id;
}

//----------------------------------------------------------------------
// METHOD: Script::scriptStatus  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
RWBoolean Script::scriptStatus() 
{
    if (lastActionStatus == actionstat_pass ||
        lastActionStatus == actionstat_ok)
        return TRUE;
    else
        return FALSE;
}

//----------------------------------------------------------------------
// METHOD: pathVerify  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void pathVerify (RWCollectable* key, RWCollectable* c, void* scriptObject)
{
    TypedPath* p = (TypedPath*) c;
    Script* script = (Script*) scriptObject;
    RWBoolean foundInvalid = FALSE;

    PathType actualPathType;
    RWBoolean foundTarget, foundParentWithoutTarget;
    p->checkPath(actualPathType, foundTarget, foundParentWithoutTarget);

    switch (p->purpose_) {

    // Parent directory must exist, target must not.
    case   purpose_datacheck:
    case   purpose_cmr: 
    case   purpose_free:
    {
#ifdef VERIFY_CMR_PATH
        if (foundTarget || !foundParentWithoutTarget ||
            (actualPathType != pathtype_directory))
            foundInvalid = TRUE;
#endif

        break;
    }

    // Target must exist as existing file.
    case   purpose_synth:
    case   purpose_script:
    case   purpose_config:
    case   purpose_status:
    {
        if (!foundTarget || (actualPathType != pathtype_file))
            foundInvalid = TRUE;

        break;
    }

    // Target must exist as existing directory.
    case   purpose_native:
    {
        if (actualPathType != pathtype_directory) {
            foundInvalid = TRUE;
            return;
        }
        if (!foundTarget && !foundParentWithoutTarget) {
            foundInvalid = TRUE;
            return;
        }
        // If the final dir path doesn't exist, but the parent does,
        // build the last path element.
        // NOTE: no attempt is made to later remove that created
        // directory.
        if (!foundTarget) {
            RWCString buildTarget;
            buildTarget = "mkdir ";
            buildTarget = p->path_;
            system(buildTarget);
        }
        break;
    }

    // Invalid cases.
    case   purpose_invalid:
    default:
        foundInvalid = TRUE;
    }

    if (foundInvalid) {
        fprintf (stderr, "Error: path %s found to be invalid.\n",
                 (char*) p->path_);
        script->object_status = status_invalid;
        script->lastActionStatus = actionstat_fail;
    }
}
//----------------------------------------------------------------------
// METHOD: replaceTildeWithHome  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void replaceTildeWithHome (RWCollectable* key, RWCollectable* c, void* dummy)
{
    TypedPath* p = (TypedPath*) c;

    if (p->path_[0] == '~') {
        p->path_.remove(0, 1); // remove the tilde
        p->path_.prepend(getenv("HOME")); // insert home path
    }
}

//----------------------------------------------------------------------
// METHOD: Script::verifyScript  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
RWBoolean Script::verifyScript()
{
    // Verify the paths.
    pathDict_.dictionary_.applyToKeyAndValue(replaceTildeWithHome, NULL);
    pathDict_.dictionary_.applyToKeyAndValue(pathVerify, this);
    
    // Verify the actions
    // ------------------

    // Each path in action has been declared, and is of the correct type.

    // Labels must exists or be created to be destroyed.

    // No worm labels are destoryed.

    // Verify erases
    // -------------
    
    // Make sure only working temporary files are erased.


    // Volume Sets
    // -----------
    
    // Availability of requested added volumes.
    
}

//----------------------------------------------------------------------
// METHOD: Script::analyzeSystemReturn  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
int  Script::analyzeSystemReturn(int systemReturn, StatusImpact impact)
{
    if (WIFEXITED(systemReturn)) {
        int exitValue = WEXITSTATUS(systemReturn);
        if (impact != impact_none) {
            switch (exitValue) {
            case 0:
                if (impact == impact_absolute)
                    lastActionStatus = actionstat_ok;
                break;
            case 2:
                lastActionStatus = actionstat_fail;
                break;
            case 3:
                if (impact == impact_absolute)
                    lastActionStatus = actionstat_pass;
                break;
            case 1:
            default:
                lastActionStatus = actionstat_abort;
            }
        }
        
        return exitValue;
    }
    else
        lastActionStatus = actionstat_abort;

    return 1;
}

//----------------------------------------------------------------------
// METHOD: Script::addStrOption  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::addStrOption (char* command, char* arg, RWCString& appendToThis)
{
    appendToThis += " -";
    appendToThis += command;
    appendToThis += " ";
    appendToThis += arg;
    appendToThis += " ";
}

//----------------------------------------------------------------------
// METHOD: Script::addIdOption  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::addIdOption (char* command, char* id, RWCString& appendToThis)
{
    RWCString key (id);
    TypedPath path;
    if (pathDict_.retrieve(key, path))
    {
        appendToThis += " -";
        appendToThis += command;
        appendToThis += " ";
        appendToThis += path.path_;
        appendToThis += " ";
    }
    else
    {
        fprintf (stderr, "key %s not found.\n", id);
        fprintf (stderr, "Error in script suspected.\n");
        object_status = status_invalid;
    }
}
//----------------------------------------------------------------------
// METHOD: Script::addOption  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::addOption (char* command, RWCString& appendToThis)
{
    appendToThis += " -";
    appendToThis += command;
    appendToThis += " ";
}


//----------------------------------------------------------------------
// METHOD: Script::file_datacheck  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::file_datacheck (char* id, char* path) {
    VERIFY_OBJECT_STATUS();
    if (trace_)
        printf ("file datacheck %s: %s\n", id, path);
        
    TypedPath tpath(purpose_datacheck);
    tpath.path_ = path;

#ifdef ADD_ROOT_TO_RELATIVE
    addArcTestRootIfRelative(tpath.path_);
#endif
    tpath.pathtype_ = pathtype_file;
    RWCString* key = new RWCString (id);
    pathDict_.insert(*key, tpath);
}
//----------------------------------------------------------------------
// METHOD: Script::file_native  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::file_native (char* id, char* path) {
    VERIFY_OBJECT_STATUS();

    if (strlen(id) == 0)
    {
        fprintf (stderr, "null id sent to Script::file_native.\n");
        object_status = status_invalid;
        return;
    }
    if (trace_)
        printf ("file native %s: %s\n", id, path);
        
    TypedPath tpath (purpose_native);
    tpath.path_ = path;
#ifdef ADD_ROOT_TO_RELATIVE
    addArcTestRootIfRelative(tpath.path_);
#endif
    tpath.pathtype_ = pathtype_directory;
    RWCString* key = new RWCString (id);
    pathDict_.insert(*key, tpath);
}

//----------------------------------------------------------------------
// METHOD: Script::file_cmr  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::file_cmr (char* id, char* path) {
    VERIFY_OBJECT_STATUS();
    if (trace_)
        printf ("file cmr %s: %s\n", id, path);
        
    TypedPath tpath (purpose_cmr);
    tpath.path_ = path;
#ifdef ADD_ROOT_TO_RELATIVE
    addArcTestRootIfRelative(tpath.path_);
#endif
    tpath.pathtype_ = pathtype_file;
    RWCString* key = new RWCString (id);
    pathDict_.insert(*key, tpath);
}
void Script::file_synth (char* id, char* path) {
    VERIFY_OBJECT_STATUS();
    if (trace_)
        printf ("file synth %s: %s\n", id, path);
        
    TypedPath tpath (purpose_synth);
    tpath.path_ = path;
#ifdef ADD_ROOT_TO_RELATIVE
    addArcTestRootIfRelative(tpath.path_);
#endif
    tpath.pathtype_ = pathtype_file;
    RWCString* key = new RWCString (id);
    pathDict_.insert(*key, tpath);
}

//----------------------------------------------------------------------
// METHOD: Script::file_free  --trp
//----------------------------------------------------------------------
void Script::file_free (char* id, char* path) {
    VERIFY_OBJECT_STATUS();
    if (trace_)
        printf ("file free %s: %s\n", id, path);
        
    TypedPath tpath (purpose_free);
    tpath.path_ = path;
#ifdef ADD_ROOT_TO_RELATIVE
    addArcTestRootIfRelative(tpath.path_);
#endif
    tpath.pathtype_ = pathtype_file;
    RWCString* key = new RWCString (id);
    pathDict_.insert(*key, tpath);
}

//----------------------------------------------------------------------
// METHOD: Script::file_script  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::file_script (char* id, char* path) {
    VERIFY_OBJECT_STATUS();
    if (trace_)
    printf ("file script %s: %s\n", id, path);
    
    TypedPath tpath (purpose_script);
    tpath.path_ = path;
    tpath.pathtype_ = pathtype_file;
    RWCString* key = new RWCString (id);
    pathDict_.insert(*key, tpath);
}
//----------------------------------------------------------------------
// METHOD: Script::file_config  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::file_config (char* id, char* path) {
    VERIFY_OBJECT_STATUS();
    if (trace_)
    printf ("file config %s: %s\n", id, path);
        
    TypedPath tpath (purpose_config);
    tpath.path_ = path;
    tpath.pathtype_ = pathtype_file;
    RWCString* key = new RWCString (id);
    pathDict_.insert(*key, tpath);
}
//----------------------------------------------------------------------
// METHOD: Script::file_status  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::file_status (char* id, char* path) {
    VERIFY_OBJECT_STATUS();
    if (trace_)
    printf ("file status %s: %s\n", id, path);
       
    TypedPath tpath (purpose_status);
    tpath.path_ = path;
    tpath.pathtype_ = pathtype_file;
    RWCString* key = new RWCString (id);
    pathDict_.insert(*key, tpath);
}
//----------------------------------------------------------------------
// METHOD: Script::file_end  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::file_end(){
    VERIFY_OBJECT_STATUS();
}

//----------------------------------------------------------------------
// METHOD: Script:export_start:  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::export_start()
{
    VERIFY_OBJECT_STATUS();
    if (usePurify)
        action = new ScriptAction(P_EXPORT_PROG_NAME);
    else if (useCoverage)
        action = new ScriptAction(C_EXPORT_PROG_NAME);
    else
        action = new ScriptAction(EXPORT_PROG_NAME);
    action->actionType = action_export;
    action->commandLine_ = action->executableName_;
    if (arcLibTrace_)
        action->commandLine_ += " -t 1 ";
    else
        action->commandLine_ += " ";
}
    
//----------------------------------------------------------------------
// METHOD: Script::export_fileset  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::export_fileset (char* filesetName)
{
    VERIFY_OBJECT_STATUS();
    if (trace_)
        printf ("export fileset %s\n", filesetName);

    action->commandLine_ += " -f ";
    action->commandLine_ += filesetName;
    action->commandLine_ += " ";
}
//----------------------------------------------------------------------
// METHOD: Script::export_volset  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::export_volset (char* volName)
{
    VERIFY_OBJECT_STATUS();
    if (trace_)
        printf ("export fileset %s\n", volName);
    
    action->commandLine_ += "-v ";
    action->commandLine_ += volName;
    action->commandLine_ += " ";        
}
    
//----------------------------------------------------------------------
// METHOD: Script::export_native  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::export_native (char* id)
{
    VERIFY_OBJECT_STATUS();
    if (trace_)
        printf ("export native %s\n", id);
    
    RWCString key (id);
    TypedPath path;
    if (pathDict_.retrieve(key, path))
    {
        action->commandLine_ += "-d ";
        action->commandLine_ += path.path_;
        action->commandLine_ += " ";
    }
    else
    {
        fprintf (stderr, "export native key %s not found.\n", id);
        fprintf (stderr, "Error in script suspected.\n");
        object_status = status_invalid;
    }
}
//----------------------------------------------------------------------
// METHOD: Script::export_cmr  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::export_cmr (char* id)
{
    VERIFY_OBJECT_STATUS();
    if (trace_)
        printf ("export cmr %s\n", id);

    RWCString key (id);
    TypedPath path;
    if (pathDict_.retrieve(id, path))
    {
        action->commandLine_ += "-c ";
        action->commandLine_ += path.path_;
        action->commandLine_ += " ";
    }
    else
    {
        fprintf (stderr, "export native key %s not found.\n", id);
        fprintf (stderr, "Error in script suspected.\n");
        object_status = status_invalid;
    }
}

//----------------------------------------------------------------------
// METHOD: Script::export_config  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::export_config(char* id)
{
    VERIFY_OBJECT_STATUS();
    if (trace_)
        printf ("export config %s\n", id);
    
    
    RWCString key (id);
    TypedPath path;
    if (pathDict_.retrieve(id, path))
    {
        action->commandLine_ += "-a ";
        action->commandLine_ += path.path_;
        action->commandLine_ += " ";
    }
    else
    {
        fprintf (stderr, "export native key %s not found.\n", id);
        fprintf (stderr, "Error in script suspected.\n");
        object_status = status_invalid;
    }
}
//----------------------------------------------------------------------
// METHOD: Script::export_status  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::export_status(char* id)
{
    VERIFY_OBJECT_STATUS();
    if (trace_)
        printf ("export status %s\n", id);
    
    RWCString key (id);
    TypedPath path;
    if (pathDict_.retrieve(id, path))
    {
        action->commandLine_ += "-s ";
        action->commandLine_ += path.path_;
        action->commandLine_ += " ";
    }
    else
    {
        fprintf (stderr, "export native key %s not found.\n", id);
        fprintf (stderr, "Error in script suspected.\n");
        object_status = status_invalid;
    }
}
//----------------------------------------------------------------------
// METHOD: Script::export_synth  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::export_synth(char* id)
{
    VERIFY_OBJECT_STATUS();
    if (trace_)
        printf ("export synth %s\n", id);
    
    RWCString key (id);
    TypedPath path;
    if (pathDict_.retrieve(id, path))
    {
        action->commandLine_ += "-D ";
        action->commandLine_ += path.path_;
        action->commandLine_ += " ";
    }
    else
    {
        fprintf (stderr, "export native key %s not found.\n", id);
        fprintf (stderr, "Error in script suspected.\n");
        object_status = status_invalid;
    }
}

void Script::export_free (char* id)
{
    VERIFY_OBJECT_STATUS();
    if (trace_)
        printf ("export free %s\n", id);
    
    RWCString key (id);
    TypedPath path;
    if (pathDict_.retrieve(id, path))
    {
        action->commandLine_ += "-e ";
        action->commandLine_ += path.path_;
        action->commandLine_ += " ";
    }
    else
    {
        fprintf (stderr, "export native key %s not found.\n", id);
        fprintf (stderr, "Error in script suspected.\n");
        object_status = status_invalid;
    }
}

//----------------------------------------------------------------------
// METHOD: Script::export_end  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::export_end()
{
    VERIFY_OBJECT_STATUS();
    ensureConfigAndStatus();
    action->execute_option = execute_ifOk;
    action->statusImpact = impact_toFailOnly;
    actionQueue.insert (action);
}


//----------------------------------------------------------------------
// METHOD: Script::label_* methods  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::label_start () {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("label start\n");

    if (usePurify)
        action = new ScriptAction(P_EXPORT_PROG_NAME);
    else if (useCoverage)
        action = new ScriptAction(C_EXPORT_PROG_NAME);
    else
        action = new ScriptAction(EXPORT_PROG_NAME);

    action->actionType = action_label;
    action->commandLine_ = action->executableName_;
    action->commandLine_ += " -E -d \"/dummy\" ";
}
void Script::label_config (char* id) {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("label %s\n", id);
    addIdOption ("a", id, action->commandLine_);
}
void Script::label_status (char* id) {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("label status %s\n", id);

    addIdOption ("s", id, action->commandLine_);
}
void Script::label_media (char* mediaType) {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("label media %s\n", mediaType);
    addStrOption ("m", mediaType, action->commandLine_);
}
void Script::label_density (char* density) {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("label density %s\n", density);
    addStrOption ("z", density, action->commandLine_);
}
void Script::label_volumeSet (char* volSetName) {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("label volset %s\n", volSetName);
    addStrOption ("v", volSetName, action->commandLine_);
}
void Script::label_fileset (char* fileSetName) {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("label fileset %s\n", fileSetName);
    addStrOption ("f", fileSetName, action->commandLine_);
}
void Script::label_addVolume (char* volName) {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("label add vol %s\n", volName);
    addStrOption ("g ", volName, action->commandLine_);
}
void Script::label_end () {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("label \n");

    ensureConfigAndStatus();

    action->execute_option = execute_ifOk;
    action->statusImpact = impact_toFailOnly;
    actionQueue.insert (action);
}


void Script::grow_start () {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("grow start\n");
    if (usePurify)
        action = new ScriptAction(P_EXPORT_PROG_NAME);
    else if (useCoverage)
        action = new ScriptAction(C_EXPORT_PROG_NAME);
    else
        action = new ScriptAction(EXPORT_PROG_NAME);
    action->actionType = action_label;
    action->commandLine_ = action->executableName_;
    action->commandLine_ += " -E ";
}
void Script::grow_config (char* id) {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("grow config %s\n", id);

    addIdOption ("a", id, action->commandLine_);
}
void Script::grow_status (char* id) {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("grow status %s\n", id);
    addIdOption ("s", id, action->commandLine_);
}
void Script::grow_volumeSet (char* volSetName) {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("grow volset %s\n", volSetName);
    addStrOption ("v", volSetName, action->commandLine_);
}
void Script::grow_fileset (char* fileSetName) {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("grow fileset %s\n", fileSetName);
    addStrOption ("f", fileSetName, action->commandLine_);
}
void Script::grow_addVolume (char* volName) {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("grow add vol %s\n", volName);
    addStrOption ("a", volName, action->commandLine_);
}
void Script::grow_end () {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("grow end\n");

    ensureConfigAndStatus();
    action->execute_option = execute_ifOk;
    action->statusImpact = impact_toFailOnly;
    actionQueue.insert (action);
}


void Script::onPass_start () {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("onPass start\n");
}
void Script::onPass_erase (char* id) {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("onPass %s\n", id);
    
    RWCString key (id);
    TypedPath path;
    if (pathDict_.retrieve(id, path))
    {
        if (path.purpose_ == purpose_script)
            {fprintf (stderr, "can't rm scripts.\n");return;}
        if (path.purpose_ == purpose_config)
            {fprintf (stderr, "can't rm config files.\n");return;}
        if (path.purpose_ == purpose_status)
            {fprintf (stderr, "can't rm status files.\n");return;}

        action = new ScriptAction(ERASE_PROG_NAME);
        action->actionType = action_erase;
        action->commandLine_ = action->executableName_;
        action->commandLine_ += " ";
        action->commandLine_ += path.path_;
        // If the path is native, kill the path contents rather than the path.
        if (path.purpose_ == purpose_native)
            action->commandLine_ += "/*";
        action->execute_option = execute_ifOk;
        action->statusImpact = impact_none;
        actionQueue.insert (action);
    }
    else
    {
        fprintf (stderr, "erase if ok key %s not found.\n", id);
        fprintf (stderr, "Error in script suspected.\n");
        object_status = status_invalid;
    }
}
void Script::onPass_unLabel (char* volSetName) {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("onPass %s\n", volSetName);

    if (usePurify)
        action = new ScriptAction(P_EXPORT_PROG_NAME);
    else if (useCoverage)
        action = new ScriptAction(C_EXPORT_PROG_NAME);
    else
        action = new ScriptAction(EXPORT_PROG_NAME);

    action->actionType = action_unlabel;
    action->commandLine_ = action->executableName_;
    addStrOption ("k -E -d \"/dummy\" -v ", volSetName, action->commandLine_);
    ensureConfigAndStatus();
    action->statusImpact = impact_none;
    action->execute_option = execute_ifOk;
    actionQueue.insert (action);
}
void Script::onPass_stop () {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("onPass stop\n");
    action = new ScriptAction("echo");
    action->actionType = action_special;    
    action->specialAction = special_stop;
    action->execute_option = execute_ifOk;
    action->statusImpact = impact_toFailOnly;
    actionQueue.insert (action);
}
void Script::onPass_end () {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("onPass end\n");
}
void Script::onFail_start () {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("onFail start\n");
}
void Script::onFail_erase (char* id) {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("onFail %s\n", id);
    
    RWCString key (id);
    TypedPath path;
    if (pathDict_.retrieve(id, path))
    {
        if (path.purpose_ == purpose_script)
            {fprintf (stderr, "can't rm scripts.\n");return;}
        if (path.purpose_ == purpose_config)
            {fprintf (stderr, "can't rm config files.\n");return;}
        if (path.purpose_ == purpose_status)
            {fprintf (stderr, "can't rm status files.\n");return;}

        action = new ScriptAction(ERASE_PROG_NAME);
        action->actionType = action_erase;
        action->commandLine_ = action->executableName_;
        action->commandLine_ += " ";
        action->commandLine_ += path.path_;
        // If the path is native, kill the path contents rather than the path.
        if (path.purpose_ == purpose_native)
            action->commandLine_ += "/*";
        action->execute_option = execute_ifNotOk;
        action->statusImpact = impact_none;
        actionQueue.insert (action);
    }
    else
    {
        fprintf (stderr, "onFail erase key  %s not found.\n", id);
        fprintf (stderr, "Error in script suspected.\n");
        object_status = status_invalid;
    }
}
void Script::onFail_unLabel (char* volSetName) {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("onFail unlabel %s\n", volSetName);
    if (usePurify)
        action = new ScriptAction(P_EXPORT_PROG_NAME);
    else if (useCoverage)
        action = new ScriptAction(C_EXPORT_PROG_NAME);
    else
        action = new ScriptAction(EXPORT_PROG_NAME);

    action->actionType = action_unlabel;
    action->commandLine_ = action->executableName_;
    addStrOption ("k -E -d \"/dummy\" -v", volSetName, action->commandLine_);
    ensureConfigAndStatus();
    action->execute_option = execute_ifNotOk;
    action->statusImpact = impact_none;
    actionQueue.insert (action);
}
void Script::onFail_continue () {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("onFail continue\n");
        
    action = new ScriptAction("echo ");
    action->actionType = action_special;    
    action->specialAction = special_continue;
    action->execute_option = execute_ifNotOk;

    // not obvious: the status will have already been impacted by special action.
    action->statusImpact = impact_none;
    actionQueue.insert (action);
}
void Script::onFail_stop () {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("onFail stop\n");
        
    action = new ScriptAction("echo ");
    action->actionType = action_special;    
    action->specialAction = special_stop;
    action->execute_option = execute_ifNotOk;

    // not obvious: the status will have already been impacted by special action.
    action->statusImpact = impact_none;
    actionQueue.insert (action);
}
void Script::onFail_end () {
    VERIFY_OBJECT_STATUS();
    if (trace_) printf ("onFail end\n");
}

//----------------------------------------------------------------------
// METHOD: Script::generate_start  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::generate_start()
{
    VERIFY_OBJECT_STATUS();
    if (trace_)
    printf ("generate start\n");
    
    action = new ScriptAction(GENERATE_PROG_NAME);
    action->actionType = action_invalid;
    action->commandLine_ = action->executableName_;
    action->commandLine_ += " ";
}
//----------------------------------------------------------------------
// METHOD: Script::generate_native  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::generate_native (char* id)
{
    VERIFY_OBJECT_STATUS();
    if (trace_)
    printf ("generate native %s\n", id);
    

    RWCString key (id);
    TypedPath path;
    if (pathDict_.retrieve(id, path))
    {
        action->commandLine_ += " "; 
        action->commandLine_ += path.path_;
        action->commandLine_ += " ";
    }
    else
    {
        fprintf (stderr, "generate native key %s not found.\n", id);
        fprintf (stderr, "Error in script suspected.\n");
        object_status = status_invalid;
    }
}
//----------------------------------------------------------------------
// METHOD: Script::generate_script  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::generate_script (char* id)
{
    VERIFY_OBJECT_STATUS();
    if (trace_)
    printf ("generate script %s\n", id);
    

    RWCString key (id);
    TypedPath path;
    if (pathDict_.retrieve(id, path))
    {
        action->commandLine_ += " -r  ";
        action->commandLine_ += path.path_;
        action->commandLine_ += " ";
    }
    else
    {
        fprintf (stderr, "generate script key %s not found.\n", id);
        fprintf (stderr, "Error in script suspected.\n");
        object_status = status_invalid;
    }
}
//----------------------------------------------------------------------
// METHOD: Script::generate_datacheck  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::generate_datacheck (char* id)
{
    VERIFY_OBJECT_STATUS();
    if (trace_)
    printf ("generate datacheck %s\n", id);
    

    RWCString key (id);
    TypedPath path;
    if (pathDict_.retrieve(id, path))
    {
        action->commandLine_ += " ";
        action->commandLine_ += path.path_;
        action->commandLine_ += " ";
    }
    else
    {
        fprintf (stderr, "generate datacheck key %s not found.\n", id);
        fprintf (stderr, "Error in script suspected.\n");
        object_status = status_invalid;
    }
}
//----------------------------------------------------------------------
// METHOD: Script::generate_end  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::generate_end()
{
    VERIFY_OBJECT_STATUS();
    if (trace_)
    printf ("generate end\n");
    

    action->commandLine_ += " -f";
    action->execute_option = execute_ifOk;
    action->statusImpact = impact_toFailOnly;
    actionQueue.insert (action);
}
    
 

//----------------------------------------------------------------------
// METHOD: Script::import_start  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::import_start()
{
    VERIFY_OBJECT_STATUS();
    if (usePurify)
        action = new ScriptAction(P_IMPORT_PROG_NAME);
    else if (useCoverage)
        action = new ScriptAction(C_IMPORT_PROG_NAME);
    else
        action = new ScriptAction(IMPORT_PROG_NAME);

    action->actionType = action_import;
    action->commandLine_ = action->executableName_;
    action->commandLine_ += " -P ";
        if (arcLibTrace_)
        action->commandLine_ += " -P ";
    else
        action->commandLine_ += " -P -t 1 ";

}

//----------------------------------------------------------------------
// METHOD: Script::import_fileset  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::import_fileset (char* filesetName) {
    VERIFY_OBJECT_STATUS();
    if (trace_)
    printf ("import fileset %s\n", filesetName);
    
    action->commandLine_ += " -f ";
    action->commandLine_ += filesetName;
    action->commandLine_ += " ";        
}
//----------------------------------------------------------------------
// METHOD: Script::import_volset  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::import_volset (char* volName) {
    VERIFY_OBJECT_STATUS();
    if (trace_)
    printf ("import volset %s\n", volName);
    
    action->commandLine_ += "-v ";
    action->commandLine_ += volName;
    action->commandLine_ += " ";        
}
//----------------------------------------------------------------------
// METHOD: Script::import_native  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::import_native (char* id) {
    VERIFY_OBJECT_STATUS();
    if (trace_)
    printf ("import native %s\n", id);
    
    
    RWCString key (id);
    TypedPath path;
    if (pathDict_.retrieve(id, path))
    {
        action->commandLine_ += "-r ";
        action->commandLine_ += path.path_;
        action->commandLine_ += " ";
    }
    else
    {
        fprintf (stderr, "generate datacheck key %s not found.\n", id);
        fprintf (stderr, "Error in script suspected.\n");
        object_status = status_invalid;
    }
}

//----------------------------------------------------------------------
// METHOD: Script::import_cmr  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::import_cmr (char* id) {
    VERIFY_OBJECT_STATUS();
    if (trace_)
    printf ("import cmr %s\n", id);
    

    RWCString key (id);
    TypedPath path;
    if (pathDict_.retrieve(id, path))
    {
        action->commandLine_ += "-c ";
        action->commandLine_ += path.path_;
        action->commandLine_ += " ";
    }
    else
    {
        fprintf (stderr, "export native key %s not found.\n", id);
        fprintf (stderr, "Error in script suspected.\n");
        object_status = status_invalid;
    }
}
void Script::import_free (char* id) {
    VERIFY_OBJECT_STATUS();
    if (trace_)
    printf ("import free %s\n", id);
    

    RWCString key (id);
    TypedPath path;
    if (pathDict_.retrieve(id, path))
    {
        action->commandLine_ += "-e ";
        action->commandLine_ += path.path_;
        action->commandLine_ += " ";
    }
    else
    {
        fprintf (stderr, "export native key %s not found.\n", id);
        fprintf (stderr, "Error in script suspected.\n");
        object_status = status_invalid;
    }
}
void Script::import_config(char* id)
{
    VERIFY_OBJECT_STATUS();
    if (trace_)
        printf ("import config %s\n", id);
    
    
    RWCString key (id);
    TypedPath path;
    if (pathDict_.retrieve(id, path))
    {
        action->commandLine_ += "-a ";
        action->commandLine_ += path.path_;
        action->commandLine_ += " ";
    }
    else
    {
        fprintf (stderr, "import native key %s not found.\n", id);
        fprintf (stderr, "Error in script suspected.\n");
        object_status = status_invalid;
    }
}

void Script::import_status(char* id)
{
    VERIFY_OBJECT_STATUS();
    if (trace_)
        printf ("import status %s\n", id);
    
    RWCString key (id);
    TypedPath path;
    if (pathDict_.retrieve(id, path))
    {
        action->commandLine_ += "-s ";
        action->commandLine_ += path.path_;
        action->commandLine_ += " ";
    }
    else
    {
        fprintf (stderr, "import native key %s not found.\n", id);
        fprintf (stderr, "Error in script suspected.\n");
        object_status = status_invalid;
    }
}
void Script::import_synth()
{
    VERIFY_OBJECT_STATUS();
    if (trace_)
        printf ("import synth\n");
    
    action->commandLine_ += "-R ";
}

//----------------------------------------------------------------------
// METHOD: Script::import_end  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::import_end()
{
    VERIFY_OBJECT_STATUS();
    ensureConfigAndStatus();
    action->execute_option = execute_ifOk;
    action->statusImpact = impact_toFailOnly;
    actionQueue.insert (action);
}


//----------------------------------------------------------------------
// METHOD: Script::fscomp_start  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::fscomp_start()
{
    VERIFY_OBJECT_STATUS();
    action = new ScriptAction(FSCOMP_PROG_NAME);
    action->actionType = action_invalid;
    action->commandLine_ = action->executableName_;
    action->commandLine_ += " ";
}    
//----------------------------------------------------------------------
// METHOD: Script::fscomp_datacheck  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::fscomp_datacheck(char* id)
{
    VERIFY_OBJECT_STATUS();
     RWCString key (id);
    TypedPath path;
    if (pathDict_.retrieve(id, path))
    {
        fscompDataCheck_ =  path.path_;
    }
    else
    {
        fprintf (stderr, "fscomp datacheck key %s not found.\n", id);
        fprintf (stderr, "Error in script suspected.\n");
        object_status = status_invalid;
    }
}
//----------------------------------------------------------------------
// METHOD: Script::fscomp_native  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::fscomp_native(char* id)
{
    VERIFY_OBJECT_STATUS();
    RWCString key (id);
    TypedPath path;
    if (pathDict_.retrieve(id, path))
    {
        fscompNative_ =  path.path_;
    }
    else
    {
        fprintf (stderr, "export native key %s not found.\n", id);
        fprintf (stderr, "Error in script suspected.\n");
        object_status = status_invalid;
    }
}
//----------------------------------------------------------------------
// METHOD: Script::  --trp
//
// ARGS:
//
// RETURNS:
//
// PRE-CONDITIONS:
// POST-CONDITIONS:
//
// ERRORS:
// NOTES:
//----------------------------------------------------------------------
void Script::fscomp_end()
{
    VERIFY_OBJECT_STATUS();
    action->commandLine_ = action->executableName_;
    action->commandLine_ += " ";
    action->commandLine_ += fscompDataCheck_;
    action->commandLine_ += " ";
    action->commandLine_ += fscompNative_;
    action->commandLine_ += " ";
    action->execute_option = execute_ifOk;
    action->statusImpact = impact_toFailOnly;
    actionQueue.insert (action);
}





