//-*-c++-*-
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <rw/cstring.h>
#include "acScript.h"

int trace =0;

void usage();
main (int argc,CHAR_PTR argv[])
{
    RWCString scriptFile, outputFile;
    RWBoolean haveScript;
    RWBoolean trace = FALSE;
    RWBoolean debugOnly = FALSE;
    RWBoolean verbose = FALSE;
    RWBoolean printUsage = FALSE;

    RWCString density = "512";
    RWCString type = "mo";
    RWCString volSetName = "synthDefaultV";
    RWCString fileSetName = "synthDevaultf";
    RWCString cmrName;
    
    int        numNewVols = 0;
    RWBoolean useCmr = TRUE;
    RWBoolean saveCmr = FALSE;
    RWBoolean saveVolSet = FALSE;
    RWBoolean saveScript = FALSE;
    RWBoolean usePurifyVersion = FALSE;
    RWBoolean useCoverageVersion = FALSE;
    RWBoolean repeatUntilFailure = FALSE;
    RWCString synthFileName;
    RWBoolean performFreeSpaceCheck = TRUE;
        
    char c;
    
    while((c = getopt(argc,argv,"s:o:tdvD:T:V:m:n:f:CcKSiRPEy:")) != EOF) {
        switch(c) {
        case 'D':
            if ( (strcmp (optarg, "512") == 0) || (strcmp (optarg, "1024") == 0))
                density = optarg;
            else {
                printUsage = TRUE;
                fprintf (stderr,
                         "\nError: option \"-D\", density, must be 512 or 1024\n\n");
            }
            break;
        case 'T':
            if ( (strcmp (optarg, "mo") == 0) || (strcmp (optarg, "worm") == 0))
                type = optarg;
            else {
                printUsage = TRUE;
                fprintf (stderr, "\nError: option \"-T\", type, must be mo or worm\n\n");
            }
            break;
        case 'n':
            numNewVols = atoi(optarg);
            if (numNewVols == 0) {
                printUsage = TRUE;
                fprintf (stderr,
                         "\nError: option \"-n\", num new vols, must be a positive whole number.\n");
            }
            break;
        case 'f':
            fileSetName = optarg;
            break;
        case 'V':
            volSetName = optarg;
            break;
        case 'C':
            useCmr = FALSE;
            break;
        case 'c':
            saveCmr = TRUE;
            break;
        case 'm':
            cmrName = optarg;
        case 'K':
            saveVolSet = TRUE;
            break;
        case 'S':
            saveScript = TRUE;
            break;
        case 'E':
            performFreeSpaceCheck = FALSE;
            break;
        case 'y':
            synthFileName = optarg;
            break;
        case 'P':
            usePurifyVersion = TRUE;
            break;
        case 'i':
            useCoverageVersion = TRUE;
            break;
        case 'R':
            repeatUntilFailure = TRUE;
            break;
        case 's':
            scriptFile = optarg;
            haveScript = TRUE;
            break;
        case 'o':
            outputFile = optarg;
            break;
        case 't':
            trace = TRUE;
            break;
        case 'd':
             debugOnly = TRUE;
            break;
        case 'v':
            verbose = TRUE;
            break;
        default:
            fprintf (stderr, "Unknown option %c.\n\n", c);
            printUsage = TRUE;
            break;
        }
    }

    if (printUsage) {
        usage();
        exit (1);
    }

    // If no script file is specified, build one per other command line
    // arguments.
    if (!haveScript || scriptFile.isNull()) {
        if (synthFileName.isNull()) {
            fprintf (stderr,
                     "\nERROR: A synth file name is required if a script is not provided.\n\n");
            usage();
            exit(1);
        }
        
        scriptFile = volSetName;
        scriptFile += ".script";
        if (saveScript)
            fprintf (stderr, "Now creating archive script file %s.\n",
                     (char*) scriptFile);
        else
             fprintf (stderr, "Now creating temporary archive script file %s.\n",
                     (char*) scriptFile);
        FILE* script = fopen ((char*) scriptFile, "w");
        if (script == NULL) {
            fprintf (stderr, "Cannot create script file %s.  Will now terminate.\n\n",
                     (char*)scriptFile);
            exit(1);
        }
        
        fprintf (script, "script %s\n", (char*)volSetName);
        
        fprintf (script, "\n");
        fprintf (script, "paths {");
        fprintf (script, "      synth   synth1          \"%s\"\n", (char*)synthFileName);
        fprintf (script, "      config  testConfig      \"~/nsr/regress/acrc\"\n");
        fprintf (script, "      status  jbstat          \"~/nsr/regress/jbstat\"\n");
        if (cmrName.isNull())
            fprintf (script, "  cmr     cmr1            \"%s.cmr\"\n", (char*) volSetName);
        else
            fprintf (script, "  cmr     cmr1            \"%s\"\n", (char*) cmrName);
            
        if (performFreeSpaceCheck) {
            RWCString freeSpaceFileName = volSetName + "_freeSpace";
            fprintf (script, "        free    free1           \"%s\"\n",
                     (char*)freeSpaceFileName);
        }
        
        fprintf (script, "}\n");
        fprintf (script, "\n");

        if (numNewVols > 0) {
            fprintf (script, "label {        \n");
            fprintf (script, "  volumeSetName   \"%s\"\n", (char*)volSetName);
            fprintf (script, "  fileSetName     \"%s\"\n", (char*)fileSetName);
            fprintf (script, "  media           %s\n", (char*) type);
            fprintf (script, "  density         %s\n", (char*)density);

            for (int i = 1; i <= numNewVols; i++) {
                RWCString vol;
                vol = volSetName;
                char incr[10];
                sprintf (incr, "%i", i);
                vol +=  incr;
                fprintf (script, "        addVolume       \"%s\"\n", (char*)vol);
            }
            fprintf (script, "}\n");
            fprintf (script, "\n");
        }

        fprintf (script, "export {\n");
        fprintf (script, "      volumeSetName   \"%s\"\n", (char*)volSetName);
        fprintf (script, "      fileSetName     \"%s\"\n", (char*)fileSetName);
        if (useCmr) {
            fprintf (script, "        writing         cmr1\n");
        }
        
        if (performFreeSpaceCheck) {
            fprintf (script, "        free            free1\n");
        }
        fprintf (script, "      synth           synth1\n");
        fprintf (script, "}\n");
        fprintf (script, "\n");
        fprintf (script, "import {\n");
        fprintf (script, "      volumeSetName   \"%s\"\n", (char*)volSetName);
        fprintf (script, "      fileSetName     \"%s\"\n", (char*)fileSetName);
        if (useCmr) {
            fprintf (script, "  comparing       cmr1\n");
        }
        if (performFreeSpaceCheck) {
            fprintf (script, "        free            free1\n");
        }
        fprintf (script, "      synth\n");
        fprintf (script, "}\n");
        fprintf (script, "\n");

        if ( (useCmr && (!saveCmr)) || (!saveVolSet) ){
            fprintf (script, "onPass {\n");
        }
        if (useCmr && (!saveCmr)) {
            fprintf (script, "  erase cmr1\n");
        }
        if (!saveVolSet) {
            fprintf (script, "  unLabel \"%s\"\n", (char*)volSetName);
        }
        if ( (useCmr && (!saveCmr)) || (!saveVolSet) ){
            fprintf (script, "}\n");
        }

        if (!saveVolSet) {
            fprintf (script, "onFail {\n");
            fprintf (script, "  unLabel \"%s\"\n", (char*)volSetName);
            fprintf (script, "}\n");
        }
        fclose(script);
     }
    

    Script myScript((char*) scriptFile);
    if (trace) {
//        myScript.turnOnTrace();
        myScript.turnOnArcLibTrace();
    }
    
    if (debugOnly)
        myScript.turnOnDebug();
    if (!outputFile.isNull())
        myScript.setRedirectOutputFile((char*)outputFile);
    if (usePurifyVersion)
        myScript.usePurifyVersion();
    if (useCoverageVersion)
        myScript.useCoverageVersion();
 
    myScript.loadScript();
    if (verbose)
        myScript.dump();
    myScript.verifyScript();
    if (repeatUntilFailure)
        myScript.executeScriptUntilFailure();
    else
        myScript.executeScript();
    if (myScript.scriptStatus()) {
        printf ("-----------------------------\n");
        printf ("test %s PASSED.\n",  myScript.scriptId());
        printf ("-----------------------------\n");
        if (!outputFile.isNull()){
            RWCString rmOutput;
            rmOutput = "rm ";

            rmOutput += outputFile;
            system ((char*) rmOutput);
        }
        exit(3);
    }
    else {
        printf ("-----------------------------\n");
        printf ("test %s FAILED.\n", myScript.scriptId());
        if (!outputFile.isNull())
            printf ("%s contains output.\n", (char*) outputFile);
        printf ("-----------------------------\n");

        exit(2);
    }
}


void usage ()
{
    printf ("Usage:\n\n");
    printf ("arcTest\t[-s scriptFile][-o output][-t][-d][-v][-D den][-T type][-V][-y syn]\n");
    printf ("\t[-f name][-n numVols][-C][-c][-m name][-K][-S][-p][-i]\n\n");
    printf ("------ One of the Following (-s, -y) is Manditory --------\n");
    printf ("-s   scriptFile   Path resolving to input arcTest script.\n");
    printf ("-y   filename     Synth file name.\n");
    printf ("\n");
    printf ("[-o  output]      Optional. Capture stdout & stderr to output file.\n");
    printf ("                            (Passed test causes output file to be deleted.)\n");
    printf ("[-t]              Optional. Turn on trace output.\n");
    printf ("[-d]              Optional. Show command lines, but don't execute.\n");
    printf ("[-v]              Optional. Turn on Script object contents dumping.\n\n");
    printf ("[-P]              Optional. Use Purify versions (default is no)\n");
    printf ("[-i]              Optional. Use Coverage versions of code\n");
    printf ("[-R]              Optional. Repeat export/import actions until a failure is found.\n");
    printf ("------ Options only valid when '-y' is used -----------\n");
    printf ("[-D  {512|1024}]  Optional. density (default 512)\n");
    printf ("[-T  {mo |worm}]  Optional. type (default mo)\n");
    printf ("[-V  name]        Optional. volume set name (default synthv)\n");
    printf ("[-f  name]        Optional. file set name (default synthf)\n");
    printf ("[-n  integer]     Optional. create a new volume set (int = num vols to add)\n");
    printf ("                             (default, use existing volume set)\n");
    printf ("[-C]              Optional. DON'T create and check cmr (default creates cmr)\n");
    printf ("[-c]              Optional. Save the CMR file\n");
    printf ("[-m  name]        Optional. Name of cmr file (default is <volSetName>.cmr). \n");
    printf ("[-K]              Optional. DON'T kill volume set after (default kills volset)\n");
    printf ("[-S]              Optional. Save the Script (default kills script).\n");
    
    printf ("Examples:\n\n");
    printf ("\tarcTest -s script//myTest.arc -o myTest.out -d\n\n");
    printf ("\tarcTest -y script//myFS.sync -o myTest.out -D1024 -n 1\n\n");
    printf ("The -s option uses a prewritten arc script, and the -y option creates\n");
    printf ("The -s option uses a prewritten arc script, and the -y option creates\n");

//    printf ("ENVIRONMENT VARIABLES:\n\n");
//    printf ("ARCTEST_ROOT:      Defines the root for created relative path items in script,\n");
//    printf ("                   including types 'native', 'cmr'.  For instance a script\n");
//    printf ("                   path of \"dirs\" with an ARCTEST_ROOT \"/tmp\" will result\n");
//    printf ("                   in a native path of \"/tmp/dirs\".  Absolute paths are not\n");
//    printf ("                   impacted by ARCTEST_ROOT.\n");
}

