#ifndef __CMR_H__ //-*-c++-*-
#define __CMR_H__

// -----------------------------------------------------------------------------
//  cmr.h
//  'C'ommon 'M'etadata 'R'epresentation tools, used for udf archive validation.
//
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//
//  CLASSES:
//
//-----------------------------------------------------------------------------
// $Source: /projects/HMS/nsr/udfAncil/msVerify/cmr.h,v $
// $Revision: 1.8 $     $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 96/01/10 11:17:06 $
//-----------------------------------------------------------------------------

//#include <stdlib.h>

// The "meta" nsr include file manages the includes of other include files,
// directed by the defines in your makefile.
#include "../nsrLib.h"

// Rogue wave tools
#include <rw/bintree.h>
#include <rw/collect.h>
#include <rw/rwstring.h>
#include <rw/stackcol.h>
#include <rw/slistcol.h>
#include <rw/queuecol.h>
#include <rw/cstring.h>
#include <rw/rwfile.h>
#include <rw/hashdict.h>
#include <rw/collstr.h>
#include <rw/ordcltn.h>

#define VERIFY_OBJECT_STATUS() \
        if (object_status == status_invalid) return
#define VERIFY_OBJECT_STATUS_R(A) \
        if (object_status == status_invalid) return(A)


//----------------------------------------------------------------------
// Class: CmrError     --trp
//
//  Error writing to file or stderr,  with severity (fatal aborts).
//
// PUBLIC DATA MEMBERS:
// PUBLIC METHODS:
//  isValid()
//      Returns TRUE if the object is in a valid state.
//  declareErrorFile()
//      Opens an error output file per name string.  If not used, 
//      stderr receives error output.
//  logError()
//      Log an error string to output, with severity.  Severity string
//      is included in the output.  Fatal severity immediately aborts
//      the program.
//
// USAGE NOTES:
//----------------------------------------------------------------------
// COLLABORATORS:
// PRIVATE METHODS:
//
// IMPLEMENTATION NOTES:
//      The current CmrError interface allows for more sophisticated 
//      processing to be added, without changed object usage.  For 
//      instance, an error stack can be added, or variable string args.
//----------------------------------------------------------------------
typedef enum {errorFatal, errorWarning, errorInform} ErrorSeverity;
class CmrError
{
private:
    FILE *errFile_;
    RWCString fileName_;
    RWBoolean fileIsOpen;
    
public:    
    CmrError::CmrError ();
    CmrError::~CmrError (){if (fileIsOpen) fclose(errFile_);};

    RWBoolean isValid();
    void declareErrorFile(char *errFileName);
    void logError(ErrorSeverity severity, char *errorText);
};

//----------------------------------------------------------------------
// Class: BaseCmr     --trp
//
//  Base class for all cmr classes (except CmrError, which BaseCmr owns).
//  Allows for run time object type checking, consistent object
//  validation where reasonable, and consistent error logging.
//
// PUBLIC DATA MEMBERS:
//  RWCString   identity_       Name of object class (e.g. "CmrDiff").
//  CmrError    error           Error logging object.
//  
//
// PUBLIC METHODS:
//  isValid()
//      Returns TRUE if the object is in a valid state.
//
// USAGE NOTES:
//  This class is pure virtual.
//
//----------------------------------------------------------------------
// COLLABORATORS:
//  OWNS CmrError, RWCString
//
// PRIVATE METHODS:
// IMPLEMENTATION NOTES:
//
//----------------------------------------------------------------------
typedef enum
{
    status_uninitialized,
    status_normal,
    status_invalid
} ObjectStatus;

class BaseCmr {
public:
    RWCString identity_;
    CmrError error;
    RWCString errFileName_;
    ObjectStatus object_status;


    BaseCmr(char *identity);
    ~BaseCmr(){};
    
    virtual RWBoolean isValid() = 0;
    void declareErrorFile(char* errFileName)
    {
        if (errFileName_.length() > 0 &&
            errFileName_ != errFileName) {
            fprintf (stderr,
                     "An error file \"%s\" has already been declared\n",
                     (char*) errFileName_);
        }
        else {
            errFileName_ = errFileName;
            error.declareErrorFile(errFileName_);
        }
    }
};

//----------------------------------------------------------------------
// Class: LineBuffer     --trp
//
//  Abstraction of an output line of text.
//  
//  
//
// PUBLIC DATA MEMBERS:
//  
//  type   member   <descr>
//
// PUBLIC METHODS:
//  isValid()
//      Returns TRUE if the object is in a valid state.
//  declareFlushFile()
//      Saves pointer to a file which receives buffer.
//  flushToOutput()
//      Send string buffer to file, re-init buffer.
//  setTabSize()
//      Define the amount of chars per increment/decrement.
//  declareFlushFile()
//      Specifify the file to receive lines.
//  append()
//      Add a string to the line buffer, don't add delimiting space.
//  incrementIndent()
//      Shift subsequent lines tabSize right.
//  decrementIndent() 
//      Shift subsequent lines tabSize left - error if decrement past
//      zero.
//
// USAGE NOTES:
//  addChunk()
//      Add a string to buffer, followed by a delimiting space 
//      (a superfluous space will occur at eol).
//  addNumber()
//      Add a number to buffer, followed by a delimiting space.
//      Dec, oct or hex base may be optionaly specified.
//  setIndent
//      Absolute indention set.
//  getIndent
//      Get the current indent level (first visible char will appear
//      at (indent * tabSize)).
//  
//
//----------------------------------------------------------------------
// COLLABORATORS:
//  OWNS RWCString, RWFile
//
// PRIVATE METHODS:
//  indent ()
//      Pad indent * tabSize leading spaces to the head of the line 
//      buffer.
//  checkForIndent ()
//      When adding strings to the line buffer, check to see if the 
//      leading indention spaces need to be added.  If so, add them.
//
// IMPLEMENTATION NOTES:
//----------------------------------------------------------------------
#define DEFAULT_TABSIZE 2
typedef enum {num_decimal, num_octal, num_hex} NumBase;
typedef enum {str_stringize, str_token} StringType;
class LineBuffer :public BaseCmr
{
public:
    LineBuffer ();
    ~LineBuffer (){};
    RWBoolean isValid();
    
    void LineBuffer::setTabSize (int tabSize);
    void LineBuffer::declareFlushFile (RWFile * file);
    void LineBuffer::flushToOutput ();
    void LineBuffer::append (const char *string);
    void LineBuffer::addChunk (const char *string, const StringType stringType = str_token);

    void LineBuffer::addNumber (int val, NumBase base = num_decimal);
    void LineBuffer::addNumber (short val, NumBase base = num_decimal);
    void LineBuffer::addNumber (long val, NumBase base = num_decimal);
    void LineBuffer::addNumber (unsigned long val, NumBase base = num_decimal);

    void LineBuffer::setIndent (int indent) {indent_ = indent;}
    void LineBuffer::getIndent (int *indent) {*indent = indent_;}
    void LineBuffer::incrementIndent ();
    void LineBuffer::decrementIndent ();
private:
    CmrError error;

    RWCString buffer_;
    RWFile *outputFilep_;
    int tabSize_;
    int indent_;
    int needIndent_;

    void LineBuffer::indent ();
    void LineBuffer::checkForIndent ();
};


//----------------------------------------------------------------------
// Class: LogLevel     --trp
//
//  Label, indention and printed status of an outline-style line.
//  Is currently used exclusively by the Log class.
//
// PUBLIC DATA MEMBERS:
//  RWCString   name_           Level name string.
//  RWBoolean   needToPrint_    TRUE if the level has not yet printed.
//  RWBoolean   indent_         Level indention.
//
// PUBLIC METHODS:
//  operator<<
//      Set the level name string.
//
// USAGE NOTES:
//----------------------------------------------------------------------
// COLLABORATORS:
//  OWNS: RWCString
//
// PRIVATE METHODS:
// IMPLEMENTATION NOTES:
//   LogLevel is a RWCollectable derived object.
//----------------------------------------------------------------------
class LogLevel : public RWCollectable
{
    RWDECLARE_COLLECTABLE(LogLevel)
public:
    RWCString name_;
    RWBoolean needToPrint_;
    RWBoolean indent_;

    LogLevel();
    ~LogLevel();
        
    RWBoolean isValid() {return TRUE;}
        
    void operator<<(char *name){
            name_ = name;
    }
    void operator<<(RWCString name) {
            name_ = name;
    }
};

//----------------------------------------------------------------------
// Class: Log     --trp
//
//  Provide outline-style log file writing.
//  Logging outline levels may be added or removed, but are only printed
//  if an actual logging item is specified through addNumber or addChunk,
//  followed by endLine.
//
// PUBLIC DATA MEMBERS:
//  LineBuffer  line_           Abstraction of a text line buffer.
//  RW Stack    levelStringStack 
//                              A temporary stack of level strings. 
//  RWOrdered   referenceLabels 
//                              A set of labels, which may be inserted
//                              once, then accessed from LogLevel scope.
// PUBLIC METHODS:
//
// USAGE NOTES:
//----------------------------------------------------------------------
// COLLABORATORS:
//  OWNS LogLevel, RWSlistCollectablesStack, RWOrdered, LineBuffer,
//       RWCString
//
// PRIVATE METHODS:
// IMPLEMENTATION NOTES:
//  Internal Notes go here (special cases, algorithm issues, etc.)
//
//----------------------------------------------------------------------
class Log :public BaseCmr
{
public:
    Log(char *logFileName);
    ~Log();

    RWBoolean isValid() {return TRUE;}
    void beginSection(char *section);
    void endSection ();
    void checkSectionWrites();
    
    void addNumber (int val, NumBase base = num_decimal);
    void addNumber (short val, NumBase base = num_decimal);
    void addNumber (long val, NumBase base = num_decimal);
    void addNumber (unsigned long val, NumBase base = num_decimal);
    void addChunk (const char *string, const StringType = str_token);
    void endLine() {line_.flushToOutput();}
    operator<<(char *s) {line_.addChunk(s);}
    operator<<(RWCString s) {line_.addChunk( (char*) s);}

    LineBuffer line_;
    RWSlistCollectablesStack *levelStringStack_;
    void insertReferenceLabel(char *s);
    RWOrdered referenceLabels;
    RWBoolean haveDetail;
    RWBoolean hasReceivedDetails() {return haveDetail;}
    
private:
    CmrError error;
    RWSlistCollectablesStack levelStack_;
    RWCString sectionLabel_;
    RWFile *logFile_;
    RWBoolean needTobeginSection_;
};


//----------------------------------------------------------------------
// CLASS: CmrTools     --trp
//
//  A loose collection of somewhat generic tools, used by the cmr
//  classes.
//
// PUBLIC DATA MEMBERS:
//  none
//
// PUBLIC METHODS:
//  elistCompare()
//      Compare two NSR elist objects, return TRUE/FALSE whether the
//      objects are identical, and optionaly log any differences.
//  getPlistTypeDesc()
//      Given an NSR path element, provide a string describing the
//      path element type.
//  plistCompare()
//      Compare two NSR plist objects, return TRUE/FALSE whether the
//      objects are identical, and optionaly log any differences. 
//
// USAGE NOTES:
//
//----------------------------------------------------------------------
// COLLABORATORS:
//
// PRIVATE METHODS:
// IMPLEMENTATION NOTES:
//----------------------------------------------------------------------
class CmrTools
{
public:
    static RWBoolean attrCompare(const Attributes attr1,
                                 const Attributes attr2,
                                 Log *log);
    static RWBoolean elistCompare(ExtentList &elist1,
                           ExtentList &elist2,
                           Log *log);
    static void getPlistTypeDesc(int type, RWCString& typeDesc);
    static RWBoolean plistCompare(PathElementList &plist1,
                           PathElementList &plist2,
                           Log *log);
};

//----------------------------------------------------------------------
// CLASS: CmrWriter     --trp
//
//  Given access to a UDF export representation, write a representative
//  CMR file.
//
// PUBLIC DATA MEMBERS:
//  type   member   <descr>
//  type   member   <descr>
//
// PUBLIC METHODS:
//  isValid()
//      Returns TRUE if the object is in a valid state.
//  beginDir()
//      Start the scope of a new directory, include name and attr.
//  endDir()
//      End the current directory scope.
//  file()
//      Define a file.
//  symlink()
//      Define a symbolic link.
//  useShortReservedWords()
//      Use two letter, rather than default full-sized reserved words.
//  setTabSize()
//      Define the tab size for new directories, rahter than default.
//
// USAGE NOTES:
//  1.  A cmrWriter object is constructed with the output file
//          e.g. cmrWriter cmr("myfile").
//      The rogue wave RWFile class is used for file io.
//  2.  Options may be set, such as tab size and short token words.
//  3.  Four functions are used to open & close directories, and 
//      specify files and symbolic links.  For instance, these calls 
//      
//         cmr.beginDir ("dir1", dir1_attributes);
//         cmr.file("file1", file1_attributes, file1_extentList);
//         cmr.symlink ("sym1", sym1_attributes, sym1_pathElementList);
//         cmr.endDir();
//
//      would write CMR like this
//
//          begin_dir "dir1" <attributes>
//          begin_entries
//              begin_file "file1" <attributes> <extent list> end_file
//              begin_symlink "sym1" <attributes <path list> end_symlink
//          end_entries
//          end_dir
//----------------------------------------------------------------------
// COLLABORATORS:
//  OWNS RWFile, RWCString, LineBuffer, Attributes (see append* members)
//
// PRIVATE METHODS:
//  writeAttributes()
//      Generate a string to describe the contents of an NSR attr object.
//  writeExtentList()
//      Generate a string of one or more extent descriptions.
//  writePathElementList()
//      Generate a string of path elements, including type string for
//      roots.
//
// IMPLEMENTATION NOTES:
//----------------------------------------------------------------------
typedef enum {shortWords = 0, longWords = 1} ReservedWordsType;
typedef enum
  {
    tok_begin_fileset = 0,
    tok_end_fileset,
    tok_begin_dir,
    tok_end_dir,
    tok_begin_entries,
    tok_end_entries,
    tok_begin_file,
    tok_end_file,
    tok_permit,
    tok_links,
    tok_owners,
    tok_times,
    tok_extent,
    tok_begin_common,
    tok_end_common,
    tok_begin_symlink,
    tok_end_symlink,
    tok_begin_path,
    tok_end_path,
    tok_begin_allocation,
    tok_end_allocation,
  }
Token;
class CmrWriter :public BaseCmr
{
private:
    CmrError error;

    NSR::Boolean firstEntry;
    ReservedWordsType reservedWordsType;
    RWFile *outputFile;
    LineBuffer *line;

    void CmrWriter::writeAttributes (const Attributes& attr);
    void CmrWriter::writeExtentList (const ExtentList& elist);
    void CmrWriter::writePathElementList (const PathElementList& plist);
    char *CmrWriter::getTok (const Token tok);

public:
    CmrWriter (const char *outputFileName);
    ~CmrWriter ();
    RWBoolean isValid() {return TRUE;}

    // The 4 basic functions, used to define directories, files and symlinks.
    void CmrWriter::beginDir (const char *dirName, const Attributes& attr);
    void CmrWriter::endDir ();
    void CmrWriter::file (const char *fileName, const Attributes& attr,
                     const ExtentList& elist);
    void CmrWriter::symlink (const char *symlinkName, const Attributes& attr,
                        const PathElementList& plist);
    void CmrWriter::unsupportedFileType(const char *fileName, const FileType& fileType) {};
    

    // Two functions used to set options from default.
    // The tab size default may be found in the lineBuffer class.
    void CmrWriter::useShortReservedWords ();
    void CmrWriter::setTabSize (const int tabSize);
};

//----------------------------------------------------------------------
// CLASS: CmrReader     --trp
//
//  Read a CMR file, calling applicatin callback functions with 
//  directory, file and symlink information.
//
// PUBLIC DATA MEMBERS:
//  ExtentList* elist_          Pointer to an NSR extent list.
//  PathElementList* plist_     Pointer to an NSR path element list.
// 
// PUBLIC METHODS:
//  setInFile
//      Define the input CMR file.
//  setCallBackFunctions()
//      Define the application callback functions.
//
// USAGE NOTES:
//  cmrparse
//      Initiate the CMR file read.  The name "cmrparse" comes from the
//      use of flex/bison.  Please note that the cmrparse member function
//      is bison(yacc) generated, and will not be found in the 
//      cmrReader.C file.  See the Makefile "sed" line for file names.
//  
//  I've compiled and run this example.  Assuming the flex, bison, rogue wave,
//  cmr.a, and archive library files are available for linking, and you have a 
//  syntax valid file specified in the constructor (e.g. "tree.txt"), you 
//  should have no trouble running this.  Here is how I built my file, 
//  called "foo.C":
//
//    g++ -D_INCLUDE__STDC__ -I../devs -I.. -I../devsTest
//      -D_INCLUDE__STDC__ -DNSR_VERIFY -DMSTEST -D_HP_ALIASES_ -c foo.C
//    g++ -o foo foo.o cmr.a msLib.o acptlib.o -ly -lfl -lrwtool
//
// #include "cmr.h"
// int trace =0; // Required by the archive library.
//
// void *myStartDir (void *parent, char *dirName, Attributes& attr){
//     printf ("dir %s\n", dirName);
//     // Simple example, doesn't care about directory parents.
//     return NULL;
// }
// void myEndDir (void *parent) {printf ("end of some dir\n");}
// void myFile (char *fileName, Attributes& attr, ExtentList& elist){
//     printf ("file %s\n", fileName);
// }
// void mySymlink (char *symLinkName, Attributes& attr, 
//                 PathElementList& plist){
//     printf ("file %s\n", symLinkName);
// }
//
// main ()
// { 
//     CmrReader myReader ("tree.txt", myStartDir, myEndDir, myFile, 
//                         mySymlink);
//     myReader.cmrparse();
// }
//----------------------------------------------------------------------
// COLLABORATORS:
//  RWCString, RWGStack, Attributes (especially the setAttributes member)
//
// PRIVATE METHODS:
//  
//  startDir ()
//      Start a new directory.
//  endDir ()
//      End the current directory.
//  sendDir ()
//      Send a directory definition (excluding children) to app.
//  startFile ()
//      Start a file definition.
//  endFile ()
//      Send the now-defined file to the app.
//  startSymlink ()
//      Start a symlink definition.
//  endSymLink ()
//      Send the now-defined symlink to the app.
//  commonInfo ()
//      Use the Attributes::setAttributes member to populate an
//      attributes object per a string written by the
//      Attributes::appendAttributes member.
//  permit ()
//      Define the mode portion of the current attr.
//  links ()
//      obsoleted
//      Define the  linkCount portion of the current attr.
//  ownership ()
//      Define the uid & gid of the current attr.
//  times ()
//      Define a, m, ctimes of current attr.
//  beginPath ()
//      Start a new plist.
//  pathEntry ()
//      Define one plist element & append to list.
//  beginAllocation ()
//      Start a new elist.
//  extent ()
//      Define one elist element & append to elist.
//
// IMPLEMENTATION NOTES:
//
//----------------------------------------------------------------------
// Four function pointer types are defined here, corresponding to the four
// functions which must be implemented by the application code.  The
// VoidPtr type is used only for pointers into the application data structure.
typedef void *VoidPtr;
typedef VoidPtr (*StartDirFunc) (VoidPtr, char *, Attributes&);
typedef void (*EndDirFunc) (VoidPtr);
typedef void (*FileFunc) (char *, Attributes&, ExtentList&);
typedef void (*SymlinkFunc) (char *, Attributes&, PathElementList&);

// CmrState is used as a secondary input file grammar check, as well as an
// internal programming check.  It makes sure that for instance a file is not
// declared in the midst of a directory or symlink definition.  The macro
// "VERIFY_CMR_STATE" is used throughout the CmrReader implementation to 
// facilitate this checking.
typedef enum {cmr_default, cmr_dir, cmr_file, cmr_symlink} cmrState;

// CmrReader maintains a stack of void pointers to directory nodes.  The
// rogue wave generic stack class is used to manage that stack.  This
// macro is required for the class (see "tools.h++" sections 15 & 132).
declare (RWGStack, VoidPtr)

// cmrin is used by the lexer/parser code.  CmrReader assigns an input file
// to it, but does not otherwise use it.
extern FILE *cmrin;

class CmrReader :public BaseCmr
{
private:

    // cmrState provides additional syntax and program checking.  See the
    // implementation file for more details.
    cmrState cmrState_;

    FILE *inputFile_;

    // The user function pointer members, typedefed above.
    StartDirFunc userStartDir_;
    EndDirFunc userEndDir_;
    FileFunc userFile_;
    SymlinkFunc userSymlink_;

    // stack of pointers to client code data structure.
    RWGStack (VoidPtr) parentDirStack;
    VoidPtr thisDir;

    // Directory, file and symlink information.
    RWCString name_;
    Attributes attr_;

    // Member functions called by the parser.  In a nutshell, these members
    // package directory open, directory close, file and symlink info from
    // the parser, then call the user functions.  State is checked
    // via CmrState.
    void CmrReader::startDir (const char *dirName);
    void CmrReader::endDir ();
    void CmrReader::sendDir ();
    void CmrReader::startFile (const char *fileName);
    void CmrReader::endFile ();
    void CmrReader::startSymlink (const char *symLinkName);
    void CmrReader::endSymLink ();
    void commonInfo(const char* common);
    void CmrReader::beginPath ();
    void CmrReader::pathEntry (const char *entryName);
    void CmrReader::beginAllocation ();
    void CmrReader::extent (const char *uvid, const char *startSector,
                            const char *length);
         
public:
    ExtentList *elist_;
    PathElementList *plist_;

    // In case you are confused by the mere 2 overt members (constructor, parse):
    // Construction assigns the CMR file to be read, and the four user
    // functions to manage that data.  The application codes only other
    // interface to CmrReader is to initiate the parse.
    CmrReader (char *inputFileName,
               StartDirFunc userStartDir, EndDirFunc userEndDir,
               FileFunc userFile, SymlinkFunc userSymlink);
    CmrReader ();
    ~CmrReader();
    RWBoolean isValid() {return TRUE;}
    void CmrReader::setInFile(char *inputFileName);
    void CmrReader::setCallBackFunctions(
        StartDirFunc userStartDir, EndDirFunc userEndDir,
        FileFunc userFile, SymlinkFunc userSymlink);
    int CmrReader::cmrparse ();
};

//----------------------------------------------------------------------
// Class: ValNode     --trp
//
//  Node validation base class, used as-is for directories.
//
// PUBLIC DATA MEMBERS:
//  RWCString   name_           Node name.
//  NodeType    type_           node type.
//
// PUBLIC METHODS:
//  getTypeDesc()
//      Given a nodetype, provide a description of that nodetype.
//  compareTo()
//      ValNode ordering function.  Is used by rogue wave to sort.
//  isCommonEqualWithLog()
//      Compares class local valnode to external valnode, returning
//      boolean equality, and optionally logging differences.
//  isEqual()
//      Returns boolean valnode local equality to another valnode. 
//      Is used by rogue wave for uniqueness checking.  Calls 
//      isEqualWithLog, with NULL as the log parameter.
//  isEqualWithLog()
//      Returns boolean valnode local equality to another valnode. 
//      Optionally logs differences.  This function is virtual.
//  operator<<()
//      Means of defining Valnode name and attributes.
//
// USAGE NOTES:
//  name
//      Returns a char* to the name data.
//
//----------------------------------------------------------------------
// COLLABORATORS:
//   OWNS RWCString, NodeType
// PRIVATE METHODS:
// 
// IMPLEMENTATION NOTES:
//----------------------------------------------------------------------
#include "nsrLib.h"
#include "cmr.h"
typedef enum 
{
    nodetype_unknown =0,
    nodetype_file,
    nodetype_symlink,
    nodetype_dir,
    nodetype_void
} NodeType;
class ValNode :  public RWCollectable
{
private:
    Attributes attr_;
    RWCString name_;
    
public:
    ValNode(NodeType type, char *name, Attributes& attr)
            : type_ (type)
    {
        name_ = name;
        attr_ = attr;
    }
    ~ValNode() {}
    
    NodeType type_;

    static void getTypeDesc(NodeType& type, RWCString &typeDesc);
    int ValNode::compareTo(const RWCollectable* c) const;

    RWBoolean isCommonEqualWithLog (const RWCollectable* c, Log *log);
    virtual RWBoolean isEqual (const RWCollectable* c);
    virtual RWBoolean isEqualWithLog (const RWCollectable* c, Log *log);

    void operator<<(char *name);
    void operator<<(Attributes attr);
    void operator<<(ExtentList elist);
    void operator<<(PathElementList plist);
    char * name() {return (char *) name_;}

private:
    ExtentList elist_;
    RWBoolean elistIsEqualWithLog(ExtentList &elist, Log *log);
    PathElementList plist_;
    RWBoolean plistIsEqualWithLog(PathElementList &plist, Log *log);
};
//----------------------------------------------------------------------
// Class: ValDirectory     --trp
//
// Container for two comparison sides of a directory. 
//
// A ValDirectory object owns two ordered lists of entries for the 
// current directory, for the driver & rider representations.  The
// directory name and a pointer into application data are owned here.
//
// PUBLIC DATA MEMBERS:
//  RWBinaryTree  driverEntries_  Sorted list of dir entries.
//  RWBinaryTree  riderEntries_   Sorted list of dir entries.
//  void*         appDirPointer_  Pointer to application directory data.
//
// PUBLIC METHODS:
//
// USAGE NOTES:
//  isValid
//      Returns TRUE if the object is in a valid state.
//  operator<<
//      means of defining name and appDirPointer.
//  name
//      Returns char* to name.
//
//----------------------------------------------------------------------
// COLLABORATORS:
//
// PRIVATE METHODS:
//
// IMPLEMENTATION NOTES:
//
//----------------------------------------------------------------------

class ValDirectory : public RWCollectable, public BaseCmr
{
public:
    ValDirectory();
    ~ValDirectory();

    RWBoolean ValDirectory::isEqual (const RWCollectable* c);

    void operator<<(char *name) {name_ = name;}
    void operator<<(void *p) {appDirPointer_ = p;}
    
    RWBoolean isValid() {return TRUE;}
    RWBinaryTree driverEntries_;
    RWBinaryTree riderEntries_;
    void *appDirPointer_;
    char * name() {return (char *) name_;}
    
private:
    RWCString name_;
};

//----------------------------------------------------------------------
// CLASS: CmrDiff     --trp
//  The CmrDiff class is used to validate UDF exports via a cmr file and an
//  on-line comparison metadata representation.
//
//  Since two representations of the same UDF export are compared by the
//  CmrDiff object, the individual directories and directory entries must
//  be synchronized and differentiable in the code and difference log output.
//  The abstraction of a "driver" and a "rider" are used throughout the classes
//  to represent the active (driver) and passive (rider) representations.
//  Since the cmr file is parsed by the cmr reader class, the cmr represetation
//  necessarily is the driver (I don't know how to create a passive
//  parser with lex & yacc).
//
// PUBLIC DATA MEMBERS:
//  none
//
// PUBLIC METHODS:
//  beginDir()
//      Is called by the driver representation to define a new dir.
//  endDir()
//      Is called by the driver representation to end current dir.
//  file()
//      Is called by the driver representation to define a new file.
//  symlink()
//      Is called by the driver representation to define a new symlink.
//
// USAGE NOTES:
//  For validation, "driver" directory entries are compared to "rider" entries.
//  At the time of this programming, an n-ary tree representation 
//  is used for comparision.  However in order to provide for flexible class
//  use, two function pointers to application functions provide access to the
//  "rider" directory tree representation.
//  
//  CmrDiff receives a begin directory function call from the reader, and
//  constructs a new ValDirectory object for that directory.  To avoid storing
//  all information for that directory and child directories in a recursive
//  data structure, it is assumed that the "rider" has ad hoc access, and thus
//  all entries for that directory are stored into the rider entries list 
//  immediately.  "driver" entries for this directory will be accrued throughout
//  subsequent cmr reader returns.  
//----------------------------------------------------------------------
// COLLABORATORS:
//  <collaborating_class> (e.g., OWNS: and/or USES:)
//
// PRIVATE METHODS:
//
// IMPLEMENTATION NOTES:
//  Application function pointers:
//   Three function pointers must be supplied by the application code to 
//   perform root access, other dir access, and directory end.  See comments
//   in the implementation for details.
// 
//  General:
//   class CmrDiff owns a stack of ValDirectory objects and methods to
//   populate and request validation of those directories.  The admin
//   of the application representation function pointers, cmr input file
//   name, output logfile name and pointer to application root data are
//   maintained here.
//
//   appDirPointer points to a class or structure used by the application
//   rider access function.  CmrDiff assumes that the rider access function
//   allocates memory for each entry.
//-----------------------------------------------------------------------------

typedef void (*RiderAccessFunc)(char *, RWBinaryTree *, void **);
typedef void (*RiderRootAccessFunc) (RWCString &, Attributes &, void **);
typedef void (*RiderEndDirectory)(void *);

class CmrDiff : public  BaseCmr
{
public:
    CmrDiff(RiderAccessFunc riderAccess,
            RiderRootAccessFunc riderRootAccess,
            RiderEndDirectory riderEndDirectory,
            char *driverDesc = "driver",
            char *riderDesc = "rider",
            char *logFileName = "cmrDiff.log");
    ~CmrDiff();

    RWBoolean isValid() {return TRUE;}

    VoidPtr beginDir(VoidPtr parent, char *dirName, Attributes& attr);
    void endDir(void *closingDir);
    void file(char *fileName, Attributes& attr, ExtentList& elist);
    void symlink(char *symLinkName, Attributes& attr, PathElementList& plist);
    RWBoolean areEqual() {return !(log_->hasReceivedDetails());}
    

private:
    int isFirstDirectory;
    Log *log_;

    RWCString riderDesc_;
    RWCString driverDesc_;
    RWSlistCollectablesStack dirStack_;
    RiderAccessFunc riderAccess_;
    RiderRootAccessFunc riderRootAccess_;
    RiderEndDirectory riderEndDirectory_;
    void *riderRoot_;
};

#endif // _CMR_H_
