//----------------------------------------------------------------------------
// (c) Copyright 1995  Hewlett-Packard Company.
// The contents of this software are proprietary and confidential to
// the Hewlett-Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transter to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the Government  is
//      subject to restrictions as set forth in subdivision (b)
//      (3)  (ii)  of the Rights in Technical Data and Computer
//      Software clause at 52.227-7013.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304
//----------------------------------------------------------------------------


%{
#include "cmr.h"

extern int yylex();
extern void yyerror(char *s);

%}

%union {
        char *string;
}

%token BEGIN_FILESET END_FILESET BEGIN_DIR END_DIR BEGIN_ENTRIES  END_ENTRIES
%token BEGIN_FILE END_FILE EXTENT 
%token BEGIN_COMMON END_COMMON
%token BEGIN_SYMLINK END_SYMLINK BEGIN_PATH END_PATH
%token BEGIN_ALLOCATION END_ALLOCATION

%token <string> QSTRING INTEGER
%type <string> qstring

%%

    /* file set */
 file_set:              begin_fileset fileset_contents end_fileset
                        ;

 begin_fileset:         BEGIN_FILESET
                        ;

 end_fileset:           END_FILESET             { }
                        ;

 fileset_contents:      directory 
                        | fileset_contents directory
                        ;

    /* directory */

 directory:             begin_directory common_info directory_contents end_directory
                        ;

 begin_directory:       BEGIN_DIR qstring       
    {   
        startDir($2);
        free($2);
    }
                        ;


 end_directory:         END_DIR
    {
        endDir();
    }
                        ;
 
 directory_contents:    /* nothing - empty directory */
        
                        /* also nothing, but with entry open & close scope */
                        | begin_entries end_entries

                        | begin_entries directory_entries end_entries
                        ;

 begin_entries:         BEGIN_ENTRIES
    {
        sendDir();
    }
                        ;

 end_entries:           END_ENTRIES             { }
                        ;

 directory_entries:     directory_entry
                        | directory_entries directory_entry
                        ;

 directory_entry:       directory
                        | file
                        | symbolic_link
                        ;
 
    /* file */

 file:                  begin_file common_info allocation end_file
                        | begin_file common_info end_file
                        ;

 begin_file:            BEGIN_FILE qstring      
    { 
        startFile($2);
        free($2);
    }
                        ;

 end_file:              END_FILE
    { 
        endFile();
    }
                        ;

    /* symbolic link */
 
 symbolic_link:         begin_symlink common_info path_element_list end_symlink
                        ;

 begin_symlink:         BEGIN_SYMLINK qstring
    { 
        startSymlink($2);
        free($2);
    }
                        ;

 end_symlink:           END_SYMLINK
    { 
        endSymLink();
    }
                        ;

    /* common info */

common_info:            begin_common qstring end_common
    {
        commonInfo($2);
        free($2);
    }

 begin_common:          BEGIN_COMMON
                        ;

 end_common:            END_COMMON
                        ;

    /* path element list */

 path_element_list:     begin_path path_entry_list end_path
    { 
    }
                        ;

 begin_path:            BEGIN_PATH
    { 
        beginPath();
    }
                        ;

 end_path:              END_PATH
                        ;

 path_entry_list:       path_entry
                        | path_entry_list path_entry
                        ;

 path_entry:            qstring
    { 
        pathEntry($1);
        free($1);
    }
                        ;

    /* ALLOCATION (extents) */

 allocation:            begin_allocation end_allocation
                        | begin_allocation extent_list end_allocation
                        ;

 begin_allocation:      BEGIN_ALLOCATION
    { 
        beginAllocation();
    }
                        ;

 end_allocation:        END_ALLOCATION
                        ;

 extent_list:           extent
                        | extent_list extent
                        ;

 extent:                EXTENT INTEGER INTEGER INTEGER 
    { 
        extent($2, $3, $4);
        free($2);
        free($3);
        free($4);        
    }
                        ;

    /* atomics */

 qstring:               QSTRING {$$ = $1;}
                        ;

    /* ---------------------- END BISON CODE---------------- */

%%
