%{
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "scriptY.h"
#include "acScript.h"

#ifdef TRACE
#define LEXTRACE(A) printf ("lex: %s\n", (A))
#else
#define LEXTRACE(A)
#endif

int nLine = 1;

%}

ws                      [ \t]+
comment                 #.*
qstring                 \"[^\"\n]*[\"\n]
token                   [a-zA-Z][a-zA-Z0-9_]*
nl                      \n
integer                 [0-9]+

%x sFILE 
%x sGEN 
%x sEX
%x sLA
%s sGR
%x sIM 
%x sOP
%x sOF
%x sFSC
%%

<sFSC,sFILE,sGEN,sEX,sIM,sLA,sGR,sOP,sOF,INITIAL>\n { nLine++; }
                        ;

<sFSC,sFILE,sGEN,sEX,sIM,sLA,sGR,sOP,sOF,INITIAL>{ws}      ;
<sFSC,sFILE,sGEN,sEX,sIM,sLA,sGR,sOP,sOF,INITIAL>{comment} ;


<sLA,sEX>{integer}      {
                                scriptlval.string = strdup(scripttext);
                                return TOKEN;
                        }

<sFSC,sFILE,sGEN,sEX,sIM,sLA,sGR,sOP,sOF,INITIAL>{qstring}  {   
                                /* skip open quote */
                                scriptlval.string = strdup(scripttext+1);
                                if(scriptlval.string[scriptleng-2] != '"')
                                        fprintf(stderr, 
                                        "Unterminated character string",
                                        (char *)0);
                                else
                                        /* remove close quote */
                                        scriptlval.string[scriptleng-2] = '\0';
                                LEXTRACE("qstring");
                                LEXTRACE(scriptlval.string);
                                return QSTRING;
                        }


        /* The '{' and '}' chars open & close scope. */
                
<sFSC,sFILE,sGEN,sEX,sIM,sLA,sGR,sOP,sOF>[\{]                   {
                                LEXTRACE("open");
                                return OPEN_SCOPE;
                        }

<sFILE>[\}]             { 
                                LEXTRACE("close files");
                                BEGIN(INITIAL); 
                                return CLOSE_FILES;
                        }

<sGEN>[\}]              { 
                                LEXTRACE("close generate");
                                BEGIN(INITIAL); 
                                return CLOSE_GENERATE;
                        }

<sEX>[\}]               { 
                                LEXTRACE("close export");
                                BEGIN(INITIAL); 
                                return CLOSE_EXPORT;
                        }
<sGR>[\}]               {
                                LEXTRACE("close grow");
                                BEGIN(INITIAL); 
                                return CLOSE_GROW;
                        }
 
<sLA>[\}]               {
                                LEXTRACE("close label");
                                BEGIN(INITIAL); 
                                return CLOSE_LABEL;
                        }

<sIM>[\}]               { 
                                LEXTRACE("close import");
                                BEGIN(INITIAL); 
                                return CLOSE_IMPORT;
                        }
<sOP>[\}]               { 
                                LEXTRACE("close on pass");
                                BEGIN(INITIAL); 
                                return CLOSE_ON_PASS;
                        }
<sOF>[\}]               { 
                                LEXTRACE("close erase");
                                BEGIN(INITIAL); 
                                return CLOSE_ON_FAIL;
                        }

<sFSC>[\}]              { 
                                LEXTRACE("close fscomp");
                                BEGIN(INITIAL); 
                                return CLOSE_FSCOMP;
                        }

        /* script id */

<INITIAL>script         {
                                LEXTRACE("script");
                                return SCRIPT;
                        }

        /* files declaration */

<INITIAL>paths          {LEXTRACE("files");BEGIN(sFILE); return FILES;}
<sFILE>datacheck        {LEXTRACE("datacheck"); return FILE_DATACHECK;}
<sFILE>native           {LEXTRACE("native"); return FILE_NATIVE;}
<sFILE>cmr              {LEXTRACE("cmr"); return FILE_CMR;}
<sFILE>script           {LEXTRACE("script"); return FILE_SCRIPT;}
<sFILE>config           {LEXTRACE("config"); return FILE_CONFIG;}
<sFILE>status           {LEXTRACE("status"); return FILE_STATUS;}
<sFILE>synth            {LEXTRACE("synth"); return FILE_SYNTH;}
<sFILE>free             {LEXTRACE("free"); return FILE_FREE;}

        /* generate declaration */

<INITIAL>generate       {LEXTRACE("generate"); BEGIN(sGEN); return GENERATE;}
<sGEN>script            {LEXTRACE("script"); return GEN_SCRIPT;}
<sGEN>native            {LEXTRACE("native"); return GEN_NATIVE;}
<sGEN>datacheck         {LEXTRACE("datacheck"); return GEN_DATACHECK;}

        /* export declaration (pardon typed state name) */

<INITIAL>export         {LEXTRACE("export"); BEGIN(sEX); return EXPORT;}
<sEX>fileSetName        {LEXTRACE("fileSet"); return EX_FILESET;}
<sEX>volumeSetName      {LEXTRACE("volset"); return EX_VOLSET;}
<sEX>from               {LEXTRACE("native"); return EX_NATIVE;}
<sEX>writing            {LEXTRACE("write"); return EX_CMR;}
<sEX>config             {LEXTRACE("config"); return EX_CONFIG;}
<sEX>status             {LEXTRACE("status"); return EX_STATUS;}
<sEX>synth              {LEXTRACE("synth"); return EX_SYNTH;}
<sEX>free               {LEXTRACE("free"); return EX_FREE;}

        /* label declaration */
<INITIAL>label          {LEXTRACE("label"); BEGIN(sLA); return LABEL;}
<sLA>config             {LEXTRACE("config"); return LA_CONFIG;}
<sLA>status             {LEXTRACE("status"); return LA_STATUS;}
<sLA>media              {LEXTRACE("media"); return LA_MEDIA;}
<sLA>density            {LEXTRACE("density"); return LA_DENSITY;}
<sLA>volumeSetName      {LEXTRACE("vol set name"); return LA_VOLSET;}
<sLA>fileSetName        {LEXTRACE("file set name"); return LA_FILESET;}
<sLA>addVolume          {LEXTRACE("addVolume"); return LA_ADD_VOLUME;}

        /* grow declaration */
<INITIAL>grow           {LEXTRACE("export"); BEGIN(sGR); return GROW;}
<sGR>config             {LEXTRACE("config"); return GR_CONFIG;}
<sGR>status             {LEXTRACE("status"); return GR_STATUS;}
<sGR>volumeSet          {LEXTRACE("vol set"); return GR_VOLSET;}
<sGR>fileSetName        {LEXTRACE("file set name"); return GR_FILESET;}
<sGR>addVolume          {LEXTRACE("addVolume"); return GR_ADD_VOLUME;}

        /* import declaration, reuses export tokens also */

<INITIAL>import         {LEXTRACE("import"); BEGIN(sIM); return IMPORT;}
<sIM>fileSetName        {LEXTRACE("fileSet"); return IM_FILESET;}
<sIM>volumeSetName      {LEXTRACE("volset"); return IM_VOLSET;}
<sIM>to                 {LEXTRACE("native"); return IM_NATIVE;}
<sIM>comparing          {LEXTRACE("cmr"); return IM_CMR;}
<sIM>config             {LEXTRACE("config"); return IM_CONFIG;}
<sIM>status             {LEXTRACE("status"); return IM_STATUS;}
<sIM>synth              {LEXTRACE("synth"); return IM_SYNTH;}
<sIM>free               {LEXTRACE("free"); return IM_FREE;}

        /* fscomp declaration */
<INITIAL>fscomp         {LEXTRACE("fscomp"); BEGIN(sFSC); return FSCOMP;}
<sFSC>datacheck         {LEXTRACE("datacheck"); BEGIN(sFSC); return FSC_DATACHECK;}
<sFSC>native            {LEXTRACE("native"); BEGIN(sFSC); return FSC_NATIVE;}

        /* on pass declaration */
<INITIAL>onPass         {LEXTRACE("onPass"); BEGIN(sOP); return ON_PASS;}
<sOP>erase              {LEXTRACE("op erase"); return OP_ERASE;}
<sOP>unLabel            {LEXTRACE("op unlabel"); return OP_UNLABEL;}
<sOP>continue           {LEXTRACE("op continue"); return OP_CONTINUE;}
<sOP>stop               {LEXTRACE("op stop"); return OP_STOP;}

        /* on fail declaration */
<INITIAL>onFail         {LEXTRACE("onFail"); BEGIN(sOF); return ON_FAIL;}
<sOF>erase              {LEXTRACE("OF erase"); return OF_ERASE;}
<sOF>unLabel            {LEXTRACE("OF unlabel"); return OF_UNLABEL;}
<sOF>continue           {LEXTRACE("of continue"); return OF_CONTINUE;}
<sOF>stop               {LEXTRACE("of stop"); return OF_STOP;}

        /* -- atomics -- */
<sFILE,sGEN,sFSC,sEX,sIM,sGR,sLA,sOF,sOP,INITIAL>{token} {
                                LEXTRACE("token");
                                LEXTRACE(scripttext);  
                                scriptlval.string = strdup(scripttext);
                                return TOKEN;
                        }

%%

void scripterror (char *s) 
{
        fprintf (stderr, "%s on line %d, at %s\n", s, nLine, scripttext);
        Script::invalidateObject();
}

int scriptwrap()
{
    return 1; // tell parser to stop at eof
}
