%{
#include "acScript.h"

extern int yylex();
extern void yyerror(char *s);

#ifdef TRACE
#define YACCTRACE(A) printf ("yacc: %s\n", (A))
#else
#define YACCTRACE(A)
#endif

%}

%union {
        char *string;
}

%token OPEN_SCOPE CLOSE_FILES CLOSE_GENERATE CLOSE_EXPORT CLOSE_IMPORT CLOSE_ERASE 
%token  CLOSE_LABEL CLOSE_GROW CLOSE_ON_PASS CLOSE_ON_FAIL
%token SCRIPT
%token FILES FILE_DATACHECK FILE_NATIVE FILE_CMR FILE_SCRIPT FILE_CONFIG FILE_STATUS
%token FILE_SYNTH FILE_FREE
%token GENERATE GEN_SCRIPT GEN_NATIVE GEN_DATACHECK
%token DESTROY
%token EXPORT EX_FILESET EX_VOLSET  EX_NATIVE EX_CMR EX_CONFIG EX_STATUS EX_SYNTH EX_FREE
%token IMPORT IM_FILESET IM_VOLSET IM_NATIVE IM_CMR IM_CONFIG IM_STATUS IM_SYNTH IM_FREE
%token LABEL LA_CONFIG LA_STATUS LA_MEDIA LA_DENSITY LA_ADD_VOLUME LA_VOLSET LA_FILESET
%token GROW GR_CONFIG GR_STATUS GR_ADD_VOLUME GR_VOLSET GR_FILESET
%token FSCOMP FSC_DATACHECK FSC_NATIVE CLOSE_FSCOMP
%token ON_PASS OP_ERASE OP_UNLABEL OP_CONTINUE OP_STOP
%token ON_FAIL OF_ERASE OF_UNLABEL OF_CONTINUE OF_STOP

%token <string> QSTRING TOKEN INTEGER
%type <string> qstring token

%%

        /* script id */

script:                 script_id
                        | script files_declaration
                        | script generate_declaration
                        | script export_declaration
                        | script label_declaration
                        | script grow_declaration
                        | script import_declaration
                        | script fscomp_declaration
                        | script on_pass_declaration
                        | script on_fail_declaration
                        ;

script_id:              SCRIPT token
                        {
                                YACCTRACE("script_id");
                                YACCTRACE($2);
                                scriptId($2);
                        }


        /* declarations */

files_declaration:      files_start    OPEN_SCOPE files_contents    files_end
generate_declaration:   generate_start OPEN_SCOPE generate_contents generate_end
export_declaration:     export_start   OPEN_SCOPE export_contents   export_end
label_declaration:      label_start    OPEN_SCOPE label_contents    label_end
grow_declaration:       grow_start     OPEN_SCOPE grow_contents     grow_end
import_declaration:     import_start   OPEN_SCOPE import_contents   import_end
fscomp_declaration:     fscomp_start   OPEN_SCOPE fscomp_contents   fscomp_end
on_pass_declaration:    on_pass_start  OPEN_SCOPE on_pass_contents  on_pass_end
on_fail_declaration:    on_fail_start  OPEN_SCOPE on_fail_contents  on_fail_end


files_start:            FILES    {YACCTRACE("files_start");file_start();}
generate_start:         GENERATE {YACCTRACE("generate_start"); generate_start();}
export_start:           EXPORT   {YACCTRACE("export_start"); export_start();}
label_start:            LABEL    {YACCTRACE("label_start"); label_start();}
grow_start:             GROW     {YACCTRACE("grow_start"); grow_start();}
import_start:           IMPORT   {YACCTRACE("import_start"); import_start();}
fscomp_start:           FSCOMP   {YACCTRACE("fscomp_start"); fscomp_start();}
on_pass_start:          ON_PASS  {YACCTRACE("on_pass_start"); onPass_start();}
on_fail_start:          ON_FAIL  {YACCTRACE("on_fail_start"); onFail_start();}

files_end:              CLOSE_FILES     {YACCTRACE("files_end");file_end();}
generate_end:           CLOSE_GENERATE  {YACCTRACE("generate_end");generate_end();}
export_end:             CLOSE_EXPORT    {YACCTRACE("export_end");export_end();}
label_end:              CLOSE_LABEL     {YACCTRACE("label_end");label_end();}
grow_end:               CLOSE_GROW      {YACCTRACE("grow_end");grow_end();}
import_end:             CLOSE_IMPORT    {YACCTRACE("import_end");import_end();}
fscomp_end:             CLOSE_FSCOMP    {YACCTRACE("fscomp_end");fscomp_end();}
on_pass_end:            CLOSE_ON_PASS   {YACCTRACE("on_pass_end");onPass_end()}
on_fail_end:            CLOSE_ON_FAIL   {YACCTRACE("on_fail_end");onFail_end()}

        /* declaration contents */

files_contents:         file_datacheck
                        | file_native
                        | file_cmr
                        | file_script
                        | file_config
                        | file_status
                        | file_free
                        | files_contents file_datacheck
                        | files_contents file_native
                        | files_contents file_cmr
                        | files_contents file_script
                        | files_contents file_config
                        | files_contents file_status
                        | files_contents file_free
                        ;
file_datacheck:         FILE_DATACHECK token qstring {
                                YACCTRACE("file_datacheck");
                                file_datacheck ($2, $3);
                        };
file_native:            FILE_NATIVE token qstring {
                                YACCTRACE("file_native");
                                file_native ($2, $3);
                        };
file_cmr:               FILE_CMR token qstring {
                                YACCTRACE("file_cmr");
                                file_cmr ($2, $3);
                        };
file_cmr:               FILE_SYNTH token qstring {
                                YACCTRACE("file_cmr");
                                file_synth ($2, $3);
                        };
file_script:            FILE_SCRIPT token qstring {
                                YACCTRACE("file_script");
                                file_script ($2, $3);
                        };
file_config:            FILE_CONFIG token qstring {
                                YACCTRACE("file_config");
                                file_config ($2, $3);
                        };
file_status:            FILE_STATUS token qstring {
                                YACCTRACE("file_status");
                                file_status ($2, $3);
                        };
file_free:              FILE_FREE token qstring {
                                YACCTRACE("file_free");
                                file_free ($2, $3);
                        };

generate_contents:      generate_native
                        | generate_script
                        | generate_datacheck
                        | generate_contents generate_native
                        | generate_contents generate_script
                        | generate_contents generate_datacheck
                        ;
generate_native:        GEN_NATIVE token {
                                YACCTRACE("generate_native");
                                generate_native($2);
                        };
generate_script:        GEN_SCRIPT token {
                                YACCTRACE("generate_script");
                                generate_script($2);
                        };
generate_datacheck:     GEN_DATACHECK token {
                                YACCTRACE("generate_datacheck");
                                generate_datacheck($2);
                        };

export_contents:        export_fileset
                        | export_volset
                        | export_native
                        | export_cmr
                        | export_config
                        | export_status
                        | export_synth
                        | export_free
                        | export_contents export_fileset
                        | export_contents export_volset
                        | export_contents export_native
                        | export_contents export_cmr
                        | export_contents export_config
                        | export_contents export_status
                        | export_contents export_synth
                        | export_contents export_free
                      ;

export_fileset:         EX_FILESET qstring {
                                YACCTRACE("export_fileset");
                                export_fileset($2);
                        }
export_volset:          EX_VOLSET qstring {
                                YACCTRACE("export_volset");
                                export_volset($2);
                        }
export_native:          EX_NATIVE token {
                                YACCTRACE("export_native");
                                export_native($2);
                        }
export_cmr:             EX_CMR token {
                                YACCTRACE("export_write");
                                export_cmr($2);
                        }
export_config:          EX_CONFIG token {
                                YACCTRACE("export_config");
                                export_config($2);
                        }
export_status:          EX_STATUS token {
                                YACCTRACE("export_status");
                                export_status($2);
                        }
export_synth:           EX_SYNTH token {
                                YACCTRACE("export_synth");
                                export_synth($2);
                        }
export_free:            EX_FREE token {
                                YACCTRACE("export_free");
                                export_free($2);
                        }

label_contents:         label_volumeSet
                        | label_contents label_config
                        | label_contents label_status
                        | label_contents label_media
                        | label_contents label_density
                        | label_contents label_volumeSet
                        | label_contents label_fileset
                        | label_contents label_addVolume
                        ;
label_config:           LA_CONFIG token {
                                YACCTRACE("label_config");
                                label_config($2);
                        }
label_status:           LA_STATUS token {
                                YACCTRACE("label_status");
                                label_status($2);
                        }
label_media:            LA_MEDIA token {
                                YACCTRACE("label_media");
                                label_media($2);
                        }
label_density:          LA_DENSITY token {
                                YACCTRACE("label_density");
                                label_density($2);
                        }
label_volumeSet:        LA_VOLSET qstring {
                                YACCTRACE("label_volumeSet");
                                label_volumeSet($2);
                        }
label_fileset:          LA_FILESET qstring {
                                YACCTRACE("label_fileset");
                                label_fileset($2);
                        }
label_addVolume:        LA_ADD_VOLUME qstring {
                                YACCTRACE("label_addVolume");
                                label_addVolume($2);
                        }


grow_contents:          grow_volumeSet
                        | grow_contents grow_config
                        | grow_contents grow_status
                        | grow_contents grow_volumeSet
                        | grow_contents grow_addVolume
                        ;
grow_config:            GR_CONFIG token {
                                YACCTRACE("grow_config");
                                grow_config($2);
                        }
grow_status:            GR_STATUS token {
                                YACCTRACE("grow_status");
                                grow_status($2);
                        }
grow_volumeSet:         GR_VOLSET qstring {
                                YACCTRACE("grow_volumeSet");
                                grow_volumeSet($2);
                        }
grow_addVolume:         GR_ADD_VOLUME qstring {
                                YACCTRACE("grow_addVolume");
                                grow_addVolume($2);
                        }


import_contents:        import_fileset
                        | import_volset
                        | import_native
                        | import_cmr
                        | import_config
                        | import_status
                        | import_synth
                        | import_free
                        | import_contents import_fileset
                        | import_contents import_volset
                        | import_contents import_native
                        | import_contents import_cmr
                        | import_contents import_config
                        | import_contents import_status
                        | import_contents import_synth
                        | import_contents import_free
                        ;
import_fileset:         IM_FILESET  qstring {
                                YACCTRACE("import_fileset");
                                import_fileset ($2);
                        };
import_volset:          IM_VOLSET qstring{
                                YACCTRACE("import_volset");
                                import_volset ($2);
                        };
import_native:          IM_NATIVE token {
                                YACCTRACE("import_native");
                                import_native($2);
                        };
import_cmr:             IM_CMR token {
                                YACCTRACE("import_verify");
                                import_cmr ($2);
                        };
import_config:          IM_CONFIG token {
                                YACCTRACE("import_config");
                                import_config ($2);
                        };
import_status:          IM_STATUS token {
                                YACCTRACE("import_status");
                                import_status ($2);
                        };
import_synth:           IM_SYNTH {
                                YACCTRACE("import_synth");
                                import_synth ();
                        };
import_free:            IM_FREE token {
                                YACCTRACE("import_free");
                                import_free ($2);
                        };

fscomp_contents:        fscomp_datacheck fscomp_native
                        | fscomp_native fscomp_datacheck

fscomp_datacheck:       FSC_DATACHECK token {
                                YACCTRACE("fscomp_datacheck");
                                fscomp_datacheck($2);
                        };
fscomp_native:          FSC_NATIVE token {
                                YACCTRACE("fscomp_native");
                                fscomp_native($2);
                        };


on_pass_contents:       on_pass_erase
                        | on_pass_unLabel
                        | on_pass_stop
                        | on_pass_contents on_pass_erase
                        | on_pass_contents on_pass_unLabel
                        | on_pass_contents on_pass_stop
on_pass_erase:          OP_ERASE token {
                                YACCTRACE("onPass_erase");
                                onPass_erase($2);
                        };
on_pass_unLabel:        OP_UNLABEL qstring {
                                YACCTRACE("onPass_unLabel");
                                onPass_unLabel($2);
                        };
on_pass_stop:           OP_STOP  {
                                YACCTRACE("onPass_stop");
                                onPass_stop();
                        };

on_fail_contents:       on_fail_erase
                        | on_fail_unLabel
                        | on_fail_continue
                        | on_fail_stop
                        | on_fail_contents on_fail_erase
                        | on_fail_contents on_fail_unLabel
                        | on_fail_contents on_fail_continue
                        | on_fail_contents on_fail_stop
on_fail_erase:          OF_ERASE token {
                                YACCTRACE("onFail_erase");
                                onFail_erase($2);
                        };
on_fail_unLabel:        OF_UNLABEL qstring {
                                YACCTRACE("onFail_unLabel");
                                onFail_unLabel($2);
                        };
on_fail_continue:       OF_CONTINUE  {
                                YACCTRACE("onFail_continue");
                                onFail_continue();
                        };
on_fail_stop:           OF_STOP {
                                YACCTRACE("onFail_stop");
                                onFail_stop();
                        };


    /* atomics */

qstring:                QSTRING {$$ = $1;}
                        ;

token:                  TOKEN {$$ = $1;}
                        ;

%%
