// udftocmr.C

// $Source: /projects/HMS/nsr/udfAncil/msVerify/udftocmr.cpp,v $
// $Revision: 1.1 $     $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 96/06/13 23:05:29 $

//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <rw/cstring.h>
#include "nsrLib.h"
#include "../msTest/arcImEx.h"
#include "cmr.h"

//-----------------------------------------------------------------------------
// Some macros
//-----------------------------------------------------------------------------

#define MAX_IOSIZE      262144          // 256K chunks

//-----------------------------------------------------------------------------
// Function prototypes
//-----------------------------------------------------------------------------

static int  writeCmr ( TreeDB& tree,TreeNode& root, CmrWriter& cmrWriter );
static void freeTree( TreeNode& node );
static void Usage(void);

//-----------------------------------------------------------------------------
// Global variables
//-----------------------------------------------------------------------------

// The trace global variable seems to be needed for the Archive object.

int trace = 0;

//-----------------------------------------------------------------------------
// arcRead.C:  This file contains source for a simple test to use the NSR
//             MSLIB Archive object in conjunction with the hpACDev autochanger
//             device object to read an nsr volume.  By default, it will simply 
//             echo to screen a list of filenames and some simple information 
//             about each file.  If the [ -r directory ] option is specified,
//             the files will actually be restored to the given directory.
//
// Usage:
//
// arcRead [-a acrcFile] [-s jbstatFile] [-r directory ] [-f fsname] -v vsname
//
// [-a acrcFile]:    Optional.  Pathname of the acrc configuration file.
// [-s jbStatFile]:  Optional.  Pathname of the jukebox status file.
// [-r directory]:   Optional.  If this option is given, the test will
//                   restore the files to the specified directory.  Note
//                   that you must ensure that the target directory has
//                   sufficient free space.
// -f fsname:        Optional.  The file set name to import, quoted if 
//                   necessary.  If not specified, the most recent fileset is
//                   imported from the given volume set.
// -v vsname:        Mandatory.  The volume set name, quoted if necessary.
//
// Example:  
//
// arcRead -a /users/jonh/.acrc -s /tmp/jbstat.1 -v "Volset 1" -f "Fileset 1"
//
// ENVIRONMENT VARIABLES:
//
// The acrc file and the jukebox status file can both be specified by 
// environment variables:
//
// ACRC_FILE:    Environment variable containing the acrc file pathname.
// JBSTAT_FILE:  Environment variable containing the jukebox status file 
//               pathname.
//
// Note that the -a and -s options will override any value in the corresponding
// environment variable, and the -a and -s options are mandatory if the
// environment variables are not set.
//-----------------------------------------------------------------------------

main (int argc,CHAR_PTR argv[])
{
    int                 c;
    int                 rval = 0;
    int                 usage = 0;
    CHAR_PTR            cptr;
    RWCString           acrcFile;
    RWCString           jbStatFile;
    RWCString           vsname;
    RWCString           fsname;
    RWCString           restoreDir;
    RWCString           cmrFileName;

    extern CHAR_PTR     optarg;                 // For getopt()
    extern int          optind,opterr;          // For getopt()

    // Get environment variables, if any.

    cptr = getenv("ACRC_FILE");
    if(cptr != NULL) {
        acrcFile = cptr;
    }
    cptr = getenv("JBSTAT_FILE");
    if(cptr != NULL) {
        jbStatFile = cptr;
    }

    // Parse command-line arguments.

    while((c = getopt(argc,argv,"a:s:r:v:f:c:")) != EOF) {
        switch(c) {
        case 'a':
            acrcFile = optarg;
            break;
        case 's':
            jbStatFile = optarg;
            break;
        case 'r':
            restoreDir = optarg;
            break;
        case 'v':
            vsname = optarg;
            break;
        case 'f':
            fsname = optarg;
            break;
        case 'c':
            cmrFileName = optarg;
            break;
        default:
            usage = 1;
            break;
        }
    }

    if(acrcFile.isNull() || jbStatFile.isNull() || vsname.isNull()
       || cmrFileName.isNull()) {
        usage = 1;
    } 

    // Make sure that the argument directory for restore, if specified,
    // exists and can be accessed.

    if(!restoreDir.isNull()) {
        if(access((CHAR_PTR) restoreDir,R_OK | W_OK | X_OK | F_OK)) {
            fprintf(stderr,"Restore directory %s not accessible!\n",
                    (CHAR_PTR) restoreDir);
            usage = 1;
        }
    }

    if(usage) {
        Usage();
        exit(2);
    }

    // If we got here, we're ready to start testing.  If the importArchive
    // function succeeds, the root TreeNode will point to the 
    // root node of the imported directory structure, and the TreeDB
    // contain the tree structure and all file information.

    NSRJukeBox          jb(acrcFile);
    TreeDB              db("/tmp/dummyfile");
    TreeNode            root;

    if(jb.error() != NSR::error_None) {
        fprintf(stderr,"Construction of NSRJukeBox failed, error is %s\n",
                (jb.error()).message());
        exit(2);
    }

    // Read the status file.

    jb.readStatusFile(jbStatFile);
    if(jb.error() != NSR::error_None) {
        fprintf(stderr,"Read of status file %s failed, error is %s\n",
                (CHAR_PTR) jbStatFile,(jb.error()).message());
        exit(2);
    }

    // Set the working volume set.

    jb.setExistingWVS(vsname);
    if(jb.error() != NSR::error_None) {
        fprintf(stderr,"Unable to set WVS to volume set \"%s\", error is %s\n",
                (CHAR_PTR) vsname,(jb.error()).message());
        exit(2);
    }

    // Do the import!

    VolsetNameMapList   vsnmList;
    VolsetNameMap *     vsnmptr;
    SidList             sidList;

//    enum whatToDoWithTree {do_udftocmr, do_udfval};
//    whatToDoWithTree = do_udftocmr;
    
    jb.importArchive(fsname,db,root);
    if(jb.error() == NSR::error_None) {
//        switch (whatToDoWithTree) {
//        case do_udftocmr:
//        {    
            CmrWriter cmrWriter((char*)cmrFileName);
            cmrWriter.useShortReservedWords ();
            cmrWriter.setTabSize (2);
            writeCmr(db, root, cmrWriter);
//            break;
//        }
//        case do_udfval:
//        {
//            break;
//        }
//        } // switch
    }

    freeTree(root);

    return rval;
}

//-----------------------------------------------------------------------------
// writeCmr     Recursively write CMR from fileset information in tree.
//
// preconditions:
//      The tree is populated with a fileset.
//
// postconditions:
//      The output file has been populated with CMR refecting fileset info in
//      the tree, or an error has terminated processing.
//-----------------------------------------------------------------------------

static int writeCmr ( TreeDB& tree,TreeNode& root, CmrWriter& cmrWriter )
{
    int                 i;
    int                 rval;
    TreeNode            tnode;
    Error               err;

    // Tell the CmrWriter of this directory, file or symlink.
    switch (root.attr_.fileType()) {

        // currently supported file types.
    case NSR::file_Directory:
        cmrWriter.beginDir(root.filename_, root.attr_);

        // recurr on children
        for(i = 0; i < root.getNumChildren(); i++) {

            err = root.getChild(i,tnode);
            if(err != NSR::error_None) {                // Parent entry
                continue;
            }
            rval = writeCmr(tree,tnode, cmrWriter);
            if(rval) {
                return rval;
            }
        }

        // Close the current directory.
        cmrWriter.endDir();    
        break;
    case NSR::file_Raw:
        cmrWriter.file (root.filename_, root.attr_, root.elist_);
        break;
    case NSR::file_SymbolicLink:
        cmrWriter.symlink(root.filename_, root.attr_, root.plist_);
        break;

        // currently unsupported file types.
    case NSR::file_SpecialBlockDevice:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_SpecialBlockDevice);
        break;
    case NSR::file_SpecialCharacterDevice:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_SpecialCharacterDevice);
        break;
    case NSR::file_ExtendedAttributes:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_ExtendedAttributes);
        break;
    case NSR::file_Fifo:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_Fifo);
        break;
    case NSR::file_Sock:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_Sock);
        break;
    case NSR::file_Unspecified:
    default:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_Unspecified);
        break;
    }

    return 0;
}


    
//-----------------------------------------------------------------------------
// FUNCTION:  Usage
//
// Prints usage information to screen.
//-----------------------------------------------------------------------------

static void
Usage( void )
{
    static CHAR_PTR usage_str[] = {
"Usage:\n",
"udftocmr [-a acrcFile] [-s jbstatFile] [-r directory ] [-f fsname] -v vsname -c cmr",
"",
"[-a acrcFile]:    Optional.  Pathname of the acrc configuration file.",
"[-s jbStatFile]:  Optional.  Pathname of the jukebox status file.",
"[-r directory]:   Optional.  If this option is given, the test will",
"                  restore the files to the specified directory.  Note",
"                  that you must ensure that the target directory has",
"                  sufficient free space.",
"-f fsname:        Optional.  The file set name to import, quoted if ",
"                  necessary.  If not specified, the most recent fileset is",
"                  imported from the given volume set.",
"-v vsname:        Mandatory.  The volume set name, quoted if necessary.",
"-c cmr:           Mandatory.  The name of the cmr file to create or overwrite.",
"",
"Example:  ",
"",
"udftocmr -a .acrc -s /tmp/jbstat.1 -v \"Volset 1\" -f \"Fileset \" -c vset1.cmr",
"",
"ENVIRONMENT VARIABLES:",
"",
"The acrc file and the jukebox status file can both be specified by ",
"environment variables:",
"",
"ACRC_FILE:    Environment variable containing the acrc file pathname.",
"JBSTAT_FILE:  Environment variable containing the jukebox status file ",
"              pathname.",
"",
"Note that the -a and -s options will override any value in the corresponding",
"environment variable, and the -a and -s options are mandatory if the",
"environment variables are not set.",
NULL
};
    int i = 0;

    while(usage_str[i] != NULL) {
        printf("%s\n",usage_str[i]);
        i++;
    }
}

//-----------------------------------------------------------------------------
// FUNCTION:  freeTree()
//
// Frees memory associated with a tree.
//-----------------------------------------------------------------------------

static void
freeTree( TreeNode& node )
{
    TreeNode    child;
    INT32       i;

    for(i = 0; i < node.getNumChildren(); i++) {
        node.getChild(i,child);
        freeTree(child);
    }
    (node.getDB())->freeNode(node);
}

