// udfval.C

// $Source: /projects/HMS/nsr/udfAncil/msVerify/udfval.cpp,v $
// $Revision: 1.6 $     $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 95/10/25 17:27:34 $

//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <rw/cstring.h>
#include "nsrLib.h"
#include "../msTest/arcImEx.h"
#include "cmr.h"

//-----------------------------------------------------------------------------
// Some macros
//-----------------------------------------------------------------------------

#define MAX_IOSIZE      262144          // 256K chunks

//-----------------------------------------------------------------------------
// Function prototypes
//-----------------------------------------------------------------------------

static int  writeCmr ( TreeDB& tree,TreeNode& root, CmrWriter& cmrWriter );
static void freeTree( TreeNode& node );
static void Usage(void);
void treeAccess (RWBoolean targetIsRoot,
                 void *voidParentDir,
                 char *dirName,
                 RWBinaryTree *dirEntries,
                 void **dirPointer);
void rootAccess (RWCString &rootName, Attributes& attr, void **rootPointer);

void *startDirFunc(VoidPtr parent, char *name, Attributes& attr);
void endDirFunc(void *dir);
void fileFunc(char *name, Attributes& attr, ExtentList& elist);
void symlinkFunc(char *name, Attributes& attr, PathElementList& plist);
void endDirectory(void *dir);

//-----------------------------------------------------------------------------
// Global variables
//-----------------------------------------------------------------------------

// The trace global variable seems to be needed for the Archive object.

int trace = 0;
CmrDiff *cmrDiff;
TreeNode* gRootptr;

//-----------------------------------------------------------------------------
// arcRead.C:  This file contains source for a simple test to use the NSR
//             MSLIB Archive object in conjunction with the hpACDev autochanger
//             device object to read an nsr volume.  By default, it will simply 
//             echo to screen a list of filenames and some simple information 
//             about each file.  If the [ -r directory ] option is specified,
//             the files will actually be restored to the given directory.
//
// Usage:
//
// arcRead [-a acrcFile] [-s jbstatFile] [-r directory ] [-f fsname] -v vsname
//
// [-a acrcFile]:    Optional.  Pathname of the acrc configuration file.
// [-s jbStatFile]:  Optional.  Pathname of the jukebox status file.
// [-r directory]:   Optional.  If this option is given, the test will
//                   restore the files to the specified directory.  Note
//                   that you must ensure that the target directory has
//                   sufficient free space.
// -f fsname:        Optional.  The file set name to import, quoted if 
//                   necessary.  If not specified, the most recent fileset is
//                   imported from the given volume set.
// -v vsname:        Mandatory.  The volume set name, quoted if necessary.
//
// Example:  
//
// arcRead -a /users/jonh/.acrc -s /tmp/jbstat.1 -v "Volset 1" -f "Fileset 1"
//
// ENVIRONMENT VARIABLES:
//
// The acrc file and the jukebox status file can both be specified by 
// environment variables:
//
// ACRC_FILE:    Environment variable containing the acrc file pathname.
// JBSTAT_FILE:  Environment variable containing the jukebox status file 
//               pathname.
//
// Note that the -a and -s options will override any value in the corresponding
// environment variable, and the -a and -s options are mandatory if the
// environment variables are not set.
//-----------------------------------------------------------------------------

main (int argc,CHAR_PTR argv[])
{
    int                 c;
    int                 rval = 0;
    int                 usage = 0;
    CHAR_PTR            cptr;
    RWCString           acrcFile;
    RWCString           jbStatFile;
    RWCString           vsname;
    RWCString           fsname;
    RWCString           restoreDir;
    RWCString           cmrFileName;
    RWCString           diffFileName;

    extern CHAR_PTR     optarg;                 // For getopt()
    extern int          optind,opterr;          // For getopt()

    // Get environment variables, if any.

    cptr = getenv("ACRC_FILE");
    if(cptr != NULL) {
        acrcFile = cptr;
    }
    cptr = getenv("JBSTAT_FILE");
    if(cptr != NULL) {
        jbStatFile = cptr;
    }

    // Parse command-line arguments.

    while((c = getopt(argc,argv,"a:s:r:v:f:c:d:")) != EOF) {
        switch(c) {
        case 'a':
            acrcFile = optarg;
            break;
        case 's':
            jbStatFile = optarg;
            break;
        case 'r':
            restoreDir = optarg;
            break;
        case 'v':
            vsname = optarg;
            break;
        case 'f':
            fsname = optarg;
            break;
        case 'c':
            cmrFileName = optarg;
            break;
        case 'd':
            diffFileName = optarg;
            break;
        default:
            usage = 1;
            break;
        }
    }

    if(acrcFile.isNull() || jbStatFile.isNull() || vsname.isNull() ||
        cmrFileName.isNull() || diffFileName.isNull() ) {
        usage = 1;
    } 

    // Make sure that the argument directory for restore, if specified,
    // exists and can be accessed.

    if(!restoreDir.isNull()) {
        if(access((CHAR_PTR) restoreDir,R_OK | W_OK | X_OK | F_OK)) {
            fprintf(stderr,"Restore directory %s not accessible!\n",
                    (CHAR_PTR) restoreDir);
            usage = 1;
        }
    }

    if(usage) {
        Usage();
        exit(2);
    }

    // If we got here, we're ready to start testing.  If the importArchive
    // function succeeds, the root TreeNode will point to the 
    // root node of the imported directory structure, and the TreeDB
    // contain the tree structure and all file information.

    NSRJukeBox          jb(acrcFile);
    TreeDB              db("/tmp/dummyfile");
    TreeNode            root;

    if(jb.error() != NSR::error_None) {
        fprintf(stderr,"Construction of NSRJukeBox failed, error is %s\n",
                (jb.error()).message());
        exit(2);
    }

    // Read the status file.

    jb.readStatusFile(jbStatFile);
    if(jb.error() != NSR::error_None) {
        fprintf(stderr,"Read of status file %s failed, error is %s\n",
                (CHAR_PTR) jbStatFile,(jb.error()).message());
        exit(2);
    }

    // Set the working volume set.

    jb.setExistingWVS(vsname);
    if(jb.error() != NSR::error_None) {
        fprintf(stderr,"Unable to set WVS to volume set \"%s\", error is %s\n",
                (CHAR_PTR) vsname,(jb.error()).message());
        exit(2);
    }

    // Do the import!

    VolsetNameMapList   vsnmList;
    VolsetNameMap *     vsnmptr;
    SidList             sidList;

//    enum whatToDoWithTree {do_udfval, do_udfval};
//    whatToDoWithTree = do_udfval;
    
    jb.importArchive(fsname,db,root);
    gRootptr = &root;

    if(jb.error() == NSR::error_None) {
//        switch (whatToDoWithTree) {
//        case do_udftocmr:
//        {    
//            char argCmrFile[] = "udfval.cmr";
//            CmrWriter cmrWriter(argCmrFile);
//            cmrWriter.useShortReservedWords ();
//            cmrWriter.setTabSize (2);
//            writeCmr(db, root, cmrWriter);
//            break;
//        }
//        case do_udfval:
//        {
    // Instantiate a CmrDiff class.
    cmrDiff  = new CmrDiff ((RiderAccessFunc) treeAccess,
                            (RiderRootAccessFunc) rootAccess,
                            (RiderEndDirectory) endDirectory,
                            (char*) cmrFileName,
                            (char*) "tree",
                            (char *) diffFileName);

    CmrReader cmrReader (cmrFileName,
                         (StartDirFunc) startDirFunc,
                         (EndDirFunc) endDirFunc,
                         (FileFunc) fileFunc,
                         (SymlinkFunc) symlinkFunc);
    cmrReader.cmrparse();
    if (cmrDiff->areEqual())
        printf ("tree and %s are equal\n", (char*)diffFileName);
    else
        printf ("tree and %s are not equal\n", (char*)diffFileName);
        

//            break;
//        }
//        } // switch
    }

    freeTree(root);

    return rval;
}

void *startDirFunc(void *parent, char *name, Attributes &attr)
{
    cmrDiff->beginDir(parent, name, attr);
}
void endDirFunc(void *dir)
{
    cmrDiff->endDir(dir);
}
void fileFunc(char *name, Attributes& attr, ExtentList& elist)
{
    cmrDiff->file(name, attr, elist);
}
void symlinkFunc(char *name, Attributes& attr, PathElementList& plist)
{
    cmrDiff->symlink(name, attr, plist);
}

//-----------------------------------------------------------------------------
// rootAccess           Provide root directory information to the CmrDiff
//                      object.
// usage
//   This is one of two functions required for use with the CmrDiff class.
//   The function declaration must agree with the "RiderRootAccessFunc"
//   function typedef in "cmrDiff.h".
//
// parameters
//   rootName     OUT   Pointer to root directory name.
//   attr         OUT   Reference to CmrDiff root attributes structure.
//   rootPointer  OUT   Pointer to the GLOBAL root treenode structure.
//-----------------------------------------------------------------------------
void rootAccess (RWCString &rootName, Attributes &attr, void **rootPointer)
{
    *rootPointer = gRootptr;
    rootName = gRootptr->filename_;
    attr = gRootptr->attr_;
}

void endDirectory(void *dir)
{
    if (dir != (void*)gRootptr)
        if ((dir < (void*)0x40000000) ||
            (dir > (void*)0x50000000)) {
            printf ("pointer (%x) out of range at %d, %s\n", (int) dir,
                    __LINE__, __FILE__);
        }
    
    if ( (dir != (void*)gRootptr) &&
         (dir != NULL)) {
        TreeNode *node = (TreeNode*) dir;
        delete node;
    }
}

//-----------------------------------------------------------------------------
// treeAccess           Access target directory per a pointer to the parent
//                      treenode, populate a list with child entries of the
//                      target directory, and return a pointer to the target
//                      directory tree node.
//
// usage
//   This is one of two functions required for use with the CmrDiff class.
//   The ValNode, FileNode and SymlinkNode classes defined in "cmrDiff.h"
//   are the required container for target entry child nodes.  The
//   function declaration must also agree with the "RiderAccessFunc"
//   typedef in "cmrDiff.h".
//
// parameters
//   voidParentDir  IN  Pointer to a treenode structure, which is parent
//                      of the target directory.
//   dirName        OUT Name of the target directory.
//   dirEntries     OUT An ordered list of target directory child entries.
//   voidTargetDir  OUT Return pointer of target directory tree node.
//-----------------------------------------------------------------------------
void treeAccess (RWBoolean targetIsRoot,
                 void *voidParentDir,
                 char *dirName,
                 RWBinaryTree *dirEntries,
                 void **voidTargetDir)
{

    // If voidParentDir is NULL we are processing the root, and thus don't
    // have a parent directory.
    // NOTE: the "new" for the root does not get free'd.
    TreeNode *targetDir;// = new TreeNode; // deleted in endDirectory.

    if (targetIsRoot) {
        targetDir = gRootptr;
    }
    else if (voidParentDir == NULL) {
        // The parent rider has not been found.
        *voidTargetDir = NULL;
        return;
    }
    else {
        
        // Get the treenode for the desired dir, from it's parent treenode.
        TreeNode *parentDir = (TreeNode *) voidParentDir;
        RWBoolean foundTargetDir = FALSE;
        targetDir = new TreeNode;
        for(int i = 0; i < parentDir->getNumChildren(); i++)
        {
            Error err = parentDir->getChild(i, *targetDir);
            if(err != NSR::error_None ||     // Parent entr
               targetDir->attr_.fileType() != NSR::file_Directory)
            {
                continue;
            }
            if (!strcmp (dirName, targetDir->filename_))
            {
                foundTargetDir = TRUE;
                break;
            }
        }

        // If the directory was not found on the rider side,
        // return, with a target dir pointer of NULL.
        if (!foundTargetDir) {
            *voidTargetDir = NULL;
            delete targetDir;
//            VoidNode *node = new VoidNode();
//            *node << dirName;
//            dirEntries->insert (node);
            return;
        }
        
    }
    *voidTargetDir = targetDir;
            
    // Iterate through each child entry of the provided parent
    // directory.
    TreeNode child;
    for(int i = 0; i < targetDir->getNumChildren(); i++)
    {
        Error err = targetDir->getChild(i, child);
        if(err != NSR::error_None) {            // Parent entry
            continue;
        }
 
        if (child.attr_.fileType() == NSR::file_Directory)
        {
            ValNode *node = new ValNode();
            *node << child.attr_;
            *node << child.filename_;
            dirEntries->insert (node);
        }
        else if (child.attr_.fileType() == NSR::file_SymbolicLink)
        {
            SymlinkNode *node = new SymlinkNode;
            *node << child.plist_;
            *node << child.attr_;
            *node << child.filename_;
            dirEntries->insert (node);
        }
        else if (child.attr_.fileType() == NSR::file_Raw)
        {
            FileNode *node = new FileNode;
            *node << child.elist_;
            *node << child.attr_;
            *node << child.filename_;
            dirEntries->insert (node);
        }
    }
}
    
//-----------------------------------------------------------------------------
// writeCmr     Recursively write CMR from fileset information in tree.
//
// preconditions:
//      The tree is populated with a fileset.
//
// postconditions:
//      The output file has been populated with CMR refecting fileset info in
//      the tree, or an error has terminated processing.
//-----------------------------------------------------------------------------

static int writeCmr ( TreeDB& tree,TreeNode& root, CmrWriter& cmrWriter )
{
    int                 i;
    int                 rval;
    TreeNode            tnode;
    Error               err;

    // Tell the CmrWriter of this directory, file or symlink.
    switch (root.attr_.fileType()) {

        // currently supported file types.
    case NSR::file_Directory:
        cmrWriter.beginDir(root.filename_, root.attr_);

        // recurr on children
        for(i = 0; i < root.getNumChildren(); i++) {

            err = root.getChild(i,tnode);
            if(err != NSR::error_None) {                // Parent entry
                continue;
            }
            rval = writeCmr(tree,tnode, cmrWriter);
            if(rval) {
                return rval;
            }
        }

        // Close the current directory.
        cmrWriter.endDir();    
        break;
    case NSR::file_Raw:
        cmrWriter.file (root.filename_, root.attr_, root.elist_);
        break;
    case NSR::file_SymbolicLink:
        cmrWriter.symlink(root.filename_, root.attr_, root.plist_);
        break;

        // currently unsupported file types.
    case NSR::file_SpecialBlockDevice:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_SpecialBlockDevice);
        break;
    case NSR::file_SpecialCharacterDevice:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_SpecialCharacterDevice);
        break;
    case NSR::file_ExtendedAttributes:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_ExtendedAttributes);
        break;
    case NSR::file_Fifo:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_Fifo);
        break;
    case NSR::file_Sock:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_Sock);
        break;
    case NSR::file_Unspecified:
    default:
        cmrWriter.unsupportedFileType(root.filename_, NSR::file_Unspecified);
        break;
    }

    return 0;
}


    
//-----------------------------------------------------------------------------
// FUNCTION:  Usage
//
// Prints usage information to screen.
//-----------------------------------------------------------------------------

static void
Usage( void )
{
    static CHAR_PTR usage_str[] = {
"Usage:\n",
"udfval -v vsname -c cmr -d log [-a acrcFile] [-s jbstatFile] ",
"       [-r directory ] [-f fsname] -v vsname",
"",
"[-a acrcFile]:    Optional.  Pathname of the acrc configuration file.",
"[-s jbStatFile]:  Optional.  Pathname of the jukebox status file.",
"[-r directory]:   Optional.  If this option is given, the test will",
"                  restore the files to the specified directory.  Note",
"                  that you must ensure that the target directory has",
"                  sufficient free space.",
"-f fsname:        Optional.  The file set name to import, quoted if ",
"                  necessary.  If not specified, the most recent fileset is",
"                  imported from the given volume set.",
"-v vsname:        Mandatory.  The volume set name, quoted if necessary.",
"-c cmr:           Mandatory.  The input cmr file name.",
"-d log:           Mandatory.  The output logfile name.",
"",
"Example:  ",
"",
"udfval -a .acrc -s /tmp/jbstat.1 -v \"Volset 1\" -f \"Fileset \" -c \"some.cmr\" -d \"some.log\"",
"",
"ENVIRONMENT VARIABLES:",
"",
"The acrc file and the jukebox status file can both be specified by ",
"environment variables:",
"",
"ACRC_FILE:    Environment variable containing the acrc file pathname.",
"JBSTAT_FILE:  Environment variable containing the jukebox status file ",
"              pathname.",
"",
"Note that the -a and -s options will override any value in the corresponding",
"environment variable, and the -a and -s options are mandatory if the",
"environment variables are not set.",
NULL
};
    int i = 0;

    while(usage_str[i] != NULL) {
        printf("%s\n",usage_str[i]);
        i++;
    }
}

//-----------------------------------------------------------------------------
// FUNCTION:  freeTree()
//
// Frees memory associated with a tree.
//-----------------------------------------------------------------------------

static void
freeTree( TreeNode& node )
{
    TreeNode    child;
    INT32       i;

    for(i = 0; i < node.getNumChildren(); i++) {
        node.getChild(i,child);
        freeTree(child);
    }
    (node.getDB())->freeNode(node);
}

