#ifndef _NSRARC_ //-*-c++-*-
#define _NSRARC_

//----------------------------------------------------------------------
//  nsrArc.h
//   Header file for nsrArc.cpp
//
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//
//  OWNERS: Jeff DiCorpo   (Original)
//          Andy Purcell   (WORM, Trustees changes)
//          Dave Van Maren (Labeling Interface)
//
//  CLASSES:
//   Archive
//
//----------------------------------------------------------------------
// $Source: /projects/HMS/nsr/udfCore/ms/nsrArc.h,v $
// $Revision: 1.61 $    $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 96/05/08 17:18:39 $
//----------------------------------------------------------------------


// The Modes of the Archive object:
enum ModeType {
    mode_Closed,                // not open
    mode_VolSet,                // Identifying and/or manipulating a volume set
    mode_GrowVS,                // Growing a volume set.
    mode_Export,                // Writing a fileset
    mode_Import                 // Reading a fileset
};

// This defines a bogus placeholding-only NetwareTrusteeInfo class,
// just to avoid a proliferation of ifdef

#ifndef TRUSTEES
class NetwareTrusteeInfo {
public:
    NetwareTrusteeInfo() {}
    ~NetwareTrusteeInfo() {}
    Boolean empty() const { return True; }
private:
};
#endif // TRUSTEES


//----------------------------------------------------------------------
// CLASS: Archive     --DVM
//
//  Reads or writes entire multi-surface UDF filesets
//
// PUBLIC DATA MEMBERS:
//  NONE
//
// PUBLIC METHODS:
//   _GENERAL_
//      error()        - return error state
//      setDevice()    - supply a device
//      openArchive()  - open, supplying volset info
//      closeArchive() - close
//   _LABELING_
//      getVolumeInfo()      - return info about a vol
//      addLabeledVolume()   - add a labeled vol
//      isCompleteVolSet()   - True if no missing vols
//      getVolSetInfo()      - return info about a volume set
//      getRootVolume()      - identify which vol in set is root
//      checkOpenIntegrity() - is logicalVol open?
//      markOpenIntegrity()  - mark logicalVol as open
//      destroyFileSet()     - render fileset inaccessible
//      destroyVolumeLabel() - make a vol unlabeled
//      getWriteProtect()    - return logicalVol status
//      setWriteProtect()    - set logicalVol writeprot status
//   _GROWING_
//      beginGrow()          - start volume set growth process
//      addUnlabeledVolume() - label and add an unlabeled vol
//      completeGrow()       - end volume set growth process
//   _EXPORT_
//      overflowSectors()    - return extra space needed for ICB
//      sizeICBTable()       - return space needed for ICBs
//      sizeDirectory()      - return space needed for a directory
//      setAvailSpace()      - supply free space for a volume
//      allocateSeqSectors() - alloc space for USE,FSD sequences
//      allocateICBTable()   - alloc space for ICBTable
//      beginExport()        - start export process
//      createDirectory()    - create a directory
//      addDirEntry()        - add a file or subdir to current dir
//      writeDirectory()     - write current dir
//      closeDirectory()     - go back to parent dir
//      writeRootICB()       - write ICB for root dir
//      writeNextICB()       - write next ICB in ICBTable
//      completeExport()     - end export process
//   _IMPORT_
//      getImplementInfo() - return info on who wrote fileset
//      beginImport()      - start import process
//      readRootICB()      - read ICB for root dir
//      readNextICB()      - read next ICB in table
//      endOfICBTable()    - at end of ICB table
//      readPath()         - read path for symlink
//      locateDirectory()  - seek to and open a dir
//      numDirEntries()    - return # entries in current dir
//      getFirstDirEntry() - get first entry in current dir
//      getNextDirEntry()  - get next entry in current dir
//      getDirEntry()      - get specified entry in current dir
//      endOfDirectory()   - at end of current dir
//      completeImport()   - end import process
//   _DEBUGGING_
//      logFile()       - sets file to log to.
//      logErrors()     - enables/disables logging of error state.
//      logAlgorithms() - enables/disables miscellaneous printf()s
//      logCalls()      - enables/disables logging of entries/exits
//
// USAGE NOTES:
//  User notes go here.
// 
//----------------------------------------------------------------------
// COLLABORATORS:  
//  Error
//  Address, Extent, ExtentList
//  FileSetInfo, FileSetList
//  Volume, VolumeTable
//  ICB_Table
//  UniqueID
//  DirectoryEntry
//  DirectoryContents
//  Directory
//  DirectoryStack
//  VolumeInfo, VolumeInfoList
//  VolumeSetInfo
//  Attributes
//  PathElement, PathElementList
//  MSDevice
//
// PRIVATE METHODS:
//
//  ((lots and lots of these))
//
// IMPLEMENTATION NOTES:
//
//----------------------------------------------------------------------
class Archive {

public:

// _GENERAL_ Methods    
    
    Archive();
    ~Archive();

    Error error() const { return err; }

    void setDevice( MSDevice *msDev );

    void openArchive( UINT16 maxNumLabeledVols,
                      const char *volSetName, 
                      UINT32 sectorSize,
                      MediaType mt );
        
    void closeArchive();

// _LABELING_ Methods    

    Boolean getVolumeInfo( sid_t sid,
                           UINT16 *nvols,
                           char *vsname,
                           char *vname,
                           UINT16 *vid, 
                           ExtentList *unerasedSpace = NULL,
                           ExtentList *erasedSpace = NULL,
                           UINT32 *dataStart = NULL);
    
    void addLabeledVolume( sid_t sid );

    Boolean isCompleteVolSet( void );

    void getVolSetInfo( VolumeSetInfo *vsip ) ;

    sid_t getRootVolume();

    Boolean checkOpenIntegrity( void );
    void markOpenIntegrity( void );

    void destroyFileSet( void );
    void destroyVolumeLabel( sid_t sid );

    WriteProtectType getWriteProtect( void );
    void setWriteProtect(WriteProtectType wp );

// _EXPORT_ Methods

    void beginGrow( void ) ;
    void addUnlabeledVolume( sid_t sid,
                             const char *volName,
                             UINT32 labelLimit,
                             const ExtentList & unerasedSpace,
                             const ExtentList & erasedSpace );
    void completeGrow( void );

// _EXPORT_ Methods.

    UINT32 overflowSectors( UINT32 noExtents );
    UINT32 overflowSectors( UINT32 noExtents,
                            const NetwareTrusteeInfo &t );

    UINT32 overflowSectors( const PathElementList &pList );
    UINT32 overflowSectors( const PathElementList &pList, 
                            const NetwareTrusteeInfo & );

    UINT32 sizeICBTable( UINT32 dirs,
                         UINT32 files,
                         UINT32 ovf );
    
    UINT32 sizeDirectory( UINT32 nEntries,
                          UINT32 sumNameLen );
    
    void setAvailSpace( sid_t sid, 
                        ExtentList &unerased,
                        ExtentList &erased );

    void beginExport( const char *fsname );

    UINT32 allocateSeqSectors();

    void allocateICBTable( UINT32 dirs,
                           UINT32 files,
                           UINT32 ovf );
    void allocateICBTable( UINT32 dirs,
                           UINT32 files,
                           UINT32 ovf, 
                           const NetwareTrusteeInfo & );

    void createDirectory( char *dname = 0 );

    void addDirEntry( Boolean isSubDir,
                      char *name,
                      UINT32 ovf );

    void writeDirectory( Extent *ext );

    void closeDirectory();

    void writeRootICB( Attributes &attr,
                       const Extent &ext );
    void writeNextICB( Attributes &attr,
                       const ExtentList &extList );
    void writeNextICB( Attributes &attr,
                       const ExtentList &extList,
                       const Extent &EAICBAddr );
    void writeNextICB( Attributes attr,
                       PathElementList pList );
    void writeNextICB( const Attributes &attr,
                       const PathElementList &pList,
                       const Extent &EAICBAddr );

#ifndef TRUSTEES
private:
#endif //NOT TRUSTEES
    void writeRootICB( Attributes &attr,
                       const Extent &ext,
                       const NetwareTrusteeInfo &t );
#ifndef TRUSTEES
public:
#endif //NOT TRUSTEES
    
#ifdef TRUSTEES
    void writeNextICB( Attributes &attr,
                       const ExtentList &extList,
                       const NetwareTrusteeInfo &t );
    void writeNextICB( Attributes attr,
                       PathElementList pList,
                       const NetwareTrusteeInfo & );
#endif //TRUSTEES
    void completeExport();

// _IMPORT_ Functions

    void getImplementInfo( ImplementationInfo * implInfo ) ;
    void getImplementInfo( INT32 fsnum,
                           ImplementationInfo * implInfo ) ;

    void beginImport( ExtentList *icbSpace ); // MO ONLY
    void beginImport( INT32 fsnum );          // WORM ONLY
    void beginImport( void );                 // WORM or MO

    void readRootICB( Attributes *attr,
                      Extent *ext );
    int readNextICB( Address    *addr,
                     Attributes *attr,
                     ExtentList *extList );
    int readNextICB( Address    *addr,
                     Attributes *attr,
                     ExtentList *extList,
                     Extent     *EAICBAddr );
#ifdef TRUSTEES
    void readRootICB( Attributes *attr,
                      Extent *ext,
                      NetwareTrusteeInfo &nti );
    int readNextICB( Address    *addr,
                     Attributes *attr,
                     ExtentList *extList, 
                     NetwareTrusteeInfo & );
#endif //TRUSTEES

    void readPath( ExtentList &elist,
                   PathElementList *pList );

    Boolean endOfICBTable();

    void locateDirectory( Extent ext );

    UINT32 numDirEntries();

    void getFirstDirEntry( Boolean *dir,
                           char *name,
                           Address *addr );
    int getNextDirEntry( Boolean *dir,
                         char *name,
                         Address *addr );
    void getDirEntry( unsigned i,
                      Boolean *dir,
                      char *name,
                      Address *addr );

    Boolean endOfDirectory();

    void completeImport();
    
// _DEBUGGING_ Functions

    void logFile(FILE * lf) ;
    void logErrors(Boolean le) ;
    void logAlgorithms(Boolean la) ;
    void logCalls(Boolean lc) ;
    
private:
    // GOES AWAY
    void createArchive( const UINT16 nvols,
                        const char *vsname, 
                        const char *fsname,
                        UINT32 size,
                        UINT32 fileDataStart);
    void setRootVolume( sid_t rootSid );
    void beginImport( INT32 fsnum,
                      ExtentList *metaDataSpace );
    // Use getVolSetInfo() instead.
    void getFileSetInfo( FileSetList *fslst );

    UINT32 S2B( UINT32 sectors ) { 
         return ( (logicalBlockSize == 512 ) ? (sectors<<9) : (sectors<<10) );
    }
    UINT32 B2S( UINT32 bytes ) { 
         return ( (logicalBlockSize == 512) ? 
                        ((bytes>>9) + ((bytes % 512) > 0)) :
                        ((bytes>>10) + ((bytes % 1024) > 0)));
    }
    // GENERAL utility stuff
    UINT16 nukeDesc( TagType descType,
                     sid_t sid,
                     UINT32 sector,
                     UINT32 addr );
    Boolean okToImport( void );
    // EA stuff
    UINT32 bytesForTrusteesEA( const NetwareTrusteeInfo & );
    UINT32 bytesForNovellAttrEA(void);
    UINT32 bytesForEAHD(void);
    void parseEA( UINT32,
                  UINT32,
                  BYTE *,
                  NetwareTrusteeInfo & );
    void loadEAs( const Address &,
                  FileEntry *fe,
                  NetwareTrusteeInfo &,
                  UINT32 & );
    void writeEAs(UINT16,
                  UINT32,
                  FileEntry *,
                  const NetwareTrusteeInfo &, 
                  Boolean &,
                  UINT32 &,
                  Attributes attr);
    void moveRoot(Extent &,
                  UINT16);
    void readSector( UINT16,
                     UINT32,
                     BYTE * );
    void readBlock( UINT16,
                    UINT32,
                    BYTE * );
    void writeSector( UINT16,
                      UINT32,
                      BYTE * );
    void writeBlock( UINT16,
                     UINT32,
                     BYTE * );
    void writeIndirectEntry(UINT16,
                            UINT32,
                            Extent &,
                            UINT32 &,
                            UINT16 &, 
                            UINT16 & );
    void getSeqExtent(UINT16,
                      UINT32,
                      Extent & );
    void readBlind(UINT16,
                   UINT32,
                   SearchDescriptor &,
                   BYTE ** );
    void getICBExtents(FileEntry *,
                       ExtentList * );
    void getICBExtentList( UINT32,
                           ExtentList &,
                           UINT32, 
                           Boolean bumpPtr = True );
    UINT32 readExtent(const Extent &,
                      BYTE * );
    UINT32 writeExtents( UINT16,
                         ICB_Table *,
                         UINT32,
                         BYTE * );
    void lastInExtent(const UINT16,
                      const Extent &,
                      UINT32 *,
                      UINT32 *,
                      UINT32 * );
    void copyFSDS( const Volume & oldRootVol );
    void getNextSeqAddress( const UINT16,
                            const Extent &,
                            UINT32,
                            UINT32 *, 
                            UINT32 *,
                            UINT32 *,
                            UINT32 *,
                            Boolean *,
                            ICBInfo * );
    Boolean searchForDesc( TagType searchDescrType,
                           Extent & seqExtent,
                           UINT32 & vdsNum ) ;
    Boolean searchVolDescSeq(sid_t sid,
                             TagType descType,
                             VDSInfo & vdsi );

    void writeVolRecogSeq(sid_t sid);
    Boolean readVolRecogSeq(sid_t sid);
    void ensureVolRecogSeq(const Volume & vol);
    void writeTD(sid_t sid,
                 UINT32 logAddr,
                 UINT32 addr);
    void writeVDS( Volume & vol );
    void writeVolumeLabel( Volume & vol );

    void writeLVD( void );
    void writeFSD( void );
    void updateSpace( const Volume & vol );
    void writeIntegrityEntry( Volume & vol );
    void checkRootVolumeSpace( UINT32 & neededSectors );
    int nonEmptyVolumeCount(void);

    Boolean readvds(sid_t sid,
                    UINT16 *vid,
                    char *vsname,
                    char *vname,
                    Extent *fsdsExt,
                    Extent *lvidsExt, 
                    UINT16 *no_vols,
                    Extent *unerasedTblExt = NULL,
                    Extent *erasedTblExt = NULL,
                    Boolean *hasAVD = NULL,
                    WriteProtectType *lvdWriteProtect = NULL );
    
    Boolean getICBTableFromLVID(void);
    void readLVIDSeq(const Volume & vol,
                     LogicalVolumeIntegrityDesc **lvidpp,
                     int *nFileSets );

    void writeNextICB( Attributes &attr,
                       const ExtentList &extList,
                       const NetwareTrusteeInfo &,
                       const Extent &EAICBAddr,
                       ICB_Table *icbTbl, 
                       Boolean isEAfe );
    
    void writeNextICB( Attributes attr,
                       PathElementList pList,
                       const NetwareTrusteeInfo &,
                       const Extent &EAICBAddr );

    void readRootICB( Attributes *attr,
                      Extent *ext,
                      ImplementationInfo *implInfo,
                      NetwareTrusteeInfo &, 
                      Boolean readingICBTable = False );

    int readNextICB( Address    *addr,
                     Attributes *attr,
                     ExtentList *extList,
                     Extent     *EAICBAddr,
                     NetwareTrusteeInfo &trustees );

    void writeSpaceEntry( sid_t sid,
                          ExtType spcType, 
                          const ExtentList &space, 
                          const Extent &e );
    void getSpace( const Extent &spcTblExt, 
                   ExtentList &space);
    
    Boolean getRootICB(INT32 fsnum,
                       Address &rootICB);

    UINT16  sizeVolSet( VolumeTable & vt ) ;
    UINT16  numContigVols( VolumeTable & vt ) ;
    Boolean isValidRoot( UINT16 volIndex ) ;

    // +++++++++++++++++ PRIVATE DATA MEMBERS ++++++++++++++++++++++++

    UINT32              partStartSector;

    UINT32              extentsPerICB;
    UINT32              extentsPerBlock;
    UINT32              extentsPerUSE;
    UINT32              extentsPerAED;

    MSDevice            *currentDevice;

    // Current volume descriptor locations and info
    UINT16              vidCurrent;

    // Volume set descriptor locations and info
    String              volumeSetName;
    String              fileSetName;
    MediaType           media;
    UINT32              logicalBlockSize;
    UINT32              logicalSectorSize;
    VolumeTable         volumeTable;
    ICB_Table           *icbTable;
    Address             rootICB;
    UINT32              root_ovf;

    ModeType            mode;

    Directory           currentDirectory;
    unsigned            entryPtr;
    DirectoryStack      directoryStack;

    Error               err;

    Boolean             endICBTable;
    Boolean             endDirectory;

    UniqueID            uniqueID;

    UINT32              numFiles;
    UINT32              numDirs;
    Boolean             partitionSeqAllowed;
    Extent              newSeqExtent;
    Boolean             rootICBICBTable;
    Boolean             volumeSetNameChanged;
    ICBInfo             bogusICBInfo;
    static Boolean      embeddedEAs;

    // VolSet Growth vars - These hold pre-growth state
    UINT32              preGrowthNumVols;
    UINT32              preGrowthNumFileSets;
    Volume              preGrowthRootVol;


    static UINT16       incrementalSeqSectors;
    static NetwareTrusteeInfo     emptyTrustees;
    static ExtendedAttrHeaderDesc static_eahd;
#ifdef TRUSTEES
    static NovellTrusteeEA        static_ntea;
#endif // TRUSTEES
};

#endif
