//--------------------------------------------------------------*-c++-*-
//  nsrImp.cpp
//   Objects used only by the Archive object.
//
//      Unpublished Confidential Information of Hewlett-Packard
//      Company. Do not disclose.  Copyright (c) Hewlett-Packard
//      Company 1995. All rights reserved
// 
// The contents of this software are proprietary and confidential to
// the Hewlett- Packard Company, and are limited in distribution to
// those with a direct need to know.  Individuals having access to
// this software are responsible for maintaining the confidentiality
// of the content and for keeping the software secure when not in use.
// Transfer to any party is strictly forbidden other than as expressly
// permitted in writing by Hewlett-Packard Company.  Unauthorized
// transfer to or possession by any unauthorized party may be a
// criminal offense.
//
//                  RESTRICTED RIGHTS LEGEND
//
//      Use,  duplication,  or disclosure by the U. S. Government is
//      subject to restrictions as set forth in subparagraph (c) (1)
//      (ii) in the Rights in Technical Data and Computer Software
//      clause in DFARS 252.227-7013 or any other successor clause.
//
//                      HEWLETT-PACKARD COMPANY
//                      3000 Hanover St.
//                      Palo Alto, CA  94304 U.S.A.
//
//      Rights for non-DOD U.S. Government Departments and Agencies
//      are as set forth in FAR 52.227-19 (c) (1,2) or any other
//      successor clause.
//
//  CLASSES:
//   VDSInfo
//   Volume
//   VolumeTable
//   ICB_Table
//   UniqueID
//   DirectoryEntry
//   DirectoryContents
//   Directory
//   DirectoryStack
//  
//----------------------------------------------------------------------
// $Source: /projects/HMS/nsr/udfCore/ms/nsrImp.cpp,v $
// $Revision: 1.15 $    $Author: hmgr $
// $State: Exp $        $Locker:  $
// $Date: 96/04/16 16:57:20 $

#include "nsrLib.h"

NSR::VolAreaCalc::VolAreaCalc(sid_t sid, UINT16 sectorSize, UINT32 numSectors )
        : sid_  (sid),
          sectorSize_ (sectorSize),
          lastSector_ (numSectors-1) {
}

NSR::Address
NSR::VolAreaCalc::volRecogSeqLoc() const
{
    Address addr;
    addr.sid = sid_;
    addr.sector = ( sectorSize_ == 512 ?
                    LOC_VRS_512 :
                    LOC_VRS_1024 );
    return addr;
}

NSR::Address
NSR::VolAreaCalc::mainAVDLoc() const 
{
    Address addr;
    addr.sid    = sid_;
    addr.sector = LOC_AVD1;
    return addr;
}

NSR::Address
NSR::VolAreaCalc::resvAVDLoc() const
{
    Address addr;
    addr.sid    = sid_;
    addr.sector = lastSector_;
    return addr;
}

NSR::Extent
NSR::VolAreaCalc::mainVDSArea() const 
{
    Extent ext;
    ext.start.sid    = sid_;
    ext.start.sector = (sectorSize_ == 512 ?
                        LOC_MVDS_512 :
                        LOC_MVDS_1024);
    ext.len = sectorSize_ * ((sectorSize_ == 512 ?
                             LOC_RVDS_512 :
                             LOC_RVDS_1024 ) - ext.start.sector);
    return ext;
}

NSR::Extent
NSR::VolAreaCalc::resvVDSArea() const 
{
    Extent ext;
    ext.start.sid    = sid_;
    ext.start.sector = (sectorSize_ == 512 ?
                        LOC_RVDS_512 :
                        LOC_RVDS_1024);
    ext.len = sectorSize_ * ((sectorSize_ == 512 ?
                             LOC_LVID_512 :
                             LOC_LVID_1024) - ext.start.sector);
    return ext;
}

NSR::Extent
NSR::VolAreaCalc::integSeqArea() const
{
    Extent ext;
    ext.start.sid    = sid_;
    ext.start.sector = ((sectorSize_ == 512) ?
                        LOC_LVID_512 : LOC_LVID_1024);
    ext.len  = sectorSize_ * (LOC_AVD1 - ext.start.sector);
    return ext;
}

// Partition: All space between AVDPs: PartStart till lastSector.
NSR::Extent
NSR::VolAreaCalc::partitionArea() const
{
    Extent ext;
    ext.start.sid    = sid_;
    ext.start.sector = LOC_PART_START;
    ext.len = sectorSize_ * (lastSector_ - LOC_PART_START);
    return ext;
}

// PartitionAreaCalc Stuff
//----------------------------------------------------------------------
// METHOD: PartitionAreaCalc::quarterPoint()  --DVM
// METHOD: PartitionAreaCalc::halfPoint()     --DVM
// METHOD: PartitionAreaCalc::q1Length()      --DVM
// METHOD: PartitionAreaCalc::q2Length()      --DVM
// METHOD: PartitionAreaCalc::h1Length()      --DVM
// METHOD: PartitionAreaCalc::h2Length()      --DVM
//      
//  These are private methods which assist in determining partition areas.
//
//  |<--- h1Length() -------------->|<--- h2Length() -------------->|
//  |<- q1Length()->|<- q2Length()->|                               |
//  ----------------------------------------------------------------
//  |               |               |                               |
//  ----------------------------------------------------------------
//  ^               ^               ^                               ^
//  LOC_USE         |               |                               |
//     quarterPoint()               |                               |
//                        halfPoint()                               |
//                                                    fileDataStart()
//
// ARGS: none.
//
// RETURNS: These all return UINT32 values.
// 
// PRE-CONDITIONS: The object must have been properly constructed.
// 
// POST-CONDITIONS: none.
// 
// ERRORS: none.
// 
//----------------------------------------------------------------------

NSR::UINT32
NSR::PartitionAreaCalc::quarterPoint() const 
{
    return LOC_USE + ((fileDataStart_-LOC_USE)>>2);
}

NSR::UINT32
NSR::PartitionAreaCalc::halfPoint() const
{
    return LOC_USE + ((fileDataStart_-LOC_USE)>>1);
}

NSR::UINT32
NSR::PartitionAreaCalc::q1Length() const
{
    return blockLen_ * (quarterPoint() - LOC_USE);
}

NSR::UINT32
NSR::PartitionAreaCalc::q2Length() const
{
    return blockLen_ * (halfPoint() - quarterPoint());
}

NSR::UINT32
NSR::PartitionAreaCalc::h1Length() const
{
    return blockLen_ * (halfPoint() - LOC_USE);
}

NSR::UINT32
NSR::PartitionAreaCalc::h2Length() const
{
    return blockLen_ * (fileDataStart_ - halfPoint());
}

//----------------------------------------------------------------------
// METHOD: PartitionAreaCalc::PartitionAreaCalc()  --DVM
//      
//  Constructs a PartitionAreaCalc object.
//
//  Two constructor forms are provided. Either the actual
//  fileDataStart may be provided, or the addresses of the USE
//  table().
//
//  The first case is used when labeling media, where the user is
//  specifying the fileDataStart. In that case, the fileDataStart is
//  rounded down to the nearest multiple of four sectors from the
//  beginning of the partition. This makes sure that the same value
//  will be calculated from the ICBAddrs in the other case.
//
//  The second case is used when recalculating the fileDataStart (and
//  all other partition area locations and lengths) after reading the
//  volume descriptor sequence. The two ICBAddrs are obtained from the
//  partition descriptor. Again, fileDataStart is recalculated from
//  these extents, and then all methods providing area information are
//  based upon it.
//
// ARGS:
//  sid_t     sid              IN Surface Identifier
//  UINT16    blockLen         IN Block length
//  MediaType mt               IN Worm or Rewritable
//  UINT32    maxFileDataStart IN First sector of data
//  Extent    unerasedICBAddr  IN From PD
//  Extent    erasedICBAddr    IN From PD
//
// RETURNS: none.
// 
// PRE-CONDITIONS: none.
// 
// POST-CONDITIONS:
//  Object is initialized with a valid value for fileDataStart_.
// 
// ERRORS: none.
// 
// NOTES:
//
//----------------------------------------------------------------------
NSR::PartitionAreaCalc::PartitionAreaCalc(sid_t sid,
                                          UINT16 blockLen,
                                          MediaType mt,
                                          UINT32 maxFileDataStart ) :
        sid_(sid),
        mediaType_(mt),
        blockLen_(blockLen)
{
    UINT32 descAreaLen = ((maxFileDataStart - LOC_USE>>2)<<2);
    fileDataStart_ = LOC_USE + descAreaLen;
}

NSR::PartitionAreaCalc::PartitionAreaCalc(UINT16 blockLen,
                                          MediaType mt,
                                          const Extent & unerasedICBAddr,
                                          const Extent & erasedICBAddr ):
        sid_(unerasedICBAddr.start.sid),
        mediaType_(mt),
        blockLen_(blockLen)
{
    UINT32 q1Len;
    if (mediaType_ != media_Worm) {
        q1Len = (erasedICBAddr.start.sector -
                 unerasedICBAddr.start.sector);
    } else {
        q1Len = unerasedICBAddr.len;
    }
    fileDataStart_ = LOC_USE + (q1Len<<2);
}

NSR::Extent
NSR::PartitionAreaCalc::unerasedICBAddr() const 
{
    Extent ext;
    ext.start.sid = sid_;
    ext.start.sector = LOC_USE;
    if (mediaType_ != media_Worm) {
        ext.len = blockLen_;
    } else {
        ext.len = q1Length();
    }
    return ext;
}

NSR::Extent
NSR::PartitionAreaCalc::unerasedSpcTblArea() const 
{
    Extent ext;
    ext = unerasedICBAddr();
    if (mediaType_ != media_Worm) {
        ext.len = q1Length();
    } else {
        ext.len = h1Length();
    }
    return ext;
}

// This should not be called if media is WORM.
NSR::Extent
NSR::PartitionAreaCalc::erasedICBAddr() const 
{
    Extent ext;
    ext.start.sid = sid_;
    ext.start.sector = quarterPoint();
    if (mediaType_ != media_Worm) {
        ext.len = blockLen_;
    } else {
        ext.len = 0;
    }
    return ext;
}

NSR::Extent
NSR::PartitionAreaCalc::erasedSpcTblArea() const 
{
    Extent ext;
    ext = erasedICBAddr();
    if (mediaType_ != media_Worm) {
        ext.len = q2Length();
    } else {
        ext.len = 0;
    }
    return ext;
}

NSR::Extent
NSR::PartitionAreaCalc::fsdSeqArea() const 
{
    Extent ext;
    ext.start.sid = sid_;
    ext.start.sector = halfPoint();
    if (mediaType_ != media_Worm) {
        ext.len = blockLen_;
    } else {
        ext.len = h2Length();
    }
    return ext;
}

// Volume Stuff

void
NSR::Volume::free( const ExtentList &unerasedextlst, const ExtentList &erasedextlst ) 
{
    // Operate on non-erased extent list
    addfree(freeSpace,unerasedextlst);
    
    // Next, operate on erased extent list
    addfree(erasedFreeSpace,erasedextlst);
}

// NOTE that this does NOT simply ADD space to the tolist, but rather
// replaces it with the fromlist, after weeding out zero-length
// extents and sorting by sector address. -DVM 7/6/95

void 
NSR::Volume::addfree( ExtentList &toList, const ExtentList &fromList ) 
{
    UINT32 high;
    Extent tmp;
    ExtentList tmpList = fromList;

    // weed out zero length extents
    if (tmpList.getCount()>0) 
        for (int j = (tmpList.getCount()-1); j >= 0; j--)
            if (tmpList[j].len == 0) tmpList.remove(j);

    toList = tmpList;

    // sort list in order by sector, from low to high
    if ( toList.getCount() < 2 ) return;

    for (unsigned end=toList.getCount()-1; end>1; end--) {
        high = 0;
        for (unsigned i=0; i<=end; i++)
            if (toList[i].start.sector > toList[high].start.sector)
                high = i;
        tmp = toList[end];
        toList[end] = toList[high];
        toList[high] = tmp;
    }
}

NSR::Extent 
NSR::Volume::alloc( UINT32 nBlocks, Boolean contiguous ) 
{
    Extent a; a.len = 0;
    UINT32 nBytes = nBlocks * logicalBlockSize;
    UINT32 i = 0;

    // allocated from non-erased list first
    do {
        if ( freeSpace.getCount() > 0 ) {
            if ( nBytes < freeSpace[i].len ) {
                a = freeSpace[i];
                a.len = nBytes;
                freeSpace[i].start.sector += nBlocks;
                freeSpace[i].len -= nBytes;
            } else if ( (nBytes == freeSpace[i].len) 
                       || ( !( contiguous ) 
                             && ( (freeSpace[i].len/logicalBlockSize)>1 ))) {
                a = freeSpace[i];
                freeSpace.remove(i);
            } else i++;
        }
    } while (( a.len==0 ) && ( i<freeSpace.getCount() ));

    return a;
}

NSR::UINT32 
NSR::Volume::totalFree() const 
{
    UINT32 total = 0;
    int i;

    for (i=0;i<freeSpace.getCount();i++)
        total += freeSpace[i].len;
    for (i=0;i<erasedFreeSpace.getCount();i++)
        total += erasedFreeSpace[i].len;
    return total;
}

NSR::UINT32 
NSR::Volume::largestExtent() const 
{
    UINT32 largest = 0;
    int i;

    for (i=0; i<freeSpace.getCount(); i++)
        if ( freeSpace[i].len > largest ) 
            largest = freeSpace[i].len;
    
    for (i=0; i<erasedFreeSpace.getCount(); i++)
        if ( erasedFreeSpace[i].len > largest ) 
            largest = erasedFreeSpace[i].len;
    
    return largest;
}

void 
NSR::VolumeTable::copy( const VolumeTable &obj ) 
{
    count = obj.count;
    element = new Volume[size = obj.size];
    for (int i=0; i<size; i++)
        element[i] = obj.element[i];

}

void
NSR::VolumeTable::extend()
{
    Volume *tmp;
    tmp = new Volume[size+10];
    for (int i=0; i<size; i++)
        tmp[i] = element[i];
    size += 10;
    delete [] element;
    element = tmp;
}

// Removes the entry at table[x].
void
NSR::VolumeTable::remove( UINT16 x )
{
    if (x < count) {
        for (UINT16 i=x; i<(count-1); i++)
            element[i] = element[i+1];
        count--;
    }
}

NSR::UINT16
NSR::VolumeTable::vid( sid_t sid ) const
{
    for (UINT16 i=0; i<count; i++)
        if (sid == element[i].sid()) return i;
    return 0xffff;
}

void
NSR::ICB_Table::bumpPtr( UINT32 n )
{
    relOffset += n;
    
    UINT32 nSectors = table[extent].len/logicalBlockSize + 
                    ((table[extent].len % logicalBlockSize) > 0);
    while( relOffset >= nSectors ) 
    {
        if (extent+1 == table.getCount()) {
            relOffset = (table[extent].len/logicalBlockSize) - 1;
            err = error_End_of_ICB_table;
            return;
        }
        relOffset %= nSectors;
        extent++;
        nSectors = table[extent].len/logicalBlockSize + 
                   ((table[extent].len % logicalBlockSize) > 0);
    }
}

NSR::Address
NSR::ICB_Table::currentAddr()
{
    Address a = table[extent].start;
    a.sector += relOffset;

    return a;
}

NSR::Address
NSR::ICB_Table::nextAddr()
{
    Address a = table[extent].start;

    if ( (relOffset+1)/(table[extent].len/logicalBlockSize) >= 1 ) {
        if ((extent+1) == table.getCount()) {
            a = lastAddr();
            err = error_End_of_ICB_table;
            return a;
        } else
            a = table[extent+1].start;
    } else 
        a.sector += relOffset + 1;

    return a;
}

NSR::Address
NSR::ICB_Table::lastAddr()
{
    Address a = table[table.getCount()-1].start;
    a.sector += (table[table.getCount()-1].len/logicalBlockSize) - 1;

    return a;
}

//
// finds the address at a given positive offset from the current address
// This is necessary when writing EA's.
//
NSR::Address
NSR::ICB_Table::addrAtOffset(INT32 offset)
{
    UINT32 rel_offset = relOffset + offset;
    UINT32 extent_idx = extent;
    Address a;
    
    while( rel_offset >= (table[extent_idx].len/logicalBlockSize) ) 
    {
        if (extent_idx+1 >= table.getCount()) {
            err = error_End_of_ICB_table;
            a = lastAddr();
            return a;
        }
        rel_offset %= table[extent_idx].len/logicalBlockSize;
        extent_idx++;
    }
    
    a = table[extent_idx].start;
    a.sector += rel_offset;

    return a;
}

void
NSR::DirectoryContents::copy( const DirectoryContents &obj )
{
    count = obj.count;
    element = new DirectoryEntry[size = obj.size];
    for (int i=0; i<size; i++)
        element[i] = obj.element[i];
}

void
NSR::DirectoryContents::extend()
{
    DirectoryEntry *tmp;
    tmp = new DirectoryEntry[size+10];
    for (int i=0; i<size; i++)
        tmp[i] = element[i];
    size += 10;
    delete [] element;
    element = tmp;
}

void
NSR::DirectoryContents::remove( unsigned x )
{
    if (x < count) {
        for (int i=x; i<(count-1); i++) element[i] = element[i+1];
        count--;
    }
}

void
NSR::DirectoryStack::copy( const DirectoryStack &obj )
{
    count = obj.count;
    element = new Directory[size = obj.size];
    for (int i=0; i<size; i++)
        element[i] = obj.element[i];
}

void
NSR::DirectoryStack::extend()
{
    Directory *tmp;
    tmp = new Directory[size+5];
    for (int i=0; i<size; i++)
        tmp[i] = element[i];
    size += 5;
    delete [] element;
    element = tmp;
}

